'''
Created on May 19, 2010

@author: Fawkes
'''
import animations
import pygame
from pygame.locals import *
from random import randint
from sys import exit
from lostsky.bullet_scripts import catalog

class BattleEventSystem(object):
    '''
    Battle System - Main class to handle battle event logic between two units.
    '''

    def __init__(self, engine):
        '''
        Function name: __init__
        Purpose: Sets up the battle system
        '''
        self.engine = engine
        self.bg_surface = pygame.Surface(engine.surface.get_size())
        self.spell_anim_catalog = catalog.get_catalog()
        self.revive_anim = self.spell_anim_catalog['special_revive'](self.engine.surface,self.bg_surface)

        
    def battle_event(self,lhs_unit,rhs_unit,script=None):
        '''
        function name: Battle Event: 
        Purpose:    Runs a single battle event between the LHS_unit and the RHS_unit.
                    If a script is provided, follows the pre-scripted battle event
        Inputs: lhs_unit, rhs_unit - participants in the battle
                script - a dictionary of the form {lhs_hit:T/F ,
                                                   lhs_crit:T/F,
                                                    rhs_hit: T/F,
                                                     rhs_crit:T/F}
        Outputs: lhs_effect, rhs_effect - HP change caused by unit actions 
                 lhs_crit, rhs_crit - Critical hit event T/F
        
        '''
        
        # initializes a fresh sprite group
        self.chara_sprites = pygame.sprite.RenderUpdates()
        
        # Initialize the animation system
        if self.engine.options.battle_anim:
       
            if lhs_unit.animation_enable:
                lhs_anim = animations.CharaAnimData(lhs_unit.anim_frames,rhs=False)
            else:
                lhs_anim = animations.CharaAnimPlaceholder(lhs_unit,rhs=False)
            self.chara_sprites.add(lhs_anim)
            
            if rhs_unit.animation_enable:
                rhs_anim = animations.CharaAnimData(rhs_unit.anim_frames,rhs=True)
            else:
                rhs_anim = animations.CharaAnimPlaceholder(rhs_unit,rhs=True)
            self.chara_sprites.add(rhs_anim)
        
        if rhs_unit.spell_actions[rhs_unit.equipped] and hasattr(rhs_unit.spell_actions[rhs_unit.equipped],'animation'):
            anim_name = rhs_unit.spell_actions[rhs_unit.equipped].animation
            if anim_name not in self.spell_anim_catalog:
                #default to fireball animation
                rhs_spell_anim = self.spell_anim_catalog['all_fireball'](self.engine.surface,self.bg_surface)
            else: 
                rhs_spell_anim = self.spell_anim_catalog[anim_name](self.engine.surface,self.bg_surface)
        else:
            rhs_spell_anim = None

        if lhs_unit.spell_actions[lhs_unit.equipped] and hasattr(lhs_unit.spell_actions[lhs_unit.equipped],'animation'):
            anim_name = lhs_unit.spell_actions[lhs_unit.equipped].animation
            if anim_name not in self.spell_anim_catalog:
                #default to fireball animation
                lhs_spell_anim = self.spell_anim_catalog['all_fireball'](self.engine.surface,self.bg_surface)
            else: 
                lhs_spell_anim = self.spell_anim_catalog[anim_name](self.engine.surface,self.bg_surface)
        else:
            lhs_spell_anim = None

        # Calculate LHS effect
        if script:
            lhs_effect, lhs_crit = self.get_battle_effects(lhs_unit, rhs_unit, True, script['lhs_hit'], script['lhs_crit'])
        else:
            lhs_effect, lhs_crit = self.get_battle_effects(lhs_unit, rhs_unit, False)
        
        print lhs_unit.name, ' uses ', lhs_unit.spell_actions[lhs_unit.equipped].name, ' on ', rhs_unit.name
            
        text_lhs_effect = self.get_text_effect(lhs_unit, lhs_effect)
        
        # Draw LHS's action
        if self.engine.options.battle_anim:
            self.draw_idle_delay(lhs_unit, rhs_unit, lhs_anim, rhs_anim)
            self.draw_lhs_action(lhs_unit, rhs_unit, lhs_anim, rhs_anim)
            self.draw_lhs_spell(lhs_unit, rhs_unit, lhs_anim, rhs_anim,lhs_spell_anim)
            if lhs_effect != 'miss' and lhs_unit.spell_actions[lhs_unit.equipped].type == 'attack':
                if lhs_crit:
                    self.engine.sfx_system.sound_catalog['crit'].play()
                else:
                    self.engine.sfx_system.sound_catalog['hit'].play()
            elif (lhs_unit.spell_actions[lhs_unit.equipped].type == 'healing' or 
                lhs_unit.spell_actions[lhs_unit.equipped].type == 'support'):
                self.engine.sfx_system.sound_catalog['heal'].play()
            else:
                self.engine.sfx_system.sound_catalog['miss'].play()
        else:
            self.draw_map_delay(lhs_unit, rhs_unit)
    
        # Apply battle effects
        rhs_hp_before = rhs_unit.HP
        revive_anim = self.apply_battle_effects(lhs_unit, rhs_unit, lhs_effect, lhs_crit)
            
        rhs_hp_after = rhs_unit.HP
        
        # Draw effect of LHS's action
        if self.engine.options.battle_anim:
            
            # Case of healing spells that don't cure HP or for support spells
            if lhs_effect != 0:
                self.draw_lhs_effect_text(lhs_unit, rhs_unit, lhs_anim, rhs_anim, text_lhs_effect)
            if revive_anim:
                self.draw_revive_effect(lhs_unit, rhs_unit, lhs_anim, rhs_anim, 'rhs')
            self.draw_idle_delay(lhs_unit, rhs_unit, lhs_anim, rhs_anim)
        else:
            self.draw_map_lhs_effect(lhs_unit, rhs_unit, text_lhs_effect)
            if revive_anim:
                self.draw_map_revive_effect(lhs_unit, rhs_unit, 'rhs')
            rhs_unit.render_hp_change(rhs_hp_before, rhs_hp_after)
        
        
        # Case: RHS unit counterattacks
        if lhs_unit.spell_actions[lhs_unit.equipped].type == "attack" and rhs_unit.can_counterattack(lhs_unit):
        
            rhs_action = True
        
            # Calculate RHS effect
            if script:
                rhs_effect, rhs_crit = self.get_battle_effects(rhs_unit, lhs_unit, True, script['rhs_hit'], script['rhs_crit'])
            else:
                rhs_effect, rhs_crit = self.get_battle_effects(rhs_unit, lhs_unit, False)

            print rhs_unit.name, ' uses ', rhs_unit.spell_actions[rhs_unit.equipped].name, ' on ', lhs_unit.name
            
            text_rhs_effect = self.get_text_effect(rhs_unit, rhs_effect)
                
            # Draw RHS's action
            if self.engine.options.battle_anim:
                self.draw_idle_delay(lhs_unit, rhs_unit, lhs_anim, rhs_anim)
                self.draw_rhs_action(lhs_unit, rhs_unit, lhs_anim, rhs_anim)
                self.draw_rhs_spell(lhs_unit, rhs_unit, lhs_anim, rhs_anim,rhs_spell_anim)
                if rhs_effect != 'miss' and rhs_unit.spell_actions[rhs_unit.equipped].type == 'attack':
                    if rhs_crit:
                        self.engine.sfx_system.sound_catalog['crit'].play()
                    else:
                        self.engine.sfx_system.sound_catalog['hit'].play()
                elif (rhs_unit.spell_actions[rhs_unit.equipped].type == 'healing' or
                    rhs_unit.spell_actions[rhs_unit.equipped].type == 'support'):
                    self.engine.sfx_system.sound_catalog['heal'].play()
                else:
                    self.engine.sfx_system.sound_catalog['miss'].play()
            else:
                self.draw_map_delay(lhs_unit,rhs_unit)
            
            # Apply RHS unit's battle effects to LHS unit 
            lhs_hp_before = lhs_unit.HP

            revive_anim = self.apply_battle_effects(rhs_unit, lhs_unit, rhs_effect, rhs_crit)

            lhs_hp_after = lhs_unit.HP
                
            # Draw the resulting effect
            if self.engine.options.battle_anim:
                self.draw_rhs_effect_text(lhs_unit, rhs_unit, lhs_anim, rhs_anim,text_rhs_effect)
                if revive_anim:
                    self.draw_revive_effect(lhs_unit, rhs_unit, lhs_anim, rhs_anim, 'lhs')
            else:
                self.draw_map_rhs_effect(lhs_unit, rhs_unit, text_rhs_effect)
                if revive_anim:
                    self.draw_map_revive_effect(lhs_unit, rhs_unit, 'lhs')

                lhs_unit.render_hp_change(lhs_hp_before, lhs_hp_after)
                
            sc_cost_rhs = rhs_unit.spell_actions[rhs_unit.equipped].sc_cost
        
        else:
            rhs_action = False
            rhs_crit = False
            rhs_effect = 'n/a'
            sc_cost_rhs = 0
            
            
        # Wrap up the battle
        sc_cost_lhs =  lhs_unit.spell_actions[lhs_unit.equipped].sc_cost
        self.post_battle_event(lhs_unit, rhs_unit,rhs_action)
        
        # Return battle results
        return lhs_effect, rhs_effect, lhs_crit, rhs_crit, sc_cost_lhs, sc_cost_rhs
        
        
    def get_battle_effects(self,acting_unit,target,scripted,hit=None,crit=None):
        
        """
        # Fuction name: get_battle_effects
        # Purpose: Computes the damage stats of the battle of the battle
        # Inputs: lhs_unit, rhs_unit
        #         script - scripted events dictionary corresponding to {lhs_hit:T/F , lhs_crit:T/F, rhs_hit: T/F, rhs_crit:T/F}
        # Outputs:
        #   effect - Damage inflicted or HP healed
        #   critical - True if a critical hit is inflicted
        #   side - RHS/LHS for attacker and defender
        """
        
        if acting_unit.spell_actions[acting_unit.equipped].type == 'attack':
            
            # Case for scripted actions (only affects offensive actions / counterattacks)
            if scripted and hit:
                effect = acting_unit.compute_damage(target)
                if crit:
                    critical = True
                    effect = int(effect*1.5)
                else:
                    critical = False
            elif scripted and not hit:
                effect, critical = ['miss',False]
            else:
                effect, critical = acting_unit.spell_actions[acting_unit.equipped].get_effect(acting_unit,target)

        elif acting_unit.spell_actions[acting_unit.equipped].type in ('healing','support'):
            effect = acting_unit.spell_actions[acting_unit.equipped].get_effect(acting_unit,target)
            critical = False
        return effect, critical
    
    def get_text_effect(self,acting_unit,effect):
        """
        # Fuction name: get_text_effect
        # Purpose: Generates the pygame surface for a given effect value
        # Inputs: acting unit - The participant taking action
        #         effect - effect to render
        # Output: text_effect - pygame surface of the damage text
        """
        
        # Case for Offensive Combat Interactions
        if acting_unit.spell_actions[acting_unit.equipped].type == 'attack':
            
            if self.engine.options.battle_anim:
                if effect == 'miss':
                    text_effect = self.engine.cfont.render("MISS!",True,(255,0,0))
                else:    
                    text_effect = self.engine.cfont.render("%3.0f"%effect,True,(255,0,0))
            else:
                if effect == 'miss':
                    text_effect = self.engine.render_outlined_text("MISS!",self.engine.cfont,(255,0,0),(255,255,255))
                else:    
                    text_effect = self.engine.render_outlined_text(str(effect),self.engine.cfont,(255,0,0),(255,255,255))
        
        # Case for Healing Combat interactions
        elif acting_unit.spell_actions[acting_unit.equipped].type in ('healing','support'):
            if self.engine.options.battle_anim:
                text_effect = self.engine.cfont.render("%3.0f"%effect,True,(100,255,100))
            else:
                text_effect = self.engine.render_outlined_text(str(effect),self.engine.cfont,(100,255,100),(0,0,0))
        
        return text_effect
    
    def apply_battle_effects(self,acting_unit,target,effect,critical):
        
        """
        # Fuction name: apply_battle_effects
        # Purpose: Applies the battle effects to the target
        # Inputs: acting unit, target - The participants in the battle
        #         effect - effect to apply to the target
        #         critical - critical hit if applicable
        # Output: revive_anim - T/F if revive animation needs to be shown
        """
        revive_anim = False
        
        if acting_unit.spell_actions[acting_unit.equipped].type == 'attack' and effect != "miss":
            target.HP -= effect
            
            # If the defender is dead, sets HP to 0 and sets their alive flag to False
            if target.HP <= 0:
                target.HP = 0
                target.alive = False

                # 10% chance of surviving a killing blow at high spirits
                #   If it works, the target survives with 1HP.
                #   The target has their spirit reduced by 5.0
                if target.spirit_stats == "high":
                    print 'High Spirits Death Save'
                    x = randint(1,10)
                    if x == 10:
                        target.alive = True
                        target.HP = 1
                        target.spirit -= 500
                # Checks if there are any auto-revive traits equipped
                elif target.ressurected == False:
                    if (target.has_trait_property('Revive Lv.1')
                        or target.has_trait_property('Revive Lv.2')
                        or target.has_trait_property('Revive Lv.3')):
                        
                        # Revives the target and returns a flag to show the barrier animation
                        revive_anim = self.revive(target)
            
            # Status Effects Processed here
            elif target.HP > 0 and acting_unit.spell_actions[acting_unit.equipped].status_effects:
                acting_unit.spell_actions[acting_unit.equipped].give_status_effect(acting_unit,target)
                        
        elif acting_unit.spell_actions[acting_unit.equipped].type in ('healing','support'):
            target.HP += effect
            if target.HP > target.maxHP:
                target.HP = target.maxHP
                
        return revive_anim
                
    def post_battle_event(self,lhs_unit,rhs_unit,rhs_action):
        '''
        # Function Name: Post-battle Event
        # Purpose: Processes post battle stats updates like lowering spell use counts
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         rhs_action - whether the RHS unit took action (counter) during this event
        '''
        
        
        
        # Lowers life of the spell action by 1
        lhs_unit.spell_actions[lhs_unit.equipped].livesleft -= 1

        # Takes off SC cost of the spell action
        if lhs_unit.spell_actions[lhs_unit.equipped].sc_cost > 0:
            lhs_unit.spirit -= lhs_unit.spell_actions[lhs_unit.equipped].sc_cost
            if lhs_unit.spirit < 0:
                lhs_unit.spirit = 0
            
        # If lives left are 0, replaces the equipped slot with empty
        if lhs_unit.spell_actions[lhs_unit.equipped].livesleft == 0:
            lhs_unit.map.display_alert('Spell Action Used Up!',
                                       "%s's %s has run out of uses!"%(lhs_unit.name,
                                                                       lhs_unit.spell_actions[lhs_unit.equipped].namesuffix))
            
            if lhs_unit.spell_actions[lhs_unit.equipped].consumable == True:
                lhs_unit.spell_actions[lhs_unit.equipped] = None
        
        
        # Checks the opponent if they have an equipped spell and if it was used in a counterattack
        if rhs_action:
            # Lowers life of the spell action by 1
            rhs_unit.spell_actions[rhs_unit.equipped].livesleft -= 1
            
                
            # Takes off SC cost of the spell action
            if rhs_unit.spell_actions[rhs_unit.equipped].sc_cost > 0:
                rhs_unit.spirit -= rhs_unit.spell_actions[rhs_unit.equipped].sc_cost
                if rhs_unit.spirit < 0:
                    rhs_unit.spirit = 0
                
            # If lives left are 0, replaces the equipped slot with empty
            if rhs_unit.spell_actions[rhs_unit.equipped].livesleft == 0:
                rhs_unit.map.display_alert('Spell Action Used Up!',
                                           "%s's %s has run out of uses!"%(rhs_unit.name,
                                                                           rhs_unit.spell_actions[rhs_unit.equipped].namesuffix))
                if rhs_unit.spell_actions[rhs_unit.equipped].consumable == True:
                    rhs_unit.spell_actions[rhs_unit.equipped] = None
    
     
    def draw_idle_delay(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim):
        '''
        # Function Name: draw_idle_delay
        # Purpose: Half second idle delay
        #         LHS Unit Animation: Idle
        #         RHS Unit Animation: Idle
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        '''
        
        if lhs_unit.animation_enable:
            lhs_anim.set_mode('idle')
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('idle')

        max_frames = 30

        for frame_num in xrange(0,max_frames):
            # Prevents lockup of the system during animation
            self.check_exit()
            self.render_background(lhs_unit, rhs_unit)
            self.render_combat_gui()
            self.render_combat_data(lhs_unit, rhs_unit)
            
            self.chara_sprites.update()
            self.chara_sprites.draw(self.engine.surface)

            # Defaults to map sized sprite if no animation available
            #if lhs_unit.animation_enable:
            #    lhs_anim.render(self.engine.surface)
            #else:
            #    self.engine.surface.blit(lhs_unit.image, (210,280),(175,0,35,35))
            #if rhs_unit.animation_enable:
            #    rhs_anim.render(self.engine.surface)
            #else:
            #    self.engine.surface.blit(rhs_unit.image, (595,280),(280,0,35,35))
            
            pygame.display.flip()
            self.engine.clock.tick(60)
            
    def draw_lhs_action(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim):
        '''
        # Function Name: draw_lhs_action
        # Purpose: Draw the LHS unit activating the spell
        #         LHS Unit Animation: Physical/Magical Spell Activation
        #         RHS Unit Animation: Idle
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        '''
        if lhs_unit.animation_enable:
            lhs_anim.set_mode(lhs_unit.spell_actions[lhs_unit.equipped].attacktype)
            max_frames = lhs_anim.FRAME_REPEAT*len(lhs_anim.anim_dict[lhs_anim.mode]['frames'])
        else:
            max_frames = 1
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('idle')
            
            
        # Draws first frame
        self.render_background(lhs_unit, rhs_unit,self.bg_surface)
        self.render_combat_gui(self.bg_surface)
        self.render_combat_data(lhs_unit, rhs_unit,self.bg_surface)
        
        self.engine.surface.blit(self.bg_surface,(0,0))
        self.chara_sprites.update()
        self.chara_sprites.draw(self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)
    
        # Runs once through the LHS unit's activation animation
        for frame_num in xrange(1,max_frames):
            self.check_exit()
            
            self.chara_sprites.clear(self.engine.surface, self.bg_surface)
            self.chara_sprites.update()
            rects = self.chara_sprites.draw(self.engine.surface)
            
            pygame.display.update(rects)
            self.engine.clock.tick(60)
    
    def draw_lhs_spell(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim,lhs_spell_anim):
        '''
        # Function Name: draw_lhs_spell
        # Purpose: Draw the LHS unit casting the spell and spell bullets
        #         LHS Unit Animation: Physical/Magical Spell Casting
        #         RHS Unit Animation: Idle
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        '''
        
        if lhs_unit.animation_enable:
            lhs_anim.set_mode("cast_"+lhs_unit.spell_actions[lhs_unit.equipped].attacktype)
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('idle')
        
        self.render_background(lhs_unit, rhs_unit,self.bg_surface)
        self.render_combat_gui(self.bg_surface)
        self.render_combat_data(lhs_unit, rhs_unit,self.bg_surface)
        
        self.engine.surface.blit(self.bg_surface,(0,0))
        self.chara_sprites.update()
        self.chara_sprites.draw(self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)

        num_sfx = len(lhs_spell_anim.sfx_timings)

        #Initialize sound effects
        if lhs_spell_anim.sfx_timings:
            current_timing, current_effect = lhs_spell_anim.sfx_timings[0]
            sfx_index = 0
        else:
            current_timing = None
            current_effect = None
            sfx_index = 0

        t0 = pygame.time.get_ticks()

        for frame_counter in xrange(0,lhs_spell_anim.max_frames):
            self.check_exit()

            rhs_mode = False

            # Play a sound effect if the current sound effect is triggered on this frame
            if current_effect and frame_counter >= current_timing:
                self.engine.sfx_system.sound_catalog[current_effect].play()
                sfx_index += 1
                if sfx_index < num_sfx:
                    current_timing, current_effect = lhs_spell_anim.sfx_timings[sfx_index]
                else:
                    current_timing = None
                    current_effect = None


            self.chara_sprites.clear(self.engine.surface, self.bg_surface)
            lhs_spell_anim.clear()
            self.chara_sprites.update()
            rects = self.chara_sprites.draw(self.engine.surface)
            rects += lhs_spell_anim.update(t0,rhs_mode)

            pygame.display.update(rects)
            t0 = pygame.time.get_ticks()
            self.engine.clock.tick(60)
         
    def draw_lhs_effect_text(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim,text_lhs_effect):
        '''
        # Function Name: draw_lhs_effect_text
        # Purpose: Draw the effect the LHS unit had on the RHS unit for 1 s
        #         LHS Unit Animation: Idle
        #         RHS Unit Animation: Idle
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        #         text_lhs_effect - effect of LHS's action on RHS unit
        '''
        frame_num = 0
        if lhs_unit.animation_enable:
            lhs_anim.set_mode('idle')
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('idle')

        max_frames = 60

        if rhs_unit.invincible and lhs_unit.spell_actions[lhs_unit.equipped].type == 'attack':
            self.revive_animation(False)


        for frame_num in xrange(0,max_frames):
            # Prevents lockup of the system during animation
            self.check_exit()
            self.render_background(lhs_unit, rhs_unit)
            self.render_combat_gui()
            self.render_combat_data(lhs_unit, rhs_unit)
            
            self.chara_sprites.update()
            self.chara_sprites.draw(self.engine.surface)

            # Draws damage caused by LHS unit
            self.engine.surface.blit(text_lhs_effect, (610-text_lhs_effect.get_width()/2, 145))
        
            pygame.display.flip()
            self.engine.clock.tick(60)
         
    def draw_rhs_action(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim):
        '''
        # Function Name: draw_rhs_action
        # Purpose: Draw the RHS unit activating the spell
        #         LHS Unit Animation: Idle
        #         RHS Unit Animation: Physical/Magical Spell Activation
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        '''
        
        if lhs_unit.animation_enable:
            lhs_anim.set_mode('idle')
        if rhs_unit.animation_enable:
            rhs_anim.set_mode(rhs_unit.spell_actions[rhs_unit.equipped].attacktype)
            max_frames = rhs_anim.FRAME_REPEAT*len(rhs_anim.anim_dict[rhs_anim.mode]['frames'])
        else:
            max_frames = 1
            
            
        # Draws first frame
        self.render_background(lhs_unit, rhs_unit,self.bg_surface)
        self.render_combat_gui(self.bg_surface)
        self.render_combat_data(lhs_unit, rhs_unit,self.bg_surface)
        
        self.engine.surface.blit(self.bg_surface,(0,0))
        self.chara_sprites.update()
        self.chara_sprites.draw(self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)
    
        # Runs once through the RHS unit's activation animation
        for frame_num in xrange(0,max_frames):
            self.check_exit()
            
            self.chara_sprites.clear(self.engine.surface, self.bg_surface)
            self.chara_sprites.update()
            rects = self.chara_sprites.draw(self.engine.surface)
            
            pygame.display.update(rects)
            self.engine.clock.tick(60)
    
    def draw_rhs_spell(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim,rhs_spell_anim):
        '''
        # Function Name: draw_rhs_spell
        # Purpose: Draw the RHS unit casting the spell and spell bullets
        #         LHS Unit Animation: Idle
        #         RHS Unit Animation: Physical/Magical Spell Casting
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        '''
        
        if lhs_unit.animation_enable:
            lhs_anim.set_mode('idle')
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('cast_'+rhs_unit.spell_actions[rhs_unit.equipped].attacktype)
        
        self.render_background(lhs_unit, rhs_unit,self.bg_surface)
        self.render_combat_gui(self.bg_surface)
        self.render_combat_data(lhs_unit, rhs_unit,self.bg_surface)
        
        self.engine.surface.blit(self.bg_surface,(0,0))
        self.chara_sprites.update()
        self.chara_sprites.draw(self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)


        # Initialize Sound Effects
        num_sfx = len(rhs_spell_anim.sfx_timings)

        if rhs_spell_anim.sfx_timings:
            current_timing, current_effect = rhs_spell_anim.sfx_timings[0]
            sfx_index = 0
        else:
            current_timing = None
            current_effect = None
            sfx_index = 0

        t0 = pygame.time.get_ticks()

        for frame_counter in xrange(0,rhs_spell_anim.max_frames):
            self.check_exit()

            rhs_mode = True

            # Play a sound effect if the current sound effect is triggered on this frame
            if current_effect and frame_counter >= current_timing:
                self.engine.sfx_system.sound_catalog[current_effect].play()
                sfx_index += 1
                if sfx_index < num_sfx:
                    current_timing, current_effect = rhs_spell_anim.sfx_timings[sfx_index]
                else:
                    current_timing = None
                    current_effect = None

            self.chara_sprites.clear(self.engine.surface, self.bg_surface)
            rhs_spell_anim.clear()
            self.chara_sprites.update()
            rects = self.chara_sprites.draw(self.engine.surface)
            rects += rhs_spell_anim.update(t0,rhs_mode)

            pygame.display.update(rects)
            t0 = pygame.time.get_ticks()
            self.engine.clock.tick(60)
    
    def draw_rhs_effect_text(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim,text_rhs_effect):
        '''
        # Function Name: draw_rhs_effect_text
        # Purpose: Draw the effect the RHS unit had on the LHS unit for 1 s
        #         LHS Unit Animation: Idle
        #         RHS Unit Animation: Idle
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        #         text_rhs_effect - effect of RHS's action on LHS unit
        '''
        
        frame_num = 0
        if lhs_unit.animation_enable:
            lhs_anim.set_mode('idle')
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('idle')

        max_frames = 60

        if lhs_unit.invincible and rhs_unit.spell_actions[rhs_unit.equipped].type == 'attack':
            self.revive_animation(True)

        for frame_num in xrange(0,max_frames):
            # Prevents lockup of the system during animation
            self.check_exit()
            self.render_background(lhs_unit, rhs_unit)
            self.render_combat_gui()
            self.render_combat_data(lhs_unit, rhs_unit)
            
            
            self.chara_sprites.update()
            self.chara_sprites.draw(self.engine.surface)

            # Defaults to map sized sprite if no animation available
            
            # if lhs_unit.animation_enable:
                # lhs_anim.render(self.engine.surface)
            # else:
                # self.engine.surface.blit(lhs_unit.image, (210,280),(175,0,35,35))
            # if rhs_unit.animation_enable:
                # rhs_anim.render(self.engine.surface)
            # else:
                # self.engine.surface.blit(rhs_unit.image, (595,280),(280,0,35,35))
            
            # Draws damage caused by LHS unit
            self.engine.surface.blit(text_rhs_effect, (225-text_rhs_effect.get_width()/2, 145))
        
            pygame.display.flip()
            self.engine.clock.tick(60)
    
    def draw_revive_effect(self,lhs_unit,rhs_unit,lhs_anim,rhs_anim,revive):
        '''
        # Function Name: draw_rhs_spell
        # Purpose: Draw an animation indicating that a character has 
        #          executed a revive trait (counterbomb)
        #         LHS Unit Animation: Idle
        #         RHS Unit Animation: Idle
        #
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_anim,rhs_anim - animation data for units in battle
        #         revive - 'lhs' or 'rhs'
        '''
        
        if lhs_unit.animation_enable:
            lhs_anim.set_mode('idle')
        if rhs_unit.animation_enable:
            rhs_anim.set_mode('idle')
            
        if revive == 'lhs':
            trait_name = self.engine.render_outlined_text(lhs_unit.traits[0][0].name+"!",self.engine.cfont,(255,0,0),(255,255,255))
        
            # Draw as if spell anim cast by RHS unit
            rhs_mode = True
            
        else:
            
            trait_name = self.engine.render_outlined_text(rhs_unit.traits[0][0].name+"!",self.engine.cfont,(255,0,0),(255,255,255))
        
            # Draw as if spell anim cast by LHS unit
            rhs_mode = False
        
        self.render_background(lhs_unit, rhs_unit,self.bg_surface)
        self.render_combat_gui(self.bg_surface)
        self.render_combat_data(lhs_unit, rhs_unit,self.bg_surface)
        if revive == 'lhs':
           self.bg_surface.blit(trait_name, (225-trait_name.get_width()/2, 145))

        else:
            self.bg_surface.blit(trait_name, (610-trait_name.get_width()/2, 145))

        self.engine.surface.blit(self.bg_surface,(0,0))
        self.chara_sprites.update()
        self.chara_sprites.draw(self.engine.surface)
        self.revive_animation(rhs_mode)

    def revive_animation(self, rhs_mode):

        """
        function name: draws the revive animation
        inputs: rhs_mode - True if the animation is played while the RHS unit is taking its action
        """

        pygame.display.flip()
        self.engine.clock.tick(60)


        t0 = pygame.time.get_ticks()
            
        for frame_counter in xrange(0,self.revive_anim.max_frames):
            self.check_exit()
            
            self.chara_sprites.clear(self.engine.surface, self.bg_surface)
            self.revive_anim.clear()
            self.chara_sprites.update()
            rects = self.chara_sprites.draw(self.engine.surface)
            rects += self.revive_anim.update(t0,rhs_mode)

            pygame.display.update(rects)
            t0 = pygame.time.get_ticks()
            self.engine.clock.tick(60)
        
        
        # Resets revive animation for next time
        self.revive_anim.anim_sprite.current_frame = 0

    def check_exit(self):
        """
        # Function Name: check_exit
        # Purpose: Checks if the user has exited in the middle of a battle (prevents
        # game from freezing up)
        """
        for event in pygame.event.get():
            if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                exit()
        
    def revive(self,user):
        """
        # Function Name: revive_execute
        # Purpose: Does whatever the trait is designed to do before the battle
        # Inputs: user - The person who is using this trait
        """
    
        print "Revive attempted for "+user.name
        if 'Revive Lv.1' in user.traits[0][0].properties:
            # Chance to revive
            chance = 40
            # HP to revive to (multiply to Max HP)
            HP_ressurect = 0.05
            # SC pentalty
            SC_penalty = 3.0
            
        if 'Revive Lv.2' in user.traits[0][0].properties:
            chance = 60
            HP_ressurect = 0.20
            SC_penalty = 4.0
        
        if 'Revive Lv.3' in user.traits[0][0].properties:
            chance = 100
            HP_ressurect = 1
            SC_penalty = 9.9
        
        if randint(0,100) < chance:
            print "Revive successful!"
            user.alive = True
            user.ressurected = True
            user.spirit -= SC_penalty
            recovery = int(HP_ressurect*user.maxHP)
            user.HP = recovery
            
            revive_success = True
        else:
            revive_success = False
            
        if user.spirit < 0.0:
            user.spirit = 0.0
            
        return revive_success
    
    def render_background(self,lhs_unit,rhs_unit,target_surface = None):
        '''
        # Function Name: draw_background
        # Purpose: Draws the background corresponding to the terrain that each unit is standing on.
        '''
        if not target_surface:
            target_surface = self.engine.surface
        
        target_surface.blit(self.engine.battle_bg,(0,0))
        
        #self.engine.surface.blit(lhs_unit.map.terrainmap[tuple(lhs_unit.location_tile)][0].bg_img,(0,0))
        #self.engine.surface.blit(rhs_unit.map.terrainmap[tuple(rhs_unit.location_tile)][0].bg_img,(420,0))
        
    def render_unit_stats(self,unit,rhs,target_surface = None):
        """
        # Function Name: plot_battle_lhs
        # Purpose: Plots the unit's portrait, name, HP, four stats, active spell, and active spell's mods to the
        # left side of the display panel
        """
        if rhs:
            offset = 420
        else:
            offset = 0
        
        if not target_surface:
            target_surface = self.engine.surface
        
        
        text_name = self.engine.bfont.render(unit.name,True, (0,0,0))
        text_line1 = self.engine.sfont.render("HP: %3.0f/%3.0f"%(unit.HP,unit.maxHP),True, (0,0,0))
        text_line2 = self.engine.sfont.render("LV: %2.0f    EXP: %3.0f/100"%(unit.level,unit.exp),True, (0,0,0))
        
        # Trait lines
        if unit.traits[0][0]:
            text_line3_str = "Action Trait: "+ unit.traits[0][0].name
        else:
            text_line3_str = "Action Trait: None"
            
        if unit.traits[1][0]:
            text_line4_str = 'Support Trait: '+unit.traits[1][0].name
        else:
            text_line4_str = 'Support Trait: None'
            
        text_line3 = self.engine.sfont.render(text_line3_str,True, (0,0,0))
        text_line4 = self.engine.sfont.render(text_line4_str,True, (0,0,0))
            
        text_line5 = self.engine.sfont.render("Spirit Charge: %3.0f"%(unit.spirit),True, (0,0,0))

        if unit.spell_actions[unit.equipped]:
            # Attack type
            target_surface.blit(self.engine.spell_icons_big, (150+offset,552),unit.spell_actions[unit.equipped].a_type_big)            
            # Damage type
            target_surface.blit(self.engine.spell_icons_big, (175+offset,552),unit.spell_actions[unit.equipped].d_type_big)
            # Spell Type
            target_surface.blit(self.engine.spell_icons_big, (200+offset,552),unit.spell_actions[unit.equipped].al_type_big)
            # Counterattack
            target_surface.blit(self.engine.spell_icons_big, (225+offset,552),unit.spell_actions[unit.equipped].c_type_big)
            
        target_surface.blit(unit.av, (10+offset, 500))
        target_surface.blit(text_name, (150+offset,500))
        target_surface.blit(text_line1, (150+offset,520))
        target_surface.blit(text_line2, (150+offset,535))
        target_surface.blit(text_line3, (150+offset,575))
        target_surface.blit(text_line4, (150+offset,590))
        target_surface.blit(text_line5, (150+offset,605))

   

    def render_combat_gui(self,target_surface = None):
        """
        # Function Name: render combat gui
        # Purpose: renders the background objects
        """
        
        if not target_surface:
            target_surface = self.engine.surface
        
        # Displays the two battle panels

        target_surface.blit(self.engine.battle_panel, (175,290))
        target_surface.blit(self.engine.battle_panel, (560,290))
        
    def render_combat_data(self,lhs_unit,rhs_unit,target_surface = None):
        """
        # Function Name: render combat
        # Purpose: Renders the menu boards, the two battle panels, and data
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        """
        if not target_surface:
            target_surface = self.engine.surface
        
        
        self_prefix = self.engine.sfont.render(" - "+lhs_unit.spell_actions[lhs_unit.equipped].nameprefix+" - ",True,(0,0,0))
        self_suffix = self.engine.bfont.render(lhs_unit.spell_actions[lhs_unit.equipped].namesuffix,True,(0,0,0))
        
       
        # Checks if the target is carrying a spell
        if rhs_unit.spell_actions[rhs_unit.equipped]:
            target_prefix = self.engine.sfont.render(" - "+rhs_unit.spell_actions[rhs_unit.equipped].nameprefix+" - ",True,(0,0,0))
            target_suffix = self.engine.bfont.render(rhs_unit.spell_actions[rhs_unit.equipped].namesuffix,True,(0,0,0))
        else:
            target_prefix = self.engine.sfont.render(" - None - ",True,(0,0,0))
            target_suffix = self.engine.bfont.render("",True,(0,0,0))
        
        
        # Gets the half-widths of the various text objects
        hw_self_prefix = self_prefix.get_width()/2
        hw_self_suffix = self_suffix.get_width()/2
        hw_target_prefix = target_prefix.get_width()/2
        hw_target_suffix = target_suffix.get_width()/2
        
        target_surface.blit(self.engine.battle_board,(0,490))
        target_surface.blit(self.engine.battle_top,(0,0))

        # Plot spell names in top right
        target_surface.blit(self_prefix, (140-hw_self_prefix,25))
        target_surface.blit(self_suffix, (140-hw_self_suffix,55))
        target_surface.blit(target_prefix, (700-hw_target_prefix,25))
        target_surface.blit(target_suffix, (700-hw_target_suffix,55))
        
        # Plots stats in bottom right
        self.render_unit_stats(lhs_unit, False, target_surface)
        self.render_unit_stats(rhs_unit, True, target_surface)

    # Methods for when battle animations are turned off go here
    def draw_map_delay(self,lhs_unit,rhs_unit):
        """
        # Function Name: draw_map_delay
        # Purpose: Draws a half second delay on the battle map, rendering the data
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        """
        
        # Half a second pause
        for i in xrange(0,30):
            lhs_unit.map.render_background()
            lhs_unit.map.render_all_units()
            lhs_unit.map.render_cursor()
            self.render_combat_data(lhs_unit,rhs_unit)
            pygame.display.flip()
            self.engine.clock.tick(60)
    
    def draw_map_lhs_effect(self,lhs_unit,rhs_unit,text_lhs_effect):
        """
        # Function Name: draw_map_lhs_effect
        # Purpose: Draws the effect of the lhs unit's action on the battle map
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         lhs_effect_text - effect of lhs unit's action on rhs_unit
        """
        
        # Draw damage for 1 second
        effect_image_width = text_lhs_effect.get_width()/2
        for i in xrange(0,60):
            lhs_unit.map.render_background()
            lhs_unit.map.render_all_units()
            lhs_unit.map.render_cursor()
            self.render_combat_data(lhs_unit,rhs_unit)
            
            self.engine.surface.blit(text_lhs_effect,
                                         (rhs_unit.location_pixel.x+18-effect_image_width-35*rhs_unit.map.screen_shift.x,
                                          rhs_unit.location_pixel.y-25-35*rhs_unit.map.screen_shift.y))
            pygame.display.flip()
            self.engine.clock.tick(60)            
            
        
    def draw_map_rhs_effect(self,lhs_unit,rhs_unit,text_rhs_effect):
        """
        # Function Name: draw_map_rhs_effect
        # Purpose: Draws the effect of the rhs unit's action on the battle map
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         text_rhs_effect - effect of rhs unit's action on lhs_unit
        """
        
        # Draw damage for 1 second
        effect_image_width = text_rhs_effect.get_width()/2
        for i in xrange(0,60):
            lhs_unit.map.render_background()
            lhs_unit.map.render_all_units()
            lhs_unit.map.render_cursor()
            self.render_combat_data(lhs_unit,rhs_unit)
            
            self.engine.surface.blit(text_rhs_effect,
                                         (lhs_unit.location_pixel.x+18-effect_image_width-35*rhs_unit.map.screen_shift.x,
                                          lhs_unit.location_pixel.y-25-35*rhs_unit.map.screen_shift.y))
            pygame.display.flip()
            self.engine.clock.tick(60)            

    def draw_map_revive_effect(self,lhs_unit,rhs_unit,revive):
        """
        # Function Name: draw_map_lhs_effect
        # Purpose: Draws the effect of the lhs unit's action on the battle map
        # Inputs: lhs_unit,rhs_unit -  participants in the battle
        #         revive - unit to revive
        """
        
        # Draw trait name for 1 second
        if revive == 'lhs':
            trait_name = self.engine.render_outlined_text(lhs_unit.traits[0][0].name+"!",self.engine.cfont,(255,0,0),(255,255,255))
        else:
            trait_name = self.engine.render_outlined_text(rhs_unit.traits[0][0].name+"!",self.engine.cfont,(255,0,0),(255,255,255))
        
        for i in xrange(0,60):
            lhs_unit.map.render_background()
            lhs_unit.map.render_all_units()
            lhs_unit.map.render_cursor()
            self.render_combat_data(lhs_unit,rhs_unit)
            
 
            if revive == 'lhs':
                self.engine.surface.blit(trait_name,
                                         (lhs_unit.location_pixel.x+18-trait_name.get_width()/2-35*rhs_unit.map.screen_shift.x,
                                          lhs_unit.location_pixel.y-25-35*rhs_unit.map.screen_shift.y))
            
            else:
                self.engine.surface.blit(trait_name,
                                         (rhs_unit.location_pixel.x+18-trait_name.get_width()/2-35*rhs_unit.map.screen_shift.x,
                                          rhs_unit.location_pixel.y-25-35*rhs_unit.map.screen_shift.y))

            pygame.display.flip()
            self.engine.clock.tick(60)            
            
        