import pygame
from pygame.locals import *
from sys import exit
from lostsky.core.linalg import Vector2
from lostsky.core.engine import Engine
from lostsky.core.utils import split_lines
from random import randint
import string

class Map(object):

    def __init__(self,engine,terrainfile,pre_map_MAE=None,mid_mission_MAE_list=[],post_map_MAE=None):

        """
        # Function Name: __init__
        # Purpose: Intializes the map
        # Inputs: engine = associates the map with the game system's engine
        #         map_size = the size of the map in terms of tile numbers
        #         pre_map_MAE = pre-mission map action event
        #         post_map_MAE = post-mission map action event
        """
        
        self.all_units = {}
        self.all_units_total = {}       # Tracks all units in the map, including those that are dead
        self.all_units_by_name = {}
        self.reserve_units = {}
        self.unit_id = 1
        self.all_landmarks = {}
        self.all_ssps = {}
        self.cursor_pos = Vector2(0,0)
        self.screen_shift = Vector2(0,0)
        self.engine = engine
        self.terrainfilename = terrainfile
        self.team1=[]
        self.team2=[]
        
        # Sprite groups
        self.sg_unitcircles = pygame.sprite.RenderUpdates()
        self.sg_units = pygame.sprite.RenderUpdates()
        self.sg_status = pygame.sprite.RenderUpdates()
        self.sg_moving_unit = pygame.sprite.RenderUpdates()
        
        # Pre-map Map Action Event
        self.pre_map_MAE = pre_map_MAE
        if pre_map_MAE != None:
            self.pre_map_MAE.map = self
        # Post-map Map Action Event
        self.post_map_MAE = post_map_MAE
        if post_map_MAE != None:
            self.post_map_MAE.map = self
        
        # Mid-mission Map Action Events 
        self.all_mid_mission_MAEs = []
        for MAE in mid_mission_MAE_list:
            self.add_MAE(MAE)
            
        # Custom Variables checked/set by map actions
        self.cust_var = {}
        self.conv_images = {}
        
        # Highlighted tiles
        self.all_highlighted_tiles = []

        # New party members list
        self.new_party_members = []
        
        # Flag for conversation only maps. Instantly jumps to the end of a battle after the prestart section
        self.nobattle = False
        #Initializes terrain map
        self.generate_terrain()
        self.setup_background()

        # Turn counter
        self.turn_count = 0
        
        # Frame number
        self.framenum = 0
        
        # Background Overlays for time of day info
        self.bg_overlay = False
        
        # Enable Display of Cursor
        self.enable_cursor = True
        
        # Enable Display of stats/battle panel
        self.enable_stats_panel = True
        
        # Animation Counter - heartbeat for animated objects on the map
        self.animation_counter = 0
        
        # Items received (Items received during battle are added in at the end)
        self.items_received = []
        
    def add_landmark(self,landmark):
        """
        # Function Name: Add landmark
        # Purpose: Adds an landmark to the map
        # Inputs: Landmark - map action event to be added
        """
    
        self.all_landmarks[landmark.name] = landmark
        landmark.map = self
        # Adds in any prohibited tiles from landmarks
        if not landmark.can_pass:
            self.map_walk_prohibited.extend(landmark.occupied_tiles)    
         
         
    def add_MAE(self,MAE):
        """
        # Function Name: Add MAE
        # Purpose: Adds an MAE to the map
        # Inputs: MAE - map action event to be added
        """
    
        self.all_mid_mission_MAEs.append(MAE)
        MAE.map = self
        
    def check_all_MAE(self):
        """
        # Function Name: Check all MAEs
        # Purpose: Check all map action events and execute them if needed
        """
        
        for MAE in self.all_mid_mission_MAEs:
            if MAE.done == False:
                MAE.check_triggers()
            
        
    def store_unit(self,unit,team):
        """
        # Function Name: store unit
        # Purpose: Stores the unit to the map's master dictionary of units and
        #       assigns each an id key. Also runs the unit's get moves / update stats
        #       commands.
        # Inputs: unit = The unit to be assigned.
        #         team = the team to be assigned to
        """
    
        self.all_units_total[unit.name] = unit
        self.all_units_by_name[unit.name] = unit
        unit.id = self.unit_id
        unit.map = self
        unit.update_stats()
        unit.HP = unit.maxHP
        self.store_team(unit,team)
        self.sg_units.add(unit.sprite)
        self.sg_unitcircles.add(unit.circle)
        self.sg_status.add(unit.status_bubble)
        
    def store_team(self,unit,team):
        """
        # Function Name: store team
        # Purpose: Stores a unit to a specific team
        # Inputs: unit = The unit to be assigned.
        #         team = The team to which a unit is to be assigned (1,2)
        """
        if team == 1:
            self.team1.append(unit)
            unit.team = 1
            unit.team_color = (255,0,0)
            
        elif team == 2:
            self.team2.append(unit)
            unit.team = 2
            unit.team_color = (0,0,255)

    def get_all_moves(self):
        """
        # Function Name: get all moves
        # Purpose: iterates through all units and runs their get moves method
        """
        for unit in self.all_units_by_name.values():
            unit.get_moves_path()
            
    def generate_terrain(self):
        """
        # Function: generate terrain
        # Purpose: Loads the map's terrain text file and generates the terrain dictionary    #
        """
        
        
        def str_2_tuple(coord):
            """
            Converts a string of the form '(x, y, z)' to a tuple of integers
            """
            # Remove linebreaks
            coord = string.strip(coord)
            # Splits the string by commas
            split_string = string.split(coord[1:-1],', ')
            
            return tuple([int(entry) for entry in split_string])
        
        self.terrainmap={}
        self.layer_2_terrain={}
        self.upper_layer={}
        self.map_walk_prohibited = []
        self.map_fly_prohibited = []

        # Loads the terrain file
        tilefile = open(self.terrainfilename)

        # Reads the first two lines that correspond to the map's size in x and y
        
        self.map_x = int(tilefile.readline())
        self.map_y = int(tilefile.readline())
        
        # Reads each set of lines which correspond to a line of terrain tiles
        for y in xrange(0,self.map_y):
            line = tilefile.readline()
            for x in xrange(0,self.map_x):
                 
                 # Matches terrain line
                 terrain = self.engine.terrain_data_by_symbol[line[x]]
                 self.terrainmap[(x,y)] = [terrain,[],[]]
                 if not terrain.walk:
                     self.map_walk_prohibited.append((x,y))
                 if not terrain.fly:
                     self.map_fly_prohibited.append((x,y))

        # Defines the shift vectors in the format [Vector, Corresponding Binary Value]
        side_vectors = [[Vector2(-1,0),1],[Vector2(0,-1),2],[Vector2(1,0),4],[Vector2(0,1),8]]
        corner_vectors = [[Vector2(-1,-1),1],[Vector2(1,-1),2],[Vector2(1,1),4],[Vector2(-1,1),8]]
        


        #####################################################################
        # Generates the terrain transition values
        # Algorithm based on: David Michael's Terrain Transitions Article
        #   http://www.gamedev.net/reference/articles/article934.asp
        #####################################################################

        for x in xrange(0,self.map_x):
            for y in xrange(0,self.map_y):

                # Note: Deep water always has lower priority than everything else so we don't iterate over that
                for layer in xrange(1,len(self.engine.terrain_types)):
                    # Transition variables for side and corner transitions - 4 bit binary number
                    s = 0
                    c = 0

                    # Checks if the tile type has a higher placement priority than the current tile
                    if layer > self.terrainmap[(x,y)][0].ident:
                        # check sides
                        for vector in side_vectors:
                            new_pos = tuple(Vector2(x,y)+vector[0])
                            # Only performs operations if the new position is actually on the map
                            if self.terrainmap.has_key(new_pos):
                                # if the position is of the same type as the one being processed,
                                # bitwise OR it with the associated binary value
                                if self.terrainmap[new_pos][0].ident == layer:
                                    s |= vector[1]

                        # check corners
                        for vector in corner_vectors:
                            new_pos = tuple(Vector2(x,y)+vector[0])
                            # Only performs operations if the new position is actually on the map
                            if self.terrainmap.has_key(new_pos):
                                # if the position is of the same type as the one being processed,
                                # bitwise OR it with the associated binary value
                                if self.terrainmap[new_pos][0].ident == layer:
                                    c |= vector[1]

                    self.terrainmap[(x,y)][1].append(s)
                    self.terrainmap[(x,y)][2].append(c)

                            # Loads cliff layer data
        for y in xrange(0,self.map_y):
            line = tilefile.readline()
            # Tile within horizontal slice
            for x, tile in enumerate(string.split(line,'|')):
                # x indicates an empty tile
                if tile and tile != 'x' and tile != "x\n" and tile != 'x\r\n':
                    layer_2_tile_id = str_2_tuple(tile)

                    # Adds the cliff tile to the map
                    self.layer_2_terrain[(x,y)] = layer_2_tile_id


                    # Sets the terrain for the tile according to the overide case
                    if self.engine.layer_2_terrain_data[layer_2_tile_id]:
                        # Override case: Replace with a non-walkable cliff tile

                        overwrite_symbol = self.engine.layer_2_terrain_data[layer_2_tile_id]

                        self.terrainmap[(x,y)].append(self.terrainmap[(x,y)][0])

                        self.terrainmap[(x,y)][0] = self.engine.terrain_data_by_symbol[overwrite_symbol]


                        if self.terrainmap[(x,y)][0].walk and (x,y) in self.map_walk_prohibited:
                            self.map_walk_prohibited.remove((x,y))
                        if self.terrainmap[(x,y)][0].fly and (x,y) in self.map_fly_prohibited:
                            self.map_fly_prohibited.remove((x,y))


                        if not self.terrainmap[(x,y)][0].walk:
                            self.map_walk_prohibited.append((x,y))
                        if not self.terrainmap[(x,y)][0].fly:
                            self.map_fly_prohibited.append((x,y))

        tilefile.close()

                   
    def add_ssp(self,ssp):
        """
        # Function Name: add_ssp
        # Purpose: Adds a spirit source point to the map
        # Inputs: ssp - spirit source point to be added
        """
        self.all_ssps[tuple(ssp.location)] = ssp
        # Adds the SSP to the landmark set
        self.add_landmark(ssp)
    
    def update_ssps(self):
        """
        # Function Name: add_ssp
        # Purpose: Updates status of all Spirit Source Points
        """
        for ssp in self.all_ssps.values():
            # Checks if there's a unit at each ssp location
            unit_name = self.check_occupancy(ssp.location)
            if unit_name:
                ssp.capture_state = self.all_units_by_name[unit_name].team
            
    def check_occupancy(self,coords):
        """
        # Function Name: check_occupancy
        # Purpose: Checks if a unit occupies a certain location
        """
        for unit in self.all_units_by_name.values():
            if tuple(coords) == tuple(unit.location_tile):
                return unit.name
        
        # If no unit has been detected, returns False
        return False

    def run_status_effects(self,unit):
        """
        # Function Name: run_status_effects
        # Purpose: Checks and executs effects of any statuses a unit might have
        """
        for status_effect in unit.status.keys():
            print "Executing %s" % status_effect
            
            hp_before = unit.HP
            
            # Checks if a unit has recovered
            recovery = self.engine.status_effects_catalog[status_effect].check_recovery(unit)
            
            if recovery:
                unit.remove_status(status_effect)
                self.display_alert("Status Effect Ended!", "%s's %s status has worn off."%(unit.name,status_effect))
            else:
                
                hp_before = unit.HP
                sc_before = unit.spirit

                # Executes effects of the S.E.
                effect = self.engine.status_effects_catalog[status_effect].execute_effect(unit)
            
                # if unit's HP is different show the change.
                if self.engine.status_effects_catalog[status_effect].HP_change:
                    # Draw HP difference as green if it is a recovery, red if it is damage
                    if effect > 0:
                        text_effect = self.engine.render_outlined_text(str(effect),self.engine.cfont,(100,255,100),(0,0,0))
                    else:
                        text_effect = self.engine.render_outlined_text(str(-1*effect),self.engine.cfont,(255,0,0),(255,255,255))
                elif self.engine.status_effects_catalog[status_effect].SC_change:


                    # Draw SC difference as lightblue if it is recovery
                    if effect > 0:
                        text_effect = self.engine.render_outlined_text(str(effect),self.engine.cfont,(128,255,255),(0,0,0))
                    else:
                        text_effect = self.engine.render_outlined_text(str(-1*effect),self.engine.cfont,(255,0,255),(255,255,255))



                if self.engine.status_effects_catalog[status_effect].show_change and effect != 0:
                    self.center_on(unit)
                    
                    text_status_effect = self.engine.bfont.render(status_effect,True,(0,0,0))
                    
                    self.render_background()
                    self.render_all_units()
                    self.render_cursor()
                    self.engine.surface.blit(self.engine.map_spell_board,(175,0))
                    self.engine.surface.blit(text_status_effect,(420-text_status_effect.get_width()/2,25))
                    self.engine.surface.blit(self.engine.menu_board,(0,490))
                    unit.plot_stats()
                    self.engine.surface.blit(text_effect,((unit.location_pixel.x+18-text_effect.get_width()/2,unit.location_pixel.y-25)-self.screen_shift*self.engine.tilesize))
                    self.engine.pause(1)
                    
                if self.engine.status_effects_catalog[status_effect].HP_change and (hp_before != unit.HP):
                    unit.render_hp_change(hp_before,unit.HP)

                if self.engine.status_effects_catalog[status_effect].SC_change and (sc_before != unit.spirit):
                    unit.render_sc_change(sc_before,unit.spirit)
            
        if unit.alive == False:
            if unit.deathquote:
                rand_num = randint(0,len(unit.deathquote)-1)
                self.say(unit.deathquote[rand_num]['line'],unit.name,unit.deathquote[rand_num]['portrait'])
            
            self.kill(unit)


    def run_ssp_sc_regen(self, team):
        """
        run_ssp_sc_regen

        Checks if any of the SSPs are occupied by any units. If so and if their SC is under a certain value,
        regenerate a constant amount of SC as long as they are on the SSP.
        """
        sc_regen_value = 50
        sc_max = 500
        if team == 1:
            current_team = self.team1
        else:
            current_team = self.team2

        text_status_effect = self.engine.bfont.render("Spirit Source Bonus!",True,(0,0,0))

        for ssp_coord in self.all_ssps.keys():

            unit_name = self.check_occupancy(ssp_coord)

            if (unit_name and self.all_units_by_name[unit_name] in current_team
                and self.all_units_by_name[unit_name].spirit < sc_max):
                unit = self.all_units_by_name[unit_name]



                sc_before = unit.spirit
                unit.spirit += sc_regen_value
                if unit.spirit > sc_max:
                    unit.spirit = sc_max

                unit.check_spirit_range()

                delta_sc = unit.spirit - sc_before

                text_effect = self.engine.render_outlined_text(str(delta_sc),self.engine.cfont,(128,255,255),(0,0,0))

                # Renders SC change
                self.center_on(unit)
                self.render_background()
                self.render_all_units()
                self.render_cursor()
                self.engine.surface.blit(self.engine.map_spell_board,(175,0))
                self.engine.surface.blit(text_status_effect,(420-text_status_effect.get_width()/2,25))
                self.engine.surface.blit(self.engine.menu_board,(0,490))
                unit.plot_stats()
                self.engine.surface.blit(text_effect,((unit.location_pixel.x+18-text_effect.get_width()/2,unit.location_pixel.y-25)-self.screen_shift*self.engine.tilesize))
                self.engine.pause(1)

                unit.render_sc_change(sc_before, unit.spirit)


    #############################
    # Graphical Methods
    #############################

    def center_on(self, unit,text_board=False):
        """
        # Function Name: center_on
        # Purpose: Centers the current viewfield on a unit
        # Inputs: Unit - The unit to be centered upon
        #              Text_board - True if using the text display board, False if using the standard map
        """
        self.center_cursor(unit.location_tile,text_board)
        
    def center_cursor(self,destination,text_board=False):
        """
        # Function Name: center_cursor
        # Purpose: Centers the current viewfield on a set of coords
        # Inputs: Unit - The unit to be centered upon
        #              Text_board - True if using the text display board, False if using the standard map
        # Function modification by Beanbeanman29 - 8/20
        """
    
        self.cursor_pos = Vector2(tuple(destination))
            
        # Includes a correction so that the shift doesn't go too far
        final_x = min(max(0,int(destination.x-12)),self.map_x-24)
        final_y = min(max(0,int(destination.y-7)),self.map_y-14)
        # Relative displacement vector between initial and final positions
        delta_pos = Vector2(final_x,final_y) - self.screen_shift

        # Only draw animation if the new position of the screen is different than the old position
        if delta_pos.get_magnitude() > 0:
            
            velocity = 0.4*delta_pos.get_normalized()
            
            max_frames = int(delta_pos.get_magnitude()/velocity.get_magnitude())
            for frame in xrange(0,max_frames):
                self.screen_shift += velocity
                    
                self.render_background()
                self.render_all_units()
                self.render_cursor()
                if text_board == False:
                    self.engine.surface.blit(self.engine.menu_board,(0,490))
                else: 
                    self.engine.surface.blit(self.engine.text_board,(0,490))
                
                pygame.display.flip()
                self.engine.clock.tick(60)
            
            # Locks in final screen position to final quantities since the values may not be whole numbers at the end
            # of the animation
            self.screen_shift.x = final_x
            self.screen_shift.y = final_y
            
    def kill(self,unit):
        """
        # Function Name: Kill
        # Purpose: Removes a unit from the map
        """
    
        # Removes unit
        del self.all_units_by_name[unit.name]
        if unit.team == 1:
            self.team1.remove(unit)
        else:
            self.team2.remove(unit)
            
        # Remove sprite from unit sprite group
        self.sg_units.remove(unit.sprite)
        self.sg_unitcircles.remove(unit.circle)
        self.sg_status.remove(unit.status_bubble)
    
    def setup_background(self):
        """
        # Function Name: setup background
        # Purpose: renders a background with all the tiles.
        """
        self.background = pygame.Surface((self.map_x*35,self.map_y*35))
        
        def render_tile(x,y):
            """
            # Function Name: render_tile
            # Purpose: renders a tile at location (x,y)
            """
            # Draws base background tile
            # If the terrain has been overwritten, draw the original background tile first
            if len(self.terrainmap[(x,y)]) == 3:
                self.background.blit(self.engine.terrain_img,(x*35,y*35),(0,self.terrainmap[(x,y)][0].ident*70,35,35))
            elif len(self.terrainmap[(x,y)]) == 4:
                self.background.blit(self.engine.terrain_img,(x*35,y*35),(0,self.terrainmap[(x,y)][3].ident*70,35,35))


            for terrain_type in xrange(1,len(self.engine.terrain_types)):
                # Draws Corner Transitions
                if self.terrainmap[(x,y)][2][terrain_type-1]!=0:                    
                    self.background.blit(self.engine.terrain_img, (x*35,y*35),(int(self.terrainmap[(x,y)][2][terrain_type-1]*35),terrain_type*70+35,35,35))
                # Draws Side transitions
                if self.terrainmap[(x,y)][1][terrain_type-1]!=0:                    
                    self.background.blit(self.engine.terrain_img, (x*35,y*35),(int(self.terrainmap[(x,y)][1][terrain_type-1]*35),terrain_type*70,35,35))
                # Draw layer 2 tiles
                if (x,y) in self.layer_2_terrain.keys():
                    self.background.blit(self.engine.layer_2_img,(x*35,y*35),(self.layer_2_terrain[(x,y)][0]*35,self.layer_2_terrain[(x,y)][1]*35,35,35))
                
        [render_tile(x,y) for x in xrange(0,self.map_x) for y in xrange(0,self.map_y)]
    
    def render_menu_panel(self):
        """
        # Function Name: render menu panel
        # Purpose: draws the menu panel at the bottom of the screen
        """
        if self.enable_stats_panel:
            self.engine.surface.blit(self.engine.menu_board,(0,490))
        else:    
            self.engine.surface.blit(self.engine.text_board,(0,490))
    
    def render_background(self):
        """
        # Function Name: render background
        # Purpose: draws background onto the screen
        #          surface = target surface to render to (Default is game screen surface)
        """
        xmin = self.screen_shift.x
        xmax = min(self.screen_shift.x+24,self.map_x)
        ymin = self.screen_shift.y
        ymax = min(self.screen_shift.y+14,self.map_y)
        
        self.engine.surface.blit(self.background,(0,0),(int(xmin*35),int(ymin*35),840,490))
                
        self.render_landmarks([xmin,xmax,ymin,ymax])
        
        if self.engine.options.grid == True:
            for x in xrange(0,24):
                pygame.draw.line(self.engine.surface,(200,200,200),(x*35,0),(x*35,490))
            for y in xrange(0,14):
                pygame.draw.line(self.engine.surface,(200,200,200),(0,y*35),(840,y*35))

        # Background Color Tint Overlays 
        if self.bg_overlay == "Night":
            self.engine.surface.blit(self.engine.night_overlay,(0,0))
        elif self.bg_overlay == "Sunset":
            self.engine.surface.blit(self.engine.sunset_overlay,(0,0))
                
        # Updates animation counter
        self.animation_counter += 1
        if self.animation_counter == 121:
            self.animation_counter = 0
                
                
    def render_landmarks(self,bounds):
        """
        # Function Name: Render Landmark
        # Purpose: Draws all the landmarks and highlights on the screen
        #          Bounds = the current xmin,xmax,ymin,ymax boundaries of the world map
        """
        xmin, xmax, ymin, ymax = bounds
        
        # Increases bounds by 35 pixels in all directions to allow for scrolling behavior
        xmin -= 35
        xmax += 35
        ymin -= 35
        ymax += 35
        
        for landmark in self.all_landmarks.values():
            
            # checks if the landmark is within the bounds of the currently viewed map
            if landmark.location.x >= xmin and landmark.location.x < xmax and landmark.location.y >= ymin and landmark.location.y < ymax:
                landmark.render()
                
        for highlight_tile in self.all_highlighted_tiles:
            if highlight_tile.x >= xmin and highlight_tile.x < xmax and highlight_tile.y >= ymin and highlight_tile.y < ymax:
                x,y = highlight_tile*35
                self.engine.surface.blit(self.engine.highlight_tile, ((x,y)-self.screen_shift*35))
        
        
    def render_deploy_locations(self):
        """
        # Function Name: Render Deploy Locations
        # Purpose: Draws the valid deployment locations onto the map
        """
        
        xmin = int(self.screen_shift.x)
        xmax = min(int(self.screen_shift.x)+24,self.map_x)
        ymin = int(self.screen_shift.y)
        ymax = min(int(self.screen_shift.y)+14,self.map_y)
        bounds = [xmin,xmax,ymin,ymax]
        
        [self.engine.surface.blit(self.engine.deploy_tile,((location[0]-self.screen_shift.x)*35,(location[1]-self.screen_shift.y)*35)) for location in self.deploy_locations if location[0] >= bounds[0] and location[0] < bounds[1] and location[1] >= bounds[2] and location[1] < bounds[3]]
         
         
    def render_all_units(self, bubbles = True):
        """
        # Function Name: render all units
        # Purpose: iterates through all units and runs their render method
        # Inputs: bubbles - T/F enable rendering of status bubbles
        """
        self.sg_unitcircles.update()
        self.sg_units.update()
        self.sg_status.update()
        
        rects = self.sg_unitcircles.draw(self.engine.surface)
        rects += self.sg_units.draw(self.engine.surface)
        if bubbles:
            rects += self.sg_status.draw(self.engine.surface)
        return rects

    def render_current_terrain_data(self):
        """
        # Function Name: render current terrain data
        # Purpose: draws information about the current tile under the unit's cursor
        """
        
        # Draws current terrain tile info
        text_current_terrain_type = self.engine.bfont.render(self.terrainmap[tuple(self.cursor_pos)][0].name,True, (0,0,0))
        text_current_terrain_tdef = self.engine.sfont.render("TDEF: %2.1f"%self.terrainmap[tuple(self.cursor_pos)][0].TDEF,True, (0,0,0))
        text_current_terrain_agl = self.engine.sfont.render("AGL: "+str(self.terrainmap[tuple(self.cursor_pos)][0].aglmod),True, (0,0,0))
        text_current_terrain_acc = self.engine.sfont.render("ACC: "+str(self.terrainmap[tuple(self.cursor_pos)][0].accmod),True, (0,0,0))
        
        self.engine.surface.blit(text_current_terrain_type,(450,505))
        self.engine.surface.blit(self.engine.terrain_icon,(450,523),(self.terrainmap[tuple(self.cursor_pos)][0].icon*50,0,50,50))
        self.engine.surface.blit(text_current_terrain_tdef,(450,575))
        self.engine.surface.blit(text_current_terrain_agl,(450,587))
        self.engine.surface.blit(text_current_terrain_acc,(450,599))

    def render_turn_number(self):
        """
        # Function Name: render turn number
        # Purpose: draws the current turn in the infobox
        """
        text_turn_num = self.engine.bfont.render("Turn Number: %s"%(str(self.turn_count)),True,(0,0,0))
        self.engine.surface.blit(text_turn_num,(600,505))

    def render_cursor(self):
        """
        # Function Name: render cursor
        # Purpose: plots the cursor update
        """
        if self.enable_cursor:
            self.engine.surface.blit(self.engine.cursor_img,
                                 (self.cursor_pos*self.engine.tilesize-self.screen_shift*self.engine.tilesize))

    def say(self,string,speaker="",portrait=None):
        """
        # Function name: say
        # Purpose: Displays a line of text and awaits for the player to press Z to continue
        # Inputs: string = The string of text to be displayed
        #         speaker = The name of the person saying it
        """
        # calls the line splitter function to generate the events
        if portrait:
            lines = split_lines(string,75)
        else:
            lines = split_lines(string)
        
        text_lines = []
        if speaker:
            text_speaker = self.engine.bfont.render("- "+speaker+" -",True,(0,0,0))
        else:
            text_speaker = self.engine.bfont.render(speaker,True,(0,0,0))
        
        # generates the text for each line
        for line in lines:
            text_lines.append(self.engine.mfont.render(line,True,(0,0,0)))
        menu_flag = True
        
        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN and (event.key == K_z or event.key == K_RETURN):
                    menu_flag = False
                        
            # Hold down C to skip through the game.
            keys = pygame.key.get_pressed()
            if keys[K_c]:
                menu_flag = False
                
            self.render_background()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.text_board,(0,490))
            for conversation_image in self.conv_images.values():
                self.engine.surface.blit(conversation_image[0],conversation_image[1])

            if portrait:
                self.engine.surface.blit(self.engine.portrait_catalog[portrait],(10,500))
                self.engine.surface.blit(text_speaker,(140,505))
                [self.engine.surface.blit(text_line,(140,530+20*index)) for index,text_line in enumerate(text_lines)]
            else:
                self.engine.surface.blit(text_speaker,(30,505))
                [self.engine.surface.blit(text_line,(30,530+20*index)) for index,text_line in enumerate(text_lines)]
            pygame.display.flip()
            self.engine.clock.tick(60)
    
    def choice(self,query,responses):
        """
        # Function name: Choice
        # Purpose: Displays a question for the player to answer
        # Inputs: string = The string of text to be displayed (e.g. do you like pie?)
        #         responses = A list of (preferably less than 3) options given to the player
        """
        text_query = self.engine.bfont.render(query,True,(0,0,0))
        text_responses = []
        for response in responses:
            text_responses.append(self.engine.sfont.render(response,True,(0,0,0)))
            
        menu_flag = True
        menu_pos = 0
        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_UP or event.key == K_LEFT:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(text_responses)-1
                    if event.key == K_DOWN or event.key == K_RIGHT:
                        if menu_pos < len(text_responses)-1:
                            menu_pos += 1
                        elif menu_pos == len(text_responses)-1:
                            menu_pos = 0
                    
                    if (event.key == K_z or event.key == K_RETURN):
                        menu_flag = False
                        return responses[menu_pos]
                        
            if menu_flag:
                self.render_background()
                self.render_all_units()
                self.render_cursor()
                self.engine.surface.blit(self.engine.text_board,(0,490))
                self.engine.surface.blit(text_query,(30,520))
                for i in xrange(0,len(text_responses)):
                    self.engine.surface.blit(text_responses[i],(50,540+20*i))
                pygame.draw.circle(self.engine.surface, (0,0,0), (40,548+20*menu_pos),5)
                pygame.display.flip()
                self.engine.clock.tick(60)
    
    
    #############################
    # Interaction Methods
    #############################    

    
    def user_input(self):
        """
        # Function Name: user input
        # Purpose: top level interaction. Checks if there is any input by the user
        #   and allows the user to move around the cursor with the arrow keys and select a unit
        #   if there is one under the mouse
        """
        arrowkeys = False
        for event in pygame.event.get():
            if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                exit()
            if event.type == KEYDOWN:
                if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                    self.cursor_arrows(event)
                    arrowkeys = True
                    self.framenum = 0
                if event.key == K_z or event.key == K_RETURN:
		    
                    selected = self.cursor_key_search()
                    
                    # Sends the player to the unit's top level menu
                    if selected is not False:
                        if self.all_units_by_name[selected].team == 1:    
                            self.all_units_by_name[selected].menu_loop()
                            
                        else:
                            self.all_units_by_name[selected].stats_loop()
                    
                    # Sends the player to the map's menu
                    elif selected == False:
                        self.map_menu()
                        
                        
                if event.key == K_a:
                    
                    # Checks to make sure the counter is within the bounds of the team 1 list
                    if self.ally_center > len(self.team1) - 1:
                        self.ally_center = len(self.team1) - 1
                        
                    if self.ally_center > 0:
                        self.ally_center -= 1
                    else:
                        self.ally_center = len(self.team1) - 1
                    self.center_on(self.team1[self.ally_center])
                    
                if event.key == K_s:
                    
                    # Checks to make sure the counter is within the bounds of the team 1 list
                    if self.ally_center > len(self.team1) - 1:
                        self.ally_center = len(self.team1) - 1
                    
                    if self.ally_center < len(self.team1) - 1:
                        self.ally_center += 1
                    else:
                        self.ally_center = 0
                    self.center_on(self.team1[self.ally_center])
                
                if event.key == K_q and self.team2:
                    # Checks to make sure the counter is within the bounds of the team 2 list
                    if self.enemy_center > len(self.team2) - 1:
                        self.enemy_center = len(self.team2) - 1
                    
                    if self.enemy_center > 0:
                        self.enemy_center -= 1
                    else:
                        self.enemy_center = len(self.team2) - 1
                    self.center_on(self.team2[self.enemy_center])
                if event.key == K_w and self.team2:
                    # Checks to make sure the counter is within the bounds of the team 2 list
                    if self.enemy_center > len(self.team2) - 1:
                        self.enemy_center = len(self.team2) - 1
                    
                    if self.enemy_center < len(self.team2) - 1:
                        self.enemy_center += 1
                    else:
                        self.enemy_center = 0
                    self.center_on(self.team2[self.enemy_center])
    
        # Use C to toggle displaying movement range
        keys = pygame.key.get_pressed()
        if keys[K_c]:
            self.move_plot = True
        else:
            self.move_plot = False
    
        # if there is not a tap detected, check if the key is being held down
        if arrowkeys == False and self.framenum == 9:
            self.cursor_arrows_hold()

    
    def cursor_key_search(self):
        """
        # Function Name: cursor key search
        # Purpose: checks if there is a unit under the cursor
        # Output: Unit name if there is a unit under the cursor, False otherwise
        """
        # Checks if there is a unit under the cursor
        return self.check_occupancy(self.cursor_pos)
        
    def cursor_arrows(self,event):
        """
        # Function Name: cursor_arrows
        # Purpose: allows a user to move the arrows around by tapping down a key
        """
        if event.key == K_LEFT:
            # Checks if the x position of the cursor is at the edge of the map.
            if self.cursor_pos.x > 0:
                self.cursor_pos.x += -1
            # If it is at the edge of the screen, but not the map, do a screen shift
            if self.cursor_pos.x > 0 and self.cursor_pos.x - self.screen_shift.x == 0:
                self.screen_shift.x += -1
        
        elif event.key == K_RIGHT:
            if self.cursor_pos.x < self.map_x-1:
                self.cursor_pos.x += +1
            if self.cursor_pos.x < self.map_x-1 and self.cursor_pos.x - self.screen_shift.x == self.engine.size_x-1:
                self.screen_shift.x += +1
                
        if event.key == K_UP:
            if self.cursor_pos.y > 0:
                self.cursor_pos.y += -1
            if self.cursor_pos.y > 0 and self.cursor_pos.y - self.screen_shift.y == 0:
                self.screen_shift.y += -1
                
        elif event.key == K_DOWN:
            if self.cursor_pos.y < self.map_y-1:
                self.cursor_pos.y += +1
            if self.cursor_pos.y < self.map_y-1 and self.cursor_pos.y - self.screen_shift.y == self.engine.size_y-5:
                self.screen_shift.y += +1


    def cursor_arrows_hold(self):
        """
        # Function Name: cursor_arrows_hold
        # Purpose: allows a user to move the arrows around by holding the key
        """
        key = pygame.key.get_pressed()

        if key[K_LEFT]:
            # Checks if the x position of the cursor is at the edge of the map.
            if self.cursor_pos.x > 0:
                self.cursor_pos.x += -1
            # If it is at the edge of the screen, but not the map, do a screen shift
            if self.cursor_pos.x > 0 and self.cursor_pos.x - self.screen_shift.x == 0:
                self.screen_shift.x += -1
            
            # Reset Frame counter to 6
            self.framenum = 6
            
        elif key[K_RIGHT]:
            if self.cursor_pos.x < self.map_x-1:
                self.cursor_pos.x += +1
            if self.cursor_pos.x < self.map_x-1 and self.cursor_pos.x - self.screen_shift.x == self.engine.size_x-1:
                self.screen_shift.x += +1
            
            # Reset Frame counter to 6
            self.framenum = 6
                
        if key[K_UP]:
            if self.cursor_pos.y > 0:
                self.cursor_pos.y += -1
            if self.cursor_pos.y > 0 and self.cursor_pos.y - self.screen_shift.y == 0:
                self.screen_shift.y += -1
                
            # Reset Frame counter to 6
            self.framenum = 6
            
        elif key[K_DOWN]:
            if self.cursor_pos.y < self.map_y-1:
                self.cursor_pos.y += +1
            if self.cursor_pos.y < self.map_y-1 and self.cursor_pos.y - self.screen_shift.y == self.engine.size_y-5:
                self.screen_shift.y += +1
    
            # Reset Frame counter to 6
            self.framenum = 6
            
        pygame.event.clear()


    def map_menu(self):

        """
        # Function Name: map menu
        # Purpose: Map level menu (End Turn, Unit Stats, Options)
        """
        menu_flag = True
        menu_pos = 0
        text_stats = self.engine.bfont.render("UNIT STATS",True, (0,0,0))
        text_options = self.engine.bfont.render("OPTIONS",True,(0,0,0))
        text_mission_end = self.engine.bfont.render("QUIT MISSION",True, (0,0,0))
        text_turn_end = self.engine.bfont.render("END TURN",True, (0,0,0))        
        text_cancel = self.engine.bfont.render("CANCEL",True, (0,0,0))

        while menu_flag:

            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # if either the Up or Right keys are pressed, and the menu position
                    # is not at the top, move the menu position up one.
                    if ( event.key == K_UP or event.key == K_LEFT )and menu_pos > 0:
                        menu_pos -= 1
                    elif ( event.key == K_UP or event.key == K_LEFT )and menu_pos == 0:
                        menu_pos = 4
                    if ( event.key == K_DOWN or event.key == K_RIGHT ) and menu_pos < 4:
                        menu_pos += 1
                    elif ( event.key == K_DOWN or event.key == K_RIGHT ) and menu_pos == 4:
                        menu_pos = 0
                    if event.key == K_z or event.key == K_RETURN:
                        if menu_pos == 0:
                            end_menu = self.unit_stats_loop()
                            if end_menu == True:
                                menu_flag = False
                        elif menu_pos == 1:
                            self.engine.options_menu()
                        elif menu_pos == 2:
                            quit = self.quit_mission_menu()
                            if quit:
                                menu_flag = False
                                # Ends the mission
                                self.victory_condition = True
                                self.battle_end = 'team2victory'
                            
                        elif menu_pos == 3:
                            self.allmoved1 = True
                            menu_flag = False
                        elif menu_pos == 4:
                            menu_flag = False
                    if ( event.key == K_x ):
                        menu_flag = False
            
            self.render_background()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))

            if menu_flag:
                pygame.draw.circle(self.engine.surface, (0,0,0), (440,513+20*menu_pos),5)

                self.engine.surface.blit(text_stats, (450,505))
                self.engine.surface.blit(text_options, (450,525))
                self.engine.surface.blit(text_mission_end, (450,545))
                self.engine.surface.blit(text_turn_end, (450,565))                
                self.engine.surface.blit(text_cancel, (450,585))

            pygame.display.flip()
            self.engine.clock.tick(60)

    def quit_mission_menu(self):
        """
        # Function Name: quit_mission_menu
        # Purpose: Allows player to select if they want to quit the mission (counts as a failure)
        # output: True if player has selected to quit mission / False if they have not
        """
        menu_flag = True
        menu_pos = False
        text_quit_query = self.engine.bfont.render("Are you sure you want to quit?",True, (0,0,0))
        text_yes = self.engine.bfont.render("Yes",True,(0,0,0))
        text_no = self.engine.bfont.render("No",True, (0,0,0))

        while menu_flag:

            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # if either the Up or Right keys are pressed, switch the position
                    if event.key in (K_UP,K_DOWN,K_RIGHT,K_LEFT):
                        menu_pos = not(menu_pos)
                    if event.key == K_z or event.key == K_RETURN:
                        return menu_pos
                    # X automatically cancels
                    if ( event.key == K_x ):
                        return False
            
            self.render_background()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))

            if menu_flag:
                pygame.draw.circle(self.engine.surface, (0,0,0), (440,533+20*menu_pos),5)

                self.engine.surface.blit(text_quit_query, (440,505))
                self.engine.surface.blit(text_no, (450,525))
                self.engine.surface.blit(text_yes, (450,545))
                

            pygame.display.flip()
            self.engine.clock.tick(60)
        
        

    def prestart(self):
        """
        # Function Name: Prestart
        # Purpose: Runs any prestart events if applicable
        """
    
        self.currentplayer = 1
        
        for unit in self.team1:
            unit.moved = False
            unit.turnend = False

        self.allmoved1 = False
        
        # Resets list of items received
        self.items_received = []
        
        if self.pre_map_MAE:
            self.pre_map_MAE.execute()
        self.check_all_MAE()
            
    def display_alert(self,title,line):
        """
        # Function Name: Displays alert
        # Purpose: Displays an alert message
        # Inputs:   Title - Title of alert
        #           Line - Alert text
        """    
        menu_flag = True
        
        alert_image = self.engine.small_text_board.copy()
        text_title = self.engine.bfont.render(title,True,(0,0,0))
        text_instructions = self.engine.sfont.render("Press Z to continue.",True,(0,0,0))
        lines = split_lines(line,50)
        
        # Draws the lines on the alert image
        alert_image.blit(text_title,(20,20))
        alert_image.blit(text_instructions,(215,110))
        count = 0
        for line in lines:
            alert_image.blit(self.engine.sfont.render(line,True,(0,0,0)),(30,40+20*count))
            count += 1
        
        while menu_flag:
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN and (event.key == K_z or event.key == K_RETURN):
                    menu_flag = False
                            
            self.render_background()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))
            self.engine.surface.blit(alert_image,(245,140))
            pygame.display.flip()
            self.engine.clock.tick(60)
    
    def deploy_screen_main(self):

        """
        # Function Name: deploy screen main
        # Purpose: Top level deploy screen
        """

        menu_flag = True
        menu_pos = 0
        text_header = self.engine.bfont.render("Set up characters!",True,(0,0,0))
        text_place = self.engine.sfont.render("Place Character",True,(0,0,0))
        text_party = self.engine.sfont.render("Party",True,(0,0,0))
        text_view = self.engine.sfont.render("View Map",True,(0,0,0))
        text_done = self.engine.sfont.render("Done",True,(0,0,0))
        text_done_gray = self.engine.sfont.render("Done",True,(100,100,100))
        
        self.engine.fade_to('black')
        # Clears out the units that are not auto-set to deploy to a certain spot
        # Deletes any unit not in the current map
        for unit in list(self.team1):
            if unit.name not in self.preset_units.keys():
                self.kill(unit)
                del(self.all_units_total[unit.name])
                

        # Places pre_deployed units
        for unit_name in self.preset_units.keys():
            # Checks if the unit is already on the map
            if unit_name not in self.all_units_by_name.keys():
                self.store_unit(self.engine.player_units_by_name[unit_name],1)
            self.all_units_by_name[unit_name].update_location(*self.preset_units[unit_name])
                
        # Tracking for Deployment Numbers
        self.num_deployed = len(self.team1)
        # Flag to see if this is the first time that the unit cap has been hit
        self.all_deployed = False
        
        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))

        while menu_flag:
            
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_UP or event.key == K_LEFT:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 3
                    if event.key == K_DOWN or event.key == K_RIGHT:
                        if menu_pos < 3:
                            menu_pos += 1
                        elif menu_pos == 3:
                            menu_pos = 0 
                    if event.key == K_z or event.key == K_RETURN:
                        if menu_pos == 0:
                            self.deploy_screen_select()
                            # Updates number of deployed units
                            text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))

                        elif menu_pos == 1:
                            self.engine.worldmap.party_menu()
                        elif menu_pos == 2:
                            self.deploy_screen_view()
                        # Only allows exiting of the menu if there's at least one unit on the map
                        elif menu_pos == 3 and self.team1:
                            menu_flag = False
                        
            if menu_flag:
                self.render_background()
                self.render_deploy_locations()
                self.render_all_units()
                self.render_cursor()
                self.engine.surface.blit(self.engine.text_board,(0,490))
                self.engine.surface.blit(text_header,(30,505))
                self.engine.surface.blit(text_num_deployed,(30,525))
                self.engine.surface.blit(text_place,(50,545))
                self.engine.surface.blit(text_party,(50,565))
                self.engine.surface.blit(text_view,(50,585))
                if self.team1:
                    self.engine.surface.blit(text_done,(50,605))
                else:
                    self.engine.surface.blit(text_done_gray,(50,605))
                pygame.draw.circle(self.engine.surface, (0,0,0), (40,552+20*menu_pos),5)
                
                pygame.display.flip()
                self.engine.clock.tick(60)
                
                
    def deploy_screen_select(self):
        """
        # Function Name: deploy screen select
        # Purpose: Player selects a unit to deploy
        """
        
        menu_flag = True
        menu_pos = 0
        
        text_select = self.engine.bfont.render("Select Unit:",True,(0,0,0))
        
        # Updates number of deployed units
        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))

        directions_header = self.engine.bfont.render("Directions",True,(0,0,0))
        directions1 = self.engine.sfont.render("Z - Select",True,(0,0,0))
        directions2 = self.engine.sfont.render("X - Cancel",True,(0,0,0))
        directions3 = self.engine.sfont.render("C - Remove",True,(0,0,0))
        
        while menu_flag:
            
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_UP or event.key == K_LEFT:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(self.engine.player_units)-1
                    if event.key == K_DOWN or event.key == K_RIGHT:
                        if menu_pos < len(self.engine.player_units)-1:
                            menu_pos += 1
                        elif menu_pos == len(self.engine.player_units)-1:
                            menu_pos = 0 
                            
                    # Place a unit
                    if event.key == K_z or event.key == K_RETURN:
                        # Only non-preset units can be modified
                        if self.engine.player_units[menu_pos].name not in self.preset_units.keys():
                            self.deploy_screen_place(self.engine.player_units[menu_pos])
                                
                        # Updates number of deployed units
                        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))

                        # The first time the unit cap is reached, return the player to the front menu
                        if self.num_deployed == self.max_deployed_units and not self.all_deployed:
                            menu_flag = False
                            self.all_deployed = True

                    # Remove the unit
                    if event.key == K_c:
                        # Only non-preset units can be modified
                        if self.engine.player_units[menu_pos].name not in self.preset_units.keys():
                            if self.engine.player_units[menu_pos] in self.team1:
                                self.kill(self.engine.player_units[menu_pos])
                                del(self.all_units_total[self.engine.player_units[menu_pos].name])
                                
                                self.num_deployed -= 1
                        # Updates number of deployed units
                        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))

                    if event.key == K_x:
                        menu_flag = False
                        
            if menu_flag:
                self.render_background()
                self.render_deploy_locations()
                self.render_all_units()
                self.render_cursor()
                
                self.engine.surface.blit(self.engine.menu_board,(0,490))
                
                self.engine.surface.blit(text_select,(450,505))
                self.engine.surface.blit(text_num_deployed,(450,525))
                self.engine.surface.blit(directions_header,(820-directions_header.get_width(),505))
                self.engine.surface.blit(directions1,(820-directions1.get_width(),525))
                self.engine.surface.blit(directions2,(820-directions2.get_width(),545))
                self.engine.surface.blit(directions3,(820-directions3.get_width(),565))
                
                
                
                # Draw LHS Unit, Name
                if self.engine.player_units[menu_pos-1] in self.team1:
                    self.engine.surface.blit(self.engine.player_units[menu_pos-1].image,(535,545),(105,0,35,35))
                else:
                    self.engine.surface.blit(self.engine.player_units[menu_pos-1].image,(535,545),(0,0,35,35))
                name_img = self.engine.sfont.render(self.engine.player_units[menu_pos-1].name,True,(0,0,0))
                self.engine.surface.blit(name_img,(553-name_img.get_width()/2,582))
                    
                # Draw center unit, name, plot stats
                if self.engine.player_units[menu_pos] in self.team1:
                    self.engine.surface.blit(self.engine.player_units[menu_pos].image,(605,560),(105,0,35,35))
                    # If the unit is on the map, center on them if cursor is not already on that position
                    if self.cursor_pos != self.engine.player_units[menu_pos].location_tile:
                        self.center_on(self.engine.player_units[menu_pos])
                else:
                    self.engine.surface.blit(self.engine.player_units[menu_pos].image,(605,560),(0,0,35,35))
                    
                
                name_img = self.engine.sfont.render(self.engine.player_units[menu_pos].name,True,(0,0,0))
                self.engine.surface.blit(name_img,(623-name_img.get_width()/2,597))
                self.engine.player_units[menu_pos].plot_stats()
                self.engine.surface.blit(self.engine.cursor_img,(605,560))
                
                # Draw RHS Unit, name
                if menu_pos+1 == len(self.engine.player_units):
                    if self.engine.player_units[0] in self.team1:
                        self.engine.surface.blit(self.engine.player_units[0].image,(685,545),(105,0,35,35))
                    else:
                        self.engine.surface.blit(self.engine.player_units[0].image,(685,545),(0,0,35,35))
                    name_img = self.engine.sfont.render(self.engine.player_units[0].name,True,(0,0,0))
                else:
                    if self.engine.player_units[menu_pos+1] in self.team1:
                        self.engine.surface.blit(self.engine.player_units[menu_pos+1].image,(685,545),(105,0,35,35))
                    else:
                        self.engine.surface.blit(self.engine.player_units[menu_pos+1].image,(685,545),(0,0,35,35))
                    name_img = self.engine.sfont.render(self.engine.player_units[menu_pos+1].name,True,(0,0,0))
                
                self.engine.surface.blit(name_img,(703-name_img.get_width()/2,582))
                
                pygame.display.flip()
                self.engine.clock.tick(60)
                
    def deploy_screen_view(self):
                            
        """
        # Function Name: deploy screen view
        # Purpose: View the map in deploy screen
        """
    
        menu_flag = True
        
        self.get_all_moves()
        while menu_flag:
            if self.framenum == 10:
                self.framenum = 0
            self.framenum += 1

            arrowkeys = False
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        self.cursor_arrows(event)
                        arrowkeys = True
                        self.framenum = 0
                    if event.key == K_z or event.key == K_RETURN:
                
                        selected = self.cursor_key_search()
                        
                        # Sends the player to the unit's top level menu
                        if selected is not False:
                            self.all_units_by_name[selected].stats_loop()
                            
                    if event.key == K_a and self.team1:
                        if self.ally_center > 0:
                            self.ally_center -= 1
                        else:
                            self.ally_center = len(self.team1) - 1
                        self.center_on(self.team1[self.ally_center])
                        
                    if event.key == K_s and self.team1:
                        if self.ally_center < len(self.team1) - 1:
                            self.ally_center += 1
                        else:
                            self.ally_center = 0
                        self.center_on(self.team1[self.ally_center])
                    
                    if event.key == K_q:
                        if self.enemy_center > 0:
                            self.enemy_center -= 1
                        else:
                            self.enemy_center = len(self.team2) - 1
                        self.center_on(self.team2[self.enemy_center])
                        
                    if event.key == K_w:
                        if self.enemy_center < len(self.team2) - 1:
                            self.enemy_center += 1
                        else:
                            self.enemy_center = 0
                        self.center_on(self.team2[self.enemy_center])

                    if event.key == K_x:
                        menu_flag = False
                        
            # Use C to toggle displaying movement range
            keys = pygame.key.get_pressed()
            if keys[K_c]:
                self.move_plot = True
            else:
                self.move_plot = False

            # if there is not a tap detected, check if the key is being held down
            if arrowkeys == False and self.framenum == 9:
                self.cursor_arrows_hold()
                self.framenum = 6
                
            # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
            selected = False
            selected = self.cursor_key_search()

            self.render_background()
            
            # Renders the unit's valid moves if C is pressed.
            if selected != False and self.move_plot == True:
                self.all_units_by_name[selected].plot_moves_and_attacks()
                
            self.render_deploy_locations()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))
            self.render_current_terrain_data()

            if selected != False:
                self.all_units_by_name[selected].plot_stats()
                    
            pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                               %(self.cursor_pos.x,self.cursor_pos.y,self.screen_shift.x,self.screen_shift.y))
            
            pygame.display.flip()
            self.engine.clock.tick(60)
            
    def deploy_screen_place(self,unit):
        
        """
        # Function Name: deploy screen place
        # Purpose: Place unit on map
        """

        directions_header = self.engine.bfont.render("Directions",True,(0,0,0))
        directions1 = self.engine.sfont.render("Z - Place unit",True,(0,0,0))
        directions2 = self.engine.sfont.render("X - Cancel",True,(0,0,0))
        directions3 = self.engine.sfont.render("C - Remove unit under cursor",True,(0,0,0))
        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))

        menu_flag = True
        self.get_all_moves()
        while menu_flag:
            if self.framenum == 10:
                self.framenum = 0
            self.framenum += 1

            arrowkeys = False
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        self.cursor_arrows(event)
                        arrowkeys = True
                        self.framenum = 0
                    if event.key == K_z or event.key == K_RETURN:
                        # A unit is placed if the location is a valid deploy location and no other unit is
                        # occupying it and the number of units has not hit the max unit cap
                        if tuple(self.cursor_pos) in self.deploy_locations and not self.cursor_key_search() and unit not in self.team1 and self.num_deployed < self.max_deployed_units:
                            
                            self.store_unit(unit,1)
                            unit.turnend = False
                            unit.update_location(*tuple(self.cursor_pos))
                            menu_flag = False
                            self.num_deployed += 1
                            
                        # A unit is placed if the location is a valid deploy location and another allied unit is occupying it and the unit underneath is not itself, replace it with this one
                        elif tuple(self.cursor_pos) in self.deploy_locations and self.cursor_key_search() and selected != unit.name: 
                            selected = self.cursor_key_search()
                            replace_unit = self.all_units_by_name[selected]
                            if replace_unit in self.team1 and replace_unit.name not in self.preset_units.keys():
                                
                                # Deletes old unit
                                self.kill(replace_unit)
                                del(self.all_units_total[replace_unit.name])
                                self.num_deployed -= 1
                                
                                # Adds in the new unit to the list
                                if unit not in self.team1:
                                    self.store_unit(unit,1)
                                    unit.turnend = False
                                    self.num_deployed += 1
                                # Updates the location for cases if unit is already deployed or if unit is new to the map
                                unit.update_location(*tuple(self.cursor_pos))
                                menu_flag = False
                        
                        # A unit is moved if it has already been deployed and the cursor is over a valid location with nobody occupying it:
                        elif tuple(self.cursor_pos) in self.deploy_locations and unit in self.team1 and not self.cursor_key_search():
                            # Updates the location for cases if unit is already deployed or if unit is new to the map
                            unit.update_location(*tuple(self.cursor_pos))
                            menu_flag = False
                        
                        # Updates number of deployed units
                        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))
   
                                    

                    if event.key == K_c:
                        selected = False
                        selected = self.cursor_key_search()
                        if selected:
                            delete_unit = self.all_units_by_name[selected]
                            if delete_unit in self.team1 and selected not in self.preset_units.keys():
                                self.kill(delete_unit)
                                del(self.all_units_total[delete_unit.name])
                                
                                self.num_deployed -= 1
                        # Updates number of deployed units
                        text_num_deployed = self.engine.sfont.render("Number of Units: %s/%s"%(str(self.num_deployed),str(self.max_deployed_units)),True,(0,0,0))
      
                    if event.key == K_x:
                        menu_flag = False

            # if there is not a tap detected, check if the key is being held down
            if arrowkeys == False and self.framenum == 9:
                self.cursor_arrows_hold()
                self.framenum = 6
                
            if menu_flag:
                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected = False
                selected = self.cursor_key_search()
    
                self.render_background()
                
                # Renders the unit's valid moves if C is pressed.
                if selected != False and self.move_plot == True:
                    self.all_units_by_name[selected].plot_moves_and_attacks()
                    
                self.render_deploy_locations()
                self.render_all_units()
                self.engine.surface.blit(unit.image,(self.cursor_pos*self.engine.tilesize-self.screen_shift*self.engine.tilesize),(140,0,35,35))
                self.render_cursor()
                self.engine.surface.blit(self.engine.menu_board,(0,490))
                self.engine.surface.blit(directions_header,(440,505))
                self.engine.surface.blit(text_num_deployed,(440,525))
                self.engine.surface.blit(directions1,(450,545))
                self.engine.surface.blit(directions2,(450,565))
                self.engine.surface.blit(directions3,(450,585))
    
                if selected != False:
                    self.all_units_by_name[selected].plot_stats()
                        
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                                   %(self.cursor_pos.x,self.cursor_pos.y,self.screen_shift.x,self.screen_shift.y))
                
                pygame.display.flip()
                self.engine.clock.tick(60)
            
    def player_turn(self):
        """
        # Function Name: player_turn
        # Purpose: Handles the human player's turn 
        """
        
        self.currentplayer = 1
        
        # Increments the turn counter by 1 at the start of the turn
        self.turn_count += 1

        for unit in self.team1:
            unit.moved = False
            unit.turnend = False

        self.allmoved1 = False
        
        # Checks all map action events
        self.check_all_MAE()

        self.center_on(self.team1[0])
        
        self.render_background()
        self.render_all_units()
        self.engine.surface.blit(self.engine.menu_board,(0,490))
        pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                   %(self.cursor_pos.x,self.cursor_pos.y,self.screen_shift.x,self.screen_shift.y))

        # Displays a message for player 1.
        self.engine.surface.blit(self.engine.p1turn, (105,140))
        pygame.display.flip()
        # Waits one second
        self.engine.pause(1)

        self.trait_turn_process(self.team1)
        [self.run_status_effects(unit) for unit in self.team1 if unit.status.keys()]
        self.run_ssp_sc_regen(1)
        
        # Checks if any status effects managed to complete the mission objective

        self.update_ssps()
        self.battle_end = self.objective.check(self)
        if self.battle_end != False:
            self.victory_condition = True
        
        if self.team1:
            self.center_on(self.team1[0])
        
        # Clears the events before the turn starts.
        pygame.event.clear()

        self.framenum = 0
        
        self.get_all_moves()
        while self.allmoved1 == False and not self.victory_condition:
            
            if self.framenum == 10:
                self.framenum = 0
            self.framenum += 1
            
            self.user_input()

            # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
            selected = False
            selected = self.cursor_key_search()

            self.render_background()
            
            # Renders the unit's valid moves if C is pressed.
            if selected != False and self.move_plot == True:
                self.all_units_by_name[selected].plot_moves_and_attacks()
                
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))
            self.render_current_terrain_data()
            self.render_turn_number()

            if selected != False:
                self.all_units_by_name[selected].plot_stats()
                    
                        
            pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                               %(self.cursor_pos.x,self.cursor_pos.y,self.screen_shift.x,self.screen_shift.y))
            
            pygame.display.flip()
            self.engine.clock.tick(60)

            unitsended = 0
            # Checks if all units have ended their turn.
            for unit in self.team1:
                if unit.turnend == True or unit.alive == False:
                    unitsended += 1

            self.update_ssps()
            self.battle_end = self.objective.check(self)
            if self.battle_end != False:
                self.victory_condition = True

            # If all the units that can move have turn ended, end player 1's turn
            if self.engine.options.turn_end == True:
                if len(self.team1) == unitsended:
                    self.allmoved1 = True
                
            # Checks all map action events
            self.check_all_MAE()
                
        # Ends the remaining characters in team 1.
        for unit in self.team1:
            unit.turnend = True

    def ai_turn(self):
        """
        # Function Name: ai_turn
        # Purpose: Runs the AI's turn
        """
        

        self.currentplayer = 2

        if self.victory_condition == False:
            
            self.currentplayer = 2
            
            for unit in self.team2:
                unit.moved = False
                unit.turnend = False

            self.allmoved2 = False
        
            self.render_background()
            self.render_all_units()
            self.engine.surface.blit(self.engine.menu_board,(0,490))
            pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                       %(self.cursor_pos.x,self.cursor_pos.y,self.screen_shift.x,self.screen_shift.y))

            # Displays a message for player 2.
            self.engine.surface.blit(self.engine.p2turn, (105,140))
            pygame.display.flip()
            # Waits two seconds
            self.engine.pause(1)
            
            self.trait_turn_process(self.team2)
            # Processes status effects
            [self.run_status_effects(unit) for unit in self.team2 if unit.status.keys()]
            self.run_ssp_sc_regen(2)

            
        
        while self.allmoved2 == False and self.victory_condition == False:    
            
            unit_move_list = []
            
            # FSM state updates for everyone first
            
            [unit.ai.update_state() for unit in self.team2]
            
            # 1st Priority: All retreating units take their turn
            [unit_move_list.append(unit) for unit in self.team2 
                if (unit.ai.current_state.name == 'AttackRetreat'
                    or unit.ai.current_state.name == 'PursuitRetreat')
                and unit.turnend == False
                and unit not in unit_move_list]
                    
            # 2nd Priority: Healers take their turn
            [unit_move_list.append(unit) for unit in self.team2
                if (unit.ai.current_state.name == 'HealerSOS'
                    or unit.ai.current_state.name == 'HealerStandby')
                and unit.turnend == False
                and unit not in unit_move_list]     
            
            # 3rd Priority: Everyone Else
            [unit_move_list.append(unit) for unit in self.team2
                if unit.turnend == False
                and unit not in unit_move_list]     
            
            # Executes all units in priority order
            for unit in list(unit_move_list):
                
                if unit.alive == True and self.victory_condition == False:
                    pygame.display.flip()
                    self.engine.clock.tick(60)
                    
                    if unit.alive:
                        
                        # Asks the unit to carry out move
                        unit.ai.execute_turn()
                                     
                    unit.turnend = True
                
                    pygame.display.flip()
                    self.engine.clock.tick(60)
                    
                    # Checks if the battle is over and if one side's objectives have been fulfilled

                    self.update_ssps()
                    self.battle_end = self.objective.check(self)

                    if self.battle_end != False:
                        self.victory_condition = True

                    # Checks all map action events
                    self.check_all_MAE()

            self.allmoved2 = True
            pygame.display.flip()
            self.engine.clock.tick(60)
                

        # Ends the remaining characters in team 2.
        if self.team2 != []:
            for unit in self.team2:
                unit.turnend = True
    def victory_scene(self):
        """
        # function name: victory scene
        # Purpose: sets up and runs the runs the post mission MAE 
        """


        if self.post_map_MAE:

            self.engine.fade_to('black',0.5)

            for unit_name in self.required_survivors:
                if unit_name not in self.all_units_total.keys():
                    self.store_unit(self.engine.player_units_by_name[unit_name],1)

                unit = self.all_units_total[unit_name]
                unit.spirit_stats = 'normal'
                unit.clear_status()
                unit.turnend = False

                if unit.alive == False:
                    unit.alive = True
                    unit.HP = unit.maxHP
                    self.all_units_by_name[unit.name] = unit
                    if unit.team == 1:
                        self.team1.append(unit)
                    elif unit.team == 2:
                        self.team2.append(unit)
                    # Adds unit sprite back in
                    self.sg_units.add(unit.sprite)
                    self.sg_unitcircles.add(unit.circle)
                    self.sg_status.add(unit.status_bubble)

            for unit in self.team1:
                # If a player unit is not required for the ending scenes, remove it from the map
                if unit.name not in self.required_survivors:
                    self.kill(unit)

            # Does any rearranging of units before the 
            self.post_map_MAE.pre_exec()

            self.post_map_MAE.fade_from_color('black', 0.5)

            self.post_map_MAE.execute()
            pygame.mixer.music.stop()
            
    def render_mission_end_banner(self):
        """
        # function name: render_mission_end_banner
        # purpose: Draw the animation of the "Victory / Mission Failed" banner at the end of a mission
        """
        if self.battle_end == 'team1victory':
            mission_end_banner = self.engine.mission_victory
        else:
            mission_end_banner = self.engine.mission_failure

        smoothstep = lambda v: (v*v*(3-2*v))
        
        start_pos = Vector2(-525,140)
        
        frame_count = 45
        step_vector = Vector2(1,0)
        for t in xrange(0,frame_count+1):
            v = float(t)/float(frame_count)
            v = smoothstep(v)
            scale_term = int(630*v)
            intermediate_step = scale_term*step_vector
            current_pos = start_pos + intermediate_step
        
            self.render_background()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))
            self.engine.surface.blit(mission_end_banner,current_pos)
                
            pygame.display.flip()
            self.engine.clock.tick(60)
        self.engine.pause(0.5)
        frame_count = 45
        start_pos = current_pos.copy()
        for t in xrange(0,frame_count+1):
            v = float(t)/float(frame_count)
            v = smoothstep(v)
            scale_term = int(730*v)
            intermediate_step = scale_term*step_vector
            current_pos = start_pos + intermediate_step
        
            self.render_background()
            self.render_all_units()
            self.render_cursor()
            self.engine.surface.blit(self.engine.menu_board,(0,490))
            self.engine.surface.blit(mission_end_banner,current_pos)
                
            pygame.display.flip()
            self.engine.clock.tick(60)
               
            
    def turn_loop(self):
        """
        # Function Name: turn loop
        # Purpose: The main loop in a battle map that drives the game by alternating between player turns.
        """
    
        self.prestart()
        
        # Values for "Center to the next unit"
        self.ally_center = 0
        self.enemy_center = 0
        self.turn_count = 0
        self.allmoved1 = False
        self.allmoved2 = False
        self.move_plot = False

        # If the event is a conversation event, exits the function immediately without executing the 
        # battle events
        
        if self.nobattle == False:
            self.victory_condition = False
            self.battle_end = False
            
            self.display_alert('Objective',self.objective.desc)
            if self.enable_deploy:
                self.deploy_screen_main()
            
            # Debug mode: Win instantly
            if self.engine.single_turn_win == True:
                self.victory_condition = True
                self.battle_end = 'team1victory'
                
            # Alternates between player and AI turns
            while not self.victory_condition:
                # Player's turn
                self.player_turn()
                        
                # If team 2 is defeated but the objective hasn't been completed, 
                # jump back to player 1
                if self.team2:
                    self.ai_turn()
                
            self.render_mission_end_banner()
            
            # Completion MAE event
            [unit.clear_status() for unit in self.team1]
            if self.battle_end == 'team1victory':
                self.victory_scene()
          
          
        # Resets all units to max HP / spirit 300 / alive
        self.mission_end()
        
        # Resets all SSPs
        for ssp in self.all_ssps.values():
            ssp.capture_state = ssp.starting_state
        
        # Removes any added player characters in case the mission is failed
        if self.objective and self.battle_end != 'team1victory':
            
            for unit in self.new_party_members:
                
                # clears out all spell actions, traits, exp, tp stuff
                unit.reset_state()
                
                # Delete unit from party roster
                del(self.engine.player_units[self.engine.player_units.index(unit)])
                del(self.engine.player_units_by_name[unit.name])
                # Delete unit from save data
                del(self.engine.player.all_unit_data[unit.name])
                # Delete unit name from save data
                del(self.engine.player.party_members[self.engine.player.party_members.index(unit.name)])
                
            
        # Returns a completion string if mission was a success
        if self.objective != None:
            return self.battle_end
            
    def unit_stats_loop(self):
        
        
        """
        # Function Name: unit_stats_loop
        # Purpose: birdseye view of all the ally and enemy units
        """
    
        menu_flag = True
        menu_pos = [0,0]
        text_ally_header = self.engine.bfont.render("Player Units",True,(0,0,0))
        text_enemy_header = self.engine.bfont.render("Enemy Units",True,(0,0,0))
        text_ally_obj_header = self.engine.bfont.render("Player Objective",True,(0,0,0))
        text_enemy_obj_header = self.engine.bfont.render("Mission Failure Condition",True,(0,0,0))
        text_ally_obj = self.engine.sfont.render(self.objective.goal,True,(0,0,0))
        text_enemy_obj = self.engine.sfont.render(self.objective.enemy_goal,True,(0,0,0))
        str_ally_status,str_enemy_status = self.objective.report(self)
        text_ally_status = self.engine.sfont.render(str_ally_status,True,(0,0,0))
        text_enemy_status = self.engine.sfont.render(str_enemy_status,True,(0,0,0))
        
        text_more = self.engine.sfont.render("[More]",True,(0,0,0))
        shift_ally = 0
        shift_enemy = 0
        
        while menu_flag:

            text_ally_units = []
            text_enemy_units = []

            # Generates text data for the ally units
            for i in xrange(0,min(6,len(self.team1))):
                text_ally_units.append([self.team1[i+shift_ally].name])
                text_ally_units[i].append(self.engine.bfont.render(self.team1[i+shift_ally].name,True,(0,0,0)))
                text_ally_units[i].append(self.engine.sfont.render("HP: %3.0f/%3.0f"%(self.team1[i+shift_ally].HP,self.team1[i+shift_ally].maxHP),True,(0,0,0)))
                text_ally_units[i].append(self.engine.sfont.render("LV: "+str(self.team1[i+shift_ally].level),True,(0,0,0)))
                text_ally_units[i].append(self.engine.sfont.render("EXP: %3.0f/100"%self.team1[i+shift_ally].exp,True,(0,0,0)))
                if self.team1[i+shift_ally].spell_actions[self.team1[i+shift_ally].equipped]:
                    text_ally_units[i].append(self.engine.sfont.render("Active: "+self.team1[i+shift_ally].spell_actions[self.team1[i+shift_ally].equipped].namesuffix,True,(0,0,0)))
                else:
                    text_ally_units[i].append(self.engine.sfont.render("Active: None",True,(0,0,0)))
                    
                
            # Generates text data for the enemy units
            for i in xrange(0,min(6,len(self.team2))):
            
                text_enemy_units.append([self.team2[i+shift_enemy].name])
                text_enemy_units[i].append(self.engine.bfont.render(self.team2[i+shift_enemy].name,True,(0,0,0)))
                text_enemy_units[i].append(self.engine.sfont.render("HP: %3.0f/%3.0f"%(self.team2[i+shift_enemy].HP,self.team2[i+shift_enemy].maxHP),True,(0,0,0)))
                text_enemy_units[i].append(self.engine.sfont.render("LV: "+str(self.team2[i+shift_enemy].level),True,(0,0,0)))
                text_enemy_units[i].append(self.engine.sfont.render("EXP: %3.0f/100"%self.team2[i+shift_enemy].exp,True,(0,0,0)))
                if self.team2[i+shift_enemy].spell_actions[self.team2[i+shift_enemy].equipped]:
                    text_enemy_units[i].append(self.engine.sfont.render("Active: "+self.team2[i+shift_enemy].spell_actions[self.team2[i+shift_enemy].equipped].namesuffix,True,(0,0,0)))
                else: 
                    text_enemy_units[i].append(self.engine.sfont.render("Active: None",True,(0,0,0)))
                    
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # Toggles sides
                    # Only allows checking the other side if there are units on the other side
                    if (event.key == K_LEFT or event.key == K_RIGHT) and self.team2:
                        if menu_pos[0] == 0:
                            menu_pos[0] = 1
                            menu_pos[1] = shift_enemy
                            if menu_pos[1] < shift_enemy:
                                menu_pos[1] = shift_enemy
                            if menu_pos[1] > len(self.team2)-1:
                                menu_pos[1] = len(self.team2)-1
                                shift_enemy = max(0,len(self.team2)-6)
                        elif menu_pos[0] == 1:
                            menu_pos[0] = 0
                            menu_pos[1] = shift_ally
                            if menu_pos[1] < shift_ally:
                                menu_pos[1] = shift_ally
                            if menu_pos[1] > len(self.team1)-1:
                                menu_pos[1] = len(self.team1)-1
                                shift_ally = max(0,len(self.team1)-6)
                                
                            
                    if event.key == K_UP:
                        # Top of the list of units: Jumps to bottom
                        if menu_pos[1] == 0:
                            if menu_pos[0] == 0:
                                menu_pos[1] = len(self.team1)-1
                                shift_ally = max(0,len(self.team1)-6)
                            if menu_pos[0] == 1:
                                menu_pos[1] = len(self.team2)-1
                                shift_enemy = max(0,len(self.team2)-6)
                        # Top of interval, advances shift up
                        elif menu_pos[0] == 0 and menu_pos[1] == shift_ally:
                            menu_pos[1] -= 1
                            shift_ally -= 1
                        elif menu_pos[0] == 1 and menu_pos[1] == shift_enemy:
                            menu_pos[1] -= 1
                            shift_enemy -= 1
                        # Within the interval, advances cursor only        
                        elif menu_pos[1] > 0:
                            menu_pos[1] -= 1
                            
                    if event.key == K_DOWN:
                        # Bottom of the list of units: Jumps to top
                        if (menu_pos[0] == 0 and menu_pos[1] == len(self.team1)-1):
                            menu_pos[1] = 0
                            shift_ally = 0

                        elif (menu_pos[0] == 1 and menu_pos[1] == len(self.team2)-1):
                            menu_pos[1] = 0
                            shift_enemy = 0
                            
                        # Bottom of the interval. Advance shift downward    
                        elif menu_pos[0] == 0 and menu_pos[1] == 5+shift_ally:
                            menu_pos[1] += 1
                            shift_ally += 1
                            
                        elif menu_pos[0] == 1 and menu_pos[1] == 5+shift_enemy:
                            menu_pos[1] += 1
                            shift_enemy += 1
                        
                        # Between intervals. Advances downward.
                        else:
                            menu_pos[1] += 1
                    if event.key == K_z or event.key == K_RETURN:
                        if menu_pos[0] == 0:
                            self.team1[menu_pos[1]].stats_loop()
                        elif menu_pos[0] == 1 and self.team2:
                            self.team2[menu_pos[1]].stats_loop()
                        
                    
                    if event.key == K_x:
                        menu_flag = False
                    if event.key == K_c:
                        menu_flag = False
                        if menu_pos[0] == 0:
                            self.center_on(self.team1[menu_pos[1]])
                        elif menu_pos[0] == 1:
                            self.center_on(self.team2[menu_pos[1]])
                        # Automatically exits the map_menu loop
                        return True
                    
            
            if menu_flag:
                # Background, Headers
                self.engine.surface.blit(self.engine.stats_bg,(0,0))
                self.engine.surface.blit(text_ally_header,(50,50))
                self.engine.surface.blit(text_enemy_header,(480,50))
                self.engine.surface.blit(text_ally_obj_header,(50,480))
                self.engine.surface.blit(text_ally_obj,(60,500))
                self.engine.surface.blit(text_ally_status,(60,520))
                self.engine.surface.blit(text_enemy_obj_header,(480,480))
                self.engine.surface.blit(text_enemy_obj,(490,500))
                self.engine.surface.blit(text_enemy_status,(490,520))
                
                
                # Displays "More" if there are more units to scroll through
                if shift_ally < len(self.team1)-6 and len(self.team1) > 6:
                    self.engine.surface.blit(text_more,(350,450))
                if shift_ally > 0:
                    self.engine.surface.blit(text_more,(350,70))
                
                # Displays "More" if there are more units to scroll through
                if shift_enemy < len(self.team2)-6 and len(self.team2) > 6:
                    self.engine.surface.blit(text_more,(750,450))
                if shift_enemy > 0:
                    self.engine.surface.blit(text_more,(750,70))
                    
                
                for i in xrange(0,min(6,len(self.team1))):
                    self.engine.surface.blit(text_ally_units[i][1],(60,80+60*i))
                    self.engine.surface.blit(self.all_units_by_name[text_ally_units[i][0]].image,(70,100+60*i),(0,0,35,35))
                    self.engine.surface.blit(text_ally_units[i][2],(120,100+60*i))
                    self.engine.surface.blit(text_ally_units[i][3],(120,120+60*i))
                    self.engine.surface.blit(text_ally_units[i][4],(220,120+60*i))
                    self.engine.surface.blit(text_ally_units[i][5],(220,100+60*i))
                
                for i in xrange(0,min(6,len(self.team2))):
                    self.engine.surface.blit(text_enemy_units[i][1],(490,80+60*i))
                    self.engine.surface.blit(self.all_units_by_name[text_enemy_units[i][0]].image,(500,100+60*i),(0,0,35,35))
                    self.engine.surface.blit(text_enemy_units[i][2],(550,100+60*i))
                    self.engine.surface.blit(text_enemy_units[i][3],(550,120+60*i))
                    self.engine.surface.blit(text_enemy_units[i][4],(650,120+60*i))
                    self.engine.surface.blit(text_enemy_units[i][5],(650,100+60*i))
                    
                if menu_pos[0] == 1:
                    cursor_y = 88+60*(menu_pos[1]-shift_enemy)
                else:
                    cursor_y = 88+60*(menu_pos[1]-shift_ally)
                pygame.draw.circle(self.engine.surface, (0,0,0), (50+430*menu_pos[0],cursor_y),5)
                    
                pygame.display.flip()
                self.engine.clock.tick(60)
                
        # Does not automatically exits the map_menu loop
        return False
                
    def trait_turn_process(self,team):
        """
        # Function Name: trait_turn_process
        # Purpose: For every unit in a team, processes their "Every Turn" traits
        # Inputs: Team (team1,team2)
        """
        
        for unit in team:
            # Checks action
            if (unit.traits[0][0]
                and 'Every Turn' in unit.traits[0][0].properties
                and unit.traits[0][0].turn_check(unit)):
                    self.center_on(unit)
                    self.render_background()
                    self.render_all_units()
                    self.render_cursor()
                    pygame.display.flip()
                    unit.traits[0][0].turn_execute(unit)
            # Checks support
            if (unit.traits[1][0]
                 and unit.traits[1][0].turn_check(unit) 
                 and 'Every Turn' in unit.traits[1][0].properties):
                self.center_on(unit)
                self.render_background()
                self.render_all_units()
                self.render_cursor()
                pygame.display.flip()
                unit.traits[1][0].turn_execute(unit)
 
    def mission_end(self):
        
        """
        # Function Name: Mission end
        # Purpose: Process post map reset stuff. At the end of each battle, resets each unit's HP/Spirit settings, 
        #           and de-associates them with the map
        """
        
        for unit in self.all_units_total.values():

            unit.HP = unit.maxHP
            unit.alive = True
            unit.spirit = 300
            unit.map = None
            unit.ressurected = False
            unit.spirit_stats = 'normal'
            unit.clear_status()
            unit.draw_status = 0
            # Fully recharges any spell card class
            for spell in unit.spell_actions:
                if spell and spell.consumable == False:
                    spell.livesleft = spell.lives
        
        
class Landmark(object):

    def __init__(self,name,location,size,img_coords,can_pass):

        """
        # Function Name: __init__
        # Purpose: Creates a landmark object
        # Inputs: Name - the name of the object
        #         Img_coords - coordinates (X,Y) within the landmarks image file
        #         Size: Size of the object in (X,Y) tiles
        #         Location: Top Left coordinate of the image
        #         Can_pass: whether the landmark can be passed through
        """
        self.name = name
        self.location = Vector2(location)
        self.location_pixel = self.location*35
        size_x,size_y = size
        self.size = Vector2(size)
        self.occupied_tiles = []
        [self.occupied_tiles.append((int(self.location.x+delta_x),int(self.location.y+delta_y))) for delta_x in xrange(0,size_x) for delta_y in xrange(0,size_y)]
        
        self.img_coords = Vector2(img_coords)
        self.can_pass = can_pass
        
    def render(self):
    
        """
        # Function Name: render
        # Purpose: draws the landmark on the screen at its location
        """
    
        x,y = self.location_pixel
        self.map.engine.surface.blit(self.map.engine.landmark_img, ((x,y)-self.map.screen_shift*35),(self.img_coords.x,self.img_coords.y,35*self.size.x,35*self.size.y))
        
        
class SpiritSourcePoint(Landmark):

    def __init__(self,name,location,preset):
        """
        # Function Name: __init__
        # Purpose: Creates a landmark object
        # Inputs: Name - the name of the object
        #         Location: coordinates of SSP
        #         Preset:  Preset capture state (0 unclaimed, 1 ally, 2 enemy)
        """
        
        self.name = name
        self.location = Vector2(location)

        # Capture state of the location:
            # 0 - Unclaimed
            # 1 - Ally
            # 2 - Enemy
        self.capture_state = preset
        self.starting_state = preset
        Landmark.__init__(self,name,location,(1,1),(0,105),True)
        
    def render(self):
        
        """
        # Function Name: render
        # Purpose: draws the landmark on the screen at its location
        """
    
        x,y = self.location_pixel
        if self.capture_state == 0:
            img_coords = (0,70,35,35)
        elif self.capture_state == 1:
            img_coords = (35,70,35,35)
        elif self.capture_state == 2:
            img_coords = (70,70,35,35)
        
        #print [x,y,img_coords]
        
        self.map.engine.surface.blit(self.map.engine.landmark_img, ((x,y)-self.map.screen_shift*35),img_coords)
    