import pygame
import os
from pygame.locals import *
from sys import exit
from linalg import Vector2
from math import floor, sin, cos, pi
from lostsky.battle.spells import *
from lostsky.battle.animations import *
# from engine import Engine


class Unit(object):
    """
    # defines the basic class of characters
    """

    def __init__(self,name,image,avatar,moves,growthvector,level,skill_rank,
                 chartype,unitclass,animation,deathquote):
    
        """
        # Function Name: __init__
        # Purpose: Creates a character unit
        # Inputs: Name = Unit's name
        #         Image = Unit's Sprite Image
        #         Avatar = Unit's Portrait Image
        #         Moves = Unit's movement range
        #         Growthvector = A list of five numbers for the growth rates [HP,STR,DEF,MAG,MDEF,AGL,ACC]
        #         Level = Unit's Starting Level
        #         Preference = Preferred attack type
        #         Skill_rank = List of 4 numbers 0-5 representing a unit's skill in a certain magic school
        #              Input format:  [NAT,ELE,SPI,FOR] 
        #         Chartype: Character type - PC, Enemy, Boss
        #         unitclass = Unit's class
        #         animation = animation ID string. None if none are available
        #         deathquotes = Sayings to display upon a unit being defeated
        """
        
        self.image = image                      # Unit's Sprite Image
        self.av = avatar                        # Unit's Portrait
        self.location_tile =Vector2(0, 0)       # Unit's Current Location in the Map Matrix
        self.location_pixel = Vector2(0, 0)     # Unit's Current Location on screen
        self.moves = moves                      # Unit's Range
        self.name = name                        # Unit's Name
        self.validmoves = {}                    # All the valid move vectors
        self.validmoves_path = {}
        self.validmoves[(0,0)]=1                # Sets the first entry to be a delta_r of (0,0)
        self.skill_rank = {'Natural':skill_rank[0],'Elemental':skill_rank[1],'Spiritual':skill_rank[2],'Force':skill_rank[3]}            
        self.chartype = chartype                # Sets the character type: PC, Enemy, Boss
        self.ressurected = False                # Tracks whether a unit has ressurected in a map or not
        self.map = None                         # Unit's associated map
        self.unitclass = unitclass              # Unit's class
        if animation:
            self.animation_enable = True           # Whether a Unit's animations are enabled
            self.anim_id_string = animation        # Key to the unit's animation set in the engine
        else:
            self.animation_enable = False
            self.anim_id_string = None
        
        self.spell_actions = [None, None, None, None, None]
        # Sets the default equipped spell position
        self.equipped = 0

        #Stats
        self.level = level
        self.starting_level = level
        self.growth = growthvector
        self.update_stats()
        self.HP = self.maxHP
        self.exp = 0
        self.alive = True
        self.team = 1 # Default to allies
        self.prohibited_tiles = []
        
        #self.spirit = 3.0                          # All units start with 3.0 levels of spirit charge
        self.spirit = 300
        
        self.spirit_stats = 'normal'               # High Spirits / Low Spirits Status

        #Turn Indicators. Default is that unit is dormant until activated by the map
        self.moved = True
        self.turnend = True

        #Sets up trait Slots
        # traits are stored in a list as follows: [[Action,Empty,Empty,Empty,Empty],[Support,Empty,Empty,Empty,Empty]] 
        self.traits = [[None, None, None, None, None],[None, None, None, None, None]]
        self.reserve_traits = [[],[]]
       
        self.trait_points = level*2
        
        # Status effects
        # Format - {'Status effect': # turns}
        self.status = {}
        self.draw_status = 0
       
        # Death Quotes
        self.deathquote = deathquote


        # Invincibility status (e.g. for Misaki)
        self.invincible = False
        
        self.sprite = UnitMapSprite(self)     # Unit map sprite
        self.circle = UnitCircleSprite(self)  # Unit team circle
        self.status_bubble = UnitStatusSprite(self)   # Unit Status Bubble sprite
        
       
    #############################
    # Computational Methods
    #############################

    def add_spell(self,spell):
        """
        # Function Name: add spell
        # Purpose: Gives a unit a spell at the first available position
        # Inputs: spell = The spell to be assigned
        # Output: assign = True if assigning was successful, 'inventoryerror' if all the slots are full

        """
        
        assign = False
        spell.get_attack_range(self)

        if None in self.spell_actions:
            self.spell_actions[self.spell_actions.index(None)] = spell
            assign = True

        return assign
    
    def add_trait(self,trait):
    
        """
        # Function Name: add trait
        # Purpose: Gives a unit a new trait
        # Inputs: trait = The trait to be assigned
        # Output: assign = True if assigning was successful, False if all the slots are full
        """
        
        assign = False

        # Assigns the trait to the first open slot
        if trait.variation == 'Action' and None in self.traits[0]:
            self.traits[0][self.traits[0].index(None)] = trait
            assign = True
        if trait.variation == 'Support' and None in self.traits[1]:
            self.traits[1][self.traits[1].index(None)] = trait
            assign = True

        for spell in self.spell_actions:
            if spell:
                spell.get_attack_range(self)

        return assign
    
    def swap_trait(self,trait_pos):
        """
        # Function Name: swap_trait
        # Purpose: swaps a standby trait to an active trait
        # Inputs: trait_pos - position of desired standby trait
        """
        
        # Temporarily stores the targetted swap's contents
        temp = self.traits[trait_pos[0]][0]

        # Assigns the trait from standby to active
        self.traits[trait_pos[0]][0] = self.traits[trait_pos[0]][trait_pos[1]+1]

        # Returns the active trait to standby
        self.traits[trait_pos[0]][trait_pos[1]+1] = temp

        # Resorts standby traits
        standby = []
        for trait in self.traits[trait_pos[0]][1:5]:
            if not trait:
                standby.append([1,None,None])
            else:
                standby.append([0,trait.name,trait])
                
        standby.sort()
        for i in (0,1,2,3):
            self.traits[trait_pos[0]][i+1] = standby[i][2]
        
        for spell in self.spell_actions:
            if spell:
                spell.get_attack_range(self)
        
    
    def reserve_trait(self,trait_pos):
        """
        # Function Name: reserve_trait
        # Purpose: unequips a trait and sends it to the reserve traits list
        # Inputs: trait_pos - position of desired standby trait
        """
            
        # Sends the trait to reserve
        self.reserve_traits[trait_pos[0]].append(self.traits[trait_pos[0]][trait_pos[1]])

        # Unequips the selected trait
        self.traits[trait_pos[0]][trait_pos[1]] = None
        
        # Resorts standby traits
        standby = []
        for trait in self.traits[trait_pos[0]][1:5]:
            if not trait:
                standby.append([1,None,None])
            else:
                standby.append([0,trait.name,trait])
                
        standby.sort()
        for i in (0,1,2,3):
            self.traits[trait_pos[0]][i+1] = standby[i][2]
    
        for spell in self.spell_actions:
            if spell:
                spell.get_attack_range(self)

    def has_trait_property(self, trait_property):
        """
        # Function Name - has_trait_property
        # Purpose: Checks if a certain property is in either the unit's traits
        # Input:    property - property to check if equipped
        # Output:    True if the property is in one of the unit's equipped traits
        #            False if the property is not in either of the unit's equipped traits.
        """
        
        # Checks Action
        if self.traits[0][0] and trait_property in self.traits[0][0].properties:
            return True
        # Checks Support
        elif self.traits[1][0] and trait_property in self.traits[1][0].properties:
            return True
        else:
            return False
            
    def get_prohibited(self):
        """
        # Function Name: get_prohibited
        # Purpose: From a unit's range, compute a list of positions the unit is allowed to pass through
        """
        
        if self.team == 1:
            self.prohibited_tiles = [tuple(unit.location_tile - self.location_tile) for unit in self.map.team2]
        if self.team == 2:
            self.prohibited_tiles = [tuple(unit.location_tile - self.location_tile) for unit in self.map.team1]

    def calculate_max_moves(self):
        """
        calculate_max_moves
        Calculates total available movement distance
        """

        maxmoves = self.moves
        # Adds in trait effects to the movement range
        if self.traits[0][0]:
            maxmoves+= self.traits[0][0].movemod_add
        if self.traits[1][0]:
            maxmoves+= self.traits[1][0].movemod_add

        # If unit has movement down, range is halved to a minimum of 1 movement
        if 'Movement Down' in self.status.keys():
            maxmoves /= 2
            if maxmoves <= 0:
                maxmoves = 1

        return maxmoves

    def get_moves_path(self):
                
        """
        # Function Name: get moves path
        # Purpose: From a unit's range, compute a list of allowed delta-position vectors
        # Inputs: None
        # Outputs: None
        # Special Thanks to Jon Wyrick for help in the development of this algorithm
        # Modified by bbm29 to support non-unit cost tiles.
        """
        
        # Start position is your current location
        start = (0,0)
        # How many moves can you make
        maxmoves = self.calculate_max_moves()

        # Check movement trait properties
        if self.has_trait_property('Flight'):
            flight = True
        else:
            flight = False

        if self.has_trait_property('Swimming'):
            swimming = True
        else:
            swimming = False

        self.validmoves = {start:(0,0)}

        if self.moves == 0:

            self.validmoves_path = self.validmoves.copy()
            self.get_valid_spell_range()
            return


        wait_list = []      #The tiles on the wait list(Vector2, tuple, cost)
        wait_list_pos = []  #The tile(Vector2) which is on the wait list
        to_check = [start]
        current_moves = [start]

        # Define unit vectors
        up = Vector2(0,-1)
        down = Vector2(0,1)
        left = Vector2(-1,0)
        right = Vector2(1,0)

        unit_vectors = [up,down,left,right]
        
        moves_left = maxmoves

        # Get the location of the prohibited tiles
        self.get_prohibited()
        
        # Initialization - Unless prohibited, a unit is allowed to move at least one step in each direction
        #   The very first move is treated as if the unit is flying so that you never get stuck
        #   in a tile if everything around you is too expensive to move into
        for unit_vector in unit_vectors:
            # processes each unit vector +/- x, +/- y
            new_pos = tuple(Vector2(start) + unit_vector)
            new_map_pos = tuple(new_pos + self.location_tile)
            # we only add the new position as a place of interest if
            #   1) It is not in all_moves, which means we have visited that spot in equal or fewer moves
            #   2) It is not in current_moves, which means we have visited that spot in the same move number
            #   3) It is not in the list of prohibited moves in units that you are now allowed to pass through
            #       Note: You ARE allowed to walk through allies.
            #   4) It is not a tile that you are not allowed to enter
            #   5) It is not a tile currently on the wait list
            #   6) Tile is on the map 
            if (not self.validmoves.has_key(new_pos) and (new_pos not in current_moves)
                and (new_pos not in self.prohibited_tiles) 
                and (new_map_pos not in self.map.map_walk_prohibited
                     or (flight == True 
                         and new_map_pos not in self.map.map_fly_prohibited)
                     or (swimming == True
                         and self.map.terrainmap[new_map_pos][0].name == 'Deep Water')
                and wait_list_pos.count(new_pos) == 0)
                and self.map.terrainmap.has_key(new_map_pos)):
                self.validmoves[new_pos] = tuple(unit_vector)
                current_moves.append(new_pos)
                        
        # Subtracts off the initial moves
        moves_left -= 1


        # Computes moves if there are more moves left to check 
        while moves_left > 0:
            
            #Check if tiles in wait list are ready to be validated
            for tile in wait_list:
                tile[2] -=1
                if tile[2] == 1:
                    self.validmoves[tile[0]] = tuple(tile[1])
            # Sets the current set of moves to be the ones to check in the next iteration
            to_check = current_moves
            current_moves = []
            
            for position in to_check:

                for unit_vector in unit_vectors:
                    # processes each unit vector +/- x, +/- y
                    new_pos = tuple(Vector2(position) + unit_vector)
                    
                    new_map_pos = tuple(new_pos + self.location_tile)
                            
                    # we only add the new position as a place of interest if
                    #   1) It is not in self.validmoves, which means we have visited that spot in equal or fewer moves
                    #   2) It is not in current_moves, which means we have visited that spot in the same move number
                    #   3) It is not in the list of prohibited moves in units that you are now allowed to pass through
                    #       Note: You ARE allowed to walk through allies.
                    #   4) It is not a tile that you are not allowed to enter
                    #   5) It is not a tile currently on the wait list
                    #   6) Tile is on the map
                    if (not self.validmoves.has_key(new_pos)
                        and (new_pos not in current_moves)
                        and (new_pos not in self.prohibited_tiles)
                        and self.map.terrainmap.has_key(new_map_pos)
                        and (new_map_pos not in self.map.map_walk_prohibited 
                             or (flight == True 
                                 and new_map_pos not in self.map.map_fly_prohibited)
                             or (swimming == True
                                 and self.map.terrainmap[new_map_pos][0].name == 'Deep Water')
                        and new_pos not in wait_list_pos)):
                        # Sets any non-1 movements to the side
                        if not flight:

                            # If unit has a swimming trait and the tile is water, treat it as a cost 1 tile
                            if (self.map.terrainmap[new_map_pos][0].cost == 1 or
                                (swimming and self.map.terrainmap[new_map_pos][0].name == 'Deep Water')):
                                self.validmoves[new_pos] = tuple(unit_vector)
                                current_moves.append(new_pos)
                            elif new_pos not in wait_list_pos:
                                wait_list.append([new_pos, tuple(unit_vector), self.map.terrainmap[new_map_pos][0].cost])
                                wait_list_pos.append(new_pos)
                        else:                            
                            self.validmoves[new_pos] = tuple(unit_vector)
                            current_moves.append(new_pos)
                                    
            #Add ready tiles to current_moves
            for tile in wait_list:
                if tile[2] == 1:
                    current_moves.append(tile[0])
            moves_left -= 1
            
        # Save a copy of valid moves before removing coordinates
        self.validmoves_path = self.validmoves.copy()
        # removes any coordinate someone else is standing on
        for other_unit in self.map.all_units_by_name.values():
            delta_pos = other_unit.location_tile - self.location_tile
            if self.validmoves.has_key(tuple(delta_pos)) and delta_pos != (0,0): # for case of comparison with itself
                del self.validmoves[tuple(delta_pos)]
    
        # Updates the valid spell range display
        self.get_valid_spell_range()

    def get_valid_spell_range(self):
        
        """
        # Function Name: get valid spell range
        # Purpose: From a unit's spell range, compute a list of all allowed spell positions
        """
        
        # generates a valid attack range beyond movement range
        self.valid_spell_range = []
        
        # Checks if a spell is equipped
        if self.spell_actions[self.equipped]:
    
            # Define unit vectors
            up = Vector2(0,-1)
            down = Vector2(0,1)
            left = Vector2(-1,0)
            right = Vector2(1,0)

            unit_vectors = [up,down,left,right]
            
            to_check = self.validmoves.keys()
            current_attacks = []
            attackrange = self.spell_actions[self.equipped].spellrange
            if self.has_trait_property('Extend Max Range'):
                attackrange += 1

            while attackrange > 0:
                for position in to_check:
                    for unit_vector in unit_vectors:
                        # processes each unit vector +/- x, +/- y
                        new_pos = tuple(Vector2(position) + unit_vector)
                        # Conditions for adding a new position
                        # 1. New Position is not already in validmoves. We only care about the range beyond the valid
                        #   moves
                        # 2. New Position is not already in current moves. This means that position is already covered in this turn.
                        # 3. New position is not already in valid attacks. This means we already covered that position previously
                        # 
                        if not self.validmoves.has_key(new_pos) and (new_pos not in current_attacks) and (new_pos not in self.valid_spell_range):
                            new_map_pos = tuple(new_pos + self.location_tile)

                            # Checks if the tile is even on the map (Important if the character is near an edge)
                            if self.map.terrainmap.has_key(new_map_pos):
                                self.valid_spell_range.append(new_pos)
                                current_attacks.append(new_pos)
                                
                attackrange -= 1
                to_check = current_attacks
                
                # Clears out the current list of moves
                current_attacks = []
                          
    
    def get_path(self,destination):
        """
        # Function name: Construct path
        # Purpose: generates a path from current position to any new position
        # Input: Destination - next location
        # Output: Returns a forward moving list of unit vector tuples that construct the path 
        """
        
    
        #print self.validmoves_path
        path = [self.validmoves_path[destination]] 
        # Gets the first step back
        next_pos = Vector2(destination) - Vector2(self.validmoves_path[destination])
        # Continues stepping until we reach the origin
        while next_pos != Vector2((0,0)):
            path.append(self.validmoves_path[tuple(next_pos)])
            # Steps one more step back
            next_pos = Vector2(next_pos) - Vector2(self.validmoves_path[tuple(next_pos)])
        # Reverses the list to get forward moving
        path.reverse()
        return path
                
    
    def update_location(self,x,y):
        
        """
        # Function Name: update location
        # Purpose: Moves a unit to a location and updates both the unit's location
        #       on the tile grid as well as the pixel location
        # Inputs: x,y = location in terms of tile positions
        """
        
        self.location_tile = Vector2(x,y)
        self.location_pixel = self.location_tile*35

    
    def level_up(self):
        """
        # Function Name: level_up
        # Purpose: Levels a unit up
        """
        
        HP_before = self.maxHP
        self.update_stats()
        HP_gain = self.maxHP - HP_before
        self.HP += HP_gain
        
        self.trait_points += 2
        

    def update_stats(self):
    
        """
        # Function Name: update stats
        # Purpose: Using a unit's growthvector,
        # computes the stats of a unit for a certain level
        # Growth given as [HP, STR, DEF, MAG, MDEF, AGL, ACC]
        """
    
        # HP = 5*(HPgrow) + floor(HPgrow*.5*LV)
        self.maxHP = 5*self.growth[0]+int(floor(self.growth[0]*self.level/2))
        # STR = 2+(STRgrow)+floor((STR or MAGgrow)*LV/20)
        self.STR = 2+self.growth[1]+int(floor(self.growth[1]*self.level/20))
        # DEF = 2 +(DEFgrow) + floor((DEFgrow or MDEFgrow)*LV/25)
        self.DEF = 2+self.growth[2]+int(floor(self.growth[2]*self.level/25))
        # MAG = Int(5 * Growth Mod * Level /40);
        self.MAG = 2+self.growth[3]+int(floor(self.growth[3]*self.level/20))
        # MDEF = 2 +(DEFgrow or MDEFgrow) + floor((DEFgrow or MDEFgrow)*LV/25)
        self.MDEF = 2+self.growth[4]+int(floor(self.growth[4]*self.level/25))
        # AGL = 10 + floor(AGLgrow*LV*8/50)
        self.AGL = 10+int(floor(self.growth[5]*self.level*8/50))
        # ACC = 5+floor(ACCgrow*LV/10)
        self.ACC = 5 + int(floor(self.growth[6]*self.level/10))

    
    def compute_trait_bonus(self):
        """
        # Function Name: Compute trait
        # Purpose: Compute sum total of trait bonuses from support and active traits
        # Inputs: None
        # Output: trait_mods: Summed stat bonuses from both equipped traits
        """
        
        # Sums the trait bonuses for support and active
        trait_mods = [0,0,0,0,0,0]
        for i in xrange(0,6):
            if self.traits[0][0]:
                trait_mods[i] += self.traits[0][0].statmods[i]
            if self.traits[1][0]:
                trait_mods[i] += self.traits[1][0].statmods[i]
        return trait_mods
    
    def compute_status_effect_bonus(self):
        """
        # Function Name: compute_status_effect_bonus
        # Purpose: Compute sum total of status effect bonuses from unit's status effects
        # Inputs: None
        # Output: se_mods - Status effect bonus total
        """
        
        se_mods = [0,0,0,0,0,0]
        # Iterates over all status effects
        for se in self.status.keys():
            # Iterates over all stats changes caused by status effect
            for index, value in enumerate(self.map.engine.status_effects_catalog[se].statmods):
                se_mods[index] += value
                
        return se_mods
                
    def compute_total_stat_mods(self):
        """
        # Function Name: Compute total stat mods
        # Purpose: Compute the sum of the stat mods from the traits and status effects
        # Output: self_total_mods
        """
        self_trait_mods = self.compute_trait_bonus()
        self_se_mods = self.compute_status_effect_bonus()
        self_total_mods = [sum(mod) for mod in zip(self_trait_mods,self_se_mods)]
        return self_total_mods
        
    def compute_hit_bonuses(self,target):
        """
        # Function Name: Compute hit bonuses
        # Purpose: Compute the total bonuses/penalties from traits and status effects to hit%
        # Output: hit_bonus - How much to add to unit's % hit
        """
        
        hit_bonus = 0

        ####
        # Effect from Traits
        ####
        
        if self.has_trait_property("Homing Attack"):
            hit_bonus += 15
        
        if target.has_trait_property("Mirage"):
            hit_bonus -= 15

        ####
        # Effect from Status Effects
        ####
        hit_bonus += sum([self.map.engine.status_effects_catalog[se].hitmod for se in self.status.keys()])
        hit_bonus -= sum([self.map.engine.status_effects_catalog[se].evamod for se in target.status.keys()])

        return hit_bonus
        
    def compute_crit_bonuses(self,target):
        """
        # Function Name: Compute crit bonuses
        # Purpose: Compute the total bonuses/penalties from traits and status effects to crit%
        # Output: hit_bonus - How much to add to unit's % crit
        """
        crit_bonus = 0

        if self.has_trait_property("Critical+ Lv.3"):
            crit_bonus += 20
        elif self.has_trait_property("Critical+ Lv.2"):
            crit_bonus += 15
        elif self.has_trait_property("Critical+ Lv.1"):
            crit_bonus += 10
        
        if self.has_trait_property("Critical- Lv.3"):
            crit_bonus -= 20
        elif self.has_trait_property("Critical- Lv.2"):
            crit_bonus -= 15
        elif self.has_trait_property("Critical- Lv.1"):
            crit_bonus -= 10
        ####
        # Effect from Status Effects
        ####
        crit_bonus += sum([self.map.engine.status_effects_catalog[se].critmod for se in self.status.keys()])
 
        return crit_bonus
        
        
    def compute_spell_relation(self,target):
        """
        # Function Name: Compute spell relation
        # Purpose: Checks an attack type against a defense type and return advantage or disadvantage
        #          See: http://wiki.featheredmelody.com/Story_of_a_Lost_Sky/Spell_Type_Relationships
        # Inputs: User, target
        # Outputs: dmgmod = Float number to multiply damage by (Range: 0.75 -> 1.25)
        #          critmod = Number to add to the critical percent
        """
        
        strengths = {"Natural":"Elemental","Elemental":"Spiritual","Spiritual":"Force","Force":"Natural"}
        weaknesses = {"Natural":"Force","Elemental":"Natural","Spiritual":"Elemental","Force":"Spiritual"}
        
        # Checks to make sure target has something equipped
        if target.spell_actions[target.equipped]:
        
            # User rank in spell (Skill Rank + Spell Intrinsic Rank)
            self_rank = self.skill_rank[self.spell_actions[self.equipped].affinity] + self.spell_actions[self.equipped].spell_rank
            target_rank = target.skill_rank[target.spell_actions[target.equipped].affinity] + target.spell_actions[target.equipped].spell_rank
            
            # Takes the difference between two spell ranks
            difference = self_rank - target_rank
            
            # Case for the spell being strong against the target
            if target.spell_actions[target.equipped].affinity == strengths[self.spell_actions[self.equipped].affinity]:
                if difference < -2:
                    relation = 'Neutral'
                    dmgmod = 1.0
                    critmod = 0
                elif difference >= -2 and difference <= 2:
                    relation = 'Strong'
                    dmgmod = 1.15
                    critmod = 15
                elif difference > 2:
                    relation = 'Very Strong'
                    dmgmod = 1.25
                    critmod = 25
                    
            # Case for the spell being weak against the target
            elif target.spell_actions[target.equipped].affinity == weaknesses[self.spell_actions[self.equipped].affinity]:
                if difference <-2:
                    relation = 'Very Weak'
                    dmgmod = 0.75
                    critmod = -25
                elif difference >= -2 and difference <= 2:
                    relation = 'Weak'
                    dmgmod = 0.85
                    critmod = -15
                elif difference > 2:
                    relation = 'Neutral'
                    dmgmod = 1.0
                    critmod = 0
                
            # Case for the spell being neither weak nor strong against the target
            else:
                relation = 'Neutral'
                dmgmod = 1.0
                critmod = 0
                
        # Case target does not have an equipped spell action
        else:
                relation = 'Strong'
                dmgmod = 1.15
                critmod = 15
        
            
        return relation, dmgmod,critmod
        
        
    def compute_threshold(self,target):
    
        """
        # Function Name: Compute_base_threshold
        # Purpose: Returns hit threshold for an encounter
        # Inputs: Target
        # Output: Threshold = 100% - probability of hitting a target
        """


        if target.invincible:
            return 100

        # Trait / SE bonuses
        self_total_mods = self.compute_total_stat_mods()
        target_total_mods = target.compute_total_stat_mods()
        
        self_hit_bonus = self.compute_hit_bonuses(target)
    
        # THRESHOLD = defender's AGL*trait % Mod + defender's AGLmod + defender's terrain AGLmod - attacker's ACC - attacker's ACCmod - attacker's terrain AGLmod    
        if target.spell_actions[target.equipped]:
            base_agl = target.AGL*(1.0+target_total_mods[5]) + target.spell_actions[target.equipped].aglmod + self.map.terrainmap[tuple(target.location_tile)][0].aglmod
        
        else:
            base_agl = target.AGL*(1.0+target_total_mods[5]) + self.map.terrainmap[tuple(target.location_tile)][0].aglmod
        
        base_acc = self.ACC*(1.0+self_total_mods[4]) + self.spell_actions[self.equipped].accmod + self.map.terrainmap[tuple(self.location_tile)][0].accmod
        
        targetthreshold = int(base_agl - base_acc)
        targetthreshold -= self_hit_bonus
        
        # Spirits Bonus: +/- 10% to Evasion and +/- 5% to Accuracy:
        if self.spirit_stats == "high":
            targetthreshold -= 10
        if target.spirit_stats == "high":
            targetthreshold += 10
        if self.spirit_stats == "low":
            targetthreshold += 10
        if target.spirit_stats == "low":
            targetthreshold -= 10
        
        
        return targetthreshold

    def compute_hitpercent(self,target):
        
        """
        # Function Name: compute_hitpercent
        # Purpose: Determine the probability of hitting with an attack
        #           as well as the probability the enemy unit will counterattack
        # Inputs: enemy = the enemy unit
        # Outputs: percent1,percent2 = Probability of successful attack, counterattack respectively as strings
        #           if either unit is out of range, it will return the string "N/A"
        """
        
        # Computes the delta positions 
        delta_pos_attack = target.location_tile - self.location_tile
        delta_pos_counter = self.location_tile - target.location_tile

        # checks if the enemy is in range
        if tuple(delta_pos_attack) in self.spell_actions[self.equipped].validattacks:


            # Gets the threshold of the target
            base_threshold_target = self.compute_threshold(target)
            # Note that the hit percent is not allowed to be greater than 100 or less than 0
            percent1 = str(max(min(100 - base_threshold_target,100),0))
            
        else:
            percent1 = 'N/A'

        # checks if you are in range of the enemy's counterattack if available
        if (target.spell_actions[target.equipped] and target.spell_actions[target.equipped].counterattack == True and
            tuple(delta_pos_counter) in target.spell_actions[target.equipped].validattacks):

            # Gets the threshold of yourself
            base_threshold_self = target.compute_threshold(self)
            percent2 = str(max(min(100 - base_threshold_self,100),0))
        else:
            percent2 = 'N/A'

        return percent1,percent2

    
    def compute_crit_threshold(self,target):
        """
        # Function Name: Compute_crit_threshold
        # Purpose: Returns crit threshold for an encounter
        # Inputs: Target
        # Output: Threshold =  probability of critical against target
        """
        
        # Gets the relationship between the two spell types
        relationship, dmgmod, rel_critmod = self.compute_spell_relation(target)
    
        # Gets trait / se mods
        self_total_mods = self.compute_total_stat_mods()
        target_total_mods = target.compute_total_stat_mods()
        
        
        self_crit_bonus = self.compute_crit_bonuses(target)
        
        # CRIT THRESHOLD = self's ACC - average(target's DEF and MDEF) + self's SC Crit mod
        critthreshold = self.ACC*(1.0+self_total_mods[4]) - ((target.DEF*(1.0+target_total_mods[1]) + target.MDEF*(1.0+target_total_mods[3]))/2) + self.spell_actions[self.equipped].critmod
        
        # Adds in spell relation and equipped trait mods
        critthreshold += rel_critmod
        critthreshold += self_crit_bonus
            
        # Spirits Bonus: +/- 10% to Critical / Critical evasion
        if self.spirit_stats == "high":
            critthreshold += 10
        elif self.spirit_stats == "low":
            critthreshold -= 10
        if target.spirit_stats == "high":
            critthreshold -= 10
        elif target.spirit_stats == "low":
            critthreshold += 10

        # Crit threshold bounded by [0, 100]
        if critthreshold < 0:
            critthreshold = 0
        elif critthreshold > 100:
            critthreshold = 100
        else:
            critthreshold = int(critthreshold)
        
        return critthreshold
    
    
    def compute_critpercent(self,target):
        """
        # Function Name: compute_critpercent
        # Purpose: Determine the probability of scoring a critical hit with an attack,
        #           as well as the probability the enemy will score a critical hit
        # Inputs: enemy = the enemy unit
        # Outputs: crit1,crit2 = Probability of successful critical, counterattack critical respectively as strings
        #           if either unit is out of range, it will return the string "N/A"
        """
            
        # Computes the delta positions 
        delta_pos_attack = target.location_tile - self.location_tile
        delta_pos_counter = self.location_tile - target.location_tile

        # checks if the enemy is in range
        if tuple(delta_pos_attack) in self.spell_actions[self.equipped].validattacks:
            crit1 = str(self.compute_crit_threshold(target))
        else:
            crit1 = 'N/A'
        
        # checks if you are in range of the target's counterattack if available
        if (target.spell_actions[target.equipped]
            and target.spell_actions[target.equipped].counterattack == True
            and tuple(delta_pos_counter) in target.spell_actions[target.equipped].validattacks):
            crit2 = str(target.compute_crit_threshold(self))
        else:
            crit2 = 'N/A'

        return crit1,crit2

    
        
    def compute_damage(self,target):
        """
        # Function Name: compute_damage
        # Purpose: Computes the estimated damage of an attack
        # Inputs: target = the caster and the target
        # Outputs: damage = Estimated damage done by the spell 
        """
        
        # Gets the relationship between the two spell types
        relationship, dmgmod, rel_critmod = self.compute_spell_relation(target)
    
        # Gets trait/se bonuses
        self_total_mods = self.compute_total_stat_mods()
        target_total_mods = target.compute_total_stat_mods()
        
        # Effect calculation for physical attack type spells
        # Damage = (Attacker's STR + Attacker's STRmod)*Attacker's Base Effect 
        if self.spell_actions[self.equipped].attacktype == 'physical':
            effect = (self.STR*(1.0+self_total_mods[0]) + self.spell_actions[self.equipped].strmod)*self.spell_actions[self.equipped].effect

        # Effect calculation for magical attack type spells
        # Effect = (Attacker's MAG + Attacker's MAGmod)*Attacker's Base Effect 
        elif self.spell_actions[self.equipped].attacktype == 'magical':
            effect = (self.MAG*(1.0+self_total_mods[2]) + self.spell_actions[self.equipped].magmod)*self.spell_actions[self.equipped].effect 
                
        # If the defender has no equipped spell or is holding an item, the default shield effect is 1
        if target.spell_actions[target.equipped]:
            
            # Shield calculation for physical damage type spells
            # Shield = (Defender's DEF + Defender's DEFmod) * TDEF * Defender's Base Shield
            if self.spell_actions[self.equipped].damagetype == 'physical':
                shield = (target.DEF*(1.0+target_total_mods[1]) + target.spell_actions[target.equipped].defmod)*self.map.terrainmap[tuple(target.location_tile)][0].TDEF*target.spell_actions[target.equipped].shield
            
            # Shield calculation for magical damage type spells
            # Shield = (Defender's MDEF + Defender's MDEFmod) * TDEF * Defender's Base Shield
            elif self.spell_actions[self.equipped].damagetype == 'magical':
                shield = (target.MDEF*(1.0+target_total_mods[3]) + target.spell_actions[target.equipped].mdefmod)*self.map.terrainmap[tuple(target.location_tile)][0].TDEF*target.spell_actions[target.equipped].shield
        else:
            # Shield calculation for physical damage type spells
            # Shield = (Defender's DEF) * TDEF
            if self.spell_actions[self.equipped].damagetype == 'physical':
                shield = (target.DEF*(1.0+target_total_mods[1]))*self.map.terrainmap[tuple(target.location_tile)][0].TDEF
        
            # Shield calculation for magical damage type spells
            # Shield = (Defender's MDEF) * TDEF
            elif self.spell_actions[self.equipped].damagetype == 'magical':
                shield = (target.MDEF*(1.0+target_total_mods[3]))*self.map.terrainmap[tuple(target.location_tile)][0].TDEF
        
            
        
        # Final damage = Effect - Shield
        damage = effect - shield
        
        # Multiplies damage with the relationship effect
        damage = damage * dmgmod
        
        if self.spirit_stats == "high":
            damage = int(damage*1.1)
        elif self.spirit_stats == "low":
            damage = int(damage*0.9)
        else:
            damage = int(damage)
        
        return damage

    
    
    def predict_damage(self,target):
        """
        # Function Name: predict_damage
        # Purpose: Determines an estimate of the amount of damage the attacker will do as well as counterattack damage
        # Inputs: enemy = the enemy unit
        # Outputs: damage1,damage2 = Damage (without critical factor), and counterattack damage respectively as strings
        #           if either unit is out of range, it will return the string "N/A"
        """
        
        # Computes the delta positions 
        delta_pos_attack = target.location_tile - self.location_tile
        delta_pos_counter = self.location_tile - target.location_tile

        if self.spell_actions[self.equipped].type == 'attack':
            # checks if the enemy is in range
            if tuple(delta_pos_attack) in self.spell_actions[self.equipped].validattacks:
                damage1 = self.compute_damage(target)
                if damage1 < 1:
                    damage1 = 1
                    
                damage1 = str(damage1)
                
            else:
                damage1 = 'N/A'
                
            # checks if you are in range of the enemy's counterattack
            if  (target.spell_actions[target.equipped] and target.spell_actions[target.equipped].counterattack == True
                 and tuple(delta_pos_counter) in target.spell_actions[target.equipped].validattacks):
                damage2 = target.compute_damage(self)
                if damage2 < 0:
                    damage2 = 1
                damage2 = str(damage2)
            else:
                damage2 = 'N/A'

                
        elif self.spell_actions[self.equipped].type in ('healing','healingitem','support'):
            damage2 = 'N/A'
            damage1 = str(self.spell_actions[self.equipped].get_effect(self,target))

        return damage1,damage2

    
    
    
    def experience(self,enemy,damage):
        """
        # Function Name: experience
        # Purpose: Determines an estimate of the amount of damage the attacker will do as well as counterattack damage
        # Inputs: enemy = the enemy unit
        #         damage = damage done
        # Formulas:
        #    LDIF = Attacker's Level - Defender's Level
        #
        #    * Base EXP = 25 * 0.75 ^ LDIF 
        #
        # NOTE: These values are rounded off to the nearest integer
        #    * Successful Story Boss Kill: 5 x Base EXP
        #    * Successful Boss Kill: 2 x Base EXP
        #    * Successful Kill: 1 x Base EXP
        #    * Successful Attack: 0.40 x Base EXP
        #    * Failed Attack: 0.10 x Base EXP 
        """

        # Compute the base EXP
        LDIF = self.level - enemy.level
        # Check to see if the target is alive first
        if self.alive == True:
            base_exp = 25. * 0.75 ** LDIF
        else:
            base_exp = 0
        
        if damage == 'miss':
            exp = int(0.10 * base_exp)
        else:
            # Checks if the enemy is still alive after the turn action.
            if enemy.alive == True:
                exp = int(0.40 * base_exp)
            else:
                if enemy.chartype == 'enemy':
                    exp = int(base_exp)
                elif enemy.chartype == 'boss' or enemy.chartype == 'pc' or enemy.chartype == 'npc':
                    exp = int(2 * base_exp)
                elif enemy.chartype == 'storyboss':
                    exp = int(5 * base_exp)

        # EXP Cap of 200 EXP
        if exp > 200:
            exp = 200
                    
        self.exp += exp

        print self.name + " got EXP: " + str(exp)
                
        level_up = False
        # Checks if there has been a successful level up:
        while self.exp >= 100:
            # Carries over extra EXP to the next level.
            self.exp = max(self.exp-100,0)
            # Increase level by 1
            self.level += 1
            # Update stats
            print self.name + " level up."
            self.level_up()
            level_up = True
            
        return exp, level_up



    
    def battle_spirit(self,enemy,damage,critical):
        """
        # Function Name: battle_spirit
        # Purpose: determines the spirit effect of the last action
        # Inputs: enemy = the enemy unit
        #         damage = damage done
        #         critical = critical hit
        # FORMULAS
        #        * Successful Attack: SC + 15
        #        * Successful Defeated Enemy: SC + 100
        #        * Successful Critical: SC + 40
        #        * Failed Attack (Miss): SC - 10
        #        * Enemy Critical: SC - 40
        #        * Resurrect (Kaguya, Mokou, etc): SC resets to 0
        """
        
        # Checks if the player is alive
        spirit_self = 0
        spirit_enemy = 0
        if self.alive == True:
            if damage == 'miss':
                spirit_self -= 10
            else:
                spirit_self += 15

                # Enemy Defeated
                if enemy.alive == False:
                    spirit_self += 100

        # Critical Hit
        if critical == True:
            spirit_self += 40
            spirit_enemy -= 40
            
        self.spirit += spirit_self
        enemy.spirit += spirit_enemy
        
        self.check_spirit_range()
        enemy.check_spirit_range()
        
        return spirit_self,spirit_enemy

    def check_spirit_range(self):
        
        """
        # Function Name: check_spirit_range
        # Purpose: checks to make sure the spirit charge values are in the appropriate range
        #    and checks if the spirit status conditions need to be updated
        """
        
        # Checks to ensure the spirit numbers remain within the range (0.00 -> 9.00
        if self.spirit > 900:
            self.spirit = 900
        if self.spirit < 0:
            self.spirit = 0
            
        # Checks spirit status conditions
        if self.spirit > 750:
            self.spirit_stats = 'high'
            if 'Low Spirit' in self.status.keys():
                self.remove_status('Low Spirit')
            if 'High Spirit' not in self.status.keys():
                self.give_status('High Spirit')
            
        elif self.spirit <= 750 and self.spirit >= 200:
            self.spirit_stats = 'normal'
            if 'High Spirit' in self.status.keys():
                self.remove_status('High Spirit')
            if 'Low Spirit' in self.status.keys():
                self.remove_status('Low Spirit')
        elif self.spirit < 200:
            self.spirit_stats = 'low'
            if 'High Spirit' in self.status.keys():
                self.remove_status('High Spirit')
            if 'Low Spirit' not in self.status.keys():
                self.give_status('Low Spirit')
        else:
            print "Spirit Range out of bounds"

    def check_spell_type_availability(self, type):
        """
        Function name: check_spell_type_availability
        Purpose; Checks if a unit has a certain type of spell. e.g. valid inputs would be "support" or "attack"
        Input:  type - spell action type
        Output: (T/F, index of spell)

        """


        for index, spell in enumerate(self.spell_actions):
            if spell and spell.type == type:
                return True, index
        return False, None

    def give_status(self,status_effect):
        """
        # Function: give_status
        # Purpose: Give unit a status effect
        # Inputs: status_effect = status effect to assign to unit
        """
        self.status[status_effect] = 0
        self.status_bubble.update_queue()
    
    def remove_status(self,status_effect):
        """
        # Function: remove_status
        # Purpose: Remove a status effect from unit
        # Inputs: status_effect = status effect to remove from unit
        """
        del self.status[status_effect]
        self.status_bubble.update_queue()
    
    def clear_status(self):    
        """
        # Function: clear_status
        # Purpose: Remove all status effects from unit
        """
        self.status = {}
        self.status_bubble.update_queue()

    def reset_state(self):     
        """
        # Function Name: reset_state
        # Purpose: clears out all spell actions, traits, exp, tp stuff
        """
        self.spell_actions = [None, None, None, None, None]
        self.traits = [[None, None, None, None, None],[None, None, None, None, None]]
        self.reserve_traits = [[],[]]
        self.exp = 0
        self.level = self.starting_level
        self.update_stats()
        self.trait_points = self.starting_level*2
        
    #############################
    # Graphical Methods
    #############################

    def plot_valid_moves(self):
    
        """
        # Function Name: plot_valid_moves
        # Purpose: Plots the valid movement positions on the screen.
        """
        
        for tile in self.validmoves.keys():
            self.map.engine.surface.blit(self.map.engine.move_tile,(Vector2(tile)*35+self.location_pixel-self.map.screen_shift*self.map.engine.tilesize))
        
    def plot_moves_and_attacks(self):
        """
        # Function Name: plot_moves_and_attacks
        # Purpose: Plots the valid movement positions on the screen, as well as the possible valid attack locations.
        """
        
        self.plot_valid_moves()
        for tile in self.valid_spell_range:
            if self.spell_actions[self.equipped].type == 'attack':
                self.map.engine.surface.blit(self.map.engine.attack_tile,(Vector2(tile)*35+self.location_pixel-self.map.screen_shift*self.map.engine.tilesize))
            elif self.spell_actions[self.equipped].type in ('healing','healingitem','support'):
                self.map.engine.surface.blit(self.map.engine.heal_tile,(Vector2(tile)*35+self.location_pixel-self.map.screen_shift*self.map.engine.tilesize))
     
    def plot_attacks(self,delta=Vector2(0,0)):

        """
        # Function Name: plot_attacks
        # Purpose: Plots the valid attack range for the equipped spell 
        # Inputs: Delta: Vector corresponding to a shift in position, occurs if the plotting is to be done during the move confirmation screen.
        """
        
        for tile in self.spell_actions[self.equipped].validattacks:
            if self.spell_actions[self.equipped].type == 'attack':
                self.map.engine.surface.blit(self.map.engine.attack_tile,(Vector2(tile)*35+self.location_pixel+delta*35-self.map.screen_shift*self.map.engine.tilesize))
            elif self.spell_actions[self.equipped].type in ('healing','healingitem','support'):
                self.map.engine.surface.blit(self.map.engine.heal_tile,(Vector2(tile)*35+self.location_pixel+delta*35-self.map.screen_shift*self.map.engine.tilesize))


    def render(self):
        """
        # Function Name: render
        # Purpose: Plots the unit's sprite as well as a status condition bubble if applicable
        """

        x,y = self.location_pixel
        self.render_team_circle()
        
        
        # Draws the normal sprite if the turn is not over.
        if self.turnend == False or self.map.currentplayer != self.team :
            self.map.engine.surface.blit(self.image, ((x,y)-self.map.screen_shift*self.map.engine.tilesize),(0,0,35,35))
        else:
            self.map.engine.surface.blit(self.image, ((x,y)-self.map.screen_shift*self.map.engine.tilesize),(105,0,35,35))
        # Plots an icon if the unit has a status effect
        if self.status.keys():
            # Displays a different bubble every 2 seconds (lifetime of animation counter).
            if self.map.animation_counter in (1,61):
                self.draw_status += 1
            # Checks if the tracking counter for status bubbles is bigger than number of status effects
            # Two cases:
            #       1. Unit is either in high / low spirits - We cap the drawing counter at #status effects and draw status effects at [1,...]
            #       2. Unit is not in high/low spirits - We cap the drawing counter at #status effects and draw status effects at [0,...]
            if (self.spirit_stats not in ('low','high') and self.draw_status > len(self.status.keys())-1) or (self.spirit_stats in ('low','high') and self.draw_status > len(self.status.keys())):
                self.draw_status = 0
                
            if self.spirit_stats in ('low','high'):
                # Draws Spirit Charge status
                if self.draw_status == 0:
                    if self.spirit_stats == "low":
                        self.map.engine.surface.blit(self.map.engine.status_spirit, (x+17,y)-self.map.screen_shift*self.map.engine.tilesize,(0,0,18,18))
                    elif self.spirit_stats == "high":
                        self.map.engine.surface.blit(self.map.engine.status_spirit, (x+17,y)-self.map.screen_shift*self.map.engine.tilesize,(18,0,18,18))
                # Draws status effects
                else:
                    self.map.engine.surface.blit(self.map.engine.status_effect_icons, (x+17,y)-self.map.screen_shift*self.map.engine.tilesize,
                        (self.map.engine.status_effects_catalog[self.status.keys()[self.draw_status-1]].icon_location*18,0,18,18))
            else:
                # Draws status effects only
                self.map.engine.surface.blit(self.map.engine.status_effect_icons, (x+17,y)-self.map.screen_shift*self.map.engine.tilesize,
                    (self.map.engine.status_effects_catalog[self.status.keys()[self.draw_status]].icon_location*18,0,18,18))
        
        else:
            if self.spirit_stats == "low":
                self.map.engine.surface.blit(self.map.engine.status_spirit, (x+17,y)-self.map.screen_shift*self.map.engine.tilesize,(0,0,18,18))
            elif self.spirit_stats == "high":
                self.map.engine.surface.blit(self.map.engine.status_spirit, (x+17,y)-self.map.screen_shift*self.map.engine.tilesize,(18,0,18,18))
        
    def render_emote(self,emotion):
        """
        # Function name: render_emote
        # Purpose: Draws an emotion bubble
        # Input: Emotion- name of emotion type
        """
        emotion_image = self.map.engine.emotion_bubbles[emotion]['image']
        frame_width = self.map.engine.emotion_bubbles[emotion]['frame_width']
        frame_height = self.map.engine.emotion_bubbles[emotion]['frame_height']
        max_frames_x = emotion_image.get_width()/frame_width
        max_frames_y = emotion_image.get_height()/frame_height 
        delay = self.map.engine.emotion_bubbles[emotion]['delay']
        
        for frame_num_y in xrange(0,max_frames_y):
            for frame_num_x in xrange(0,max_frames_x):
                
                self.map.render_background()
                self.map.render_all_units(bubbles = False)
                self.map.render_cursor()
                self.map.render_menu_panel()
                if self.map.enable_stats_panel:
                    self.plot_stats()
                
                
                # Draws emoticon frame
                self.map.engine.surface.blit(emotion_image,self.location_pixel+Vector2(15,-(frame_width-10))-35*self.map.screen_shift,(frame_width*frame_num_x,frame_height*frame_num_y,frame_width,frame_height))
                pygame.display.flip()
                self.map.engine.clock.tick(60)
                # Delay between frames
                if delay > 0:
                    self.map.engine.pause(delay)
        
    
    def render_team_circle(self):
        """
        # Function name: render_team_circle
        # Purpose: Draws the character's underneath circle
        """
        
        if self.chartype == 'boss':
            self.map.engine.surface.blit(self.map.engine.team_panels, ((self.location_pixel.x,self.location_pixel.y)-self.map.screen_shift*self.map.engine.tilesize),(70,0,35,35))
        elif self.chartype == 'npc':
            self.map.engine.surface.blit(self.map.engine.team_panels, ((self.location_pixel.x,self.location_pixel.y)-self.map.screen_shift*self.map.engine.tilesize),(105,0,35,35))            
        elif self.team == 1:
            self.map.engine.surface.blit(self.map.engine.team_panels, ((self.location_pixel.x,self.location_pixel.y)-self.map.screen_shift*self.map.engine.tilesize),(0,0,35,35))
        elif self.team == 2:
            self.map.engine.surface.blit(self.map.engine.team_panels, ((self.location_pixel.x,self.location_pixel.y)-self.map.screen_shift*self.map.engine.tilesize),(35,0,35,35))
            
    def render_startle(self):
        """
        # Function name: render_startle
        # Purpose: Has a character jump up twice in a startled position
        """
        
        # Background render
        self.map.render_background()
        self.map.render_menu_panel()
        if self.map.enable_stats_panel:
            self.plot_stats()
        background_surface = self.map.engine.surface.copy()
        
        # First frame - Update entire screen    
        self.map.sg_units.clear(self.map.engine.surface,background_surface)
        rects = self.map.render_all_units()
        self.map.render_cursor()
        pygame.display.flip()
        self.map.engine.clock.tick(60)
        
        starting_height = self.location_pixel.y
        # Freeze team color circle in place
        self.circle.hold = True
        
        # We want to have this unit drawn last so it is in the proper order when 
        # the unit jumps to the next tile up.
        # Removes unit from map group and creates a separate group for it
        startle_group = pygame.sprite.RenderUpdates()
        
        self.map.sg_units.remove(self.sprite)
        startle_group.add(self.sprite)
        
        def y(a , v, t):
            """
            # function name: y
            # purpose:
            #     Simulates constant force towards bottom of screen for.
            #     (Positive = Down)
            #     Standard kinematics equation: y(t) = y0 + v*t + a*t^2
            #     Assume y0 = 0
            # Input: a - acceleration (pixel/frame^2)
            #        v - initial speed (pixel/frame)
            #        t - time (frame)
            """
            return int(a*t**2 + v*t)
        
        # Parameters for jumping
        v = -4.0
        # Number of frames in animation
        frames = 10
        # Require jump to reach max height midway through animation
        # e.g. velocity = 0 at max height
        # v(t) = v0 + a * t 
        a = -v/frames/2
        
        for i in xrange(0,2):
            for t in xrange(0,2*frames-1):
                for event in pygame.event.get():
                    if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                        exit()
                        
                self.location_pixel.y = starting_height+y(a, v, t)

                self.map.sg_units.clear(self.map.engine.surface,background_surface)
                startle_group.clear(self.map.engine.surface,background_surface)
                self.map.sg_unitcircles.clear(self.map.engine.surface,background_surface)
                

                # Draws all updated unit positions
                rects = self.map.render_all_units(bubbles = False)
                
                # Draw this unit last
                startle_group.update()
                rects += startle_group.draw(self.map.engine.surface)
                self.map.render_cursor()
                pygame.display.update(rects)
                self.map.engine.clock.tick(60)
                
        # Adds unit back to map group
        self.map.sg_units.add(self.sprite)
                
        # Reset current position
        self.circle.hold = False
        self.location_pixel.y = starting_height
        
        self.map.render_background()
        self.map.render_all_units()
        self.map.render_cursor()
        self.map.render_menu_panel()
        if self.map.enable_stats_panel:
            self.plot_stats()
        pygame.display.flip()
        self.map.engine.clock.tick(60)
        
            
    def render_walk(self,path):
        """
        # Function name: render_walk
        # Purpose: Plots the unit's sprite moving from A to B
        # Inputs:   Unit = Unit to make walk
        #           Path = A list of tuples containing the forward moving path (retrieved from self.get_path() )
        """
        
        new_path = [Vector2(path[0])]
        # converts the input path to a list of vectors
        for step in path[1:]:
            if Vector2(step) == Vector2(new_path[-1]).normalize():
                new_path[-1] += Vector2(step)
            else:
                new_path.append(step)
        
        
        # Smoothstep function
        # See: http://sol.gfxile.net/interpolation/
        smoothstep = lambda v: (v*v*(3-2*v))
        
        start_pos = self.location_pixel
        current_pos = start_pos.copy()
        
        # Background render
        self.map.render_background()
        self.map.render_cursor()
        self.map.render_menu_panel()
        if self.map.enable_stats_panel:
            self.plot_stats()
        background_surface = self.map.engine.surface.copy()
        
        # First frame - Update entire screen    
        self.map.sg_units.clear(self.map.engine.surface,background_surface)
        rects = self.map.render_all_units()
        pygame.display.flip()
        self.map.engine.clock.tick(60)

        # Remove unit from map's main sprite groups
        self.map.sg_units.remove(self.sprite)
        self.map.sg_status.remove(self.status_bubble)

        # Add unit to map's moving unit sprite group
        self.map.sg_moving_unit.add((self.sprite, self.status_bubble))

        while new_path:
            next_step = Vector2(new_path.pop(0))
            # Sets frame count proportional to the square root of the magnitude of the step vector
            frame_count = int(10*next_step.get_magnitude()**(0.5))
            
            for t in xrange(0,int(frame_count)+1):
                
                for event in pygame.event.get():
                    if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                        exit()
                        
                v = float(t)/float(frame_count)
                v = smoothstep(v)
                scale_term = int(35*v)
                intermediate_step = scale_term*next_step
                current_pos = start_pos + intermediate_step

                # Gets updated current position
                self.location_pixel = Vector2(current_pos.x,current_pos.y)

                # Clears the background
                self.map.sg_unitcircles.clear(self.map.engine.surface,background_surface)
                self.map.sg_units.clear(self.map.engine.surface,background_surface)
                self.map.sg_status.clear(self.map.engine.surface,background_surface)
                self.map.sg_moving_unit.clear(self.map.engine.surface, background_surface)
                self.map.sg_moving_unit.update()

                # Draws all updated unit positions
                rects = self.map.render_all_units()

                # Draw this unit last so it will always be on top
                rects += self.map.sg_moving_unit.draw(self.map.engine.surface)

                pygame.display.update(rects)
                self.map.engine.clock.tick(60)
                
            start_pos = current_pos.copy()


        # Remove unit to map's moving unit sprite group
        self.map.sg_moving_unit.remove((self.sprite, self.circle, self.status_bubble))

        # Add unit back into map's main sprite groups
        self.map.sg_units.add(self.sprite)
        self.map.sg_status.add(self.status_bubble)

    
    def render_semitrans(self,delta):
        
        """
        # Function Name: render semitrans
        # Purpose: Renders a semitransparent copy of oneself at a position delta from the current location
        # Inputs: Delta (Vector2) relative to current position.
        """
        
        x,y = self.location_pixel
        x1,y1 = x+delta.x*self.map.engine.tilesize,y+delta.y*self.map.engine.tilesize
       
        # Draws a transparent copy of the sprite
        self.map.engine.surface.blit(self.image, ((x1,y1)-self.map.screen_shift*self.map.engine.tilesize),(140,0,35,35))
        
    def render_hp_change(self,hp_before,hp_after):
        """
        # Function Name: render_hp_change
        # Purpose: Draws an HP meter and shows a change in HP
        # Inputs: hp_before, hp_after - the HP values before and after a change occurs 
        """
        # Smoothstep function
        # See: http://sol.gfxile.net/interpolation/
        smoothstep = lambda v: (v*v*(3-2*v))
        meter_before = int(float(hp_before)*32/float(self.maxHP))
        meter_after = int(float(hp_after)*32/float(self.maxHP))
        delta_meter = meter_after - meter_before
        
        x,y = self.location_pixel
        
        frame_count = 60
        for t in xrange(0,frame_count):
            v = float(t)/float(frame_count)
            v = smoothstep(v)
            scale_term = int(delta_meter*v)
            
            current_percent = meter_before + scale_term
            
            # After the first action
            self.map.render_background()
            self.map.render_all_units()
            self.map.engine.surface.blit(self.map.engine.battle_board,(0,490))
            
            # Draws background of health meter
            self.map.engine.surface.blit(self.map.engine.health_meter,((x,y)-self.map.screen_shift*self.map.engine.tilesize),(35,0,35,35))
            # Draws HP bar
            self.map.engine.surface.blit(self.map.engine.health_meter,((x,y)-self.map.screen_shift*self.map.engine.tilesize),(2,0,current_percent+1,35))
            # Draws foreground of health meter
            self.map.engine.surface.blit(self.map.engine.health_meter,((x,y)-self.map.screen_shift*self.map.engine.tilesize),(0,35,35,35))
            
            pygame.display.flip()
            self.map.engine.clock.tick(60)
        self.map.engine.pause(0.5)


    def render_sc_change(self,sc_before,sc_after):
            """
            # Function Name: render_hp_change
            # Purpose: Draws an HP meter and shows a change in HP
            # Inputs: hp_before, hp_after - the HP values before and after a change occurs
            """
            # Smoothstep function
            # See: http://sol.gfxile.net/interpolation/
            smoothstep = lambda v: (v*v*(3-2*v))
            meter_before = int(float(sc_before)*32/900.0)
            meter_after = int(float(sc_after)*32/900.0)
            delta_meter = meter_after - meter_before


            x,y = self.location_pixel

            frame_count = 60
            for t in xrange(0,frame_count):
                v = float(t)/float(frame_count)
                v = smoothstep(v)
                scale_term = int(delta_meter*v)

                current_percent = meter_before + scale_term

                # After the first action
                self.map.render_background()
                self.map.render_all_units()
                self.map.engine.surface.blit(self.map.engine.battle_board,(0,490))

                # Draws background of health meter
                self.map.engine.surface.blit(self.map.engine.health_meter,((x,y)-self.map.screen_shift*self.map.engine.tilesize),(35,0,35,35))
                # Draws HP bar
                self.map.engine.surface.blit(self.map.engine.health_meter,((x,y)-self.map.screen_shift*self.map.engine.tilesize),(37, 35,current_percent+1,35))
                # Draws foreground of health meter
                self.map.engine.surface.blit(self.map.engine.health_meter,((x,y)-self.map.screen_shift*self.map.engine.tilesize),(0,35,35,35))

                pygame.display.flip()
                self.map.engine.clock.tick(60)
            self.map.engine.pause(0.5)
                        
    def plot_stats(self):

        """
        # Function Name: plot_stats
        # Purpose: Plots the unit's portrait, name, HP, four stats, active spell, and active spell's mods to the
        # left side of the display panel
        """
        
        text_name = self.map.engine.bfont.render(self.name,True, (0,0,0))
        text_line1 = self.map.engine.sfont.render("HP: %3.0f/%3.0f"%(self.HP,self.maxHP),True, (0,0,0))
        text_line2 = self.map.engine.sfont.render("LV: %2.0f    EXP: %3.0f/100"%(self.level,self.exp),True, (0,0,0))
        if self.spell_actions[self.equipped]:
            text_line3 = self.map.engine.bfont.render(("Action: "+self.spell_actions[self.equipped].namesuffix),True,(0,0,0))
        else:
            text_line3 = self.map.engine.bfont.render(("Action: None"),True,(0,0,0))

        text_line4 = self.map.engine.sfont.render("Spirit Charge: %3.0f"%(self.spirit),True, (0,0,0))

        self.map.engine.surface.blit(self.av, (10, 500))
        self.map.engine.surface.blit(text_name, (150,500))
        self.map.engine.surface.blit(text_line1, (150,520))
        self.map.engine.surface.blit(text_line2, (150,540))

        if self.spell_actions[self.equipped]:
            # Attack type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (150,557),self.spell_actions[self.equipped].a_type_big)            
            # Damage type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (175,557),self.spell_actions[self.equipped].d_type_big)
            # Spell Type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (200,557),self.spell_actions[self.equipped].al_type_big)
            # Counterattack
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (225,557),self.spell_actions[self.equipped].c_type_big)
                    
        self.map.engine.surface.blit(text_line3, (150,580))
        self.map.engine.surface.blit(text_line4, (150,600))

    def plot_battle_lhs(self):
        """
        # Function Name: plot_battle_lhs
        # Purpose: Plots the unit's portrait, name, HP, four stats, active spell, and active spell's mods to the
        # left side of the display panel
        """
        
        text_name = self.map.engine.bfont.render(self.name,True, (0,0,0))
        text_line1 = self.map.engine.sfont.render("HP: %3.0f/%3.0f"%(self.HP,self.maxHP),True, (0,0,0))
        text_line2 = self.map.engine.sfont.render("LV: %2.0f    EXP: %3.0f/100"%(self.level,self.exp),True, (0,0,0))
        
        # Trait lines
        if self.traits[0][0]:
            text_line3_str = "Action Trait: "+ self.traits[0][0].name
        else:
            text_line3_str = "Action Trait: None"
            
        if self.traits[1][0]:
            text_line4_str = 'Support Trait: '+self.traits[1][0].name
        else:
            text_line4_str = 'Support Trait: None'
            
        text_line3 = self.map.engine.sfont.render(text_line3_str,True, (0,0,0))
        text_line4 = self.map.engine.sfont.render(text_line4_str,True, (0,0,0))
            
        text_line5 = self.map.engine.sfont.render("Spirit Charge: %3.0f"%(self.spirit),True, (0,0,0))

        if self.spell_actions[self.equipped]:
            # Attack type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (150,552),self.spell_actions[self.equipped].a_type_big)            
            # Damage type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (175,552),self.spell_actions[self.equipped].d_type_big)
            # Spell Type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (200,552),self.spell_actions[self.equipped].al_type_big)
            # Counterattack
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (225,552),self.spell_actions[self.equipped].c_type_big)
            
        self.map.engine.surface.blit(self.av, (10, 500))
        self.map.engine.surface.blit(text_name, (150,500))
        self.map.engine.surface.blit(text_line1, (150,520))
        self.map.engine.surface.blit(text_line2, (150,535))
        self.map.engine.surface.blit(text_line3, (150,575))
        self.map.engine.surface.blit(text_line4, (150,590))
        self.map.engine.surface.blit(text_line5, (150,605))


    def plot_battle_rhs(self):
        """
        # Function Name: plot_stats_rhs
        # Purpose: Plots the unit's portrait, name, HP, four stats, active spell, and active spell's mods to the right
        # side of the display panel
        """

        text_name = self.map.engine.bfont.render(self.name,True, (0,0,0))
        text_line1 = self.map.engine.sfont.render("HP: %3.0f/%3.0f"%(self.HP,self.maxHP),True, (0,0,0))
        text_line2 = self.map.engine.sfont.render("LV: %2.0f    EXP: %3.0f/100"%(self.level,self.exp),True, (0,0,0))
    
        # Trait lines
        if self.traits[0][0]:
            text_line3_str = "Action Trait: "+ self.traits[0][0].name
        else:
            text_line3_str = "Action Trait: None"
            
        if self.traits[1][0]:
            text_line4_str = 'Support Trait: '+self.traits[1][0].name
        else:
            text_line4_str = 'Support Trait: None'
            
        text_line3 = self.map.engine.sfont.render(text_line3_str,True, (0,0,0))
        text_line4 = self.map.engine.sfont.render(text_line4_str,True, (0,0,0))
            
            
        text_line5 = self.map.engine.sfont.render("Spirit Charge: %3.0f"%(self.spirit),True, (0,0,0))

            
        if self.spell_actions[self.equipped]:
            # Attack type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (570,552),self.spell_actions[self.equipped].a_type_big)            
            # Damage type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (595,552),self.spell_actions[self.equipped].d_type_big)
            # Spell Type
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (620,552),self.spell_actions[self.equipped].al_type_big)
            # Counterattack
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (645,552),self.spell_actions[self.equipped].c_type_big)
            
        self.map.engine.surface.blit(self.av, (430, 500))
        self.map.engine.surface.blit(text_name, (570,500))
        self.map.engine.surface.blit(text_line1, (570,520))
        self.map.engine.surface.blit(text_line2, (570,535))
        self.map.engine.surface.blit(text_line3, (570,575))
        self.map.engine.surface.blit(text_line4, (570,590))
        self.map.engine.surface.blit(text_line5, (570,605))

    def plot_predictor(self,target):
        """
        # Function Name: plot_predictor
        # Purpose: Plots the combat stats for self attacking target
        #            Sprites
        #            HP/Level
        #            Spell Name, Rank, Relationships
        # Inputs: target - Target being attacked
        """
        # Asks for the strings for the battle stats    
        if self.spell_actions[self.equipped].type == 'attack':
            percent1,percent2 = self.compute_hitpercent(target)
            crit1,crit2 = self.compute_critpercent(target)
            damage1,damage2 = self.predict_damage(target)
            relation,dmgmod,critmod = self.compute_spell_relation(target)
            
        elif self.spell_actions[self.equipped].type in ('healing','healingitem','support'):
            
            percent1,percent2 = "100", "N/A"
            crit1,crit2 = "N/A","N/A"
            damage1,damage2 = self.predict_damage(target)
            relation = 'Neutral'

        # Generates the text objects
        text_percent1 = self.map.engine.sfont.render(percent1,True, (255,255,255))
        text_percent2 = self.map.engine.sfont.render(percent2,True, (255,255,255))
        text_crit1 = self.map.engine.sfont.render(crit1,True, (255,255,255))
        text_crit2 = self.map.engine.sfont.render(crit2,True, (255,255,255))
        text_damage1 = self.map.engine.sfont.render(damage1,True, (255,255,255))
        text_damage2 = self.map.engine.sfont.render(damage2,True, (255,255,255))
        
           
        text_name = self.map.engine.bfont.render(self.name,True,(0,0,0))
        text_target_name = self.map.engine.bfont.render(target.name,True, (0,0,0))
        
        text_spell_name = self.map.engine.sfont.render(self.spell_actions[self.equipped].namesuffix,True,(0,0,0))
        if target.spell_actions[target.equipped]:
            text_target_spell_name = self.map.engine.sfont.render((target.spell_actions[target.equipped].namesuffix),True,(0,0,0))
            text_target_total_rank = self.map.engine.sfont.render('Spell Rank: '+str(target.skill_rank[target.spell_actions[target.equipped].affinity] + target.spell_actions[target.equipped].spell_rank),True,(0,0,0))
        else:
            text_target_spell_name = self.map.engine.sfont.render("None",True,(0,0,0))
            text_target_total_rank = self.map.engine.sfont.render('Spell Rank: 0',True,(0,0,0))
            
        text_hp = self.map.engine.sfont.render("HP: %3.0f/%3.0f - LV: %2.0f  "%(self.HP,self.maxHP,self.level),True, (0,0,0))
        text_target_hp = self.map.engine.sfont.render("HP: %3.0f/%3.0f - LV: %2.0f  "%(target.HP,target.maxHP,target.level),True, (0,0,0))
        text_spell_type = self.map.engine.sfont.render("Relation:",True,(0,0,0))
        
        text_total_rank = self.map.engine.sfont.render('Spell Rank: '+str(self.skill_rank[self.spell_actions[self.equipped].affinity] + self.spell_actions[self.equipped].spell_rank),True,(0,0,0))
        
        # Plots names
        self.map.engine.surface.blit(text_name,(440,500))
        
        # Plots sprites
        self.map.engine.surface.blit(self.image, (440,520),(0,0,35,35))
        
        # ATTACK TYPE
        self.map.engine.surface.blit(self.map.engine.spell_icons, (480,530),self.spell_actions[self.equipped].a_type_small)  
        # DAMAGE TYPE
        self.map.engine.surface.blit(self.map.engine.spell_icons, (495,530),self.spell_actions[self.equipped].d_type_small)
        # COUNTERATTACK
        self.map.engine.surface.blit(self.map.engine.spell_icons, (510,530),self.spell_actions[self.equipped].al_type_small)
        # COUNTERATTACK
        self.map.engine.surface.blit(self.map.engine.spell_icons, (525,530),self.spell_actions[self.equipped].c_type_small)
        
        self.map.engine.surface.blit(text_hp, (440,555))
        self.map.engine.surface.blit(text_spell_name, (440,570))    
        self.map.engine.surface.blit(text_total_rank,(440,587))
        
        
        if target is not self:
            
            self.map.engine.surface.blit(text_target_name,(690,500))
            self.map.engine.surface.blit(target.image, (690,520),(0,0,35,35))
            
            if target.spell_actions[target.equipped]:
                # Plots enemy spell properties
                self.map.engine.surface.blit(self.map.engine.spell_icons, (730,530),target.spell_actions[target.equipped].a_type_small)  
                # DAMAGE TYPE
                self.map.engine.surface.blit(self.map.engine.spell_icons, (745,530),target.spell_actions[target.equipped].d_type_small)
                # TYPE
                self.map.engine.surface.blit(self.map.engine.spell_icons, (760,530),target.spell_actions[target.equipped].al_type_small)
                # COUNTERATTACK
                self.map.engine.surface.blit(self.map.engine.spell_icons, (775,530),target.spell_actions[target.equipped].c_type_small)
    
            
            self.map.engine.surface.blit(text_target_hp, (690,555))
              
            self.map.engine.surface.blit(text_target_spell_name, (690,570)) 
            self.map.engine.surface.blit(text_target_total_rank,(690,587))
            
            
            # Plots the relations
            self.map.engine.surface.blit(text_spell_type,(440,605))
            self.map.engine.surface.blit(text_spell_type,(690,605))
            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (445+text_spell_type.get_width(),602),self.spell_actions[self.equipped].al_type_big)
            if target.spell_actions[target.equipped]:
                self.map.engine.surface.blit(self.map.engine.spell_icons_big, (695+text_spell_type.get_width(),602),target.spell_actions[target.equipped].al_type_big)
            
            
            if relation == 'Neutral' or self.spell_actions[self.equipped].type in ('healing','healingitem','support'):
                self.map.engine.surface.blit(self.map.engine.relation_arrows['neutral'],(465+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['neutral'],(715+text_spell_type.get_width(),602))
            elif relation == 'Strong':
                self.map.engine.surface.blit(self.map.engine.relation_arrows['up'],(465+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['down'],(715+text_spell_type.get_width(),602))
            elif relation == 'Weak':
                self.map.engine.surface.blit(self.map.engine.relation_arrows['down'],(465+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['up'],(715+text_spell_type.get_width(),602))
            elif relation == 'Very Strong':
                self.map.engine.surface.blit(self.map.engine.relation_arrows['up'],(465+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['up'],(483+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['down'],(715+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['down'],(733+text_spell_type.get_width(),602))
            elif relation == 'Very Weak':
                self.map.engine.surface.blit(self.map.engine.relation_arrows['down'],(465+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['down'],(483+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['up'],(715+text_spell_type.get_width(),602))
                self.map.engine.surface.blit(self.map.engine.relation_arrows['up'],(733+text_spell_type.get_width(),602))
            
        # Plots the center stat systems
        self.map.engine.surface.blit(self.map.engine.combatstats,(570,545))

                
        # Plots percent hit
        self.map.engine.surface.blit(text_percent1,(577,560))
        self.map.engine.surface.blit(text_percent2, (577,595))
        # Plots percent critical
        self.map.engine.surface.blit(text_crit1,(612,560))
        self.map.engine.surface.blit(text_crit2, (612,595))
        # Plots damage
        self.map.engine.surface.blit(text_damage1,(647,560))
        self.map.engine.surface.blit(text_damage2, (647,595))

        
        

    def plot_results(self,enemy,self_exp_delta,self_level_up,
                     enemy_exp_delta,enemy_level_up,self_spirit_delta,enemy_spirit_delta,
                     sc_cost_user, sc_cost_target):
        
        """
        # Function Name: plot_results
        # Purpose: Displays the battle's results
        # Inputs:   enemy = The opponent
        #           damage = How much damage the unit has taken, or if the unit was missed, the text "miss"
        #           self_exp_delta, enemy_exp_delta = experience points gained          
        #           self_spirit_delta, enemy_spirit_delta = spirit charge change
        #           sc_cost_user, sc_cost_target = spirit charge cost for spells
        """
        
        menu_flag = True

        text_results = self.map.engine.bfont.render("Results",True, (0,0,0))
        text_levelup = self.map.engine.bfont.render("Level up!",True, (0,0,0))
        text_instruct = self.map.engine.sfont.render("Press Z to Continue",True, (0,0,0))
        text_self_name = self.map.engine.bfont.render(self.name,True, (0,0,0))
        text_enemy_name = self.map.engine.bfont.render(enemy.name,True, (0,0,0))

        text_self_exp = self.map.engine.sfont.render("EXP: +"+str(self_exp_delta)+" ; "+str(self.exp)+"/100",True, (0,0,0))
        text_enemy_exp = self.map.engine.sfont.render("EXP: +"+str(enemy_exp_delta)+" ; "+str(enemy.exp)+"/100",True, (0,0,0))

        text_self_sc_cost = self.map.engine.sfont.render("Spell SC Cost: "+str(sc_cost_user),True, (0,0,0))
        text_enemy_sc_cost = self.map.engine.sfont.render("Spell SC Cost: "+str(sc_cost_target),True, (0,0,0))
        
        if self_spirit_delta >= 0:
            text_self_spirit_delta = self.map.engine.sfont.render("Battle SC Change: +%s"%(str(self_spirit_delta)),True,(0,0,0))
        else:
            text_self_spirit_delta = self.map.engine.sfont.render("Battle SC Change:  %s"%(str(self_spirit_delta)),True,(0,0,0))

        if enemy_spirit_delta >= 0:
            text_enemy_spirit_delta = self.map.engine.sfont.render("Battle SC Change: +%s"%(str(enemy_spirit_delta)),True,(0,0,0))
        else:
            text_enemy_spirit_delta = self.map.engine.sfont.render("Battle SC Change:  %s"%(str(enemy_spirit_delta)),True,(0,0,0))

        text_self_spirit_final =  self.map.engine.sfont.render("Final SC:  %s"%(str(self.spirit)),True,(0,0,0))
        text_enemy_spirit_final =  self.map.engine.sfont.render("Final SC:  %s"%(str(enemy.spirit)),True,(0,0,0))
        
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN and (event.key == K_z  or event.key == K_RETURN):
                    menu_flag = False
            if menu_flag:
                self.map.render_background()
                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                self.plot_stats()
                self.map.engine.surface.blit(self.map.engine.results_panel,(280,140))

                
                self.map.engine.surface.blit(text_results,(390,150))

                # Displays the self results
                self.map.engine.surface.blit(text_self_name,(305,170))
                self.map.engine.surface.blit(self.image, (305,190),(0,0,35,35))
                self.map.engine.surface.blit(text_self_exp,(380,195))
                self.map.engine.surface.blit(text_self_sc_cost,(380,210))
                self.map.engine.surface.blit(text_self_spirit_delta,(380,225))
                self.map.engine.surface.blit(text_self_spirit_final,(380,240))
                if self_level_up == True:
                    self.map.engine.surface.blit(text_levelup,(380,255))
                
                # Displays the enemy results
                if enemy != self:
                    self.map.engine.surface.blit(text_enemy_name,(305,290))
                    self.map.engine.surface.blit(enemy.image, (305,310),(0,0,35,35))
                    self.map.engine.surface.blit(text_enemy_exp,(380,315))
                    self.map.engine.surface.blit(text_enemy_sc_cost,(380,330))
                    self.map.engine.surface.blit(text_enemy_spirit_delta,(380,345))
                    self.map.engine.surface.blit(text_enemy_spirit_final,(380,360))
                    
                    if enemy_level_up == True:
                        self.map.engine.surface.blit(text_levelup,(380,375))


                self.map.engine.surface.blit(text_instruct,(420,395))
            pygame.display.flip()
            self.map.engine.clock.tick(60)
        

       
    def post_battle_event(self,target):
        '''
        # Function Name: Post-battle Event
        # Purpose: Processes post battle stats updates like lowering spell use counts
        # Inputs: Target - other participant in the battle
        '''
        
        # Lowers life of the spell action by 1
        self.spell_actions[self.equipped].livesleft -= 1

        # Takes off SC cost of the spell action
        if self.spell_actions[self.equipped].sc_cost > 0:
            self.spirit -= self.spell_actions[self.equipped].sc_cost
            if self.spirit < 0:
                self.spirit = 0
            
        # If lives left are 0, replaces the equipped slot with empty
        if self.spell_actions[self.equipped].livesleft == 0:
            self.map.display_alert('Spell Action Used Up!',"%s's %s has run out of uses!"%(self.name,self.spell_actions[self.equipped].namesuffix))
            if self.spell_actions[self.equipped].consumable == True:
                self.spell_actions[self.equipped] = None
        
        
        # Checks the opponent if they have an equipped spell and if it was used in a counterattack
        delta_pos = self.location_tile - target.location_tile
        if (target.spell_actions[target.equipped] and target.alive == True
           and target.spell_actions[target.equipped].livesleft > 0
           and target.spell_actions[target.equipped].counterattack == True
           and tuple(delta_pos) in target.spell_actions[target.equipped].validattacks
           and target.spirit >= target.spell_actions[target.equipped].unlock):
            # Lowers life of the spell action by 1
            target.spell_actions[target.equipped].livesleft -= 1
            
                
            # Takes off SC cost of the spell action
            if target.spell_actions[target.equipped].sc_cost > 0:
                target.spirit -= target.spell_actions[target.equipped].sc_cost
                if target.spirit < 0:
                    target.spirit = 0
                
            # If lives left are 0, replaces the equipped slot with empty
            if target.spell_actions[target.equipped].livesleft == 0:
                self.map.display_alert('Spell Action Used Up!',"%s's %s has run out of uses!"%(target.name,target.spell_actions[target.equipped].namesuffix))
                if target.spell_actions[target.equipped].consumable == True:
                    target.spell_actions[target.equipped] = None
        
    def can_counterattack(self,target):
        """
        # Fuction name: can_counterattack
        # Purpose: checks if a unit is able to counterattack (spell permits / status effects permits
        # Outputs:    True if unit is able to counter
        #             False if unit is not able to counter
        """
        # Spell Conditions
        #    - A spell is equipped
        #    - A spell has enough uses left
        #    - Spell can counterattack
        #    - SC requirements are met
        if self.spell_actions[self.equipped]: 
            can_use_spell = self.spell_actions[self.equipped].livesleft > 0 and self.spell_actions[self.equipped].counterattack == True and self.spirit >= self.spell_actions[self.equipped].unlock
        
            # Target is in range
            delta_pos = self.location_tile - target.location_tile
            target_in_range =  tuple(delta_pos) in self.spell_actions[self.equipped].validattacks
            
        else:
            can_use_spell = False
            target_in_range = False
        
        # Does not have any blocking status effects
        blocking_status_list = ['Stun']
        can_attack = True
        # If unit has at least one bocking status, unit cannot attack
        for status in blocking_status_list:
            if status in self.status.keys():
                can_attack = False

        return self.alive and can_use_spell and target_in_range and can_attack
       
    #############################
    # Character Event Methods
    #############################
    
    def map_heal(self,target,action,effect=0,bullet_id='Auto'):
    
        """
        # Function name: map_heal
        # Purpose: Plots an animation for a character healing herself. Also applicable for healingitems
        # Inputs: 
        #       target - target to show effect on
        #       action - name of healing action
        #       effect - How much HP is healed (0 if it is relying on a spell/item)
        #       bullet_id - bullet to use for animation
        """
        
        # Generates name of the spell at the top
        text_name = self.map.engine.bfont.render(action,True,(0,0,0))
        half_width = text_name.get_width()/2
    
        spell_used = False
        
        if effect == 0:
            # Designates that a spell was used in this sequence
            spell_used = True
            effect = self.spell_actions[self.equipped].get_effect(self,target)
            sc_cost = self.spell_actions[self.equipped].sc_cost
            self.spell_actions[self.equipped].livesleft -= 1
        else:
            sc_cost = 0
            
    
        # Origin
        x_o = (target.location_pixel.x)
        y_o = (target.location_pixel.y)
    
        effect_text = self.map.engine.render_outlined_text(str(effect),self.map.engine.cfont,(100,255,100),(0,0,0))

        # Play the sound effect of a healing item
        self.map.engine.sfx_system.sound_catalog['item'].play()
        
        # Shows animation if animation option is set to true
        if self.map.engine.options.battle_anim == True:
            # Shows the equipped bullet flying in a circle around the user
            # Uses parametric equation for a circle to animate
            for i in xrange(0,21):
        
                for event in pygame.event.get():
                    if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                        exit()
                
                self.map.render_background()
                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                self.map.engine.surface.blit(self.map.engine.map_spell_board,(175,0))
                self.map.engine.surface.blit(text_name,(420-half_width,25))
                self.plot_stats()
            
                x = x_o+30*sin(i*pi/5)
                y = y_o+30*cos(i*pi/5)
                
                self.map.engine.surface.blit(self.map.engine.heal_bullet,((x,y)-self.map.screen_shift*self.map.engine.tilesize))
                pygame.display.flip()
                self.map.engine.clock.tick(60)
        else:
        
            self.map.render_background()
            self.map.render_all_units()
            self.map.render_cursor()
            self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
            self.map.engine.surface.blit(self.map.engine.map_spell_board,(175,0))
            self.map.engine.surface.blit(text_name,(420-half_width,25))
            self.plot_stats()
            self.map.engine.pause(1)
            
        rhs_hp_before = target.HP
        target.HP += effect
        
        if target.HP > target.maxHP:
            target.HP = target.maxHP
            
        
        # Only renders the value of the healing if the effect is nonzero
        if effect > 0:
            self.map.render_background()
            self.map.render_all_units()
            self.map.render_cursor()
            self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
            self.map.engine.surface.blit(self.map.engine.map_spell_board,(175,0))
            self.map.engine.surface.blit(text_name,(420-half_width,25))
            self.plot_stats()
            self.map.engine.surface.blit(effect_text,((target.location_pixel.x+18-effect_text.get_width()/2,target.location_pixel.y-25)-self.map.screen_shift*self.map.engine.tilesize))
            
            pygame.display.flip()
            self.map.engine.clock.tick(60)
            self.map.engine.pause(1)
                
            rhs_hp_after = target.HP
            target.render_hp_change(rhs_hp_before, rhs_hp_after)

        if spell_used == True:
            # If lives left are 0, replaces the equipped slot with empty
            if self.spell_actions[self.equipped].livesleft == 0:
                self.map.display_alert('Spell Action Used Up!',"%s's %s has run out of uses!"%(self.name,self.spell_actions[self.equipped].namesuffix))
                if self.spell_actions[self.equipped].consumable == True:
                    self.spell_actions[self.equipped] = None
        
        return effect,sc_cost
    
    #############################
    # Interaction Methods
    #############################
    
    
    def move_to(self,destination):
        """
        # Function Name: move_to
        # Purpose: Move a unit from one place to another. Used in scripted movements
        # Inputs: destination - destination to move a unit to
        """
        
        path = []
        # Intermediate step current position
        step_pos = self.location_tile
        delta = Vector2(destination) - self.location_tile

        # Gets all the moves in X
        while delta.x != 0:
            if delta.x > 0:
                path.append(Vector2(1,0))
                step_pos.x += 1
            elif delta.x < 0:
                step_pos.x -= 1
                path.append(Vector2(-1,0))
            delta =  Vector2(destination) - step_pos
        # gets all the moves in Y
        while delta.y != 0:
            if delta.y > 0:
                step_pos.y += 1
                path.append(Vector2(0,1))
            elif delta.y < 0:
                step_pos.y -= 1
                path.append(Vector2(0,-1))
            delta = Vector2(destination) - step_pos 
        self.render_walk(path)
        
        self.update_location(*destination)
    

    def menu_loop(self):
        
        """
        # Function Name: menu_loop
        # Purpose: The unit's first level menu:
        #       Allows selection of: Move, Spell, traits Stats, Wait, and Cancel if possible
        """
        
        menu_flag = True
        menu_pos = 1


        # Conditions for moving: Has not moved yet, has not used a spell yet, is alive
        
        if self.moved == False and self.turnend == False and self.alive == True:
            text_move = self.map.engine.sfont.render("MOVE",True, (0,0,0))
        else:
            text_move = self.map.engine.sfont.render("MOVE",True, (100,100,100))

        # Conditions for using a SC or Wait: Has not used a spell yet, is alive

        if self.turnend == False and self.alive == True:
            text_action = self.map.engine.sfont.render("SPELL ACTIONS",True, (0,0,0))
            text_wait = self.map.engine.sfont.render("WAIT",True, (0,0,0))
            text_trait = self.map.engine.sfont.render("TRAITS",True, (0,0,0))
        else:
            text_action = self.map.engine.sfont.render("SPELL ACTIONS",True, (100,100,100))        
            text_wait = self.map.engine.sfont.render("WAIT",True, (100,100,100))
            text_trait = self.map.engine.sfont.render("TRAITS",True, (100,100,100))
            
        
        text_stats = self.map.engine.sfont.render("STATS",True, (0,0,0))
        text_cancel = self.map.engine.sfont.render("CANCEL",True, (0,0,0))

        while menu_flag:

            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # if either the Up or Right keys are pressed, and the menu position
                    # is not at the top, move the menu position up one.
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 1:
                            menu_pos -= 1
                        elif menu_pos == 1:
                            menu_pos = 6
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 6:
                            menu_pos += 1
                        elif menu_pos == 6:
                            menu_pos = 1
                            
                    if ( event.key == K_z or event.key == K_RETURN):

                        # This option sends the user to the character's move select loop
                        # Conditions for moving: Has not moved yet, has not used a spell yet, is alive
                        if menu_pos == 1 and self.turnend == False and self.alive == True:
                            
                            confirmedmove = False
                            confirmedmove = self.move_loop()
                            # updates location after a move has been made
                            if self.turnend == True:
                                self.map.get_all_moves()
                                text_move = self.map.engine.sfont.render("MOVE",True, (100,100,100))
                                text_action = self.map.engine.sfont.render("SPELLS",True, (100,100,100))
                                text_wait = self.map.engine.sfont.render("WAIT",True, (100,100,100))
                                menu_flag = False


                        # This option sends the user to the character's spell select loop
                        # Conditions for using an action: Has not used a spell yet, is alive
                        if menu_pos == 2 and self.turnend == False and self.alive == True:
                            confirmedattack = False
                            confirmedattack = self.spell_loop_a()
                            if confirmedattack == True or self.turnend == True:
                                self.map.get_all_moves()
                                self.turnend = True
                                text_move = self.map.engine.sfont.render("MOVE",True, (100,100,100))
                                text_action = self.map.engine.sfont.render("SPELLS",True, (100,100,100))
                                text_wait = self.map.engine.sfont.render("WAIT",True, (100,100,100))
                                text_trait = self.map.engine.sfont.render("TRAITS",True, (100,100,100))
                                menu_flag = False

                        # This option sends the user to the character's trait switching loop
                        # Conditions for swapping trait: Traits may be swapped any time before a unit commits to an action
                        if menu_pos == 3 and self.turnend == False and self.alive == True:
                            self.trait_loop()
                     
                        if menu_pos == 4:
                            self.stats_loop()
                     
                        # This option ends the character's turn
                        if menu_pos == 5 and self.turnend == False and self.alive == True:
                            self.turnend = True
                            text_move = self.map.engine.sfont.render("MOVE",True, (100,100,100))
                            text_action = self.map.engine.sfont.render("SPELLS",True, (100,100,100))
                            text_wait = self.map.engine.sfont.render("WAIT",True, (100,100,100))
                            menu_flag = False


                        # This option cancels without ending the turn
                        if menu_pos == 6:
                            menu_flag = False
                    if ( event.key == K_x ):
                        menu_flag = False

            if menu_flag:
                    
                self.map.render_background()
                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                           %(self.map.cursor_pos.x,self.map.cursor_pos.y,self.map.screen_shift.x,self.map.screen_shift.y))
        
                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected = False
                selected = self.map.cursor_key_search()
                if selected is not False:
                    self.map.all_units_by_name[selected].plot_stats()
                    
                pygame.draw.circle(self.map.engine.surface, (0,0,0), (440,498+15*menu_pos),5)
                self.map.engine.surface.blit(text_move, (450,505))
                self.map.engine.surface.blit(text_action, (450,520))
                self.map.engine.surface.blit(text_trait, (450,535))
                self.map.engine.surface.blit(text_stats, (450,550))
                self.map.engine.surface.blit(text_wait, (450,565))
                self.map.engine.surface.blit(text_cancel, (450,580))
    
                pygame.display.flip()
                self.map.engine.clock.tick(60)
    
        
    def trait_loop(self):
        """
        # Function Name: trait_loop
        # Purpose: The Unit's traits menu
        #      Allows for the selection of first trait to swap.
        """
        
        text_support = self.map.engine.bfont.render("Support Trait", True, (0,0,0))
        text_action = self.map.engine.bfont.render("Action Trait",True,(0,0,0))
        text_standby = self.map.engine.bfont.render("Standby Traits",True, (0,0,0))

        def get_trait(self):
            """
            # Function Name: get_trait
            # Purpose: Generates the traits text objects
            """
            
            text_trait = [[],[]]
            for i in (0,1):
                for trait in self.traits[i]:
                    if trait:
                        text_trait[i].append(self.map.engine.sfont.render(trait.name,True,(0,0,0)))
                    else:
                        text_trait[i].append(self.map.engine.sfont.render('EMPTY',True,(0,0,0)))
            return text_trait
        
        text_trait = get_trait(self)
        menu_flag = True
        menu_pos = [0,0]

        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # Moves the selection cursor around
                    if event.key == K_UP:
                        if menu_pos[1] > 0:
                            menu_pos[1] -= 1
                        elif menu_pos[1] == 0:
                            menu_pos[1] = 3
                    if event.key == K_DOWN:
                        # Accounts for save / cancel options
                        if menu_pos[1] < 3:
                            menu_pos[1] += 1
                        else:
                            menu_pos[1] = 0
                    if event.key == K_RIGHT or event.key == K_LEFT:
                        if menu_pos[0] == 1:
                            menu_pos[0] = 0
                        elif menu_pos[0] == 0:
                            menu_pos[0] = 1
                            if menu_pos[1]>3:
                                menu_pos[1] = 3
                            
                    if event.key == K_z or event.key == K_RETURN:
                        self.swap_trait(menu_pos)
                    
                    if event.key == K_x:
                        menu_flag = False
                        # Recomputes valid moves
                        self.get_moves_path()

                        
            if menu_flag:
                
                text_trait = get_trait(self)
                self.map.render_background()
                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                self.plot_stats()
                
                self.map.engine.surface.blit(text_standby,(450,540))
                for list_index, trait_list in enumerate(text_trait):
                    for trait_index, trait_entry in enumerate(trait_list[1:]):
                        self.map.engine.surface.blit(trait_entry, (480+140*list_index, 560+15*trait_index))

                # Trait Type Icons
                for i in xrange(0,4):
                    self.map.engine.surface.blit(self.map.engine.spell_icons, (460,560+i*15),(105,0,15,15))
                    self.map.engine.surface.blit(self.map.engine.spell_icons, (600,560+i*15),(135,0,15,15))

                # Active Traits
                # Action
                self.map.engine.surface.blit(text_action,(450,503))
                self.map.engine.surface.blit(self.map.engine.spell_icons, (460,522),(105,0,15,15))
                self.map.engine.surface.blit(text_trait[0][0],(480,522))
                # Support
                self.map.engine.surface.blit(text_support,(590,503))
                self.map.engine.surface.blit(self.map.engine.spell_icons, (600,522),(135,0,15,15))
                self.map.engine.surface.blit(text_trait[1][0],(620,522))

                pygame.draw.circle(self.map.engine.surface, (0,0,0), (453+140*menu_pos[0],568+15*menu_pos[1]),5)
                
                pygame.display.flip()
                self.map.engine.clock.tick(60)
                
                
    
    def move_loop(self):
    
        """
        # Function Name: move_loop
        # Purpose: The unit's second level move loop:
        #      Allows the player to select where they want to move a unit
        # Output: moved_flag = True if a selection has been made, False if one has not been made
        """
        
        self.start_pos = tuple(self.location_tile)
        self.get_moves_path()
        menu_flag = True
        self.map.framenum = 0 
        while menu_flag:
            # Frame counter for holding down the keys to move
            if self.map.framenum == 10:
                self.map.framenum = 0
            self.map.framenum += 1
            
            arrowkeys = False
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    
                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        self.map.cursor_arrows(event)
                        arrowkeys = True
                        # Resets the frame counter
                        self.map.framenum = 0

                    if event.key == K_x:
                        menu_flag = False
                        moved_flag = False
                        
                        self.map.center_on(self)
                        
                    if event.key == K_z or event.key == K_RETURN:

                        # If the Z key is pressed, it will calculate the vector between the current position
                        # of the cursor and the unit's location. If that vector is part of the valid moves,
                        # the unit is moved there.
                        delta_pos = self.map.cursor_pos - self.location_tile
                        new_pos = self.map.cursor_pos
                        if self.validmoves.has_key(tuple(delta_pos)):
                            path = self.get_path(tuple(delta_pos))
                            if new_pos != self.location_tile:
                                self.render_walk(path)
                            self.update_location(new_pos.x, new_pos.y)
                            self.spell_loop_a()
                            
                            if self.turnend:
                                menu_flag = False
                            else:
                                self.update_location(self.start_pos[0],self.start_pos[1])
                            
                            self.map.cursor_pos = Vector2(tuple(self.location_tile))
                                

                
            # if there is not a tap detected, check if the key is being held down
            if arrowkeys == False and self.map.framenum == 9:
                self.map.cursor_arrows_hold()

            
            if menu_flag:
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                       %(self.map.cursor_pos.x,self.map.cursor_pos.y,self.map.screen_shift.x,self.map.screen_shift.y))
    
                self.map.render_background()
            
                self.plot_valid_moves()
    
                self.map.render_all_units()
                self.map.render_cursor()
                        
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
            
                text_from = self.map.engine.bfont.render(self.map.terrainmap[tuple(self.location_tile)][0].name,True, (0,0,0))
                text_from_def = self.map.engine.sfont.render("TDEF: %2.1f"%self.map.terrainmap[tuple(self.location_tile)][0].TDEF,True, (0,0,0))
                text_from_agl = self.map.engine.sfont.render("AGL: "+str(self.map.terrainmap[tuple(self.location_tile)][0].aglmod),True, (0,0,0))
                text_from_acc = self.map.engine.sfont.render("ACC: "+str(self.map.terrainmap[tuple(self.location_tile)][0].accmod),True, (0,0,0))
              
                        
                text_to = self.map.engine.bfont.render(self.map.terrainmap[tuple(self.map.cursor_pos)][0].name,True, (0,0,0))
                text_to_def = self.map.engine.sfont.render("TDEF: %2.1f"%self.map.terrainmap[tuple(self.map.cursor_pos)][0].TDEF,True, (0,0,0))
                text_to_agl = self.map.engine.sfont.render("AGL: "+str(self.map.terrainmap[tuple(self.map.cursor_pos)][0].aglmod),True, (0,0,0))
                text_to_acc = self.map.engine.sfont.render("ACC: "+str(self.map.terrainmap[tuple(self.map.cursor_pos)][0].accmod),True, (0,0,0))
                
                # Plots the terrain data
                self.map.engine.surface.blit(text_from,(450,505))
                self.map.engine.surface.blit(self.map.engine.terrain_icon,(450,523),(self.map.terrainmap[tuple(self.location_tile)][0].icon*50,0,50,50))
                self.map.engine.surface.blit(text_from_def,(450,575))
                self.map.engine.surface.blit(text_from_acc,(450,587))
                self.map.engine.surface.blit(text_from_agl,(450,599))
                            
                self.map.engine.surface.blit(self.map.engine.move_arrow,(565,523))
                            
                self.map.engine.surface.blit(text_to,(810-text_to.get_width(),505))
                self.map.engine.surface.blit(self.map.engine.terrain_icon,(760,523),(self.map.terrainmap[tuple(self.map.cursor_pos)][0].icon*50,0,50,50))
                self.map.engine.surface.blit(text_to_def,(810-text_to_def.get_width(),575))
                self.map.engine.surface.blit(text_to_acc,(810-text_to_acc.get_width(),587))
                self.map.engine.surface.blit(text_to_agl,(810-text_to_agl.get_width(),599))
        
                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected = False
                selected = self.map.cursor_key_search()
                if selected is not False:
                    self.map.all_units_by_name[selected].plot_stats()
    
                pygame.display.flip()
                self.map.engine.clock.tick(60)

    def spell_loop_a(self):
        """
        # Function Name: Spell Loop A
        # Purpose: The unit's second level spell loop:
        #      Allows the player to select a spell to use
        # Output: attacked_flag = True if an attack has been made, False if one has not been made
        """
        
        menu_title_text = self.map.engine.sfont.render('Select Action',True, (0,0,0))
        spell_text = []
        menu_flag = True
        menu_pos = 0
        attacked_flag = False
        
        text_header_uses = self.map.engine.sfont.render('Uses',True,(0,0,0))
        text_header_sc = self.map.engine.sfont.render('Min.SC',True,(0,0,0))
        text_header_sc_cost = self.map.engine.sfont.render('SC Cost',True,(0,0,0))
        
        # Generates text strings for all spells in the inventory
        for spell_id in xrange(0,5):
            if self.spell_actions[spell_id]:
                    # If there are no uses left, changes the text to red
                if self.spell_actions[spell_id].livesleft == 0:
                    color = (255,0,0)
                # If the character has insufficient spirit charge, change text to blue
                elif self.spell_actions[spell_id].unlock > self.spirit:
                    color = (0,0,255)
                # If the character can use the attack, change text to black
                else:
                    color = (0,0,0)

                spell_text.append({'spell':True,
                                   'name':self.map.engine.sfont.render(self.spell_actions[spell_id].namesuffix,True,color),
                                   'uses':self.map.engine.sfont.render("%s/%s"%(str(self.spell_actions[spell_id].livesleft),str(self.spell_actions[spell_id].lives)),True,color),
                                   'sc_min':self.map.engine.sfont.render(str(self.spell_actions[spell_id].unlock),True,color),
                                   'sc_cost':self.map.engine.sfont.render(str(self.spell_actions[spell_id].sc_cost),True,color)})
            # no Spell available
            else:
                spell_text.append({'spell':False,'name':self.map.engine.sfont.render('Empty',True, (0,0,0))})
            
                    
        spell_text.append(self.map.engine.sfont.render('WAIT',True, (0,0,0)))
        spell_text.append(self.map.engine.sfont.render('CANCEL',True, (0,0,0)))
        
        while menu_flag:
    
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # if either the Up or Right keys are pressed, and the menu position
                    # is not at the top, move the menu position up one.
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 6
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 6:
                            menu_pos += 1
                        elif menu_pos == 6:
                            menu_pos = 0
                    if  event.key == K_z  or event.key == K_RETURN:

                        # Conditions for selecting a spellcard:
                        #   -Position of the menu is between 1 and 5,
                        #   -the spell does not correspond to an empty slot, and the number of uses left is > 0.
                        #   -the spell's spirit charge unlock is lower than or equal to the user's spirit charge
                        if menu_pos >=0 and menu_pos <=4 and self.spell_actions[menu_pos] and self.spell_actions[menu_pos].livesleft > 0 and self.spirit >= self.spell_actions[menu_pos].unlock:
                            self.equipped = menu_pos
                            
                            self.get_valid_spell_range()
                            self.spell_loop_b()
                            if self.turnend == True:
                                menu_flag = False
                        if menu_pos == 5:
                            menu_flag = False
                            self.turnend = True

                        if menu_pos == 6:
                            menu_flag = False
                    if  event.key == K_x :
                        menu_flag = False



            if menu_flag:
                
                self.map.render_background()
                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                self.plot_stats()
                
                pygame.draw.circle(self.map.engine.surface, (0,0,0), (440,523+15*menu_pos),5)
                self.map.engine.surface.blit(text_header_uses,(600,500))
                self.map.engine.surface.blit(text_header_sc,(650,500))
                self.map.engine.surface.blit(text_header_sc_cost,(700,500))
                
                for index,displayed_text in enumerate(spell_text[0:5]):
                    self.map.engine.surface.blit(displayed_text['name'],(450,515+index*15))
                    # Blits other data only if valid
                    if displayed_text['spell']:
                        self.map.engine.surface.blit(displayed_text['uses'],(600,515+index*15))
                        self.map.engine.surface.blit(displayed_text['sc_min'],(650,515+index*15))
                        self.map.engine.surface.blit(displayed_text['sc_cost'],(700,515+index*15))

                self.map.engine.surface.blit(spell_text[5], (450,590))
                self.map.engine.surface.blit(spell_text[6], (450,605))

                # SPELL SELECTION ICONS
                for i in xrange(0,5):
                    if self.spell_actions[i]:
                        # ATTACK TYPE
                        self.map.engine.surface.blit(self.map.engine.spell_icons, (755,515+15*(i)),self.spell_actions[i].a_type_small)  
                        # DAMAGE TYPE
                        self.map.engine.surface.blit(self.map.engine.spell_icons, (770,515+15*(i)),self.spell_actions[i].d_type_small)
                        # SPELL TYPE
                        self.map.engine.surface.blit(self.map.engine.spell_icons, (785,515+15*(i)),self.spell_actions[i].al_type_small)
                        # COUNTERATTACK
                        self.map.engine.surface.blit(self.map.engine.spell_icons, (800,515+15*(i)),self.spell_actions[i].c_type_small)
                  
            pygame.display.flip()
            self.map.engine.clock.tick(60)

    def spell_loop_b(self):
        """
        # Function Name: Spell Loop B
        # Purpose: The unit's third level spell loop:
        #      Allows the player to select a target to use a spell on
        # Output: attacked_flag = True if an attack has been made, False if one has not been made
        """
        
        menu_flag = True
        attacked_flag = False
        self.map.framenum = 0
        
        while menu_flag:
            
            arrowkeys = False      
            if self.map.framenum == 10:
                self.map.framenum = 0
            self.map.framenum += 1
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                        self.map.cursor_arrows(event)
                        arrowkeys = True
                        # Resets the frame counter
                        self.map.framenum = 0
                    if event.key == K_x:
                        menu_flag = False
                        self.map.center_on(self)
                    if event.key == K_z or event.key == K_RETURN:

                        # If the Z key is pressed, it will calculate the vector between the current position
                        # of the cursor and the unit's location. If that vector is part of the valid moves,
                        # the unit is moved there.
                        delta_pos = self.map.cursor_pos - self.location_tile
                        
                        if tuple(delta_pos) in self.spell_actions[self.equipped].validattacks:

                            selected = self.map.cursor_key_search()

                            # If a target is found
                            if selected is not False:
                                if self.spell_actions[self.equipped].type == 'attack':
                                    # Conditions for attack type cards:
                                    # A unit is under the cursor, the target unit is not on your team, and the target is alive
                                    if self.map.all_units_by_name[selected].team != self.team and self.map.all_units_by_name[selected].alive == True:
                                        self.spell_loop_c(selected)
                                        if self.turnend == True:
                                            menu_flag = False
                                            self.map.cursor_pos = Vector2(tuple(self.location_tile))
                                elif self.spell_actions[self.equipped].type in ('healing','healingitem','support'):
                                    # Conditions for healing type cards:
                                    # A unit is under the cursor, the target unit is on your team, the target is alive, and the [target's HP != its max HP or healing spell cures status effects]
                                    if self.map.all_units_by_name[selected].team == self.team and self.map.all_units_by_name[selected].alive == True and (self.map.all_units_by_name[selected].HP != self.map.all_units_by_name[selected].maxHP or self.spell_actions[self.equipped].status_effects):
                                        self.spell_loop_c(selected)
                                        if self.turnend == True:
                                            menu_flag = False
                                            self.map.cursor_pos = Vector2(tuple(self.location_tile))
                                            


            # if there is not a tap detected, check if the key is being held down
            if arrowkeys == False and self.map.framenum == 9:
                self.map.cursor_arrows_hold()
            
            # Plots the valid attack tiles
            if menu_flag:
            
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                           %(self.map.cursor_pos.x,self.map.cursor_pos.y,self.map.screen_shift.x,self.map.screen_shift.y))
        
                self.map.render_background()
                self.plot_attacks()
                        
                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                self.plot_stats()

                # Display for combat statistics
                # Checks if there is a unit underneath the cursor, and if there is, plots the unit's data.
                selected_stats = False
                selected_stats = self.map.cursor_key_search()
                # Displays if there is a unit under your cursor, it is not on your team for attack class spells and or on your team for healing-class spells.
                if (selected_stats is not False and
                     ((self.map.all_units_by_name[selected_stats].team != self.team and
                        self.spell_actions[self.equipped].type == 'attack') or 
                      (self.map.all_units_by_name[selected_stats].team == self.team and
                        self.spell_actions[self.equipped].type in ('healing','healingitem','support')))):
                
                    #Plots predictor system
                    self.plot_predictor(self.map.all_units_by_name[selected_stats])
                    
                        
                pygame.display.flip()
                self.map.engine.clock.tick(60)


    
    def spell_loop_c(self,selected):

        """
        # Function Name: Spell Loop C
        # Purpose: Confirmation of action.
        # Input: selected = The target the spell will be cast to
        """
        
        menu_flag = True
        menu_pos = 0
        
        target = self.map.all_units_by_name[selected]

        text_confirm = self.map.engine.bfont.render("Confirm?",True, (0,0,0))
        text_yesno = self.map.engine.sfont.render("YES          NO",True, (0,0,0))

        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                        
                    if (event.key == K_LEFT or event.key == K_DOWN) and menu_pos == 1:
                        menu_pos = 0

                    if (event.key == K_RIGHT or event.key == K_UP) and menu_pos == 0:
                        menu_pos = 1
                        
                    if event.key == K_x:
                        menu_flag = False
                        
                    if event.key == K_z or event.key == K_RETURN:

                        if menu_pos == 0:
                            self.spell_actions[self.equipped].action(self,self.map.all_units_by_name[selected])
                            menu_flag = False
                            self.map.cursor_pos = Vector2(tuple(self.location_tile))
                            self.turnend = True
                            
                        if menu_pos == 1:
                            menu_flag = False


            # Plots the valid attack tiles
            if menu_flag:
                    
                pygame.display.set_caption("Story of a Lost Sky - Pos (%1.0f,%1.0f) - Shift (%1.0f,%1.0f)"
                                           %(self.map.cursor_pos.x,self.map.cursor_pos.y,self.map.screen_shift.x,self.map.screen_shift.y))
        
                self.map.render_background()
                self.plot_attacks()

                self.map.render_all_units()
                self.map.render_cursor()
                self.map.engine.surface.blit(self.map.engine.menu_board,(0,490))
                self.plot_stats()

                # Display for combat statistics
                self.plot_predictor(target)

                self.map.engine.surface.blit(text_confirm,(585,505))
                self.map.engine.surface.blit(text_yesno,(590,525))
                
                # Plots selection cursor
                if menu_pos == 0:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (580,533),5)
                if menu_pos == 1:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (645,533),5)

                pygame.display.flip()
                self.map.engine.clock.tick(60)

        
    def stats_loop(self, wm_flag = False):
        """
        # Function name: stats_loop
        # Purpose: Displays a character's stats, traits, spells etc
        # Inputs: wm_flag - set to true to suppress terrain info (to call from worldmap)
        """
        
        menu_flag = True
        menu_pos = 1
        cursor_pos = 0
        spell_pos = 0
        
        # Basic information
        text_name = self.map.engine.bfont.render(self.name,True,(0,0,0))
        text_class = self.map.engine.sfont.render("Class: "+self.unitclass,True,(0,0,0))
        text_hp = self.map.engine.sfont.render("HP: %3.0f/%3.0f"%(self.HP,self.maxHP),True, (0,0,0))
        text_exp = self.map.engine.sfont.render("LV: %2.0f    EXP: %3.0f/100"%(self.level,self.exp),True, (0,0,0))
        text_sc = self.map.engine.sfont.render("Spirit Charge: %3.0f"%(self.spirit),True, (0,0,0))
        
        # Terrain details
        if not wm_flag:
            text_location_header = self.map.engine.bfont.render("Terrain",True,(0,0,0))
            text_loc = self.map.engine.sfont.render(self.map.terrainmap[tuple(self.location_tile)][0].name,True, (0,0,0))
            text_loc_tdef = self.map.engine.sfont.render("TDEF: %2.1f"%self.map.terrainmap[tuple(self.location_tile)][0].TDEF,True, (0,0,0))
            text_loc_agl = self.map.engine.sfont.render("AGL: "+str(self.map.terrainmap[tuple(self.location_tile)][0].aglmod),True, (0,0,0))
            text_loc_acc = self.map.engine.sfont.render("ACC: "+str(self.map.terrainmap[tuple(self.location_tile)][0].accmod),True, (0,0,0))
            
        # Raw stats
        text_stats_header = self.map.engine.bfont.render("Stats",True,(0,0,0))
        text_STR = self.map.engine.sfont.render("STR: "+str(self.STR),True,(0,0,0))
        text_DEF = self.map.engine.sfont.render("DEF: "+str(self.DEF),True,(0,0,0))
        text_MAG = self.map.engine.sfont.render("MAG: "+str(self.MAG),True,(0,0,0))
        text_MDEF = self.map.engine.sfont.render("MDEF: "+str(self.MDEF),True,(0,0,0))
        text_ACC = self.map.engine.sfont.render("ACC: "+str(self.ACC),True,(0,0,0))
        text_AGL = self.map.engine.sfont.render("AGL: "+str(self.AGL),True,(0,0,0))
        text_moves = self.map.engine.sfont.render("MOVE: "+str(self.moves),True,(0,0,0))
        
        # Spells
        text_spells_header = self.map.engine.bfont.render("Spells",True,(0,0,0))
        text_spell_data = self.get_text_spell_data()
        # Spell Skill Ranks
        text_skill_header = self.map.engine.bfont.render("Spell Skill Rank",True,(0,0,0))
        text_nat = self.map.engine.sfont.render("Natural: "+str(self.skill_rank['Natural']),True,(0,0,0))
        text_ele = self.map.engine.sfont.render("Elemental: "+str(self.skill_rank['Elemental']),True,(0,0,0))
        text_spi = self.map.engine.sfont.render("Spiritual: "+str(self.skill_rank['Spiritual']),True,(0,0,0))
        text_for = self.map.engine.sfont.render("Force: "+str(self.skill_rank['Force']),True,(0,0,0))
        text_spell_level = self.map.engine.sfont.render("Level: ",True,(0,0,0))
        
        # traits
        text_trait = self.map.engine.bfont.render(self.name+"'s Traits",True,(0,0,0))
        text_action = self.map.engine.bfont.render("Action",True,(0,0,0))
        text_support = self.map.engine.bfont.render("Support",True,(0,0,0))
        text_action_standby = self.map.engine.bfont.render("Action (Standby)",True,(0,0,0))
        text_support_standby = self.map.engine.bfont.render("Support (Standby)",True,(0,0,0))
        text_type = self.map.engine.bfont.render("Type",True,(0,0,0))
        text_desc = self.map.engine.bfont.render("Description",True,(0,0,0))
        text_trait_stats = self.map.engine.bfont.render("Stat Boosts",True,(0,0,0))
        text_properties = self.map.engine.bfont.render("Properties",True,(0,0,0))
        
        text_trait_data = self.get_text_trait_data()
        
        while menu_flag:
        
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT:
                        if menu_pos == 1:
                            menu_pos = 3
                        elif menu_pos > 1:
                            menu_pos -= 1
                    if event.key == K_RIGHT:
                        if menu_pos < 3:
                            menu_pos += 1
                        elif menu_pos == 3:
                            menu_pos = 1
                    if event.key == K_UP:
                        # Cursor Position for Trait Screen
                        if menu_pos == 2:
                            if cursor_pos == 0:
                                cursor_pos = 9
                            elif cursor_pos > 0:
                                cursor_pos -= 1
                        # Cursor Position for Spell Screen
                        elif menu_pos == 3:
                            if spell_pos == 0:
                                spell_pos = 4
                            elif spell_pos > 0:
                                spell_pos -= 1
                    if event.key == K_DOWN:
                        # Cursor Position for Trait Screen
                        if menu_pos == 2:
                            if cursor_pos == 9:
                                cursor_pos = 0
                            elif cursor_pos < 9:
                                cursor_pos += 1
                        # Cursor Position for Spell Screen
                        elif menu_pos == 3:
                            if spell_pos == 4:
                                spell_pos = 0
                            elif spell_pos < 4:
                                spell_pos += 1
                        
                    if event.key == K_x:
                        menu_flag = False
                        
            if menu_flag:
            
                # Stat page 1: Basic character stats and spells
                if menu_pos == 1:
                    # Basic Info
                    self.map.engine.surface.blit(self.map.engine.stats_bg,(0,0))
                    self.map.engine.surface.blit(self.av, (50,50))
                    self.map.engine.surface.blit(text_name,(180,50))
                    self.map.engine.surface.blit(text_class,(180,70))
                    self.map.engine.surface.blit(text_hp,(180,85))
                    self.map.engine.surface.blit(text_exp,(180,100))
                    self.map.engine.surface.blit(text_sc,(180, 115))
                    self.map.engine.surface.blit(self.image,(180,135),(0,0,35,35))
                    
                    # Terrain related items
                    if not wm_flag:
                        self.map.engine.surface.blit(text_location_header,(50,180))
                        self.map.engine.surface.blit(self.map.engine.terrain_icon,(60,205),(self.map.terrainmap[tuple(self.location_tile)][0].icon*50,0,50,50))
                        self.map.engine.surface.blit(text_loc, (60,260))
                        self.map.engine.surface.blit(text_loc_tdef, (120,206))
                        self.map.engine.surface.blit(text_loc_acc, (120,220))
                        self.map.engine.surface.blit(text_loc_agl, (120,235))
                    
                    # Raw stats
                    self.map.engine.surface.blit(text_stats_header,(50,280))
                    self.map.engine.surface.blit(text_STR,(70,300))
                    self.map.engine.surface.blit(text_DEF,(70,320))
                    
                    self.map.engine.surface.blit(text_MAG,(170,300))
                    self.map.engine.surface.blit(text_MDEF,(170,320))
                    
                    self.map.engine.surface.blit(text_ACC,(270,300))
                    self.map.engine.surface.blit(text_AGL,(270,320))
                    self.map.engine.surface.blit(text_moves,(70,340))
                    
                    # Skill Ranks
                    self.map.engine.surface.blit(text_skill_header,(50,360))
                    
                    self.map.engine.surface.blit(text_nat,(85,382))
                    self.map.engine.surface.blit(text_ele,(85,402))
                    self.map.engine.surface.blit(text_spi,(85,422))
                    self.map.engine.surface.blit(text_for,(85,442))
                    
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,380),(120,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,400),(80,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,420),(60,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,440),(100,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_rel_diag,(200,365))
                    
                    
                    # Spell_data
                    self.map.engine.surface.blit(text_spells_header,(470,50))
                   
                    
                    for index,single_spell in enumerate(text_spell_data):
                        if single_spell[0]:
                            # Spell Name
                            self.map.engine.surface.blit(single_spell[1],(480,80+100*index))
                            
                            # EFFECT/SHIELD
                            self.map.engine.surface.blit(single_spell[8],(490,100+100*index))
                            self.map.engine.surface.blit(single_spell[9],(490,120+100*index))
                            
                            # RANGE/LIFE
                            self.map.engine.surface.blit(single_spell[13],(590,100+100*index))
                            self.map.engine.surface.blit(single_spell[11],(590,120+100*index))
                            
                            # SPIRIT/RANK STARS
                            self.map.engine.surface.blit(single_spell[12],(690,100+100*index))
                            self.map.engine.surface.blit(single_spell[14],(690,120+100*index))
                            
                            self.map.engine.surface.blit(self.map.engine.level_stars,(690,140+100*index),(0,self.spell_actions[index].spell_rank*15,50,15))
                         
                            # Icons
                            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (490,140+100*(index)),self.spell_actions[index].a_type_big)
                                
                            # DAMAGE TYPE
                            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (515,140+100*(index)),self.spell_actions[index].d_type_big)
                                    
                            # SPELL TYPE
                            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (540,140+100*(index)),self.spell_actions[index].al_type_big)
                            
                            # Counterattack
                            self.map.engine.surface.blit(self.map.engine.spell_icons_big, (565,140+100*(index)),self.spell_actions[index].c_type_big)
                            
                        else:
                            self.map.engine.surface.blit(single_spell[1],(480,80+100*index))

                    
                # Stat page 2: trait descriptions
                elif menu_pos == 2:
                    
                    self.map.engine.surface.blit(self.map.engine.stats_bg,(0,0))
                    self.map.engine.surface.blit(text_trait,(50,50))
                    
                    # Action trait
                    self.map.engine.surface.blit(text_action,(60,80))
                    self.map.engine.surface.blit(text_trait_data[0][1],(75,100))
                    # Action Standby trait
                    self.map.engine.surface.blit(text_action_standby,(60,120))
                    
                    # Support trait
                    self.map.engine.surface.blit(text_support,(60,220))
                    self.map.engine.surface.blit(text_trait_data[5][1],(75,240))
                    
                    # Support Standby trait
                    self.map.engine.surface.blit(text_support_standby,(60,260))
                    
                    for i in (1,2,3,4):
                        self.map.engine.surface.blit(text_trait_data[i][1],(75,120+20*i))
                        self.map.engine.surface.blit(text_trait_data[5+i][1],(75,260+20*i))
                        
                    
                    
                    #Displays cursor
                    if cursor_pos == 0:
                        pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,107),5)
                    elif cursor_pos > 0 and cursor_pos < 5:
                        pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,127+20*cursor_pos),5)
                    elif cursor_pos == 5:
                        pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,247),5)
                    elif cursor_pos > 5:
                        pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,167+20*cursor_pos),5)
                    
                        
                    # Data display
                    if text_trait_data[cursor_pos][0]:
                        self.map.engine.draw_trait_data(text_trait_data[cursor_pos])
                                              
                                              
                # Stat page 3: spell descriptions        
                elif menu_pos == 3:
                    
                    # Spell_data
                    self.map.engine.surface.blit(self.map.engine.stats_bg,(0,0))
                    self.map.engine.surface.blit(text_spells_header,(50,50))
                    
                    self.map.engine.surface.blit(text_spell_data[0][1],(70,80))
                    self.map.engine.surface.blit(text_spell_data[1][1],(70,110))
                    self.map.engine.surface.blit(text_spell_data[2][1],(70,140))
                    self.map.engine.surface.blit(text_spell_data[3][1],(70,170))
                    self.map.engine.surface.blit(text_spell_data[4][1],(70,200))
                    
                    # Draws cursor
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (60,88+30*spell_pos),5)
                  
                    # Skill Ranks
                    self.map.engine.surface.blit(text_skill_header,(50,360))
                    
                    self.map.engine.surface.blit(text_nat,(85,382))
                    self.map.engine.surface.blit(text_ele,(85,402))
                    self.map.engine.surface.blit(text_spi,(85,422))
                    self.map.engine.surface.blit(text_for,(85,442))
                    
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,380),(120,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,400),(80,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,420),(60,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,440),(100,0,20,20))
                    self.map.engine.surface.blit(self.map.engine.spell_rel_diag,(200,365))
                    
                    # Data display
                    if self.spell_actions[spell_pos]:
                        self.map.engine.draw_spell_action_data(self.spell_actions[spell_pos],
                                                           text_spell_data[spell_pos])
                        
                            
                
                pygame.display.flip()
                self.map.engine.clock.tick(60)    
                
    
    def get_text_spell_data(self):
        """
        # Function name: get_text_spell_data
        # Purpose: From a unit's spell actions, returns a list of text objects with the data
        # Inputs: None
        # Outputs: text_spell_data - a list containing the text objects for the unit's spell actions
        """
        
        text_spell_data = [self.map.engine.get_selected_spell_data(spell_action) for spell_action in self.spell_actions]
        return text_spell_data
        
        
        
    def get_text_trait_data(self):
        """
        # Function name: get_text_trait_data
        # Purpose: From a unit's traits, returns a list of text objects with the data
        # Inputs: None
        # Outputs: text_trait_data - a list containing the text objects for the unit's traits
        """
        
        text_trait_data = []
        
        # Counter for number of traits valid
        [text_trait_data.append(self.map.engine.get_text_selected_trait(trait)) for i in (0,1) for trait in self.traits[i]]
                
        return text_trait_data
                
    def spell_swap_menu(self):
        """
        # Function name: spell_swap_menu
        # Purpose: Allows spell to be swapped out from unit
        """
        
        menu_flag = True
        menu_pos = 0
                
        # Spells
        text_spells_header = self.map.engine.bfont.render("Swap Spell Action",True,(0,0,0))
        
        text_spell_data = self.get_text_spell_data()
        text_cancel = self.map.engine.bfont.render("Cancel",True,(0,0,0))
        
        # Spell Skill Ranks
        text_skill_header = self.map.engine.bfont.render("Spell Skill Rank",True,(0,0,0))
        text_nat = self.map.engine.sfont.render("Natural: "+str(self.skill_rank['Natural']),True,(0,0,0))
        text_ele = self.map.engine.sfont.render("Elemental: "+str(self.skill_rank['Elemental']),True,(0,0,0))
        text_spi = self.map.engine.sfont.render("Spiritual: "+str(self.skill_rank['Spiritual']),True,(0,0,0))
        text_for = self.map.engine.sfont.render("Force: "+str(self.skill_rank['Force']),True,(0,0,0))
        text_active = self.map.engine.sfont.render("[Active]",True,(0,0,0))
        
        text_directions1 = self.map.engine.sfont.render("Z / Enter - Select Spell Action Slot",True,(0,0,0))
        text_directions2 = self.map.engine.sfont.render("C - Unequip Spell Action Slot",True,(0,0,0))
        text_directions3 = self.map.engine.sfont.render("A - Set as Active Spell Action",True,(0,0,0))
        text_directions4 = self.map.engine.sfont.render("X - Back",True,(0,0,0))
        
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 5
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 5:
                            menu_pos += 1
                        elif menu_pos == 5:
                            menu_pos = 0
                    if event.key == K_z  or event.key == K_RETURN:
                        if menu_pos < 5:
                            # Ask the player to choose a spell from the inventory
                            selection_result = self.map.engine.spell_swap_inventory(self)
                            if selection_result:
                                # Append the current spell to the inventory if it is an a real spell
                                if self.spell_actions[menu_pos]:
                                    self.map.engine.player.add_item(self.spell_actions[menu_pos])
                                # Replace the selected spell into the unit's slot
                                self.spell_actions[menu_pos]=selection_result[0]

                                selection_result[0].get_attack_range(self)

                                # Call for updated spell data
                                text_spell_data = self.get_text_spell_data()
                                    
                            
                        elif menu_pos == 5:
                            menu_flag = False
                    # Unequips a spell
                    if event.key == K_c and menu_pos < 5:
                        if self.spell_actions[menu_pos]:
                            self.map.engine.player.add_item(self.spell_actions[menu_pos])
                            # Replace the selected spell into the unit's slot with None
                            self.spell_actions[menu_pos] = None
                        
                        # Call for updated spell data
                        text_spell_data = self.get_text_spell_data()
                                    
                    # Sets equipped spell to active
                    if event.key == K_a and menu_pos < 5:
                        self.equipped = menu_pos
                                    
                    if event.key == K_x:
                        menu_flag = False
                        
                        
            # Stat page 3: spell descriptions        
            if menu_flag:
                
                # Spell_data
                self.map.engine.surface.blit(self.map.engine.stats_bg,(0,0))
                self.map.engine.surface.blit(text_spells_header,(50,50))
                
                for index,spell_name in enumerate(text_spell_data): 
                    self.map.engine.surface.blit(spell_name[1],(70,80+index*30))
                
                # Draws Active Spell Indicator
                self.map.engine.surface.blit(text_active,(300,80+30*(self.equipped)))
                
                self.map.engine.surface.blit(text_cancel,(70,230))
                
                # Draws cursor
                pygame.draw.circle(self.map.engine.surface, (0,0,0), (60,88+30*menu_pos),5)
              
                # Skill Ranks
                self.map.engine.surface.blit(text_skill_header,(50,360))
                
                self.map.engine.surface.blit(text_nat,(85,382))
                self.map.engine.surface.blit(text_ele,(85,402))
                self.map.engine.surface.blit(text_spi,(85,422))
                self.map.engine.surface.blit(text_for,(85,442))
                
                self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,380),(120,0,20,20))
                self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,400),(80,0,20,20))
                self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,420),(60,0,20,20))
                self.map.engine.surface.blit(self.map.engine.spell_icons_big, (60,440),(100,0,20,20))
                self.map.engine.surface.blit(self.map.engine.spell_rel_diag,(200,365))
                
                self.map.engine.surface.blit(text_directions1,(60,520))
                self.map.engine.surface.blit(text_directions2,(60,540))
                self.map.engine.surface.blit(text_directions3,(60,560))
                self.map.engine.surface.blit(text_directions4,(60,580))
                
                
                # Data display
                if menu_pos < 5 and self.spell_actions[menu_pos]:
                    self.map.engine.draw_spell_action_data(self.spell_actions[menu_pos],
                                                             text_spell_data[menu_pos])
                
                pygame.display.flip()
                self.map.engine.clock.tick(60)
                

                
    def trait_swap_menu(self):
        """
        # Function Name: trait swap menu
        # Purpose: Allows selection of traits to be swaped in and out of a unit's
        #   reserved traits
        """
        
        menu_flag = True
        menu_pos = 0

        
        text_learn = self.map.engine.bfont.render("Learn Traits",True,(0,0,0))
        text_cancel = self.map.engine.bfont.render("Cancel",True,(0,0,0))        
        
        text_trait = self.map.engine.bfont.render(self.name+"'s Traits",True,(0,0,0))
        text_action = self.map.engine.bfont.render("Action",True,(0,0,0))
        text_support = self.map.engine.bfont.render("Support",True,(0,0,0))
        text_action_standby = self.map.engine.bfont.render("Action (Standby)",True,(0,0,0))
        text_support_standby = self.map.engine.bfont.render("Support (Standby)",True,(0,0,0))
        
        text_directions1 = self.map.engine.sfont.render("Z / Enter - Select Trait Slot",True,(0,0,0))
        text_directions2 = self.map.engine.sfont.render("C - Unequip Trait Slot",True,(0,0,0))
        text_directions3 = self.map.engine.sfont.render("A - Swap Trait to Active mode",True,(0,0,0))
        text_directions4 = self.map.engine.sfont.render("X - Back",True,(0,0,0))
        
        text_trait_data = self.get_text_trait_data()
        
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 11
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 11:
                            menu_pos += 1
                        elif menu_pos == 11:
                            menu_pos = 0
                    
                    # Trait inventory menu
                    if event.key == K_z  or event.key == K_RETURN:
                        if menu_pos < 9:
                                
                            # Swap action trait
                            if menu_pos < 5:
                                selected_trait = self.trait_swap_inventory(0)
                                if selected_trait != False:
                                    temp = self.traits[0][menu_pos]
                                    self.traits[0][menu_pos] = selected_trait
                                    # If the current trait is not an empty slot,
                                    # return the trait to reserve list
                                    if temp:
                                        self.reserve_traits[0].append(temp)
                                    # Refresh text objects
                                    text_trait_data = self.get_text_trait_data()
                                        
                            # Swap support trait
                            elif (menu_pos >= 5 or menu_pos < 10):
                                selected_trait = self.trait_swap_inventory(1)
                                if selected_trait != False:
                                    temp = self.traits[1][menu_pos-5]
                                    self.traits[1][menu_pos-5] = selected_trait
                                    if temp:
                                        self.reserve_traits[1].append(temp)
                                    # Refresh text objects
                                    text_trait_data = self.get_text_trait_data()
                                    
                        elif menu_pos == 10:
                            self.map.engine.trait_learning_system.trait_learning_menu(self)
                            text_trait_data = self.get_text_trait_data()

                        elif menu_pos == 11:
                            menu_flag = False
                    
                    # Swaps a standby trait to active
                    if event.key == K_a:
                        if (menu_pos > 0 and menu_pos < 5):
                            self.swap_trait([0,menu_pos-1])
                            text_trait_data = self.get_text_trait_data()

                        elif ( menu_pos >5 or menu_pos < 10):
                            self.swap_trait([1,menu_pos-6])
                            text_trait_data = self.get_text_trait_data()
                        
                    
                    # Unequips a trait
                    if event.key == K_c and menu_pos < 10:
                        # Unequip action trait
                        if menu_pos < 5 and self.traits[0][menu_pos]:
                            self.reserve_trait([0,menu_pos])
                            text_trait_data = self.get_text_trait_data()
                        # Unequip support trait
                        elif menu_pos >= 5 and self.traits[1][menu_pos-5]:
                            self.reserve_trait([1,menu_pos-5])
                            text_trait_data = self.get_text_trait_data()
                            
                    if event.key == K_x:
                        menu_flag = False
        
            if menu_flag:
                      
                self.map.engine.surface.blit(self.map.engine.stats_bg,(0,0))
                self.map.engine.surface.blit(text_trait,(50,50))
                
                # Action trait
                self.map.engine.surface.blit(text_action,(60,80))
                self.map.engine.surface.blit(text_trait_data[0][1],(75,100))
                # Action Standby trait
                self.map.engine.surface.blit(text_action_standby,(60,120))
                
                # Support trait
                self.map.engine.surface.blit(text_support,(60,220))
                self.map.engine.surface.blit(text_trait_data[5][1],(75,240))
                
                # Support Standby trait
                self.map.engine.surface.blit(text_support_standby,(60,260))
                
                for i in (1,2,3,4):
                    self.map.engine.surface.blit(text_trait_data[i][1],(75,120+20*i))
                    self.map.engine.surface.blit(text_trait_data[5+i][1],(75,260+20*i))
                    
                self.map.engine.surface.blit(text_learn,(75,360))
                self.map.engine.surface.blit(text_cancel,(75,380))
                
                #Displays cursor
                if menu_pos == 0:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,107),5)
                elif menu_pos > 0 and menu_pos < 5:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,127+20*menu_pos),5)
                elif menu_pos == 5:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,247),5)
                elif menu_pos > 5:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (65,167+20*menu_pos),5)
                
                self.map.engine.surface.blit(text_directions1,(60,500))
                self.map.engine.surface.blit(text_directions2,(60,520))
                self.map.engine.surface.blit(text_directions3,(60,540))
                self.map.engine.surface.blit(text_directions4,(60,560))
                
                # Data display
                if menu_pos < 9 and text_trait_data[menu_pos][0]:
                    self.map.engine.draw_trait_data(text_trait_data[menu_pos])
                    
                pygame.display.flip()
                self.map.engine.clock.tick(60)
    
     
    def trait_swap_inventory(self,trait_list_pos):
        """
        # Function Name: trait swap menu
        # Purpose: Allows selection of traits to be swaped in and out of a unit's
        #   reserved traits
        # Inputs: Trait_list_pos - 0 for action, 1 for support (position of appropriate trait list)
        """
        
        menu_flag = True
        menu_pos = 0
        
        if trait_list_pos == 0:
            text_inv_header = self.map.engine.bfont.render("Reserve Action Traits",True,(0,0,0))
        else:
            text_inv_header = self.map.engine.bfont.render("Reserve Support Traits",True,(0,0,0))
        
        text_type = self.map.engine.bfont.render("Type",True,(0,0,0))
        text_desc = self.map.engine.bfont.render("Description",True,(0,0,0))
        text_trait_stats = self.map.engine.bfont.render("Stat Boosts",True,(0,0,0))
        text_properties = self.map.engine.bfont.render("Properties",True,(0,0,0))
        
        text_stats_header = self.map.engine.bfont.render("Stats",True,(0,0,0))
        
        # Gets the text for the currently selected trait
        if self.reserve_traits[trait_list_pos]:
            text_selected = self.map.engine.get_text_selected_trait(self.reserve_traits[trait_list_pos][0])
        else:
            text_selected = self.map.engine.get_text_selected_trait(None)
        
        # Generates a list of what's in the reserve traits
        text_reserve_inventory = []
        for trait in self.reserve_traits[trait_list_pos]:
            if trait:
                text_reserve_inventory.append(self.map.engine.sfont.render(trait.name,True,(0,0,0)))
            else:
                text_reserve_inventory.append(self.map.engine.sfont.render('Empty',True,(100,100,100)))
            
        while menu_flag:
        
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if (event.key == K_DOWN or event.key == K_RIGHT)  and self.reserve_traits[trait_list_pos]:
                        if menu_pos < len(self.reserve_traits[trait_list_pos])-1:
                            menu_pos += 1
                        elif menu_pos == len(self.reserve_traits[trait_list_pos])-1 :
                            menu_pos = 0 
                        # Requests new selected spell info
                        text_selected = self.map.engine.get_text_selected_trait(self.reserve_traits[trait_list_pos][menu_pos])
                        
                    if (event.key == K_UP or event.key == K_LEFT)  and self.reserve_traits[trait_list_pos]:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(self.reserve_traits[trait_list_pos])-1
                        
                        # Requests new selected spell info
                        text_selected = self.map.engine.get_text_selected_trait(self.reserve_traits[trait_list_pos][menu_pos])
                        
                    if (event.key == K_z  or event.key == K_RETURN) and self.reserve_traits[trait_list_pos]:
                        # Pops the selected trait out of reserve and returns it
                        selected_trait = self.reserve_traits[trait_list_pos].pop(menu_pos)
                        return selected_trait
                    
                    if event.key == K_x:
                        # Cancel
                        menu_flag = False
                        return False
                    
            if menu_flag:
            
                self.map.engine.surface.blit(self.map.engine.stats_bg,(0,0))
                self.map.engine.surface.blit(text_inv_header,(50,50))
    
                
                # Draws cursor
                if self.reserve_traits[trait_list_pos]:
                    pygame.draw.circle(self.map.engine.surface, (0,0,0), (60,88+20*menu_pos),5)
              
                # Displays the inventory
                item_counter = 0
                for text_inv_trait in text_reserve_inventory:
                    self.map.engine.surface.blit(text_inv_trait,(70,80+item_counter*20))
                    item_counter += 1
                
                # Data display
                if self.reserve_traits[trait_list_pos]:
                    self.map.engine.draw_trait_data(text_selected)
                    

            pygame.display.flip()
            self.map.engine.clock.tick(60)

class UnitCircleSprite(pygame.sprite.Sprite):
        
    def __init__(self,unit):
        """
        # Function: __init__
        # Purpose: Creates a sprite for the circle underneath the unit
        # Inputs: unit - unit to initialize from
        """
        pygame.sprite.Sprite.__init__(self)
        self.unit = unit
        self.hold = False   # Hold position 
                            # Used for animating vertical movements. Set this to True)

        team_panels = pygame.image.load(os.path.join('images','team_panels.png')).convert_alpha()
    
        # Initialize Sprites
        if unit.chartype == 'boss':
            self.original_image = team_panels.subsurface((70,0,35,35))
        elif self.unit.chartype == 'npc':
            self.original_image = team_panels.subsurface((105,0,35,35))
        elif self.unit.chartype == 'pc':
            self.original_image = team_panels.subsurface((0,0,35,35))
        elif self.unit.chartype == 'enemy':
            self.original_image = team_panels.subsurface((35,0,35,35))

        self.image = self.original_image
        self.rect = self.image.get_rect()


    def update(self):
        """
        # Function: update
        # Purpose: Updates sprite to current location of the unit (pixel)
        """
        if not self.hold:
            self.image = self.original_image
            self.rect.topleft = tuple(self.unit.location_pixel-self.unit.map.screen_shift*35)
        
        # Do not draw the image if unit is outside the bounds of the screen
        if (self.rect.topleft[0] >= 875 or self.rect.topleft[0] < -35
             or self.rect.topleft[1] >= 490 or self.rect.topleft[1] < -35):
            self.rect.topleft = (-35,-35)


        # Draw only a portion of the sprite if the unit is getting cut off by the menu panel
        if self.rect.topleft[1] > 455 and self.rect.topleft[1] < 490:

            cutoff_y = 35 - (self.rect.topleft[1] - 455)
            self.image = self.original_image.subsurface((0,0,35,cutoff_y))
    
    

class UnitMapSprite(pygame.sprite.Sprite):
    
    def __init__(self,unit):
        """
        # Function: __init__
        # Purpose: Creates a sprite for the map sprite
        # Inputs: unit - unit to initialize from
        """
        pygame.sprite.Sprite.__init__(self)
        self.unit = unit
        
        self.active_image = unit.image.subsurface((0,0,35,35))
        self.wait_image =  unit.image.subsurface((105,0,35,35))
        self.image = self.active_image
        self.rect = self.image.get_rect()
        

    def update(self):
        """
        # Function: update
        # Purpose: Updates sprite to current location of the unit (pixel)
        """
        if self.unit.turnend == False or self.unit.map.currentplayer != self.unit.team:
            self.image = self.active_image
        else:
            self.image = self.wait_image
        self.rect.topleft = tuple(self.unit.location_pixel-self.unit.map.screen_shift*35)
        
            
        # Do not draw the image if unit is outside the bounds of the screen
        if (self.rect.topleft[0] >= 875 or self.rect.topleft[0] < -35
             or self.rect.topleft[1] >= 490 or self.rect.topleft[1] < -35):
            
            self.rect.topleft = (-35,-35) 

        # Draw only a portion of the sprite if the unit is getting cut off by the menu panel
        if self.rect.topleft[1] > 455 and self.rect.topleft[1] < 490:

            cutoff_y = 35 - (self.rect.topleft[1] - 455)
            self.image = self.active_image.subsurface((0,0,35,cutoff_y))

class UnitStatusSprite(pygame.sprite.Sprite):

    def __init__(self,unit):
        """
        # Function: __init__
        # Purpose: Creates a sprite for the unit's status bubble
        # Inputs: unit - unit to initialize bubble for from
        """
        pygame.sprite.Sprite.__init__(self)
        self.unit = unit
        
        # Blank Surface
        self.blank = pygame.surface.Surface((18,18)).convert_alpha()
        self.blank.fill((0,0,0,0))
        self.status_queue = [self.blank]
        self.image = self.blank
        self.rect = self.image.get_rect()
        self.counter = 0

    def update_queue(self):
        
        # Generate new status queue
        self.status_queue = [self.unit.map.engine.status_effects_catalog[status_effect].icon
                            for status_effect in self.unit.status.keys()]
        if not self.status_queue:
            self.status_queue = [self.blank]
            
        # Draw next image from front of the line
        self.image = self.status_queue.pop(0)
        self.rect = self.image.get_rect()
        
    def update(self):
        # Change image every 60 frames
        if self.counter == 61:
            # Append current image to back of the line
            self.status_queue.append(self.image)
            # Draw next image from front of the line
            self.image = self.status_queue.pop(0)
            self.rect = self.image.get_rect()
            
            # Reset Counter 
            self.counter = 0
        
        self.counter += 1 
        self.rect.topleft = tuple(Vector2(17,-5)+self.unit.location_pixel-self.unit.map.screen_shift*35)
        
        # Do not draw the image if unit is outside the bounds of the screen
        if (self.rect.topleft[0] >= 840 or self.rect.topleft[0] < -35
             or self.rect.topleft[1] >= 455 or self.rect.topleft[1] < -35):
            self.rect.topleft = (-35,-35) 

