import pygame
import os
from pygame.locals import *
from sys import exit
try: from cmath import floor, sin, pi, cos
except: from math import floor, sin, pi, cos

from lostsky.core.persistentobjects import Options, PlayerData, EventData
from lostsky.worldmap.mission_manager import MissionManager
from lostsky.worldmap.trading import TradingSystem
from lostsky.worldmap.spell_synthesis import SpellSynthesisSystem
from lostsky.battle.trait import TraitLearningSystem
from lostsky.battle import status_effects
from lostsky.battle.battle_event import BattleEventSystem
from lostsky.core.utils import split_lines
from lostsky.core.sound_effects import SfxSystem

try: import cPickle as pickle
except: import pickle

class Engine(object):

    """
    # Central Game Engine
    """
    
    def __init__(self,surface,tilesize,size):
        """
        # Function Name: __init__
        # Purpose: Intializes the game engine, and loads some game engine level graphics
        # Inputs: surface = the surface to render the images on
        #         tilesize = the size of the tiles
        #         size = the size of the screen in terms of number of tiles
        """
        
        
        self.clock = pygame.time.Clock()
        
        # Splash images
        self.splash_fmp = pygame.image.load(os.path.join('images','splash_fmp.jpg'))
        
        # Cursor image
        self.cursor_img = pygame.image.load(os.path.join('images','cursor.png')).convert_alpha()
        
        # Music Library
        self.music_catalog = {}
        
        # Small Font, Big Font, Combat display font
        self.sfont = pygame.font.Font("DejaVuSans-Oblique.ttf", 12)
        self.mfont = pygame.font.Font("DejaVuSans-Oblique.ttf", 16)
        self.bfont = pygame.font.Font("VeraSeBd.ttf", 14)
        self.cfont = pygame.font.Font("VeraSeBd.ttf", 24)
        
        # The size of the window
        self.size_x = size[0]
        self.size_y = size[1]
        self.surface = surface

        # The size of each tile
        self.tilesize = tilesize
        
        # Colored Tiles
        self.move_tile = pygame.image.load(os.path.join('images','green_tile.png')).convert_alpha()
        self.attack_tile = pygame.image.load(os.path.join('images','red_tile.png')).convert_alpha()
        self.heal_tile = pygame.image.load(os.path.join('images','blue_tile.png')).convert_alpha()
        self.deploy_tile = pygame.image.load(os.path.join('images','deploy_tile.png')).convert_alpha()
        self.team_panels = pygame.image.load(os.path.join('images','team_panels.png')).convert_alpha()
        
        # Various menu panels
        self.menu_board = pygame.image.load(os.path.join('images','menuboard.png')).convert_alpha()
        self.move_arrow = pygame.image.load(os.path.join('images','move_arrow.png')).convert_alpha()
        self.battle_panel = pygame.image.load(os.path.join('images','battlepanel2.png')).convert_alpha()
        self.battle_bg = pygame.image.load(os.path.join('images','battlebg.jpg'))
        self.stats_bg = pygame.image.load(os.path.join('images','statsbg.jpg'))
        self.battle_board = pygame.image.load(os.path.join('images','battleboard.png')).convert()
        self.battle_top = pygame.image.load(os.path.join('images','battletop.png')).convert_alpha()
        self.results_panel = pygame.image.load(os.path.join('images','results_panel.png')).convert()
        self.text_board = pygame.image.load(os.path.join('images','textboard.png')).convert()
        self.map_spell_board = pygame.image.load(os.path.join('images','map_spell_panel.png')).convert_alpha()
        self.small_text_board = pygame.image.load(os.path.join('images','small_text_panel.png')).convert()
        self.tiny_text_board = pygame.image.load(os.path.join('images','tiny_text_panel.png')).convert()
        self.shop_bg = pygame.image.load(os.path.join('images','shopbg.jpg'))
        self.chapter_title = pygame.image.load(os.path.join('images','chapter_titles.png')).convert_alpha()
        self.health_meter = pygame.image.load(os.path.join('images','health_meter.png')).convert_alpha()
       
       
        # Item Categories
        self.item_category_cursor = pygame.image.load(os.path.join('images','item_category_cursor.png')).convert_alpha()
        self.item_category_header = pygame.image.load(os.path.join('images','item_category_header.png')).convert_alpha()
        self.treasure_header = pygame.image.load(os.path.join('images','treasure_header.png')).convert_alpha()
        
       
        # Team move images
        self.p1turn = pygame.image.load(os.path.join('images','player1turn.png')).convert_alpha()
        self.p2turn = pygame.image.load(os.path.join('images','player2turn.png')).convert_alpha()

        # Status images
        self.status_effect_icons = pygame.image.load(os.path.join('images','status_effect_icons.png')).convert_alpha()
        self.status_spirit = pygame.image.load(os.path.join('images','status-spirit.png')).convert_alpha()

        # Image for displaying hit/miss percent
        self.arrows = pygame.image.load(os.path.join('images','hit_arrows.png')).convert_alpha()
        self.combatstats = pygame.image.load(os.path.join('images','combatstats.png')).convert_alpha()
        
        
        # Load terrain tile file
        self.terrain_img = pygame.image.load(os.path.join('images','terraintilesv4.png')).convert_alpha()
        self.terrain_icon = pygame.image.load(os.path.join('images','terrainicons.png')).convert()
        self.layer_2_img = pygame.image.load(os.path.join('images','layer_2_tiles.png')).convert_alpha()
        
        # Spell icons
        self.spell_icons = pygame.image.load(os.path.join('images','spell_icons.png')).convert()
        self.spell_icons_big = pygame.image.load(os.path.join('images','spell_icons_big.png')).convert()
        self.spell_rel_diag = pygame.image.load(os.path.join('images','spell_type_diagram.png')).convert_alpha()
        self.level_stars = pygame.image.load(os.path.join('images','level_stars.png')).convert_alpha()
        relation_arrows = pygame.image.load(os.path.join('images','icons','advantage_icons.png')).convert_alpha()
        self.relation_arrows = {'up':relation_arrows.subsurface(0,0,20,20),
                                'down':relation_arrows.subsurface(20,0,20,20),
                                'neutral':relation_arrows.subsurface(40,0,20,20)
                                }
        
        self.heal_bullet = pygame.image.load(os.path.join('images','bullets','smallorb_magenta.png')).convert_alpha()

        # World Map objects
        self.wm_bg = pygame.image.load(os.path.join('images','worldmapbg.jpg'))
        self.region_bg = pygame.image.load(os.path.join('images','worldmap_empty.jpg'))
        self.map_cursor = pygame.image.load(os.path.join('images','map_cursor.png')).convert_alpha()
        self.player_img = pygame.image.load(os.path.join('images','map_sprites','02-youmu-d.png')).convert_alpha()
        self.location_circle = pygame.image.load(os.path.join('images','location_tile.png')).convert_alpha()
        self.event_icon = pygame.image.load(os.path.join('images','event_icon.png')).convert_alpha()
        self.path_dot = pygame.image.load(os.path.join('images','path_dot.png')).convert_alpha()
        
        # Party Management Menu Images
        self.party_bg = pygame.image.load(os.path.join('images','partybg.jpg'))
        self.unit_tile = pygame.image.load(os.path.join('images','unit_tile.png')).convert_alpha()
        
        # Loads the game options
        # Attempts to get the game options from saved data
        try:
            self.options = Options.load()
        # If it fails, creates a new options object and saves it
        except IOError:
            self.options = Options()
            self.options.save()
            
        # Save/Load text images
        self.text_save = self.bfont.render("Saving.",True,(255,255,255))
        self.text_load = self.bfont.render("Loading...",True,(255,255,255))
            
        # Title Screen Objects
        self.title = pygame.image.load(os.path.join('images','title.jpg'))
        self.title_cursor = pygame.image.load(os.path.join('images','title_cursor.png')).convert_alpha()
           
        # Mission Manager
        self.mission_manager = MissionManager(self)
        self.mission_bg = pygame.image.load(os.path.join('images','missionbg.jpg'))
           
        # Landmark images
        self.landmark_img = pygame.image.load(os.path.join('images','landmarks.png')).convert_alpha()
        self.highlight_tile = pygame.image.load(os.path.join('images','highlight_tile.png')).convert_alpha()
        
        # Victory/Defeat banners
        self.mission_victory = pygame.image.load(os.path.join('images','mission_victory.png')).convert_alpha()
        self.mission_failure = pygame.image.load(os.path.join('images','mission_failure.png')).convert_alpha()
        
        # Generates empty player data object
        self.player_units = []
        self.player_units_by_name = {}
        self.player = PlayerData()
        self.all_events_master = {}
        
        # Sound Effects System
        self.sfx_system = SfxSystem()
        
        # Battle Event system
        self.battle_event_system = BattleEventSystem(self)
        
        # Trading / Synth system stuff
        self.trading_system = TradingSystem(self)
        
        # Trait Learning System
        self.trait_learning_system = TraitLearningSystem(self)
        
        # Spell synthesis system
        self.spell_synthesis_system = SpellSynthesisSystem(self)
        
        #Debugging Shortcut Flags
        self.enable_prologue = True
        self.single_turn_win = False
        self.enable_splash = True
        self.enable_wm_tutorial = True
        self.unlock_wm = True
        self.unlock_shops = False  
        
        # Music - Currently playing song
        self.current_music = None
        
        # Map BG overlay images
        self.night_overlay = pygame.Surface((self.size_x*35,self.size_y*35))
        self.night_overlay.fill((0,0,50))
        self.night_overlay.set_alpha(150)
        self.sunset_overlay = pygame.Surface((self.size_x*35,self.size_y*35))
        self.sunset_overlay.fill((150,0,0))
        self.sunset_overlay.set_alpha(70)
        
        # Status Effects
        self.status_effects_catalog = {'High Spirit':status_effects.HighSpirit(),
                                       'Low Spirit':status_effects.LowSpirit(),
                                       'Poison':status_effects.Poison(),
                                        'Stun':status_effects.Stun(),
                                        'Spirit Drain':status_effects.SpiritDrain(),
                                        'Immobilize':status_effects.Immobilize(),
                                        'Dizzy':status_effects.Dizzy(),
                                        'Movement Down':status_effects.MovementDown(),
                                        'Binding Seal':status_effects.BindingSeal(),
                                        'STR Down':status_effects.StrDown(),
                                        'DEF Down':status_effects.DefDown(),
                                        'MAG Down':status_effects.MagDown(),
                                        'MDEF Down':status_effects.MdefDown(),
                                        'Illusion Veil':status_effects.IllusionVeil(),
                                        'Tracking Shot':status_effects.TrackingShot(),
                                        'Life Bless':status_effects.LifeBless(),
                                        'STR Up':status_effects.StrUp(),
                                        'MAG Up':status_effects.MagUp(),
                                        'DEF Up':status_effects.DefUp(),
                                        'MDEF Up':status_effects.MdefUp(),
                                        'Spirit Recharge':status_effects.SpiritRecharge(),
                                        }
        
        # Emotion Bubble Images
        
        self.treasure_icons = pygame.image.load(os.path.join('images','icons','treasureicons.png')).convert_alpha()
        self.card_icons = pygame.image.load(os.path.join('images','icons','cards.png')).convert_alpha()
        
        self.icons = {
                      # Spell Card Icons
                      'red card':self.card_icons.subsurface((0,0,12,14)),
                      'orange card':self.card_icons.subsurface((12,0,12,14)),
                      'yellow card':self.card_icons.subsurface((24,0,12,14)),
                      'green card':self.card_icons.subsurface((36,0,12,14)),
                      'teal card':self.card_icons.subsurface((48,0,12,14)),
                      'blue card':self.card_icons.subsurface((60,0,12,14)),
                      'purple card':self.card_icons.subsurface((72,0,12,14)),
                      'gray card':self.card_icons.subsurface((84,0,12,14)),
                      
                      # Crystal Icons
                      'silver crystal':self.treasure_icons.subsurface((0,0,24,24)),
                      'red crystal':self.treasure_icons.subsurface((24,0,24,24)),
                      'orange crystal':self.treasure_icons.subsurface((48,0,24,24)),
                      'yellow crystal':self.treasure_icons.subsurface((72,0,24,24)),
                      'green crystal':self.treasure_icons.subsurface((96,0,24,24)),
                      'teal crystal':self.treasure_icons.subsurface((120,0,24,24)),
                      'lightblue crystal':self.treasure_icons.subsurface((144,0,24,24)),
                      'darkblue crystal':self.treasure_icons.subsurface((168,0,24,24)),
                      'purple crystal':self.treasure_icons.subsurface((192,0,24,24)),
                      'brown crystal':self.treasure_icons.subsurface((216,0,24,24)),
                      'jade crystal':self.treasure_icons.subsurface((240,0,24,24)),
                      'ruby crystal':self.treasure_icons.subsurface((264,0,24,24)),
                      'white crystal':self.treasure_icons.subsurface((288,0,24,24)),
                      'gray crystal':self.treasure_icons.subsurface((312,0,24,24)),
                      'black crystal':self.treasure_icons.subsurface((336,0,24,24)),
                      'rainbow crystal':self.treasure_icons.subsurface((360,0,24,24)),
                      'package':self.treasure_icons.subsurface((384,0,24,24)),
                      'treasure box':self.treasure_icons.subsurface((408,0,24,24))
                      
                      
                      }

    ############################################
    # Misc. utility methods
    ############################################
        
    def play_music(self,song_name):
        """
        # Function Name: Play Music
        # Purpose: Plays a certain song
        # Inputs: song_name = name of the song
        """
        
        self.current_music = song_name
        if self.options.enable_music:
            pygame.mixer.music.load(os.path.join('music',self.music_catalog[song_name][0]))
            pygame.mixer.music.play(-1)
        
    def fade_to(self,color,time=0.25):    
        """
        # Function Name: fade_to
        # Purpose: Fades the screen to a color
        # Inputs:   color - either the strings 'black', 'white'
        #                   or a RGB tuple e.g.(0,0,255)
        #           time = time (in seconds) to fade. Default is 0.25 s
        """
        
        if color == 'black':
            color = (0,0,0)
        elif color == 'white':
            color = (255,255,255)
        # Determines the max time (in # frames)
        max_time = int(time*60)
    
        # Gets a copy of what's currently on the screen
        screen_copy = self.surface.copy()
        overlay = pygame.Surface((self.size_x*35,self.size_y*35))
        overlay.fill(color)
        for t in xrange(0,max_time):
            pygame.event.get()
            # Uses a sin(x) function to get from 0 to 255 
            overlay.set_alpha(int(255*sin(pi*t/(2*max_time))))
            self.surface.blit(screen_copy,(0,0))
            self.surface.blit(overlay,(0,0))
            pygame.display.flip()
            self.clock.tick(60)
            
    def fade_from(self,color,time=0.25):
        """
        # Function Name: fade_from
        # Purpose: Fades the screen from a color
        # Inputs:   color - either the strings 'black', 'white'
        #                   or a RGB tuple e.g.(0,0,255)
        #           time = time (in seconds) to fade. Default is 0.5 s
        """
        
        if color == 'black':
            color = (0,0,0)
        elif color == 'white':
            color = (255,255,255)
        max_time = int(time*60)
    
        screen_copy = self.surface.copy()
        overlay = pygame.Surface((self.size_x*35,self.size_y*35))
        overlay.fill(color)
        for t in xrange(0,max_time):
            pygame.event.get()
            # Uses a cos(x) function to get from 0 to 255
            overlay.set_alpha(int(255*cos(pi*t/(2*max_time))))
            self.surface.blit(screen_copy,(0,0))
            self.surface.blit(overlay,(0,0))
            pygame.display.flip()
            self.clock.tick(60)

        # Restores the screen (in case the alpha transparency locks in at 1 b/c rounding)
        self.surface.blit(screen_copy,(0,0))
        pygame.display.flip()
        self.clock.tick(60)
            
            
    def pause(self,time):
        """
        # Function Name: pause
        # Purpose: pauses the game for a specified amount of seconds
        # Inputs: time = time to pause, in seconds. May be either an integer or float
        """
        
        framecount = int(60*time)
        for i in xrange(0,framecount):
        
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                    
            pygame.display.flip()
            self.clock.tick(60)
        
    def render_outlined_text(self,line,font,text_color,outline_color):
        """
        # Function Name: render_outlined_text
        # Purpose: Create a image of a set of text that has a 1 pixel padded outlined color
        # Inputs:    line - line of text to render
        #            font - font to render with
        #            text_color - (R,G,B) 
        #            outline_color - (R,G,B)
        # Outputs: text_surface - image of outlined text
        """
        inner_text = font.render(line,True,text_color)
        outer_text = font.render(line,True,outline_color)
        text_surface = pygame.Surface((inner_text.get_width()+2,inner_text.get_height()+2),flags=SRCALPHA)
        text_surface.fill((0,0,0,0))
        text_surface.blit(outer_text,(0,0))
        text_surface.blit(outer_text,(2,0))
        text_surface.blit(outer_text,(0,2))
        text_surface.blit(outer_text,(2,2))
        text_surface.blit(inner_text,(1,1))
        
        return text_surface
        
        
    ############################################
    # Data Manipulation Methods
    ############################################
            
    def check_event_completion(self,event_list):
        """
        # Function Name: check_event_completion
        # Purpose: Checks if a list of events has been completed
        # Inputs: event_list - the list of events to check
        # Outputs: True if the events have been completed
        #          False if the events have not been completed
        """
        events_done = 0
        for event_id in event_list:
            # Case: The mission is not in the game
            if event_id not in self.all_events_master.keys():
                break
                
            if self.all_events_master[event_id].done:
                events_done += 1

        if events_done == len(event_list):
            return True
        else: 
            return False
    
    def add_event_master(self,event_list):
        """
        # Function Name: add_event_master
        # Purpose: Adds event to the master list of events
        # Inputs: event_list - the list of event to be added
        """
        
        for event in event_list:
            # Assigns the event a unique ID
            self.all_events_master[event.event_id] = event
            # Adds the event to the player event data
            self.player.all_event_data[event.event_id] = EventData(event)
            
        self.all_events_available = []
        self.all_events_completed = []
        self.all_events_sign_up = []
        
        # Updates all events on the map
        self.worldmap.update_all_events()
            
    def save_menu(self,title_mode=False):
    
        """
        # Function Name: Save/Load Menu
        # Purpose: Manages saving and loading
        # Inputs: title_mode = Load from title screen. Sets method into Load only mode.
        """
        
        # Grabs the latest update to the save data
        self.update_player_data()
        
        data_report = self.data_summary()
        text_data_report = {}
        menu_flag = True
        menu_pos = 0
        text_menu_header = self.bfont.render("Save Menu",True,(0,0,0))
        if title_mode == False:
            text_save = self.bfont.render("Save",True,(0,0,0))
        else:
            text_save = self.bfont.render("Save",True,(100,100,100))
        text_load = self.bfont.render("Load",True,(0,0,0))
        text_cancel = self.bfont.render("Cancel",True,(0,0,0))
        text_data_headers = [  self.bfont.render("Current Data",True,(0,0,0)),
                                 self.bfont.render("Slot #1",True,(0,0,0)),
                                 self.bfont.render("Slot #2",True,(0,0,0)),
                                 self.bfont.render("Slot #3",True,(0,0,0)),
                                 self.bfont.render("Slot #4",True,(0,0,0)),
                                 self.bfont.render("Auto Save #1",True,(0,0,0)),
                                 self.bfont.render("Auto Save #2",True,(0,0,0))]
        
        def update_data_text(data_report):
            """
            # Function Name: update data text
            # Purpose: Turns a data report dictionary and returns the rendered text objects
            # Inputs: data_report - a data report list
            """
            text_data_report = []
            for save_data in data_report:
                if save_data == 'No Data':
                
                    text_data_report.append([ 
                                            self.sfont.render("Avg. Level: NO DATA",True,(0,0,0)),
                                            self.sfont.render("Missions Done: NO DATA",True,(0,0,0)),
                                            self.sfont.render("Comment: NO DATA",True,(0,0,0)),
                                            False
                                           ])
                else:
                    text_data_report.append([ 
                                            self.sfont.render("Avg. Level: "+save_data[0],True,(0,0,0)),
                                            self.sfont.render("Missions Done: "+save_data[1],True,(0,0,0)),
                                            self.sfont.render("Comment: "+str(save_data[2]),True,(0,0,0)),
                                            True
                                           ])
            return text_data_report
        
        text_data_report = update_data_text(data_report)
        
        while menu_flag:
   
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if (event.key == K_z or event.key == K_RETURN) and menu_pos in (0,1):
                        
                        # Selects Save (locked out of loading from the main screen)
                        if menu_pos == 0 and title_mode == False:
                            self.slot_select_loop(text_data_report,'s',title_mode)
                            # Refreshes data reports
                            data_report = self.data_summary()
                            text_data_report = {}
                            text_data_report = update_data_text(data_report)
                            
                        # Selects Load
                        if menu_pos == 1:
                            confirm = self.slot_select_loop(text_data_report,'l',title_mode)
                            # Refreshes data reports
                            data_report = self.data_summary()
                            text_data_report = {}
                            text_data_report = update_data_text(data_report)
                            # Exits save menu after loading
                            if confirm == True:
                                menu_flag = False
                                return True
                            
                    if event.key == K_UP:
                        if menu_pos > 0:
                            menu_pos -= 1
                        else:
                            menu_pos = 2
                    if event.key == K_DOWN:
                        if menu_pos < 2:
                            menu_pos += 1
                        else:
                            menu_pos = 0
                    if ((event.key == K_z or event.key == K_RETURN) and menu_pos == 2) or event.key == K_x:
                        menu_flag = False
                        return False
                        
            if menu_flag:        
                # Background image
                self.surface.blit(self.stats_bg,(0,0))
                # Header
                self.surface.blit(text_menu_header,(50,50))
                # Renders the options
                self.surface.blit(text_save,(70,70))
                self.surface.blit(text_load,(70,90))
                self.surface.blit(text_cancel,(70,110))    
                # Draws the cursor
                pygame.draw.circle(self.surface, (0,0,0), (60,78+20*menu_pos),5)
                
                # Renders Current Data (Does not display if on the main screen)
                if title_mode == False:
                    self.surface.blit(text_data_headers[0],(50,350))
                    self.surface.blit(text_data_report[0][0],(60,370))
                    self.surface.blit(text_data_report[0][1],(60,390))
                
                
                # Renders Saved Data
                self.render_save_slots(text_data_headers, text_data_report)
                
                pygame.display.flip()
                self.clock.tick(60)
        
    def slot_select_loop(self,text_data_report,mode,title_mode):
        """
        # Function Name: slot select loop
        # Purpose: Allows player to select a save slot to save/load into
        # Inputs: text data report: current summary of saved data
        #         mode - 's' for save, 'l' for load
        #         title_mode - True for load only title mode
        # Outputs:  True if save/load done
        #           False if save/load cancelled
        """
        
        menu_flag = True
        menu_pos = 0
        text_menu_header = self.bfont.render("Save Menu",True,(0,0,0))
        if title_mode == False:
            text_save = self.bfont.render("Save",True,(0,0,0))
        else:
            text_save = self.bfont.render("Save",True,(100,100,100))
        text_load = self.bfont.render("Load",True,(0,0,0))
        text_cancel = self.bfont.render("Cancel",True,(0,0,0))
        text_data_headers = [  self.bfont.render("Current Data",True,(0,0,0)),
                                 self.bfont.render("Slot #1",True,(0,0,0)),
                                 self.bfont.render("Slot #2",True,(0,0,0)),
                                 self.bfont.render("Slot #3",True,(0,0,0)),
                                 self.bfont.render("Slot #4",True,(0,0,0)),
                                 self.bfont.render("Auto Save #1",True,(0,0,0)),
                                 self.bfont.render("Auto Save #2",True,(0,0,0))
                                 ]
        confirmed = False
        if mode == 'l':
            slot_max = 5
        else:
            slot_max = 3
        
        while menu_flag:
   
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if (event.key == K_z or event.key == K_RETURN): 
                        
                        if mode == 's' or text_data_report[menu_pos+1][3] == True:
                            confirmed = self.data_confirm_loop(text_data_report, menu_pos, mode,title_mode)
                        
                        # If confirmed, jumps back to the top level save menu and returns a True flag
                        if confirmed == True:
                            menu_flag = False
                            return True
                            
                    if event.key == K_UP:
                        if menu_pos > 0:
                            menu_pos -= 1
                        else:
                            menu_pos = slot_max
                    if event.key == K_DOWN:
                        if menu_pos < slot_max:
                            menu_pos += 1
                        else:
                            menu_pos = 0
                    if event.key == K_x:
                        menu_flag = False
                        # Returns a False flag to inidicate that no selection has been made
                        return False
                        
            if menu_flag:        
                # Background image
                self.surface.blit(self.stats_bg,(0,0))
                # Header
                self.surface.blit(text_menu_header,(50,50))
                # Renders the options
                self.surface.blit(text_save,(70,70))
                self.surface.blit(text_load,(70,90))
                self.surface.blit(text_cancel,(70,110))    
                
                # Renders Current Data except under load only mode at the title
                if title_mode == False:
                    self.surface.blit(text_data_headers[0],(50,350))
                    self.surface.blit(text_data_report[0][0],(60,370))
                    self.surface.blit(text_data_report[0][1],(60,390))
                    
                # Draws the cursor
                pygame.draw.circle(self.surface, (0,0,0), (480,58+80*menu_pos),5)
                
                # Draws the save slots
                self.render_save_slots(text_data_headers, text_data_report)
                
                pygame.display.flip()
                self.clock.tick(60)
                
    def render_save_slots(self,text_data_headers,text_data_report):
        """
        # Function Name: render_save_slots
        # Purpose: Draws the data for all the manual and auto-save slots
        # Inputs:    text_data_headers - text represenations of each slot's title
        #            text_data_report - text representations of the information about the
        #                save data to display to the player
        """
    
        # Renders Saved Data
        for i in xrange(1,7):
            self.surface.blit(text_data_headers[i],(490,-30+(80)*i))
            self.surface.blit(text_data_report[i][0],(500,-10+(80)*i))
            self.surface.blit(text_data_report[i][1],(500,10+(80)*i))
            self.surface.blit(text_data_report[i][2],(500,30+(80)*i))
            
    def data_confirm_loop(self,text_data_report,slot,mode,title_mode):
        """
        # Function Name: data confirm loop
        # Purpose: Allows player to confirm saving/loading
        # Inputs:   text data report - current summary of saved data
        #           slot - slot selected to save into
        #           mode - save / load
        #           title_mode - True for load only title mode
        """
        
        menu_flag = True
        menu_pos = True
        text_menu_header = self.bfont.render("Save Menu",True,(0,0,0))
        if title_mode == False:
            text_save = self.bfont.render("Save",True,(0,0,0))
        else:
            text_save = self.bfont.render("Save",True,(100,100,100))
        text_load = self.bfont.render("Load",True,(0,0,0))
        text_cancel = self.bfont.render("Cancel",True,(0,0,0))
        text_yes = self.bfont.render("Yes",True,(0,0,0))
        text_no = self.bfont.render("No",True,(0,0,0))
        if mode == 's':
            text_confirm_dialog = self.bfont.render("Confirm Save?",True,(0,0,0))
            text_action_process = self.bfont.render("Saving...",True,(0,0,0))
            text_action_done = self.bfont.render("Saved!",True,(0,0,0))
        else:
            text_confirm_dialog = self.bfont.render("Confirm Load?",True,(0,0,0))
            text_action_process = self.bfont.render("Loading...",True,(0,0,0))
            text_action_done = self.bfont.render("Loaded!",True,(0,0,0))
            
        text_data_headers = [  self.bfont.render("Current Data",True,(0,0,0)),
                                 self.bfont.render("Slot #1",True,(0,0,0)),
                                 self.bfont.render("Slot #2",True,(0,0,0)),
                                 self.bfont.render("Slot #3",True,(0,0,0)),
                                 self.bfont.render("Slot #4",True,(0,0,0)),
                                 self.bfont.render("Auto Save #1",True,(0,0,0)),
                                 self.bfont.render("Auto Save #2",True,(0,0,0))]
        
        while menu_flag:
   
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if (event.key == K_z or event.key == K_RETURN):
                        if menu_pos == True:
                        
                                
                                #####################
                                # Before Save / Load
                                #####################
                                
                                # Background image
                                self.surface.blit(self.stats_bg,(0,0))
                                # Header
                                self.surface.blit(text_menu_header,(50,50))
                                # Renders the options
                                self.surface.blit(text_save,(70,70))
                                self.surface.blit(text_load,(70,90))
                                self.surface.blit(text_cancel,(70,110))    
                                
                                # Renders Current Data
                                if title_mode == False:
                                    self.surface.blit(text_data_headers[0],(50,350))
                                    self.surface.blit(text_data_report[0][0],(60,370))
                                    self.surface.blit(text_data_report[0][1],(60,390))
                                
                                # Renders Saved Data
                                self.render_save_slots(text_data_headers, text_data_report)
                                
                                self.surface.blit(self.results_panel,(280,140))
                                self.surface.blit(text_confirm_dialog,(300,160))
                                self.surface.blit(text_yes,(360,180))
                                self.surface.blit(text_no,(460,180))
                                
                                # Draws the cursor
                                pygame.draw.circle(self.surface, (0,0,0), (453-100*menu_pos,188),5)
                                
                                # Renders Current Data
                                if title_mode == False:
                                    self.surface.blit(text_data_headers[0],(300,210))
                                    self.surface.blit(text_data_report[0][0],(320,230))
                                    self.surface.blit(text_data_report[0][1],(320,250))
                                    
                                self.surface.blit(text_data_headers[slot+1],(300,300))
                                self.surface.blit(text_data_report[slot+1][0],(320,320))
                                self.surface.blit(text_data_report[slot+1][1],(320,340))

                                self.surface.blit(text_action_process,(380,380))
                                
                                pygame.display.flip()
                                self.clock.tick(60)
                                
                                #####################
                                # Save / Load Action
                                #####################
                                if mode == 's':
                                    self.player.save(slot+1)
                                else:
                                    self.load_player(slot+1)
                                
                                #####################
                                # After Save/Load
                                #####################
                                
                                # Background image
                                self.surface.blit(self.stats_bg,(0,0))
                                # Header
                                self.surface.blit(text_menu_header,(50,50))
                                # Renders the options
                                self.surface.blit(text_save,(70,70))
                                self.surface.blit(text_load,(70,90))
                                self.surface.blit(text_cancel,(70,110))    
                                
                                # Renders Current Data
                                if title_mode == False:
                                    self.surface.blit(text_data_headers[0],(50,350))
                                    self.surface.blit(text_data_report[0][0],(60,370))
                                    self.surface.blit(text_data_report[0][1],(60,390))
                                
                                
                                # Renders Saved Data
                                self.render_save_slots(text_data_headers, text_data_report)
                
                                
                                self.surface.blit(self.results_panel,(280,140))
                                self.surface.blit(text_confirm_dialog,(300,160))
                                self.surface.blit(text_yes,(360,180))
                                self.surface.blit(text_no,(460,180))
                                
                                # Draws the cursor
                                pygame.draw.circle(self.surface, (0,0,0), (453-100*menu_pos,188),5)
                                
                                # Renders Current Data
                                if title_mode == False:
                                    self.surface.blit(text_data_headers[0],(300,210))
                                    self.surface.blit(text_data_report[0][0],(320,230))
                                    self.surface.blit(text_data_report[0][1],(320,250))
                                
                                self.surface.blit(text_data_headers[slot+1],(300,300))
                                self.surface.blit(text_data_report[slot+1][0],(320,320))
                                self.surface.blit(text_data_report[slot+1][1],(320,340))

                                self.surface.blit(text_action_done,(380,380))
                                
                                pygame.display.flip()
                                self.clock.tick(60)
                                self.pause(1)
                                
                                return menu_pos
                        
                        else:
                            return False
                        
                    # If there is an arrow key press, toggle yes/no
                    if event.key in (K_UP,K_DOWN,K_RIGHT,K_LEFT):
                        menu_pos = not(menu_pos)
                        
                    if event.key == K_x:
                        menu_flag = False
                        return False
                        
            if menu_flag:        
                # Background image
                self.surface.blit(self.stats_bg,(0,0))
                # Header
                self.surface.blit(text_menu_header,(50,50))
                # Renders the options
                self.surface.blit(text_save,(70,70))
                self.surface.blit(text_load,(70,90))
                self.surface.blit(text_cancel,(70,110))    
                
                # Renders Current Data
                if title_mode == False:
                    self.surface.blit(text_data_headers[0],(50,350))
                    self.surface.blit(text_data_report[0][0],(60,370))
                    self.surface.blit(text_data_report[0][1],(60,390))
                
                
                # Renders Saved Data
                self.render_save_slots(text_data_headers, text_data_report)
                
                
                
                self.surface.blit(self.results_panel,(280,140))
                self.surface.blit(text_confirm_dialog,(300,160))
                self.surface.blit(text_yes,(360,180))
                self.surface.blit(text_no,(460,180))
                
                # Draws the cursor
                pygame.draw.circle(self.surface, (0,0,0), (453-100*menu_pos,188),5)
                
                # Renders Current Data
                if title_mode == False:
                    self.surface.blit(text_data_headers[0],(300,210))
                    self.surface.blit(text_data_report[0][0],(320,230))
                    self.surface.blit(text_data_report[0][1],(320,250))
                    
                self.surface.blit(text_data_headers[slot+1],(300,300))
                self.surface.blit(text_data_report[slot+1][0],(320,320))
                self.surface.blit(text_data_report[slot+1][1],(320,340))
                
                pygame.display.flip()
                self.clock.tick(60)
                
    def options_menu(self):
        """
        # Function Name: Options Menu
        # Purpose: Manages system options
        # Inputs: None
        """
        
        menu_flag = True
        menu_pos = 0
        text_menu_header = self.bfont.render("Options Menu",True,(0,0,0))
        text_yes = self.bfont.render("Yes",True,(0,0,255))
        text_no = self.bfont.render("No",True,(255,0,0))
        text_battle_anim = self.bfont.render("Show battle animations:",True,(0,0,0))
        text_turn_end = self.bfont.render("End turn after all units move:",True,(0,0,0))
        text_enable_music = self.bfont.render("Enable music:",True,(0,0,0))
        text_show_enemy_moves = self.bfont.render("Show enemy movements:",True,(0,0,0))
        text_auto_save = self.bfont.render("Auto save before/after missions:",True,(0,0,0))
        
        text_grid = self.bfont.render("Draw grid:",True,(0,0,0))
        text_save = self.bfont.render("SAVE SETTINGS",True,(0,0,0))
        text_cancel = self.bfont.render("CANCEL",True,(0,0,0))

        # Creates a copy of the options to restore in the event of canceling
        options_copy = (self.options.battle_anim,self.options.turn_end,self.options.grid,self.options.enable_music,
            self.options.show_enemy_moves,self.options.auto_save)

        
        while menu_flag:
        
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if (event.key == K_LEFT or event.key == K_RIGHT or event.key == K_z or event.key == K_RETURN) and menu_pos <= 5:
                        if menu_pos == 0:
                            # Toggles true/false the battle animations setting
                            self.options.battle_anim = not(self.options.battle_anim)
                        elif menu_pos == 1:
                            # Toggles true/false the turn end setting
                            self.options.turn_end = not(self.options.turn_end)
                        elif menu_pos == 2:
                            # Toggles the grid setting
                            self.options.grid = not(self.options.grid)
                        elif menu_pos == 3:
                            # Toggles the music setting
                            self.options.enable_music = not(self.options.enable_music)
                            
                            if not self.options.enable_music:
                                pygame.mixer.music.stop()
                            else:
                                if self.current_music:
                                    self.play_music(self.current_music)

                        elif menu_pos == 4:
                            # Toggles the show enemy movements setting.
                            self.options.show_enemy_moves = not(self.options.show_enemy_moves)
                                    
                        elif menu_pos == 5:
                            # Toggles the show enemy movements setting.
                            self.options.auto_save = not(self.options.auto_save)
                                    
                    if event.key == K_UP:
                        if menu_pos > 0:
                            menu_pos -= 1
                        else:
                            menu_pos = 7
                    if event.key == K_DOWN:
                        if menu_pos < 7:
                            menu_pos += 1
                        else:
                            menu_pos = 0
                    if (event.key == K_z or event.key == K_RETURN) and menu_pos == 6:
                        menu_flag = False
                        self.options.save()
                    if ((event.key == K_z or event.key == K_RETURN) and menu_pos == 7) or event.key == K_x:
                        menu_flag = False
                        self.options.battle_anim = options_copy[0]
                        self.options.turn_end = options_copy[1]
                        self.options.grid = options_copy[2]
                        self.options.enable_music = options_copy[3]
                        self.options.show_enemy_moves = options_copy[4]
                        self.options.auto_save = options_copy[5]
                    
                        # Stops/Starts music if player cancels selection.
                        if not self.options.enable_music:
                            pygame.mixer.music.stop()
                        else:
                            if self.current_music:
                                self.play_music(self.current_music)
                                
            if menu_flag:        
                # Background image
                self.surface.blit(self.stats_bg,(0,0))
                # Header
                self.surface.blit(text_menu_header,(50,50))
                # Draws the cursor
                pygame.draw.circle(self.surface, (0,0,0), (60,78+30*menu_pos),5)
                
                # Generates display of the options (Yes/No)
                self.surface.blit(text_battle_anim,(70,70))
                if self.options.battle_anim:
                    self.surface.blit(text_yes,(70+text_battle_anim.get_width()+10,70))
                else:
                    self.surface.blit(text_no,(70+text_battle_anim.get_width()+10,70))
                    
                self.surface.blit(text_turn_end,(70,100))
                if self.options.turn_end:
                    self.surface.blit(text_yes,(70+text_turn_end.get_width()+10,100))
                else:
                    self.surface.blit(text_no,(70+text_turn_end.get_width()+10,100))
                    
                self.surface.blit(text_grid,(70,130))
                if self.options.grid:
                    self.surface.blit(text_yes,(70+text_grid.get_width()+10,130))
                else:
                    self.surface.blit(text_no,(70+text_grid.get_width()+10,130))
                    
                self.surface.blit(text_enable_music,(70,160))
                if self.options.enable_music:
                    self.surface.blit(text_yes,(70+text_enable_music.get_width()+10,160))
                else:
                    self.surface.blit(text_no,(70+text_enable_music.get_width()+10,160))
                    
                self.surface.blit(text_show_enemy_moves,(70,190))
                if self.options.show_enemy_moves:
                    self.surface.blit(text_yes,(70+text_show_enemy_moves.get_width()+10,190))
                else:
                    self.surface.blit(text_no,(70+text_show_enemy_moves.get_width()+10,190))
                    
                    
                self.surface.blit(text_auto_save,(70,220))
                if self.options.auto_save:
                    self.surface.blit(text_yes,(70+text_auto_save.get_width()+10,220))
                else:
                    self.surface.blit(text_no,(70+text_auto_save.get_width()+10,220))
                    
                self.surface.blit(text_save,(70,250))
                self.surface.blit(text_cancel,(70,280))    
                
                pygame.display.flip()
                self.clock.tick(60)
    
    
    def treasure_menu(self):
        """
        # Function Name: treasure menu
        # Purpose: Allows player to view available treasures
        """
            
        text_all_treasure_data = []
        # Generates treasure objects
        for treasure_key in self.player.treasures.keys():
            text_single_treasure_data = []
            # Format (Text objects)
                # [name(string), big_name,small_name, quantity, type, description]
            text_single_treasure_data.append(self.treasure_catalog[treasure_key].name)
            text_single_treasure_data.append(self.bfont.render(self.treasure_catalog[treasure_key].name,True,(0,0,0)))
            text_single_treasure_data.append(self.sfont.render(self.treasure_catalog[treasure_key].name,True,(0,0,0)))
            text_single_treasure_data.append(self.sfont.render(str(self.player.treasures[treasure_key]),True,(0,0,0)))
            text_single_treasure_data.append(self.sfont.render(self.treasure_catalog[treasure_key].type,True,(0,0,0)))
            text_single_treasure_data.append([])
            desc_lines = split_lines(self.treasure_catalog[treasure_key].desc,50)
            [text_single_treasure_data[5].append(self.sfont.render(line,True,(0,0,0))) for line in desc_lines]
            text_all_treasure_data.append(text_single_treasure_data)
            text_single_treasure_data.append(self.treasure_catalog[treasure_key].icon)
            
        text_all_treasure_data.sort()
            
        menu_flag = True
        menu_pos = 0
        text_treasure_header = self.bfont.render("Treasures",True,(0,0,0))
        text_empty = self.sfont.render("Empty",True,(0,0,0))
        text_desc = self.bfont.render("Description",True,(0,0,0))
        text_quantity = self.bfont.render("Quantity",True,(0,0,0))
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_UP or event.key == K_LEFT:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == len(text_all_treasure_data)-1:
                            menu_pos = len(text_all_treasure_data)-1
                    if event.key == K_DOWN or event.key == K_RIGHT:
                        if menu_pos < len(text_all_treasure_data)-1:
                            menu_pos += 1
                        elif menu_pos == len(text_all_treasure_data)-1:
                            menu_pos = 0 
                    if event.key == K_x:
                        # Cancel
                        menu_flag = False
            
            if menu_flag:
                
                self.surface.blit(self.stats_bg,(0,0))
                self.surface.blit(self.treasure_header,(0,0))
                self.surface.blit(text_treasure_header,(50,50))
      
                # Draws cursor
                pygame.draw.circle(self.surface, (0,0,0), (60,95+24*menu_pos),5)
              
                # Displays the inventory
                item_counter = 0
                # Case: Treasure inventory is empty
                if not text_all_treasure_data:
                    self.surface.blit(text_empty,(70,90))
                else:    
                    for text_single_treasure in text_all_treasure_data:
                        item_counter += 1
                        self.surface.blit(self.icons[text_single_treasure[6]],(70,60+24*item_counter))
                        self.surface.blit(text_single_treasure[2],(100,65+24*item_counter))
                        self.surface.blit(text_single_treasure[3],(300,65+24*item_counter))
                        
                    # Renders currently selected treasure data
                    self.surface.blit(text_all_treasure_data[menu_pos][1],(470,50))
                    self.surface.blit(text_quantity,(480,70))
                    self.surface.blit(text_all_treasure_data[menu_pos][3],(490,90))
                    self.surface.blit(text_desc,(480,110))
                    [self.surface.blit(line,(490,130+20*index)) for index,line in enumerate(text_all_treasure_data[menu_pos][5])]
                    
                pygame.display.flip()
                self.clock.tick(60)
                
    def spell_swap_inventory(self,unit):
        """
        # Function name: spell_swap_inventory
        # Purpose: Selects spell to swap into the unit from the inventory
        # Inputs: None
        # Outputs: False - if no selection has been made
        #          selected spell, inventory_list - spell and inventory if a selection has been made
        """
        
        menu_flag = True
        shift = 0
        menu_pos = [0,0]
        
        text_inv_header = ( self.bfont.render("Attack Sorcery",True,(0,0,0)),
                          self.bfont.render("Healing and Support Sorcery",True,(0,0,0)),
                          self.bfont.render("Spell Cards",True,(0,0,0)),
                          self.bfont.render("Items",True,(0,0,0)))
                          
        text_none = self.sfont.render("None",True,(0,0,0))
        text_properties = self.bfont.render("Properties",True,(0,0,0))
        text_desc = self.bfont.render("Description",True,(0,0,0))
        text_stats_header = self.bfont.render("Stats",True,(0,0,0))
        text_spell_level = self.sfont.render("Level: ",True,(0,0,0))
       
        if self.player.items[0]:
            text_selected = self.get_selected_spell_data(self.player.items[0][0])
        else:
            text_selected = self.get_selected_spell_data(None)
        
        text_can_use = self.bfont.render("Can Use:",True,(0,0,0))
        text_inventory_list = self.get_inventory_list(0)
                                  
        text_directions1 = self.sfont.render("Z / Enter - Select Spell Action",True,(0,0,0))
        text_directions2 = self.sfont.render("C - See who can equip this Spell Action",True,(0,0,0))
        text_directions3 = self.sfont.render("X - Back",True,(0,0,0))
        text_directions4 = self.sfont.render("Left/Right - Cycle through Spell Action Types",True,(0,0,0))
        
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_UP:
                        if menu_pos[1] > 0:
                            menu_pos[1] -= 1
                        elif menu_pos[1] == len(self.player.items[menu_pos[0]])-1:
                            menu_pos[1] = len(self.player.items[menu_pos[0]])-1
                        # Requests new selected spell info
                        if self.player.items[menu_pos[0]]:
                            text_selected = self.get_selected_spell_data(self.player.items[menu_pos[0]][menu_pos[1]])
                    if event.key == K_DOWN:
                        if menu_pos[1] < len(self.player.items[menu_pos[0]])-1:
                            menu_pos[1] += 1
                        elif menu_pos[1] == len(self.player.items[menu_pos[0]])-1:
                            menu_pos[1] = 0 
                        # Requests new selected spell info
                        if self.player.items[menu_pos[0]]:
                            text_selected = self.get_selected_spell_data(self.player.items[menu_pos[0]][menu_pos[1]])
                    
                    if event.key == K_LEFT:
                        if menu_pos[0] > 0:
                            menu_pos[0] -= 1
                        elif menu_pos[0] == 0:
                            menu_pos[0] = 3 
                        # Resets the vertical position of the cursor when inventory slots are swapped
                        menu_pos[1] = 0                       
                        # Fetches the new inventory list
                        text_inventory_list = self.get_inventory_list(menu_pos[0])
                        # Requests new selected spell info
                        if self.player.items[menu_pos[0]]:
                            text_selected = self.get_selected_spell_data(self.player.items[menu_pos[0]][menu_pos[1]])
                        else:
                            text_selected = self.get_selected_spell_data(None)
                        
                    if event.key == K_RIGHT:
                        if menu_pos[0] < 3:
                            menu_pos[0] += 1
                        elif menu_pos[0] == 3:
                            menu_pos[0] = 0
                        # Resets the vertical position of the cursor when inventory slots are swapped
                        menu_pos[1] = 0
                        # Fetches the new inventory list
                        text_inventory_list = self.get_inventory_list(menu_pos[0])
                        # Requests new selected spell info
                        if self.player.items[menu_pos[0]]:
                            text_selected = self.get_selected_spell_data(self.player.items[menu_pos[0]][menu_pos[1]])
                        else:
                            text_selected = self.get_selected_spell_data(None)
                            
                    
                    if (event.key == K_z or event.key == K_RETURN) and self.player.items[menu_pos[0]]:
                        if self.player.items[menu_pos[0]][menu_pos[1]]:

                            # Runs the verification
                            can_equip = self.player.items[menu_pos[0]][menu_pos[1]].check_restrictions(unit)

                            if can_equip:
                                # Pull the spell from the inventory
                                selected_spell = self.player.items[menu_pos[0]].pop(menu_pos[1])
                                return selected_spell,menu_pos[0]
                        
                    if event.key == K_x:
                        # Cancel
                        menu_flag = False
                        return False
                
            keys = pygame.key.get_pressed()
            # If you hold down C, it will display who can equip the spell
            if keys[K_c]:
                can_equip_flag = True
            else:
                can_equip_flag = False

                        
            #: spell descriptions        
            if menu_flag:
                
                # Spell_data
                self.surface.blit(self.stats_bg,(0,0))
                self.surface.blit(self.item_category_header,(0,0))
                self.surface.blit(text_inv_header[menu_pos[0]],(50,50))
                                
                # Draws Item Category cursor
                self.surface.blit(self.item_category_cursor,(51+70*menu_pos[0],-10))
              
                # Displays the inventory
                for index,item in enumerate(text_inventory_list):
                    self.surface.blit(item[1],(70,80+index*20))
                    # If the inventory is not empty, plots the usage remaining as well
                    if item[0]:
                        self.surface.blit(item[2],(320,80+index*20))
                
                
                self.surface.blit(text_directions1,(60,500))
                self.surface.blit(text_directions2,(60,520))
                self.surface.blit(text_directions3,(60,540))
                self.surface.blit(text_directions4,(60,560))
                
                # Data display for selected spell action
                if self.player.items[menu_pos[0]]:
                    current_spell = self.player.items[menu_pos[0]][menu_pos[1]]
                    # Item Cursor
                    pygame.draw.circle(self.surface, (0,0,0), (60,88+20*menu_pos[1]),5)

                    # Display spell stats/ properties
                    if not can_equip_flag:
                        self.draw_spell_action_data(current_spell,text_selected)
                    
                    # Display who can use the spell
                    else:
                        self.surface.blit(text_can_use,(470,50))
                        counter_x = 0
                        counter_y = 0
                        # Checks each unit in the player's library
                        for equip_unit in self.player_units:
                            self.surface.blit(self.unit_tile,(500+70*counter_x,82+50*counter_y))
                            
                            if current_spell.check_restrictions(equip_unit):
                                self.surface.blit(equip_unit.image,(500+counter_x*70,80+counter_y*50),(0,0,35,35))
                            else:
                                self.surface.blit(equip_unit.image,(500+counter_x*70,80+counter_y*50),(105,0,35,35))
                            counter_x += 1
                            if counter_x == 4:
                                counter_x = 0
                                counter_y += 1
                
                pygame.display.flip()
                self.clock.tick(60)
            
    def data_summary(self):
        
        """
        # Function Name: data_summary
        # Purpose: Returns a summary of all player data (including saved Data)
        #               -Average Level
        #               -Mission Complete Count
        # Outputs: data_report - report containing above information
        #                        [Current Data, Save 1-4, Autosave]
        """
    
        def player_report(player):

            """
            # Function Name: player_report
            # Purpose: Returns a summary of player data
            #               -Average Level
            #               -Mission Complete Count
            # Inputs: Player Data
            """
            
            total_unit_level = 0
            missions_completed = 0
            # Sums up total unit level
            for unit_data in player.all_unit_data.values():
                total_unit_level += unit_data.level
            # Sums up amount of missions completed
            for event_data in player.all_event_data.values():
            
                if event_data.done:
                    missions_completed += 1
                
            # Subtract 1 from final mission complete count since prologue doesn't count
            if self.all_events_master['Prologue'].done:
                missions_completed -= 1
                    
            # computes the average unit level
            level_avg = str(int(total_unit_level / len(player.all_unit_data.values())))
            
            # Comment about the save data 
            comment = player.comment
            
            return (level_avg, str(missions_completed),comment)
        
        data_report = []
        data_report.append(player_report(self.player))
        
        # Iterates over all four save slots
        for i in xrange(1,5):
            # Attempts to load the data
            try:
                data_file = file(os.path.join('data','savedata'+str(i)+'.dat'),'rb')
                # Temporarily loads the player data 
                temp_player = pickle.load(data_file)
                data_report.append(player_report(temp_player))
            except IOError:
                data_report.append("No Data")
                
                
        # Attempts to load the Autosave Slot 1
        try:
            data_file = file(os.path.join('data','autosave01.dat'),'rb')
            # Temporarily loads the player data 
            temp_player = pickle.load(data_file)
            data_report.append(player_report(temp_player))
        except IOError:
            data_report.append("No Data")
            
            
        # Attempts to load the Autosave Slot 2
        try:
            data_file = file(os.path.join('data','autosave02.dat'),'rb')
            # Temporarily loads the player data 
            temp_player = pickle.load(data_file)
            data_report.append(player_report(temp_player))
        except IOError:
            data_report.append("No Data")
                
        return data_report

    def load_player(self,slot):
    
        """
        # Function Name: load_player
        # Purpose: loads player data from file
        # Inputs: slot - save slot to load from
        """
    
        # Attempts to load game save data
        try:
            self.player = self.player.load(slot)
            
        # If it fails, aborts load
        except IOError:
            print "No Player Data Available"
            return
        
        # Flushes the state of all player units before loading
        [unit.reset_state() for unit in self.player_units_catalog.values()]
        
        # Drops all party members from current game
        self.player_units = []
        self.player_units_by_name = {}
            
        # Brings back all party members in save data
        for unit_name in self.player.party_members:
            print unit_name+" being added to party"
            added_unit = self.player_units_catalog[unit_name]
            if added_unit.map == None:
                added_unit.map = self.worldmap
            self.player_units.append(added_unit)
            self.player_units_by_name[added_unit.name] = added_unit
        
        # Update to units
        for unit in self.player_units:
            self.player.all_unit_data[unit.name].update_to_unit(unit)
            unit.HP = unit.maxHP
            
        # Adds in any new missions
        for event_id in self.all_events_master.keys():
            if event_id not in self.player.all_event_data.keys():
                print "New Mission detected! "+event_id
                # Adds the event to the player event data
                self.player.all_event_data[event_id] = EventData(self.all_events_master[event_id])


        # Update to events
        for id in self.player.all_event_data.keys():

            if id in self.all_events_master.keys():
                #Updates completion data for events
                self.player.all_event_data[id].update_to_event(self.all_events_master[id])
            else:
                # If an event no longer exists, delete it from the data
                del self.player.all_event_data[id]

        # Calls for a world map refresh
        self.worldmap.update_all_events()
                
    def update_player_data(self):
    
        """
        # Function Name: update player data
        # Purpose: updates the stored unit data
        """
    
        # Update from units
        for unit in self.player_units:
            self.player.all_unit_data[unit.name].update_from_unit(unit)
        
        # Update from events
        for event in self.all_events_master.values():
            self.player.all_event_data[event.event_id].update_from_event(event)
            
        self.worldmap.update_all_events()
            
    ############################################
    # Inventory Management Functions
    ############################################
    
    def get_inventory_list(self,list_num):
        """
        # Function Name: get_inventory_list
        # Purpose: Generates the text objects of the currently selected inventory list
        # Inputs: list - 0 for Offensive Spells
        #                1 for Healing Spells
        #                2 for Spell Cards
        #                3 for Items
        #                4 for Misc
        """
        
        text_inventory_list = []
        for item in self.player.items[list_num]:
            if not item:
                text_inventory_list.append((None,self.sfont.render('Empty',True,(100,100,100))))
            else:
                text_inventory_list.append((item.name,self.sfont.render(item.name,True,(0,0,0)),self.sfont.render(str(item.livesleft)+"/"+str(item.lives),True,(0,0,0))))
        return text_inventory_list
    
    def get_selected_spell_data(self, spell_action):
    
        """
        # Function Name: get_selected_spell_data
        # Purpose: Generates the text objects of the currently selected item in the inventory
        # Inputs: spell_action - a spell or an item
        """
    
        
        if spell_action:
            text_selected_spell  = [spell_action.name,self.bfont.render(spell_action.namesuffix,True,(0,0,0))]
            text_selected_spell.append(self.sfont.render("STR:  "+str(spell_action.strmod),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("DEF:  "+str(spell_action.defmod),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("MAG:  "+str(spell_action.magmod),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("MDEF:  "+str(spell_action.mdefmod),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("ACC:  "+str(spell_action.accmod),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("AGL:  "+str(spell_action.aglmod),True,(0,0,0)))
            
            # Healing item percent heal case
            if spell_action.type == 'healingitem' and spell_action.effect_class == 'percent':
                text_selected_spell.append(self.sfont.render("POWER:  "+str(int(spell_action.effect))+"%",True,(0,0,0)))
                
            # Everything else case
            else:
                text_selected_spell.append(self.sfont.render("POWER:  "+str(spell_action.effect),True,(0,0,0)))
            
            text_selected_spell.append(self.sfont.render("SHIELD:  "+str(spell_action.shield),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("CRIT:  "+str(spell_action.critmod),True,(0,0,0)))
            if spell_action.livesleft > 0:
                text_selected_spell.append(self.sfont.render("USES: "+str(spell_action.livesleft)+"/"+str(spell_action.lives),True,(0,0,0)))
            else:
                text_selected_spell.append(self.sfont.render("USES: "+str(spell_action.livesleft)+"/"+str(spell_action.lives),True,(255,0,0)))
            text_selected_spell.append(self.sfont.render("Req SC:  %3.0f"%spell_action.unlock,True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("RANGE:  "+str(spell_action.minrange)+" - "+str(spell_action.spellrange),True,(0,0,0)))
            text_selected_spell.append(self.sfont.render("SC Cost: %s"%(str(spell_action.sc_cost)),True,(0,0,0)))
                
            description = split_lines(spell_action.desc,40)
            text_selected_spell.append([])
            for line in description:
                text_selected_spell[15].append(self.sfont.render(line,True,(0,0,0)))

            # Status Effect 
            if spell_action.type == "attack":
                text_selected_spell.append([(status_effect, 
                                            self.sfont.render(status_effect,True,(0,0,0)),
                                            self.sfont.render(str(percent)+"%",True,(0,0,0))) 
                                            for status_effect,percent in spell_action.status_effects.items()])
            elif spell_action.type == 'support':
                 text_selected_spell.append([(status_effect, 
                                            self.sfont.render(status_effect,True,(0,0,0)),
                                            self.sfont.render("100%",True,(0,0,0))) 
                                            for status_effect in spell_action.status_effects])
            elif spell_action.type == 'healing' or spell_action.type == 'healingitem':
                 text_selected_spell.append([(status_effect,
                                            self.sfont.render(status_effect,True,(0,0,0)),
                                            self.sfont.render("100%",True,(0,0,0))) 
                                            for status_effect in spell_action.status_effects])
            else:
                print "unrecognized spell type for status effect display"
                
                

        else:
            text_selected_spell = [None,self.bfont.render("Empty",True,(0,0,0))]
        
        return text_selected_spell
        
        
    def draw_uncreated_spell(self,spell_action_key):
        """
        # Function name: draw_spell_data
        # Purpose: draws information about a single spell that has not already been created
        # Inputs: Spell action key - name of spell action
        """
        # Generates data about the spell from a temporary spell action
        temp_spell = self.spell_catalog[spell_action_key].construct_spell()
        text_spell_data = self.get_selected_spell_data(temp_spell)
        self.draw_spell_action_data(temp_spell, text_spell_data)
        
    def draw_spell_action_data(self,spell_action,text_spell_data):
        """
        # Function name: draw_spell_data
        # Purpose: draws information about a single spell on the right panel
        # Inputs: spell_action - spell action to draw
        #         text_spell_data - rendered text generated with self.get_selected_spell_data
        """
        
        # Preliminary text items
        text_none = self.sfont.render("None",True,(0,0,0))
        text_properties = self.bfont.render("Properties",True,(0,0,0))
        text_desc = self.bfont.render("Description",True,(0,0,0))
        text_stats_header = self.bfont.render("Stats",True,(0,0,0))
        text_spell_level = self.sfont.render("Level: ",True,(0,0,0))
        text_give_status = self.bfont.render("Gives Status Effects:",True,(0,0,0))
        text_remove_status = self.bfont.render("Removes Status Effects:",True,(0,0,0))
        
        self.surface.blit(text_spell_data[1],(470,50))
        self.surface.blit(text_properties,(490,80))

        # Icons
        self.surface.blit(self.spell_icons_big, (500,105),spell_action.a_type_big)
            
        # DAMAGE TYPE
        self.surface.blit(self.spell_icons_big, (525,105),spell_action.d_type_big)
                
        # SPELL TYPE
        self.surface.blit(self.spell_icons_big, (550,105),spell_action.al_type_big)
        
        # Counterattack
        self.surface.blit(self.spell_icons_big, (575,105),spell_action.c_type_big)
        self.surface.blit(text_spell_level,(500,130))
            
        self.surface.blit(self.level_stars,(540,130),(0,spell_action.spell_rank*15,50,15))
        
        
        self.surface.blit(text_desc,(490,150))
        
        # Spell Description
        [self.surface.blit(line,(500,170+count*20)) for count, line in enumerate(text_spell_data[15])]
        
        self.surface.blit(text_stats_header,(490,310))

        # EFFECT/SHIELD
        self.surface.blit(text_spell_data[8],(500,330))
        self.surface.blit(text_spell_data[9],(500,350))
        
        # RANGE/LIFE
        self.surface.blit(text_spell_data[13],(600,330))
        self.surface.blit(text_spell_data[11],(600,350))
        
        # SPIRIT Required / Cost
        self.surface.blit(text_spell_data[12],(700,330))
        self.surface.blit(text_spell_data[14],(700,350))
        
        # Raw Stat Modifiers
        
        # STR/DEF/CRIT
        self.surface.blit(text_spell_data[2],(500,370))
        self.surface.blit(text_spell_data[3],(500,390))
        self.surface.blit(text_spell_data[10],(500,410))
        # MAG/MDEF
        self.surface.blit(text_spell_data[4],(600,370))
        self.surface.blit(text_spell_data[5],(600,390))
        # AGL/ACC
        self.surface.blit(text_spell_data[6],(700,370))
        self.surface.blit(text_spell_data[7],(700,390))
    
        if spell_action.status_effects:
            if spell_action.type == "attack" or spell_action.type == 'support':
                self.surface.blit(text_give_status,(490, 430))
                
            elif spell_action.type == 'healingitem' or spell_action.type == 'healing':
                self.surface.blit(text_remove_status,(490, 430))
            else:
                print "Unrecognized spell type for drawing status effects"+spell_action.type
            
            for index, status_data in enumerate(text_spell_data[16]):
                name,text_name,percent = status_data
                
                # draw the SE icon
                self.surface.blit(self.status_effects_catalog[name].icon,(500,450+32*index))
                
                # draw the name of the SE
                self.surface.blit(text_name,(540,455+32*index))
                
                # draw the percent effectivness of the SE
                self.surface.blit(percent,(700,455+32*index))
                
                
    
        
    def draw_trait_data(self,text_selected_trait):
    
        """
        # Function name: draw_trait_data
        # Purpose: Draws the trait information on the RHS stats screen
        # Inputs: text_selected_trait - text representation of trait properties
        """
        
        # Header Data
        text_type = self.bfont.render("Type",True,(0,0,0))
        text_desc = self.bfont.render("Description",True,(0,0,0))
        text_trait_stats = self.bfont.render("Stat Boosts",True,(0,0,0))
        text_properties = self.bfont.render("Properties",True,(0,0,0))
        
        self.surface.blit(text_selected_trait[2],(470,50))
        self.surface.blit(text_type,(490,80))
        self.surface.blit(text_selected_trait[3],(500,100))
        self.surface.blit(text_desc,(490,120))
        
        [self.surface.blit(line,(500,140+20*count)) for count,line in enumerate(text_selected_trait[4])]
        
        self.surface.blit(text_trait_stats,(490,240))
        
        # STR and DEF
        self.surface.blit(text_selected_trait[5],(500,260))
        self.surface.blit(text_selected_trait[6],(500,280))
        
        # MAG and MDEF
        self.surface.blit(text_selected_trait[7],(600,260))
        self.surface.blit(text_selected_trait[8],(600,280))
        
        # AGL and ACC
        self.surface.blit(text_selected_trait[9],(700,260))
        self.surface.blit(text_selected_trait[10],(700,280))
        
        # Movement, EXP, Spirit
        self.surface.blit(text_selected_trait[11],(500,300))
        self.surface.blit(text_selected_trait[12],(600,300))
        self.surface.blit(text_selected_trait[13],(700,300))
        
        
        # Additional Properties
        self.surface.blit(text_properties,(490,320))
        [self.surface.blit(property_text,(500,340+20*count)) for count,property_text in enumerate(text_selected_trait[14])]
     
    def draw_treasure_data(self,treasure_key):

        """
        # Function name: draw_treasure_data
        # Purpose: draws information about a single treasure
        # Inputs: treasure_key - treasure id string
        """
    
        text_treasure = self.bfont.render("Treasure",True,(0,0,0))
        text_desc = self.bfont.render("Description",True,(0,0,0))
        text_single_treasure_data = self.get_single_treasure_data(treasure_key)
        # Renders currently selected treasure data
        self.surface.blit(text_single_treasure_data[0],(470,50))
        self.surface.blit(text_treasure,(470,70))
        self.surface.blit(text_desc,(480,110))
        self.surface.blit(text_single_treasure_data[1],(490,130))
        
    def get_text_selected_trait(self,trait):
                    
        """
        # Function name: get_text_selected_trait
        # Purpose: From a single trait, returns a list of text objects with the data
        # Inputs: trait - trait to be processed
        # Outputs: text_selected_trait - a list containing the text objects for the trait
        """
        
        if trait:
            text_selected_trait = [trait.name,self.sfont.render(trait.name,True,(0,0,0))]
            
            text_selected_trait.append(self.bfont.render(trait.name,True,(0,0,0)))
            text_selected_trait.append(self.sfont.render(trait.variation,True,(0,0,0)))
            text_selected_trait.append([])
            [text_selected_trait[4].append(self.sfont.render(line,True,(0,0,0))) for line in split_lines(trait.desc,50)]
            
            text_selected_trait.append(self.sfont.render("STR: "+str(trait.statmods[0]*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("DEF: "+str(trait.statmods[1]*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("MAG: "+str(trait.statmods[2]*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("MDEF: "+str(trait.statmods[3]*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("ACC: "+str(trait.statmods[4]*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("AGL: "+str(trait.statmods[5]*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("MOVE: "+str(trait.movemod_add),True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("EXP: "+str(trait.expmod*100)+"%",True,(0,0,0)))
            text_selected_trait.append(self.sfont.render("SC: "+str(trait.spiritmod*100)+"%",True,(0,0,0)))
            text_selected_trait.append([])

            
            # Displays the properties
            if trait.properties != []:
                for property in sorted(trait.properties):
                    text_selected_trait[14].append(self.sfont.render(property,True,(0,0,0)))
                    
        # Case for empty trait slot
        else:
            text_selected_trait = [None,self.sfont.render("Empty",True,(100,100,100))]

        return text_selected_trait
                   
    def splash_screen(self):
        """
        # Function Name: splash screen
        # Purpose: displays the splash screen(s)
        """
        
        self.surface.fill((255,255,255))
        pygame.display.flip()
        self.clock.tick(60)
        self.surface.blit(self.splash_fmp,(0,0))
        self.fade_from('white',0.5)
        self.pause(2.5)
        self.fade_to('white',0.5)
        
    def title_screen(self):
        """
        # Function Name: title screen
        # Purpose: Title screen menu
        """
        
        if self.enable_splash:
            self.splash_screen()
        
        text_version = self.sfont.render("Version 0.2.1",True,(255,255,255))
        menu_flag = True
        menu_pos = 0
        fade_flag = False

        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                    
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 3
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 3:
                            menu_pos += 1
                        elif menu_pos == 3:
                            menu_pos = 0
                            
                    # Selection
                    if event.key == K_z or event.key == K_RETURN:
                        if menu_pos == 0:
                            self.fade_to('black')
                            if self.enable_prologue:
                                #self.all_events_master['Prologue'].location = self.worldmap.all_locations_by_name[self.all_events_master['Prologue'].location_name]
                                self.all_events_master['Prologue'].execute()
                            
                            self.all_events_master['Prologue'].done = True
                            self.worldmap.update_all_events()
                            
                            self.fade_to('black')
                            self.launch_wm()
                    
                        elif menu_pos == 1:
                            
                            self.all_events_master['Prologue'].done = True
                            self.worldmap.update_all_events()
                            
                            self.fade_to('black')
                            confirm_load = self.save_menu(True)
                            if confirm_load == True:
                                self.launch_wm(True)
                            fade_flag = False
                    
                        elif menu_pos == 2:
                            self.fade_to('black')
                            self.options_menu()
                            fade_flag = False
                        
                        elif menu_pos == 3:
                            exit()
                        
            self.surface.blit(self.title,(0,0))
            self.surface.blit(self.title_cursor,(770,110+60*menu_pos))

            if fade_flag == False:
                self.fade_from('white')
                fade_flag = True
                
            pygame.display.update(Rect(770,110,23,203))
            self.clock.tick(60)
            
    def launch_wm(self,load_player=False):
        """
        # Function Name: launch world map
        # Purpose: start the world map
        # Inputs: load_player - False if new game, True if load from save game
        """
        
        # If new game, invokes navigate loop fresh start
        if not load_player:
            self.worldmap.navigate_loop()
        # Game has been loaded and new map is called up with saved player position
        while True:
            self.worldmap.navigate_loop(True)
            
