import pygame
from lostsky.core.utils import split_lines
from pygame.locals import *
from sys import exit

class MissionManager(object):

    def __init__(self,engine):

        """
        # Function Name: __init__
        # Purpose: Initiates the mission manager object
        # Inputs: Engine - engine that the mission manager is associated with
        """
        self.engine = engine
        
    ####################################
    # Computational Methods
    ####################################
    
    def sign_up_event(self,event):
        """
        # Function Name: sign up event
        # Purpose: Given an event, signs up for that event
        # Inputs: Event - Event to be signed up for
        """
    
        # Event is now signed up for 
        event.sign_up = True
        # Event is now no longer available
        event.available = False
        # Refreshes world map
        self.engine.worldmap.update_all_events()
        
    def cancel_event(self,event):
    
        """
        # Function Name: cancel event
        # Purpose: Given an event, cancel that event
        # Inputs: Event - Event to be cancelled
        """
        # Event is now cancelled
        event.sign_up = False
        # Event is now available
        event.available = True
        # Event has been manually cancelled
        # Keeps automatically signed up events from being readded
        event.manual_cancel = True
        # Refreshes world map
        self.engine.worldmap.update_all_events()
        
        
    def get_num_sign_ups(self):
    
        """
        # Function Name: get num sign ups
        # Purpose: get the number of signed up missions
        """
    
        if not self.engine.all_events_sign_up:
            return 0
        else:
            return len(self.engine.all_events_sign_up)
        
    def get_num_available(self):
    
        """
        # Function Name: get num available missions
        # Purpose: get the number of available missions
        """
    
        if not self.engine.all_events_available:
            return 0
        else:
            return len(self.engine.all_events_available)
        
    def get_num_completed(self):     

        """
        # Function Name: get num completed
        # Purpose: get the number of completed missions
        """
    
        if not self.engine.all_events_completed:
            return 0
        else:
            return len(self.engine.all_events_completed)
        
    ####################################
    # Graphical Methods
    ####################################
        
    def get_mission_stats(self):
    
        """
        # Function Name; Get mission stats
        # Purpose: Returns text objects of the number of missions available, signed up, and completed
        # Output: text_mission_stats - rendered text list containing above information
        """
    
        text_mission_stats = []
        text_mission_stats.append(self.engine.sfont.render("Missions Available: "+str(self.get_num_available()),True,(0,0,0)))
        text_mission_stats.append(self.engine.sfont.render("Signed Up: "+str(self.get_num_sign_ups())+" / 8",True,(0,0,0)))
        text_mission_stats.append(self.engine.sfont.render("Completed: "+str(self.get_num_completed()),True,(0,0,0)))
        return text_mission_stats
    
    def get_event_list(self,event_list):
    
        """
        # Function Name: get_event_list
        # Purpose: get a list of event names
        # Input: event_list - a list of events (e.g. all available events
        # Output: text_event_list - a list of rendered text objects containing the list of event names
        """
    
        text_event_list = []
        for event in event_list:
            text_event_list.append(self.engine.sfont.render(event.name,True,(0,0,0))) 
        
        # If event list is empty, show a No Missions Available message
        if not event_list:
            text_event_list.append(self.engine.sfont.render('No Missions Available',True,(100,100,100))) 
            
        return text_event_list
        
    def get_event_text_data(self,event):
    
        """
        # Function Name: get_event_text_data
        # Purpose: get the event name, location, type, etc
        # Input: event - the event to get information from
        # Output: text_event - rendered text of the information
        """
    
        text_event = []
        if event:
            # Event Name
            text_event.append(self.engine.sfont.render(event.name,True,(0,0,0)))
            # Event Type ("Mission"/"Conversation")
            text_event.append(self.engine.sfont.render(event.type,True,(0,0,0)))
            # Event Location
            text_event.append(self.engine.sfont.render(event.location_name,True,(0,0,0)))
            # Event Cancellable
            if event.cancellable == True:
                text_event.append(self.engine.sfont.render("Yes",True,(0,0,0)))
            else:
                text_event.append(self.engine.sfont.render("No",True,(0,0,0)))
            
            text_event.append([])
            for reward in event.map_data.rewards_list:
                if reward[0] == "spell_action":
                    text_event[4].append(self.engine.sfont.render(reward[1],True,(0,0,0))) 
                elif reward[0] == "treasure":
                    text_event[4].append(self.engine.sfont.render(self.engine.treasure_catalog[reward[1]].name,True,(0,0,0))) 
    
            if len(event.map_data.rewards_list) == 0:
                text_event[4].append(self.engine.sfont.render('None',True,(0,0,0)))
            
            text_event.append([self.engine.sfont.render(line,True,(0,0,0)) for line in
                               split_lines(event.desc,70)])
            
        return text_event
        
    ####################################
    # Interaction Methods
    ####################################
        
    def index_menu(self):
    
        """
        # Function Name: index_menu
        # Purpose: Index Menu - Select from missions, rumors, archives, return to map
        # Inputs: None
        """
    
        text_frontpage = self.engine.bfont.render("Front Page",True,(0,0,0))
        text_index = self.engine.bfont.render("Index",True,(0,0,0))
        
        text_missions = self.engine.bfont.render("Mission Sign Up",True,(0,0,0))
        text_current = self.engine.bfont.render("Current Missions",True,(0,0,0))
        text_rumors = self.engine.bfont.render("Rumors",True,(0,0,0))
        text_archives = self.engine.bfont.render("Archives",True,(0,0,0))
        text_cancel = self.engine.bfont.render("Cancel",True,(0,0,0))
        
        text_latest = self.engine.bfont.render("Top News",True,(0,0,0))
        
        current_story = self.engine.news_reports_list[0]
        for report in self.engine.news_reports_list[1:]:
            if self.engine.all_events_master[report['mission']].done == True:
                current_story = report
        
        currentstory_title = self.engine.bfont.render(current_story['name'],True,(0,0,0))
        currentstory_author = self.engine.sfont.render("By: "+current_story['author'],True,(0,0,0))
        currentstory_location = self.engine.sfont.render(current_story['location'],True,(0,0,0))
        
        currentstory_lines = split_lines(current_story['text'],70)
        text_lines = []
        [text_lines.append(self.engine.sfont.render(line,True,(0,0,0))) for line in currentstory_lines]
        
        menu_flag = 0
        menu_pos = 0
        
        while menu_flag == 0:
                
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 4
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 4:
                            menu_pos += 1
                        elif menu_pos == 4:
                            menu_pos = 0
                    if event.key == K_z or event.key == K_RETURN:
                        # Sign Up Menu
                        if menu_pos == 0:
                            self.sign_up_menu()
                        # Current Missions Menu
                        elif menu_pos == 1: 
                            self.current_missions_menu()
                        # Rumors Menu
                        elif menu_pos == 2:
                            self.rumors_menu()
                        # Archives Menu
                        elif menu_pos == 3:
                            self.archives_menu()
                            
                    if event.key == K_x or ((event.key == K_z or event.key == K_RETURN) and menu_pos == 4):
                        menu_flag = 1
                            
            if menu_flag == 0:
                self.engine.surface.blit(self.engine.mission_bg,(0,0))
                self.engine.surface.blit(text_frontpage,(35,105))
                self.engine.surface.blit(text_index,(35,150))
                self.engine.surface.blit(text_missions,(55,170))
                self.engine.surface.blit(text_current,(55,190))
                self.engine.surface.blit(text_rumors,(55,210))
                self.engine.surface.blit(text_archives,(55,230))
                self.engine.surface.blit(text_cancel,(55,250))
                
                self.engine.surface.blit(text_latest,(350,150))
                self.engine.surface.blit(currentstory_title,(360,170))
                self.engine.surface.blit(currentstory_author,(370,190))
                self.engine.surface.blit(currentstory_location,(370,210))
                for line_num in xrange(0,len(text_lines)):
                    self.engine.surface.blit(text_lines[line_num],(360,230+line_num*15))
                
                pygame.draw.circle(self.engine.surface, (0,0,0), (45,178+20*menu_pos),5)
                pygame.display.flip()
                self.engine.clock.tick(30)
    
    def sign_up_menu(self):
        """
        # Function Name: sign_up_menu
        # Purpose: Sign up menu
        """
        text_sign_up_menu = self.engine.bfont.render("Mission Sign-Up",True,(0,0,0))
        text_missions = self.engine.bfont.render("Missions",True,(0,0,0))
        
        text_name = self.engine.bfont.render("Mission Name",True,(0,0,0))
        text_type = self.engine.bfont.render("Type",True,(0,0,0))
        text_location = self.engine.bfont.render("Location",True,(0,0,0))
        text_cancellable = self.engine.bfont.render("Can Cancel?",True,(0,0,0))
        text_rewards = self.engine.bfont.render("Rewards",True,(0,0,0))
        text_desc = self.engine.bfont.render("Description",True,(0,0,0))
        
        text_directions1 = self.engine.sfont.render("Z / Enter - Sign Up",True,(0,0,0))
        text_directions2 = self.engine.sfont.render("X - Back",True,(0,0,0))
        
        text_event_list = self.get_event_list(self.engine.all_events_available)
        if self.engine.all_events_available:
            text_selected_event = self.get_event_text_data(self.engine.all_events_available[0])
        text_mission_stats = self.get_mission_stats()
        
        menu_flag = True
        menu_pos = 0

        while menu_flag:
            
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ) and self.engine.all_events_available:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(self.engine.all_events_available)-1
                        text_selected_event = self.get_event_text_data(self.engine.all_events_available[menu_pos])
        
                    if ( event.key == K_DOWN or event.key == K_RIGHT ) and self.engine.all_events_available:
                        if menu_pos < len(self.engine.all_events_available)-1:
                            menu_pos += 1
                        elif menu_pos == len(self.engine.all_events_available)-1:
                            menu_pos = 0
                        text_selected_event = self.get_event_text_data(self.engine.all_events_available[menu_pos])
            
                    if event.key == K_z or event.key == K_RETURN:
                        if self.engine.all_events_available and self.get_num_sign_ups() < 8:
                            # Sign up for the event
                            self.sign_up_event(self.engine.all_events_available[menu_pos])
                            if menu_pos > len(self.engine.all_events_available)-1:
                                menu_pos = len(self.engine.all_events_available)-1
                            
                            if self.engine.all_events_available:
                                text_selected_event = self.get_event_text_data(self.engine.all_events_available[menu_pos])
                            text_event_list = self.get_event_list(self.engine.all_events_available)
                            text_mission_stats = self.get_mission_stats()
                    
                    if event.key == K_x:
                        menu_flag = False
        
            if menu_flag:
                self.engine.surface.blit(self.engine.mission_bg,(0,0))
                self.engine.surface.blit(text_sign_up_menu,(35,105))
                self.engine.surface.blit(text_mission_stats[0],(200,108))
                self.engine.surface.blit(text_mission_stats[1],(400,108))
                self.engine.surface.blit(text_mission_stats[2],(600,108))
                
                self.engine.surface.blit(text_missions,(35,150))
                counter = 0

                [self.engine.surface.blit(event_name,(55,170+index*15)) 
                     for index, event_name in enumerate(text_event_list)]
                
                if self.engine.all_events_available:
                    
                    
                    self.engine.surface.blit(text_name,(350,150))
                    self.engine.surface.blit(text_selected_event[0],(360,170))
                    self.engine.surface.blit(text_type,(350,190))
                    self.engine.surface.blit(text_selected_event[1],(360,210))
                    self.engine.surface.blit(text_location,(350,230))
                    self.engine.surface.blit(text_selected_event[2],(360,250))
                    self.engine.surface.blit(text_cancellable,(350,270))
                    self.engine.surface.blit(text_selected_event[3],(360,290))
                    if self.engine.all_events_available[menu_pos].show_rewards:
                        self.engine.surface.blit(text_rewards,(600,150))
                        [self.engine.surface.blit(reward,(610,170+index*20)) 
                         for index, reward in enumerate(text_selected_event[4])]
                    
                    self.engine.surface.blit(text_desc,(350,310))
                    [self.engine.surface.blit(line,(360,330+20*index))
                          for index,line in enumerate(text_selected_event[5])]
                    
                    
                    pygame.draw.circle(self.engine.surface, (0,0,0), (45,178+15*menu_pos),5)
                    
                self.engine.surface.blit(text_directions1,(60,520))
                self.engine.surface.blit(text_directions2,(60,540))
                
                pygame.display.flip()
                self.engine.clock.tick(30)
                
    
    def current_missions_menu(self):
        """
        # Function Name: current missions menu
        # Purpose: Allows the player to view the current missions and cancel if desired
        """
    
        text_title = self.engine.bfont.render("Current Missions",True,(0,0,0))
        text_missions = self.engine.bfont.render("Missions",True,(0,0,0))
        
        text_name = self.engine.bfont.render("Mission Name",True,(0,0,0))
        text_type = self.engine.bfont.render("Type",True,(0,0,0))
        text_location = self.engine.bfont.render("Location",True,(0,0,0))
        text_cancellable = self.engine.bfont.render("Can Cancel?",True,(0,0,0))
        text_rewards = self.engine.bfont.render("Rewards",True,(0,0,0))
        text_desc = self.engine.bfont.render("Description",True,(0,0,0))
        
        
        text_directions1 = self.engine.sfont.render("Z / Enter - Cancel Mission",True,(0,0,0))
        text_directions2 = self.engine.sfont.render("X - Back",True,(0,0,0))
        
        text_event_list = self.get_event_list(self.engine.all_events_sign_up)
        if self.engine.all_events_sign_up:
            text_selected_event = self.get_event_text_data(self.engine.all_events_sign_up[0])
        else:
            text_selected_event = self.get_event_text_data(None)
        text_mission_stats = self.get_mission_stats()
        
        menu_flag = 0
        menu_pos = 0

        while menu_flag == 0:
            
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ) and self.engine.all_events_sign_up:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(self.engine.all_events_sign_up)-1
                        text_selected_event = self.get_event_text_data(self.engine.all_events_sign_up[menu_pos])
        
                    if ( event.key == K_DOWN or event.key == K_RIGHT ) and self.engine.all_events_sign_up:
                        if menu_pos < len(self.engine.all_events_sign_up)-1:
                            menu_pos += 1
                        elif menu_pos == len(self.engine.all_events_sign_up)-1:
                            menu_pos = 0
                        text_selected_event = self.get_event_text_data(self.engine.all_events_sign_up[menu_pos])
            
                    if event.key == K_z or event.key == K_RETURN:
                        # Checks if it is note empty and not cancellable
                        if self.engine.all_events_sign_up[menu_pos] and self.engine.all_events_sign_up[menu_pos].cancellable == True:
                            # Cancel the mission
                            self.cancel_event(self.engine.all_events_sign_up[menu_pos])
                            if menu_pos > len(self.engine.all_events_sign_up)-1:
                                menu_pos = len(self.engine.all_events_sign_up)-1
                            
                            if self.engine.all_events_sign_up:
                                text_selected_event = self.get_event_text_data(self.engine.all_events_sign_up[0])
                            text_event_list = self.get_event_list(self.engine.all_events_sign_up)
                            text_mission_stats = self.get_mission_stats()
                            
                    if event.key == K_x:
                        menu_flag = 1
        
            if menu_flag == 0:
                self.engine.surface.blit(self.engine.mission_bg,(0,0))
                self.engine.surface.blit(text_title,(35,105))
                self.engine.surface.blit(text_missions,(35,150))
                
                self.engine.surface.blit(text_mission_stats[0],(200,108))
                self.engine.surface.blit(text_mission_stats[1],(400,108))
                self.engine.surface.blit(text_mission_stats[2],(600,108))
                
                counter = 0
                for event_name in text_event_list:
                    self.engine.surface.blit(event_name,(55,170+counter*15))
                    counter += 1
                    
                if self.engine.all_events_sign_up:
                    self.engine.surface.blit(text_name,(350,150))
                    self.engine.surface.blit(text_selected_event[0],(360,170))
                    self.engine.surface.blit(text_type,(350,190))
                    self.engine.surface.blit(text_selected_event[1],(360,210))
                    self.engine.surface.blit(text_location,(350,230))
                    self.engine.surface.blit(text_selected_event[2],(360,250))
                    self.engine.surface.blit(text_cancellable,(350,270))
                    self.engine.surface.blit(text_selected_event[3],(360,290))
                    
                    if self.engine.all_events_sign_up[menu_pos].show_rewards:
                        self.engine.surface.blit(text_rewards,(600,150))
                        [self.engine.surface.blit(reward,(610,170+counter*20)) 
                         for counter,reward in enumerate(text_selected_event[4])]
                    
                    self.engine.surface.blit(text_desc,(350,310))
                    [self.engine.surface.blit(line,(360,330+20*index))
                          for index,line in enumerate(text_selected_event[5])]
                    
                    pygame.draw.circle(self.engine.surface, (0,0,0), (45,178+15*menu_pos),5)
                
                self.engine.surface.blit(text_directions1,(60,520))
                self.engine.surface.blit(text_directions2,(60,540))
                
                pygame.display.flip()
                self.engine.clock.tick(30)
                
    
    def rumors_menu(self):
    
        """
        # Function Name: rumors menu
        # Purpose: Provide a list of relevant rumors and hints
        # Inputs: None
        """
    
        text_title = self.engine.bfont.render("Rumors and Hints",True,(0,0,0))
        
        #Constructs the hint/rumor list
        all_hint_names = []
        all_hint_lines = []
        for hint in self.engine.hint_list:
            prereqs_met = True
            for prereq in hint['prereqs']:
                # If any one of the required events has not been completed, do not include the hint
                if self.engine.all_events_master[prereq].done == False:
                    prereqs_met = False
            if prereqs_met:
                all_hint_names.append({'small_title':self.engine.sfont.render(hint['name'],True,(0,0,0)),
                                       'big_title':self.engine.bfont.render(hint['name'],True,(0,0,0)),
                                       'author':self.engine.sfont.render(hint['author'],True,(0,0,0))})
                text_hint_lines = []
                for line in split_lines(hint['text'],70):
                    text_hint_lines.append(self.engine.sfont.render(line,True,(0,0,0)))
                all_hint_lines.append(text_hint_lines)
                
        menu_flag = True
        menu_pos = 0
        text_selected_hint = all_hint_lines[menu_pos]
        while menu_flag:
            
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(all_hint_names)-1
                        text_selected_hint = all_hint_lines[menu_pos]
                        
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < len(all_hint_names)-1:
                            menu_pos += 1
                        elif menu_pos == len(all_hint_names)-1:
                            menu_pos = 0
                        text_selected_hint = all_hint_lines[menu_pos]
                    if event.key == K_x:
                        menu_flag = False
                
            if menu_flag:
                self.engine.surface.blit(self.engine.mission_bg,(0,0))
                self.engine.surface.blit(text_title,(35,105))

                counter = 0
                for hint_name in all_hint_names:
                    self.engine.surface.blit(hint_name['small_title'],(55,170+counter*15))
                    counter += 1
                    
                self.engine.surface.blit(all_hint_names[menu_pos]['big_title'],(340,180))
                self.engine.surface.blit(all_hint_names[menu_pos]['author'],(360,200))
                for line_num in xrange(0,len(text_selected_hint)):
                    self.engine.surface.blit(text_selected_hint[line_num],(350,220+line_num*15))
                    
                pygame.draw.circle(self.engine.surface, (0,0,0), (45,178+15*menu_pos),5)
                pygame.display.flip()
                self.engine.clock.tick(30)
                
    def archives_menu(self):
        
        """
        # Function Name: archives menu
        # Purpose: Archive of completed Missions
        """
        text_title = self.engine.bfont.render("Archive",True,(0,0,0))
        text_missions = self.engine.bfont.render("Missions",True,(0,0,0))
        
        text_name = self.engine.bfont.render("Mission Name",True,(0,0,0))
        text_type = self.engine.bfont.render("Type",True,(0,0,0))
        text_location = self.engine.bfont.render("Location",True,(0,0,0))
        text_rewards = self.engine.bfont.render("Rewards",True,(0,0,0))
        text_desc = self.engine.bfont.render("Description",True,(0,0,0))
        
        text_directions1 = self.engine.sfont.render("X - Back",True,(0,0,0))
        
        text_event_list = self.get_event_list(self.engine.all_events_completed)
        text_selected_event = self.get_event_text_data(self.engine.all_events_completed[0])
        text_mission_stats = self.get_mission_stats()
        
        menu_flag = 0
        menu_pos = 0

        while menu_flag == 0:
            
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ) and self.engine.all_events_completed:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(self.engine.all_events_completed)-1
                        text_selected_event = self.get_event_text_data(self.engine.all_events_completed[menu_pos])
        
                    if ( event.key == K_DOWN or event.key == K_RIGHT ) and self.engine.all_events_completed:
                        if menu_pos < len(self.engine.all_events_completed)-1:
                            menu_pos += 1
                        elif menu_pos == len(self.engine.all_events_completed)-1:
                            menu_pos = 0
                        text_selected_event = self.get_event_text_data(self.engine.all_events_completed[menu_pos])
            
                    if event.key == K_x:
                        menu_flag = 1
        
            if menu_flag == 0:
                self.engine.surface.blit(self.engine.mission_bg,(0,0))
                self.engine.surface.blit(text_title,(35,105))
                self.engine.surface.blit(text_missions,(35,150))
                
                self.engine.surface.blit(text_mission_stats[0],(200,108))
                self.engine.surface.blit(text_mission_stats[1],(400,108))
                self.engine.surface.blit(text_mission_stats[2],(600,108))
                
                counter = 0
                for event_name in text_event_list:
                    self.engine.surface.blit(event_name,(55,170+counter*15))
                    counter += 1
                
                    
                if self.engine.all_events_completed[menu_pos]:
                    
                    self.engine.surface.blit(text_name,(350,150))
                    self.engine.surface.blit(text_selected_event[0],(360,170))
                    self.engine.surface.blit(text_type,(350,190))
                    self.engine.surface.blit(text_selected_event[1],(360,210))
                    self.engine.surface.blit(text_location,(350,230))
                    self.engine.surface.blit(text_selected_event[2],(360,250))
                    self.engine.surface.blit(text_rewards,(600,150))
                    
                    [self.engine.surface.blit(reward,(610,170+counter*20)) for counter,
                     reward in enumerate(text_selected_event[4])]
                        
                        
                    self.engine.surface.blit(text_desc,(350,310))
                    [self.engine.surface.blit(line,(360,330+20*index))
                          for index,line in enumerate(text_selected_event[5])]
                    
                self.engine.surface.blit(text_directions1,(60,540))
                
                pygame.draw.circle(self.engine.surface, (0,0,0), (45,178+15*menu_pos),5)
                pygame.display.flip()
                self.engine.clock.tick(30)
                
    