
# Trading system
import os
import pygame
from pygame.locals import *
from lostsky.core.utils import split_lines
from sys import exit

class TradingSystem(object):
    """
    # Top Level Trading System Manager
    # Handles:
    #       - Manages and keeps up to date all trading information
    #       - Allows viewing of trade items
    #       - Displays Misc Messages from the shop owner
    """

    def __init__(self,engine):
        """
        # Function name: __init__
        # Purpose: Creates trading system object
        # Inputs: engine - top level engine
        """
    
        # link with engine
        self.engine = engine
        # Load shopkeeper image
        self.rinnosuke = pygame.image.load(os.path.join("images","rinnosuke_shop.png")).convert_alpha()
        # A list of available trades
        self.available_trades = []
        self.available_specials = []
    
    def update_available_trades(self):
        """
        # Function name: update available trades
        # Purpose: Updates the current list of available trades
        """
        self.available_trades = []

        for trade in self.engine.trading_catalog.trading_list.values():
            prereqs_done = 0
            
            for prereq in trade.prereqs:
                if self.engine.all_events_master[prereq].done:
                    prereqs_done += 1
                else:
                    break

            if prereqs_done == len(trade.prereqs) and trade.id_string not in self.engine.player.trading_data['non_repeatable']:
                self.available_trades.append(trade.id_string)
                self.available_trades.sort()
     
    def verify_trade(self,trade):
        """
        # Function name: verify_trade
        # Purpose: Verifies that a trade can be performed
        # Inputs: treasure_inventory: Player's treasure inventory
        # Outputs: can_trade = True if player can perform this trade, false otherwise
        #          wanted_items = text objects representing items the player has / doesn't have
        """
    
        
        wanted_items = []
        can_trade = True
        for item_wanted in trade.wanted:
        
            # Case: Player either does not have the treasure or player does not have the item in sufficient quantities
            if item_wanted['id_string'] not in self.engine.player.treasures.keys() or (item_wanted['id_string'] in self.engine.player.treasures.keys() and item_wanted['quantity'] > self.engine.player.treasures[item_wanted['id_string']]):
                
                # Appends a text representation in red color
                wanted_items.append(self.engine.sfont.render(self.engine.treasure_catalog[item_wanted['id_string']].name+" x"+str(item_wanted['quantity']),True,(255,0,0)))
                
                can_trade = False
                
            # Case: Player has item in sufficient quantities
            elif item_wanted['quantity'] <= self.engine.player.treasures[item_wanted['id_string']]:
    
                # Appends a text representation in black color
                wanted_items.append(self.engine.sfont.render(self.engine.treasure_catalog[item_wanted['id_string']].name+" x"+str(item_wanted['quantity']),True,(0,0,0)))
               
        return can_trade, wanted_items
                
                
    def execute_trade(self,trade):
    
        """
        # Function name: execute trade
        # Purpose: Performs the trade
        # Inputs: trade - trade to be performed
        """
        
        # Removes wanted items
        [self.engine.player.remove_treasure(item_wanted['id_string'],item_wanted['quantity']) for item_wanted in trade.wanted]
        [self.engine.player.trading_data['found_treasures'].append(item_wanted['id_string']) for item_wanted in trade.wanted 
        if self.engine.treasure_catalog[item_wanted['id_string']].type == "Generic" and 
        item_wanted['id_string'] not in self.engine.player.trading_data['found_treasures']]
    
        # Adds offered items
        for item_offered in trade.offered:
            if item_offered['item_type'] == 'treasure':
                self.engine.player.add_treasure(item_offered['item_id'],item_offered['quantity'])
            else:
                [self.engine.player.add_item(self.engine.spell_catalog[item_offered['item_id']].construct_spell()) for counter in xrange(0,item_offered['quantity'])]


        if self.engine.player.trading_data['next_milestone'] != None:
            current_milestone = self.engine.trading_catalog.milestones[self.engine.player.trading_data['next_milestone']]

        # Checks if the next treasure collection milestone has been reached

        print self.engine.player.trading_data['next_milestone']
        print self.engine.trading_catalog.milestones

        if self.engine.player.trading_data['next_milestone'] != None and len(self.engine.player.trading_data['found_treasures']) == current_milestone:
            current_reward = self.engine.trading_catalog.rewards[self.engine.player.trading_data['next_milestone']]
            
            self.say("Congratulations, you've reached another milestone!")
            self.say(current_reward['desc'])
            
            for reward in current_reward['items_given']:
                # Gives a treasure
                if reward['item_type'] == 'treasure':
                    self.engine.player.add_treasure(reward['item_id'],reward['quantity'])
                    self.say("I'll give you this item as a reward: %s x%s"%(self.engine.treasure_catalog[reward['item_id']].name,
                                                 str(reward['quantity'])))
                
                # Gives a spell item
                else:
                    for counter in xrange(0,reward['quantity']):
                        spell_reward = self.engine.spell_catalog[reward['item_id']].construct_spell()
                        self.engine.player.add_item(spell_reward)
                    self.say("I'll give you this item as a reward: %s x%s"%(spell_reward.name,
                             str(reward['quantity'])))
            # Goes to next treasure
            self.engine.player.trading_data['next_milestone'] += 1
            # All rewards have been given
            if self.engine.player.trading_data['next_milestone'] >= len(self.engine.trading_catalog.rewards):
                self.engine.player.trading_data['next_milestone'] = None

        if not trade.repeatable:
            self.engine.player.trading_data['non_repeatable'].append(trade.id_string)
    
    def render_shop_menu(self):
        """
        # Function name: render_shop_menu
        # Purpose: Draws the main menu for the shop
        """
        
        text_menu_header = self.engine.bfont.render("Kourindou",True,(0,0,0))
        text_viewtrade = self.engine.bfont.render("View Trades",True,(0,0,0))
        text_treasurecollection = self.engine.bfont.render("Treasure Collection",True,(0,0,0))
        text_cancel = self.engine.bfont.render("Cancel",True,(0,0,0))

        # Background image
        self.engine.surface.blit(self.engine.shop_bg,(0,0))
        self.engine.surface.blit(self.rinnosuke,(70,280))
        self.engine.surface.blit(self.engine.text_board,(0,490))
        # Header
        self.engine.surface.blit(text_menu_header,(50,50))
        
        # Renders the options
        self.engine.surface.blit(text_viewtrade,(70,70))
        self.engine.surface.blit(text_treasurecollection,(70,90))
        self.engine.surface.blit(text_cancel,(70,110))
                

    def say(self,lines):
    
        """
        # Function name: say
        # Purpose: Displays a line of text and awaits for the player to press Z to continue
        # Inputs: lines = The string of text to be displayed
        """
        
        menu_flag = True
        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_z or event.key == K_RETURN:
                        menu_flag = False
            if menu_flag:
                self.render_shop_menu()
                
                # Draws the message
                self.set_rinnosuke_speech(lines)
                pygame.display.flip()
                self.engine.clock.tick(60)
                
    def set_rinnosuke_speech(self,string):
        """
        # Function name: set_rinnosuke_speech
        # Purpose: Sets rinnosuke to display lines of speech
        # Inputs: string = The string of text to be displayed
        """
    
        # calls the line splitter function to generate the events
        lines = split_lines(string)
        text_lines = []
        [text_lines.append(self.engine.mfont.render(line,True,(0,0,0))) for line in lines]
        text_speaker = self.engine.bfont.render("- Rinnosuke -",True,(0,0,0))
            
        self.engine.surface.blit(text_speaker,(30,505))
        [self.engine.surface.blit(text_line,(30,530+20*counter)) for counter,text_line in enumerate(text_lines)]
        

    def draw_treasure_data(self,treasure_key):

        """
        # Function name: draw_treasure_data
        # Purpose: draws information about a single treasure
        """
    
        text_treasure = self.engine.bfont.render("Treasure",True,(0,0,0))
        text_desc = self.engine.bfont.render("Description",True,(0,0,0))
        text_single_treasure_data = self.get_single_treasure_data(treasure_key)
        # Renders currently selected treasure data
        self.engine.surface.blit(text_single_treasure_data[0],(470,50))
        self.engine.surface.blit(text_treasure,(470,70))
        self.engine.surface.blit(text_desc,(480,110))
        [self.engine.surface.blit(line,(490,130+index*20)) for index,line in enumerate(text_single_treasure_data[1])]

    def get_single_treasure_data(self,treasure_key):
    
        """
        # Function Name: get_single_treassure_data
        # Purpose: Generates the text objects of the currently selected treasure in the inventory
        # Inputs: treasure_key - relevant treasure key
        """
    
        text_single_treasure_data = []
        # Format (Text objects)
            # [big_name, description]
        text_single_treasure_data.append(self.engine.bfont.render(self.engine.treasure_catalog[treasure_key].name,True,(0,0,0)))
        text_single_treasure_data.append([self.engine.sfont.render(line,True,(0,0,0)) for line in split_lines(self.engine.treasure_catalog[treasure_key].desc,50)])
        return text_single_treasure_data
 
    def get_text_trade_data(self,catalog,availability_list):
        
        """
        # Function name: get_text_trade_data
        # Purpose: Generates the text objects for the valid trades
        # Inputs: catalog: The catalog dictionary to develop data from
        #           availability_list - the list of id_strings keys associated with the catalog    
        # Outputs: all_trade_data - a list of all the text objects for the all the trades
        #            
        """
    
        all_trade_data = []
        for trade_key in availability_list:
            trade = catalog[trade_key]

            can_trade,text_wanted = self.verify_trade(trade)
            text_trade_data = {}
            text_trade_data['name_big'] = self.engine.bfont.render(trade.name,True,(0,0,0))
            text_trade_data['name_small'] = self.engine.sfont.render(trade.name,True,(0,0,0))
            text_trade_data['desc'] = []
            [text_trade_data['desc'].append(self.engine.sfont.render(line,True,(0,0,0))) for line in split_lines(trade.desc,50)]

            text_trade_data['wanted'] = text_wanted
            text_trade_data['offered'] = []
            
            # Gets the offered item data: text objects [Item name,quantity]
            for item_offered in trade.offered:
                if item_offered['item_type'] == 'treasure':
                    text_trade_data['offered'].append(self.engine.sfont.render(self.engine.treasure_catalog[item_offered['item_id']].name+" x"+str(item_offered['quantity']),True,(0,0,0)))
                elif item_offered['item_type'] == 'spell_action':
                    text_trade_data['offered'].append(self.engine.sfont.render(item_offered['item_id']+" x"+str(item_offered['quantity']),True,(0,0,0)))
                    
            all_trade_data.append(text_trade_data)
            
        return all_trade_data
    
    def trading_menu(self):
    
        """
        # Function name: trading_menu
        # Purpose: Top level trading menu
        """
    

        menu_flag = True
        menu_pos = 0
        
        self.intro()
        
        while menu_flag:
        
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 2
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 2:
                            menu_pos += 1
                        elif menu_pos == 2:
                            menu_pos = 0
                    if event.key == K_z or event.key == K_RETURN:

                        # Ordinary trades
                        if menu_pos == 0:
                            self.select_trade_menu()
                           
                        # Treasure Collection
                        elif menu_pos == 1:
                            self.treasure_collection()
                            
                        # Exit
                        elif menu_pos == 2:
                            menu_flag = False
                            
                    if event.key == K_x:
                        menu_flag = False
                    
            if menu_flag:
                self.render_shop_menu()
                
                # Draws the cursor
                pygame.draw.circle(self.engine.surface, (0,0,0), (60,78+20*menu_pos),5)
                
                pygame.display.flip()
                self.engine.clock.tick(60)
                
    def intro(self):
    
        """
        # Function name: intro
        # Purpose: rinnosuke's introductory speech
        """
        
        if self.engine.player.trading_data['first_time']:
            self.say("Welcome to Kourindou!")
            self.say("Here you can trade treasures that you've found for useful items.")
            self.say("As the number of treasures returned increases, I have some special rewards for you.")
            self.engine.player.trading_data['first_time'] = False
        else:
            self.say("Welcome to Kourindou!")
            self.say("I have some great deals on trades.")
   
    def select_trade_menu(self):
   
        """
        # Function name: select_trade_menu
        # Purpose: allows the player to view all the available trades
        """
        text_menu_header = self.engine.bfont.render("Available Trades",True,(0,0,0))

        text_desc = self.engine.bfont.render("Description:",True,(0,0,0))
        text_wanted = self.engine.bfont.render("Wanted:",True,(0,0,0))
        text_offered = self.engine.bfont.render("Offered:",True,(0,0,0))
        text_empty = self.engine.bfont.render("Empty",True,(0,0,0))
        menu_flag = True
        menu_pos = 0
        
        self.update_available_trades()
        all_trade_data = self.get_text_trade_data(self.engine.trading_catalog.trading_list,self.available_trades)
        
        while menu_flag:
        
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(all_trade_data)-1
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < len(all_trade_data)-1:
                            menu_pos += 1
                        elif menu_pos == len(all_trade_data)-1:
                            menu_pos = 0
                    if (event.key == K_z or event.key == K_RETURN) and all_trade_data:
                        trade_performed = self.examine_trade_menu(self.engine.trading_catalog.trading_list[self.available_trades[menu_pos]],all_trade_data[menu_pos])
                        # Refresh trade data
                        if trade_performed:
                            self.update_available_trades()
                            all_trade_data = self.get_text_trade_data(self.engine.trading_catalog.trading_list,self.available_trades)
                            if menu_pos > len(all_trade_data)-1:
                                menu_pos = len(all_trade_data)-1
                            
                        
                    if event.key == K_x:
                        menu_flag = False
                    
            if menu_flag:
                # Background image
                self.engine.surface.blit(self.engine.stats_bg,(0,0))
                # Header
                self.engine.surface.blit(text_menu_header,(50,50))
                
                # Draws all the available trades
                counter = 0
                for text_trade_data in all_trade_data:
                    self.engine.surface.blit(text_trade_data['name_small'],(70,70+counter*20))
                    counter += 1
                # Draws the word empty if there are no trades available
                if not all_trade_data:
                    self.engine.surface.blit(text_empty,(70,70))
                
                # Draw data about the current trade (only if there is one available)
                if all_trade_data:
                    # Spell Name
                    self.engine.surface.blit(all_trade_data[menu_pos]['name_big'],(480,80))
                    self.engine.surface.blit(text_desc,(480,100))

                    # Draws the description
                    [self.engine.surface.blit(line,(490,120+20*line_counter)) for line_counter, line in enumerate(all_trade_data[menu_pos]['desc'])]
                    
                    # Draws wanted items
                    self.engine.surface.blit(text_wanted,(480,300))
                    [self.engine.surface.blit(wanted_item,(490,320+20*line_counter)) for line_counter, wanted_item in enumerate(all_trade_data[menu_pos]['wanted'])]
                    
                    # Draws offered items
                    self.engine.surface.blit(text_offered,(480,400))
                    [self.engine.surface.blit(offered_item,(490,420+20*line_counter)) for line_counter, offered_item in enumerate(all_trade_data[menu_pos]['offered'])]
                        
                    # Draws the cursor
                    pygame.draw.circle(self.engine.surface, (0,0,0), (60,78+20*menu_pos),5)
                
                pygame.display.flip()
                self.engine.clock.tick(60)
    
    def examine_trade_menu(self,trade,text_trade_data):

        """
        # Function name: examine_trade_menu
        # Purpose: allows the player to view all the stats of a trade
        # Inputs:   trade - trade object to act on
        #           text_trade_data - trade's pre-computed text trade data
        # Outputs: True if a trade was performed, false if not
        """
    
        # Set up text objects
        text_desc = self.engine.bfont.render("Description:",True,(0,0,0))
        text_wanted = self.engine.bfont.render("Wanted:",True,(0,0,0))
        text_offered = self.engine.bfont.render("Offered:",True,(0,0,0))
        can_trade, text_wanted_items = self.verify_trade(trade) 

        if can_trade:
            text_confirm = self.engine.bfont.render("Confirm Trade",True,(0,0,0))
        else:
            text_confirm = self.engine.bfont.render("Confirm Trade",True,(100,100,100))
            
        text_cancel = self.engine.bfont.render("Cancel",True,(0,0,0))
        menu_flag = True
        menu_pos = 0
        num_wanted = len(trade.wanted)
        num_offered = len(trade.offered)
        # Max position = # wanted + # offered + 2 (For Execute Trade / Cancel) - 1 (for 0 indexing)
        menu_pos_max = num_wanted + num_offered + 2 - 1
        
        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = menu_pos_max
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < menu_pos_max:
                            menu_pos += 1
                        elif menu_pos == menu_pos_max:
                            menu_pos = 0
                    if (event.key == K_z or event.key == K_RETURN) and menu_pos == menu_pos_max - 1 and can_trade:
                        self.execute_trade(trade)
                        return True
                            
                    if event.key == K_x or ((event.key == K_z or event.key == K_RETURN) and menu_pos == menu_pos_max):
                        menu_flag = False
                        return False
                   
                   
            if menu_flag:
                # Background image
                self.engine.surface.blit(self.engine.stats_bg,(0,0))
                
                # Spell Name
                self.engine.surface.blit(text_trade_data['name_big'],(50,80))
                self.engine.surface.blit(text_desc,(50,100))

                # Draws the description
                [self.engine.surface.blit(line,(60,120+20*line_counter)) for line_counter, line in enumerate(text_trade_data['desc'])]
                
                # Draws wanted items
                self.engine.surface.blit(text_wanted,(50,300))
                [self.engine.surface.blit(wanted_item,(60,320+20*line_counter)) for line_counter, wanted_item in enumerate(text_wanted_items)]
                
                # Draws offered items
                self.engine.surface.blit(text_offered,(50,400))
                [self.engine.surface.blit(offered_item,(60,420+20*line_counter)) for line_counter, offered_item in enumerate(text_trade_data['offered'])]
                    
                self.engine.surface.blit(text_confirm,(60,540))
                self.engine.surface.blit(text_cancel,(60,560))
                    
                # Cursor is in the wanted items section
                if menu_pos < num_wanted:
                    # Draw information about the wanted treasure
                    self.draw_treasure_data(trade.wanted[menu_pos]['id_string'])
                    
                    # Draws the cursor
                    pygame.draw.circle(self.engine.surface, (0,0,0), (50,328+20*menu_pos),5)

                # Cursor is in the offered items section
                elif menu_pos >= num_wanted and menu_pos < num_offered + num_wanted:
                    # Draw information about the offered item
                    if trade.offered[menu_pos-num_wanted]['item_type'] == "treasure":
                        self.draw_treasure_data(trade.offered[menu_pos-num_wanted]['item_id'])
                    elif trade.offered[menu_pos-num_wanted]['item_type'] == "spell_action":
                        self.engine.draw_uncreated_spell(trade.offered[menu_pos-num_wanted]['item_id'])
                    
                    # Draw the cursor
                    pygame.draw.circle(self.engine.surface, (0,0,0), (50,428+20*(menu_pos-num_wanted)),5)
                
                # Cursor is in the last two options
                elif menu_pos >= num_offered+num_wanted:
                    # Draw the cursor
                    pygame.draw.circle(self.engine.surface, (0,0,0), (50,548+20*(menu_pos-num_wanted-num_offered)),5)
                    
                    
                pygame.display.flip()
                self.engine.clock.tick(60)


    def treasure_collection(self):
        """
        function name: treasure_collection
        purpose: Display and examine a collection of returned treasures
        """
        
        text_menu_header = self.engine.bfont.render("Treasure Collection",True,(0,0,0))
        
        menu_flag = True
        menu_pos = [0,0]
        treasure_list = [treasure.id_string for treasure in self.engine.treasure_catalog.values() 
                         if treasure.type == "Generic"]
        text_num_treasures = self.engine.bfont.render("Treasures Found: "+str(len(self.engine.player.trading_data['found_treasures'])),True,(0,0,0))
        if self.engine.player.trading_data['next_milestone'] != None:
            text_next_milestone = self.engine.bfont.render("Next Reward At: "+str(self.engine.trading_catalog.milestones[self.engine.player.trading_data['next_milestone']]),True,(0,0,0))
        else:
            text_next_milestone = self.engine.bfont.render("Next Reward At: N/A",True,(0,0,0))
        
        
        treasure_list.sort()
        
        while menu_flag:
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if event.key == K_UP:
                        if menu_pos[1] > 0:
                            menu_pos[1] -= 1
                        else:
                            menu_pos[1] = 4
                            
                    if event.key == K_LEFT:
                        if menu_pos[0] > 0:
                            menu_pos[0] -= 1
                        else:
                            menu_pos[0] = 4
                            
                    if event.key == K_DOWN:
                        if menu_pos[1] < 4:
                            menu_pos[1] += 1
                        else:
                            menu_pos[1] = 0
                        
                    if event.key == K_RIGHT:
                        if menu_pos[0] < 4:
                            menu_pos[0] += 1
                        else:
                            menu_pos[0] = 0

                    if event.key == K_x:
                        menu_flag = False
            
            if menu_flag:
            
                # Background image
                self.engine.surface.blit(self.engine.stats_bg,(0,0))

                # Header
                self.engine.surface.blit(text_menu_header,(50,50))

                # Treasures displayed as a 5x5 grid
                for y in xrange(0,5):
                    for x in xrange(0,5):
                        self.engine.surface.blit(self.engine.unit_tile,(50+70*x,82+50*y))
                        
                        treasure_id = treasure_list[5*y+x]
                        
                        if treasure_id in self.engine.player.trading_data['found_treasures']:
                            self.engine.surface.blit(self.engine.icons[self.engine.treasure_catalog[treasure_id].icon],
                                                                   (55+70*x,90+50*y))
                
                # Draws all the treasures
                if treasure_list[5*menu_pos[1]+menu_pos[0]] in self.engine.player.trading_data['found_treasures']:
                    self.draw_treasure_data(treasure_list[5*menu_pos[1]+menu_pos[0]])

                self.engine.surface.blit(text_num_treasures,(50,400))
                self.engine.surface.blit(text_next_milestone,(50,420))                        
                        
                # Draws cursor
                self.engine.surface.blit(self.engine.cursor_img,(50+70*menu_pos[0],82+50*menu_pos[1]))
                           
                pygame.display.flip()
                self.engine.clock.tick(60)
        
        

class Trade(object):

    def __init__(self,name,id_string,desc,repeatable,prereqs,wanted,offered,):


        """
        # Function name: __init__
        # Purpose: Creates a trade object
        # Inputs:
        #           name - the displayed name for the trade
        #           id_string - the key in the trade catalog
        #           desc - a description of the trade
        #           repeatable - T/F. if trade can only be done once
        #           prereqs - after what mission does this trade appear
        #           wanted - a list of items to be traded
        #           offered - a list of items that the player receives
        """
    
        self.name = name
        self.id_string = id_string
        self.desc = desc
        self.repeatable = repeatable
        self.prereqs = prereqs
        self.wanted = wanted
        self.offered = offered
        
        