# World Map Class
# Map engine: "Gensokyo Navi"
import pygame
import os
from pygame.locals import *
from sys import exit
from lostsky.core.linalg import Vector2
from lostsky.core.utils import split_lines
from lostsky.core.persistentobjects import PlayerData
from lostsky.battle.mapobj import Map
from lostsky.battle.objectives import * 
from lostsky.battle.ai_fsm import UnitAI


class Worldmap(object):

    def __init__(self,engine,region_list,region_path_list):
        """
        # Function Name: __init__
        # Purpose: Intializes world map
        # Inputs: Engine - Game system engine
        #         region_list - a list of regions in the world map
        """
    
        
        self.engine = engine
        engine.worldmap = self
        self.all_regions = {}
        self.all_regions_by_name = {}
        self.all_locations_by_name = {}
        self.conv_images = {}
        self.region_path_list = region_path_list
        default_starting_pos = Vector2(5,6)
        self.player_pos = default_starting_pos
        self.player_sprite_group = pygame.sprite.RenderUpdates()
        self.player_sprite = WorldmapSprite( 35*self.player_pos, self.engine.player_img)
        self.player_sprite_group.add(self.player_sprite)
        self.background_surface = pygame.surface.Surface((840,630))

        self.dot_positions = set()
        for region in region_list:
            self.all_regions[tuple(region.wm_coords)]=region
            self.all_regions_by_name[region.name]=region
            region.wm_parent = self
            for location in region.all_locations.values():
                self.all_locations_by_name[location.name] = location
                
        # Sets up connecting paths
        
        # List of all dots to not include a range around the region
        remove_pts = []
        for region in region_list:
            remove_pts.extend((tuple(region.wm_coords),tuple((region.wm_coords+Vector2(-1,1))),tuple((region.wm_coords+Vector2(0,1))),tuple((region.wm_coords+Vector2(0,1)))))

        for path in region_path_list:
        
            # Each path has two connecting points along with the direction key to press to get to each location
            direction_to_1 = path.locations.keys()[0]
            region_1 = path.locations[direction_to_1]
            direction_to_2 = path.locations.keys()[1]
            region_2 = path.locations[direction_to_2]

            # We assign the starting point's path directions to this path so it can be called when the character needs to move
            self.all_regions_by_name[region_2].paths[direction_to_1] = path
            self.all_regions_by_name[region_1].paths[direction_to_2] = path
            
            # Constructs destination vectors
            delta_pos = self.all_regions_by_name[region_1].wm_coords - self.all_regions_by_name[region_2].wm_coords
            path.vectors[direction_to_1] = delta_pos
            path.vectors[direction_to_2] = -1*(delta_pos.copy())
            
            # # Decomposes delta vector into path dots for the map
            for i in xrange(min(1,int(delta_pos.y)),max(int(delta_pos.y),0)):
                dot_vector = self.all_regions_by_name[region_2].wm_coords + Vector2(0,i)
                if tuple(dot_vector) not in remove_pts:
                    path.dot_positions.append(tuple(35*dot_vector))
                
            for j in xrange(min(1,int(delta_pos.x)),max(int(delta_pos.x),0)):
                dot_vector = self.all_regions_by_name[region_2].wm_coords + Vector2(j,delta_pos.y)
                if tuple(dot_vector) not in remove_pts:
                    path.dot_positions.append(tuple(35*dot_vector))
                
                
        
            
        # Region connection paths
        self.unit_associate()
        
    ############################################
    # Computational Methods
    ############################################
    
    def unit_associate(self):
    
        """
        # Function Name: unit_associate
        # Purpose: Associates the unit's map attribute with the world map
        """
    
        for unit in self.engine.player_units:
            unit.map = self
        
    def update_all_events(self):
        """
        # Function Name: unit_associate
        # Purpose: Associates the unit's map attribute with the world map
        # Outputs: new_mission = True/False if a new mission is available
        """
    
        # Clears out all location event data
        for location in self.all_locations_by_name.values():
            location.all_events = [None, None, None, None]
        
        # Clears out the available list
        self.engine.all_events_available = []
        self.engine.all_events_completed = []
        self.engine.all_events_sign_up = []
        
        # Updates all events
        for event in self.engine.all_events_master.values():
            
            # If event is done, add it to the list of completed events
            if event.done:
                
                event.available = False
                event.sign_up = False
                self.engine.all_events_completed.append(event)
            else:
        
                # Updates availability of all events
                
                # Mission is added if no prerequisites are present or
                # all prerequisites are met
                if ((not event.prereq or self.engine.check_event_completion(event.prereq)) 
                    and not event.sign_up and not event.manual_cancel):
                    
                    if not event.auto_signup:
                        event.available = True
                    else:
                        event.sign_up = True
                        
                # If the event is available, add it to the list of available events
                if event.available:
                    self.engine.all_events_available.append(event)
                
                # if the event is signed up for add it to the world map
                if event.sign_up:
                    self.engine.all_events_sign_up.append(event)
                    self.all_locations_by_name[event.location_name].add_event(event)
                    
            # Updates player data
            self.engine.player.all_event_data[event.event_id].update_from_event(event)
            
            self.update_all_paths()

    def update_all_paths(self):
    
        """
        # Function Name: update all paths
        # Purpose: Updates the unlocking of all the world map paths
        """
    
        for path in self.region_path_list:
            # A path is unlocked if there are no prerequisite missions or if all of these missions has been completed
            # The Engine's unlock_wm debug flag unlocks the whole map
            if self.engine.unlock_wm or not path.prereq or self.engine.check_event_completion(path.prereq):
                path.unlock = True
            else:
                path.unlock = False
                
        for region in self.all_regions.values():
            for path in region.location_path_list:
                # A path is unlocked if there are no prerequisite missions or if all of these missions has been completed
                # The Engine's unlock_wm debug flag unlocks the whole map
                if self.engine.unlock_wm or not path.prereq or self.engine.check_event_completion(path.prereq):
                    path.unlock = True
                else:
                    path.unlock = False
            
                
    
    def update_wm_player_data(self):
    
        """
        # Function Name: update wm player data
        # Purpose: Updates the location of the player
        """
        self.engine.player.wm_data['wm_coords'] = tuple(self.player_pos)
        self.engine.player.wm_data['in_region'] = False
        
    ############################################
    # Graphical Methods
    ############################################
    
    def render_background(self):
        """
        # Function Name: Render map to background surface
        # Purpose: Draws the background map and the locations
        """

        # Creates relevant text objects
        region_text = self.engine.sfont.render("Region",True,(0,0,0))
        region_text_halfwidth = region_text.get_width()/2
        
        # Directions
        text_directions1 = self.engine.sfont.render("A - Menu",True,(0,0,0))
        text_directions2 = self.engine.sfont.render("Z - Select",True,(0,0,0))
        
        self.background_surface.blit(self.engine.wm_bg,(0,0))
        self.background_surface.blit(self.engine.text_board,(0,490))
        self.background_surface.blit(self.engine.tiny_text_board,(735,0))
        self.background_surface.blit(text_directions1,(840-(text_directions1.get_width()+20),15))
        self.background_surface.blit(text_directions2,(840-(text_directions2.get_width()+20),30))
        
        
    
        for region in self.all_regions.values():
            draw_bubble = False
            # Checks if there is an event in the region
            for location in region.all_locations.values():
                if location.has_event():
                    draw_bubble = True
                    break
                
            if draw_bubble:
                self.background_surface.blit(self.engine.event_icon,(region.pix_coords+Vector2(10,-5)))
            
            self.background_surface.blit(region.name_small,region.name_coords)
            self.background_surface.blit(region_text,region.pix_coords+Vector2(17-region_text_halfwidth,55))

        [self.background_surface.blit(self.engine.path_dot,dot_pos) for path in self.region_path_list if path.unlock
        for dot_pos in path.dot_positions]

    def render_description(self):
        """
        Purpose: Draws the description of the current location to background surface
        """

        selected = self.all_regions[tuple(self.player_pos)]
        self.background_surface.blit(self.engine.portrait_catalog[selected.portrait],(10,500))
        self.background_surface.blit(selected.name_big,(140,505))
        desc_lines = split_lines(selected.desc)
        [self.background_surface.blit(self.engine.sfont.render(line,True,(0,0,0)),(140,525+index*20)) for index,line in enumerate(desc_lines)]

    def render_updated_map(self):
        """
        Purpose: Updates the background surface and draws everything
        """
        self.render_background()
        self.render_description()
        self.engine.surface.blit(self.background_surface, (0,0))
        self.player_sprite_group.update(self.player_pos*35)
        self.player_sprite_group.clear( self.engine.surface, self.background_surface)
        self.player_sprite_group.draw(self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)

    
    def move_to(self,direction_vector):
        """
        # Function Name: move to
        # Purpose: moves the player from point A to point B
        # Inputs: Destination - Where the player is going
        """
        #self.cursor_pos += direction_vector
        path = []
        if abs(direction_vector.x) > 0:
            path.append(Vector2(direction_vector.x,0))
        if abs(direction_vector.y) > 0:
            path.append(Vector2(0,direction_vector.y))
        
        # Smoothstep function
        # See: http://sol.gfxile.net/interpolation/
        smoothstep = lambda v: (v*v*(3-2*v))
        
        start_pos = 35*self.player_pos
        current_pos = 35*self.player_pos

        self.render_updated_map()

        for next_step in path:
        
            # Sets frame count proportional to the square root of the magnitude of the step vector
            frame_count = int(10*next_step.get_magnitude()**(1.0/2))           
            
            for t in xrange(0,int(frame_count)+1):
                v = float(t)/float(frame_count)
                v = smoothstep(v)
                scale_term = int(35*v)
                intermediate_step = scale_term*next_step
                
                current_pos = start_pos + intermediate_step

                self.player_sprite_group.update(current_pos)
                self.player_sprite_group.clear( self.engine.surface, self.background_surface)
                rects = self.player_sprite_group.draw( self.engine.surface)

                pygame.display.update(rects)
                self.engine.clock.tick(60)
                
            start_pos = current_pos.copy()
            
        self.player_pos += direction_vector

    ############################################
    # Interactive methods
    ############################################
    def say(self,string,speaker=""):
       
        """
        # Function name: say
        # Purpose: Displays a line of text and awaits for the player to press Z to continue
        # Inputs: string = The string of text to be displayed
        #         speaker = The name of the person saying it
        """
    
        # calls the line splitter function to generate the events
        lines = split_lines(string)
        text_lines = []
        if speaker != "":
            text_speaker = self.engine.bfont.render("- "+speaker+" -",True,(0,0,0))
        else:
            text_speaker = self.engine.bfont.render(speaker,True,(0,0,0))
        
        # generates the text for each line
        for line in lines:
            text_lines.append(self.engine.mfont.render(line,True,(0,0,0)))
        menu_flag = True

        self.render_background()

        self.engine.surface.blit(self.background_surface, (0,0))
        self.player_sprite_group.update(self.player_pos*35)
        self.player_sprite_group.clear( self.engine.surface, self.background_surface)
        self.player_sprite_group.draw(self.engine.surface)

        self.engine.surface.blit(text_speaker,(30,505))

        for conversation_image in self.conv_images.values():
            self.engine.surface.blit(conversation_image[0],conversation_image[1])

        for i in xrange(0,len(text_lines)):
            self.engine.surface.blit(text_lines[i],(30,530+20*i))

        pygame.display.flip()

        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN and (event.key == K_z or event.key == K_RETURN):
                    menu_flag = False
                    
            # Hold down C to skip through the dialog.
            keys = pygame.key.get_pressed()
            if keys[K_c]:
                menu_flag = False

            self.engine.clock.tick(60)
    
    def choice(self,query,responses):
        """
        # Function name: Choice
        # Purpose: Displays a question for the player to answer
        # Inputs: string = The string of text to be displayed (e.g. do you like pie?)
        #         responses = A list of (preferably less than 3) options given to the player
        """
    
        text_query = self.engine.bfont.render(query,True,(0,0,0))
        text_responses = []
        for response in responses:
            text_responses.append(self.engine.sfont.render(response,True,(0,0,0)))
            
        menu_flag = True
        menu_pos = 0


        self.render_background()
        self.engine.surface.blit(self.background_surface, (0,0))
        self.player_sprite_group.update(self.player_pos*35)
        self.player_sprite_group.clear( self.engine.surface, self.background_surface)
        self.player_sprite_group.draw(self.engine.surface)
        self.engine.surface.blit(text_query,(30,520))
        for i in xrange(0,len(text_responses)):
            self.engine.surface.blit(text_responses[i],(50,540+20*i))
        pygame.draw.circle(self.engine.surface, (0,0,0), (40,548+20*menu_pos),5)
        pygame.display.flip()

        while menu_flag:
            
            # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_UP or event.key == K_LEFT:
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = len(text_responses)-1
                    if event.key == K_DOWN or event.key == K_RIGHT:
                        if menu_pos < len(text_responses)-1:
                            menu_pos += 1
                        elif menu_pos == len(text_responses)-1:
                            menu_pos = 0
                    
                    if (event.key == K_z or event.key == K_RETURN):
                        menu_flag = False
                        return responses[menu_pos]
                        
            if menu_flag:

                self.engine.surface.blit(self.background_surface, (35, 535), (35, 535, 10, 105 ))
                pygame.draw.circle(self.engine.surface, (0,0,0), (40,548+20*menu_pos),5)

                pygame.display.update(pygame.Rect(35, 535, 20, 105 ))
                self.engine.clock.tick(60)
    
    def user_input(self):
        """
        # Function Name: user input
        # Purpose: top level interaction. Checks if there is any input by the user
        #   and allows the user to move around the cursor with the arrow keys and select a unit
        #   if there is one under the mouse
        """
    
        arrowkeys = False
        directions = {K_LEFT:'left',K_RIGHT:'right',K_UP:'up',K_DOWN:'down'}
        for event in pygame.event.get():
            if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                exit()
            if event.type == KEYDOWN:
                if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                    # Checks if the player is allowed to move:
                    #   a. A path exists
                    #   b. The path is unlocked
                    if directions[event.key] in self.all_regions[tuple(self.player_pos)].paths.keys() and self.all_regions[tuple(self.player_pos)].paths[directions[event.key]].unlock:
                        self.move_to(self.all_regions[tuple(self.player_pos)].paths[directions[event.key]].vectors[directions[event.key]])
                        self.update_wm_player_data()
                        self.last_move = directions[event.key]

                        # Refreshes world map display

                        self.render_updated_map()


                        return
                if event.key == K_z or event.key == K_RETURN:
                    confirm_load = self.all_regions[tuple(self.player_pos)].navigate_loop(incoming_direction=self.last_move)
                    if not confirm_load:
                        self.render_updated_map()

                    return confirm_load
                            
                # World Map Menu
                if event.key == K_a:
                    self.engine.fade_to('black')
                    confirm_load = self.wm_menu()
                    if not confirm_load:
                        self.render_updated_map()
                    return confirm_load
        
    def navigate_loop(self,load_player=False):
    
        """
        # Function Name: Navigate Loop
        # Purpose: Player Interaction with the world map
        # Inputs: Load Player = True if player location data is to be loaded,
        #                       False if the map starts fresh
        """
        menu_flag = True

        self.engine.play_music('overworld')
        self.framenum = 0
        self.player_pos = Vector2(5,6)  
        confirm_load = False
        
        # Tracking for player's movement on the map
        self.last_move = None
        
        # Associates all player units with the world map
        self.unit_associate()

        if load_player:
            self.player_pos = Vector2(self.engine.player.wm_data['wm_coords'])
            if self.engine.player.wm_data['in_region']:
                confirm_load = self.all_regions[tuple(self.player_pos)].navigate_loop(self.engine.player.wm_data['region_coords'])
                if confirm_load:
                    return
                
        
        self.update_wm_player_data()
        if not load_player and self.engine.enable_wm_tutorial:
            self.tutorial()


        self.render_background()
        self.render_description()
        self.engine.surface.blit(self.background_surface, (0,0))
        self.player_sprite_group.clear( self.engine.surface, self.background_surface)
        self.player_sprite_group.update(self.player_pos*35)
        rects = self.player_sprite_group.draw( self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)

        while menu_flag:
        
            if self.framenum == 9:
                self.framenum = 0
            self.framenum += 1
            
            confirm_load = self.user_input()
            if confirm_load:
                return


            self.player_sprite_group.clear( self.engine.surface, self.background_surface)
            self.player_sprite_group.update(self.player_pos*35)
            self.player_sprite_group.clear( self.engine.surface, self.background_surface)
            rects = self.player_sprite_group.draw( self.engine.surface)
            
            pygame.display.update(rects)
            self.engine.clock.tick(60)
            
    def tutorial(self):
        """
        # function name: Tutorial
        # Purpose: Provides the player with a first time tutorial of the game
        """
    
        reply_1 = self.choice("Would you like to see a tutorial on the world map?",["Yes","No"])
        if reply_1 == "Yes":
            self.say("Welcome to the world map!")
            self.say("The world map has two levels, this region level and a closer in location level.")
            self.say("At the location level, you can visit specific places.")
            self.say("If a location has a blue event bubble over it, then there is some mission or other event there.")
            self.say("You can also access the main menu from the world map.")
            reply_2 = self.choice("Which part would you like to hear about?",["Party Management","Missions","Save/Load","Finish Tutorial"])
            while reply_2 != "Finish Tutorial":
                if reply_2 == "Party Management":
                    self.conv_images['pm_menu'] = [pygame.image.load(os.path.join('images','worldmaptutorial','party_menu.png')).convert(),(228,70)]
                    self.say("This is the Party Management screen.")
                    self.say("You can use this to check the stats of each character")
                    del(self.conv_images['pm_menu'])
                    self.conv_images['si_menu'] = [pygame.image.load(os.path.join('images','worldmaptutorial','spell_inventory.png')).convert(),(228,70)]
                    self.say("You can also modify what spell actions and items each character can carry.")
                    self.say("Whenever you unequip spells or get new ones, they are automatically sorted in the categories of Attack Sorcery, Healing Sorcery, Spellcards, Healing Items.")
                    del(self.conv_images['si_menu'])
                    self.conv_images['tr_menu'] = [pygame.image.load(os.path.join('images','worldmaptutorial','equipped_traits.png')).convert(),(228,70)]
                    self.say("Another option is the traits menu.")
                    self.say("From here, you can change what traits start out active and which ones stay on standby.")
                    del(self.conv_images['tr_menu'])
                    self.say("As you gain more levels, you'll want to return here to learn new traits.")
                    self.conv_images['tr_menu2'] = [pygame.image.load(os.path.join('images','worldmaptutorial','unit_swap_traits.png')).convert(),(228,70)]
                    self.say("Newly learned traits will be placed into a unit's reserve traits and need to be swapped into standby or active to use them.")
                    del(self.conv_images['tr_menu2'])
                    self.say("That's all for the party management system!")
                elif reply_2 == "Missions":
                    self.say("The Missions menu handles all the events in the game.")
                    self.conv_images['mi_menu'] = [pygame.image.load(os.path.join('images','worldmaptutorial','mission_menu.png')).convert(),(228,70)]
                    self.say("From here, you will be able to view, sign up, and cancel currently available missions.")
                    self.say("This menu also has an archive of all the previous missions and helpful hints that appear over the course of the game.")
                    del(self.conv_images['mi_menu'])
                    
                elif reply_2 == "Save/Load":
                    self.say("The Save/Load menu allows you to record the progress of the game.")
                    self.say("Remember to save often, especially before a big battle.")
                
                reply_2 = self.choice("Which one would you like to hear about?",["Party Management","Missions","Save/Load","Finish Tutorial"])
            
    def wm_menu(self):
        
        """
        # Function Name: World Map Menu
        # Purpose: Manages access to team management, save / load, options
        """
    
        menu_flag = True
        menu_pos = 0
        
        fade_flag = False
        
        text_menu_header = self.engine.bfont.render("Main Menu",True,(0,0,0))
        text_party = self.engine.bfont.render("Party",True,(0,0,0))
        text_treasure = self.engine.bfont.render("Treasure",True,(0,0,0))
        if self.engine.unlock_shops or self.engine.check_event_completion(['CH1ST4']):
            text_synthesis = self.engine.bfont.render("Spell Synthesis",True,(0,0,0))
            text_trading = self.engine.bfont.render("Item Trading",True,(0,0,0))
        else:
            text_synthesis = self.engine.bfont.render("Spell Synthesis",True,(100,100,100))
            text_trading = self.engine.bfont.render("Item Trading",True,(100,100,100))
            
        text_mission = self.engine.bfont.render("Missions",True,(0,0,0))
        text_data = self.engine.bfont.render("Save / Load",True,(0,0,0))
        text_options = self.engine.bfont.render("Options",True,(0,0,0))
        text_exit = self.engine.bfont.render("Exit Game",True,(0,0,0))
        text_cancel = self.engine.bfont.render("Cancel",True,(0,0,0))
        
        self.engine.play_music('shop_theme')
        
        while menu_flag:
   
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    
                    if (event.key == K_z or event.key == K_RETURN):
                        
                        # Selects Party Management Menu
                        if menu_pos == 0:
                            self.party_menu()
                        
                        # Mission Management Menu
                        elif menu_pos == 1:
                            self.engine.mission_manager.index_menu()
                            
                        # Treasure Item menu
                        elif menu_pos == 2:
                            self.engine.treasure_menu()
                            
                        # Spell Synthesis menu
                        elif menu_pos == 3 and (self.engine.unlock_shops or self.engine.check_event_completion(['CH1ST4'])):
                            self.engine.spell_synthesis_system.spell_synthesis_menu()
                            
                        # Item Trading Menu
                        elif menu_pos == 4 and (self.engine.unlock_shops or self.engine.check_event_completion(['CH1ST4'])):
                            self.engine.trading_system.trading_menu()
                            
                        # Selects Save / Load Menu
                        elif menu_pos == 5:
                            confirm_load = self.engine.save_menu()
                            if confirm_load == True:
                                return True
                                
                        # Selects Options Menu
                        elif menu_pos == 6:
                            self.engine.options_menu()
                            
                        # Exits program
                        elif menu_pos == 7:
                            exit()
                                
                    if event.key == K_UP or event.key == K_LEFT:
                        if menu_pos > 0:
                            menu_pos -= 1
                        else:
                            menu_pos = 8
                    if event.key == K_DOWN or event.key == K_RIGHT:
                        if menu_pos < 8:
                            menu_pos += 1
                        else:
                            menu_pos = 0
                    if ((event.key == K_z or event.key == K_RETURN) and menu_pos == 8) or event.key == K_x:
                        menu_flag = False
                        self.engine.play_music('overworld')
                        
                        
            if menu_flag:        
                # Background image
                self.engine.surface.blit(self.engine.stats_bg,(0,0))
                # Header
                self.engine.surface.blit(text_menu_header,(50,50))
                # Renders the options
                self.engine.surface.blit(text_party,(70,70))
                self.engine.surface.blit(text_mission,(70,90))
                self.engine.surface.blit(text_treasure,(70,110))
                self.engine.surface.blit(text_synthesis,(70,130))
                self.engine.surface.blit(text_trading,(70,150))
                self.engine.surface.blit(text_data,(70,170))
                self.engine.surface.blit(text_options,(70,190))
                self.engine.surface.blit(text_exit,(70,210))
                self.engine.surface.blit(text_cancel,(70,230))    
                # Draws the cursor
                pygame.draw.circle(self.engine.surface, (0,0,0), (60,78+20*menu_pos),5)
                
                
                
                # Fade in once
                if fade_flag == False:
                    self.engine.fade_from('black')
                    fade_flag = True
                
                pygame.display.flip()
                self.engine.clock.tick(60)
    
    def party_menu(self):
    
        """
        # Party Management Menu
        # Purpose: Manages units in the party allowing management of
        #       traits and spells, etc
        """
    
        units = self.engine.player_units
        menu_flag = True
        menu_pos = [0,0]
        selected = 0
        names = []
        # Compiles a list of names
        for unit in units:
            text_name = self.engine.sfont.render(unit.name,True,(255,255,255))
            names.append([text_name,text_name.get_width()/2])
        
        text_directions1 = self.engine.sfont.render('Z - Select Character',True,(0,0,0))
        text_directions2 = self.engine.sfont.render("C - Unit's Stats",True,(0,0,0))
        text_directions3 = self.engine.sfont.render('A - Sort Units',True,(0,0,0))
        text_directions4 = self.engine.sfont.render('X - Return to Main Menu',True,(0,0,0))
        
        
        while menu_flag:
            
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT:
                        if menu_pos[0] > 0:
                            menu_pos[0] -= 1
                        elif menu_pos[0] == 0:
                            menu_pos[0] = 7
                        
                    if event.key == K_RIGHT:
                        if menu_pos[0] < 7:
                            menu_pos[0] += 1
                        elif menu_pos[0] == 7:
                            menu_pos[0] = 0
                        
                    if event.key == K_UP:
                        if menu_pos[1] > 0:
                            menu_pos[1] -= 1
                        elif menu_pos[1] == 0:
                            menu_pos[1] = 3
                    if event.key == K_DOWN: 
                        if menu_pos[1] < 3:
                            menu_pos[1] += 1
                        elif menu_pos[1] == 3:
                            menu_pos[1] = 0
                    
                    # Computes the selected unit
                    selected = menu_pos[0] + 8*menu_pos[1]
                    
                    # Calls up the options screen
                    if event.key == K_z or event.key == K_RETURN:
                        if selected < len(units):
                            self.unit_options(selected)
                    
                    # Jumps straight to stats screen
                    if event.key == K_c:
                        if selected < len(units):
                            units[selected].stats_loop(wm_flag = True)
                        
                    # Sort Units
                    if event.key == K_a:
                        self.engine.player_units = self.sort_units(units)
                        units = self.engine.player_units
                        names = []
                        # Compiles a list of names
                        for unit in units:
                            text_name = self.engine.sfont.render(unit.name,True,(255,255,255))
                            names.append([text_name,text_name.get_width()/2])
                        
                    if event.key == K_x:
                        menu_flag = False
                        
                    
                    
            self.engine.surface.blit(self.engine.party_bg,(0,0))
            self.engine.surface.blit(self.engine.menu_board,(0,490))
           
            
            # Draw the panels
            for row in xrange(0,4):
                for column in xrange(0,8):
                    self.engine.surface.blit(self.engine.unit_tile,(158+70*column,142+70*row))
                
            # Draw the units
            counter_x = 0
            counter_y = 0
            for unit in units:
            
                self.engine.surface.blit(unit.image,(158+70*counter_x,140+70*counter_y),(0,0,35,35))
                self.engine.surface.blit(names[counter_x+8*counter_y][0],(175+70*counter_x-names[counter_x+8*counter_y][1],180+70*counter_y))
                
                counter_x += 1
                # starts the next row if counter exceeds 8
                if counter_x > 7:
                    counter_x = 0
                    counter_y += 1
            
            # Draw the cursor
            self.engine.surface.blit(self.engine.cursor_img,(158+70*menu_pos[0],140+70*menu_pos[1]))
            
            self.engine.surface.blit(text_directions1,(440,505))
            self.engine.surface.blit(text_directions2,(440,520))
            self.engine.surface.blit(text_directions3,(440,535))
            self.engine.surface.blit(text_directions4,(440,550))
            
            # If the selected position is in the range of the length of the units, draw the 
            # stats of the unit
            if selected < len(units):
                units[selected].plot_stats()
                    
            pygame.display.flip()
            self.engine.clock.tick(60)
    
    def sort_units(self,units):
        """"
        # function name: sort_units
        # purpose: Allows player to sort units.
        """
        
        # Updates the unit names
        def update_unit_name_display(units):
            names = []
            # Compiles a list of names
            for unit in units:
                text_name = self.engine.sfont.render(unit.name,True,(255,255,255))
                names.append([text_name,text_name.get_width()/2])
            return names
            
        # Sorts units by a certain critieria
        def sort_units_by(criteria):
            sort_list = []
            # Compiles the criteria to be sorted for each unit (Name will be tiebreaker)
            for unit in self.engine.player_units:
                sort_criteria = {'Name':(unit.name,unit.name,unit),
                                 'Level':(unit.level,unit.name,unit),
                                 'EXP':(unit.exp,unit.name,unit),
                                 'HP':(unit.HP,unit.name,unit),
                                 'Class':(unit.unitclass,unit.name,unit),
                                 'STR':(unit.STR,unit.name,unit),
                                 'MAG':(unit.MAG,unit.name,unit),
                                 'DEF':(unit.DEF,unit.name,unit),
                                 'MDEF':(unit.MDEF,unit.name,unit),
                                 'ACC':(unit.ACC,unit.name,unit),
                                 'AGL':(unit.AGL,unit.name,unit)}
                
                sort_list.append(sort_criteria[criteria])
                
            # Sorts units
            sort_list.sort()
            # Sorts Highest to Lowest for numerical criteria
            if criteria in ('Level','HP','STR','MAG','DEF','MDEF','ACC','AGL'):
                sort_list.reverse()
                
            units_list = [unit[2] for unit in sort_list]
            return units_list
            
        names = update_unit_name_display(units)
        
        menu_flag = True
        menu_pos = [0,0]
        text_header = self.engine.bfont.render('Sort Units:',True,(0,0,0))
        criteria_list = ['Name','Class','Level','EXP','HP','STR','DEF','MDEF','ACC','AGL']
        text_criteria_list = [self.engine.sfont.render(criteria,True,(0,0,0)) for criteria in criteria_list]
        text_cancel = self.engine.sfont.render('Cancel',True,(0,0,0))
        
        while menu_flag:
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    # Handles left/right
                    if event.key == K_RIGHT or event.key == K_LEFT:
                        menu_pos[0] = not menu_pos[0]
                        
                    if event.key == K_UP:
                        if menu_pos[1] > 0:
                            menu_pos[1] -= 1
                        elif menu_pos[1] == 0:
                            menu_pos[1] = 4
                            
                    if event.key == K_DOWN:
                        if menu_pos[1] < 4:
                            menu_pos[1] += 1
                        elif menu_pos[1] == 4:
                            menu_pos[1] = 0
                    
                    # Calls up the options screen
                    if event.key == K_z or event.key == K_RETURN:
                        
                        units = sort_units_by(criteria_list[menu_pos[0]*5+menu_pos[1]])
                        names = update_unit_name_display(units)
                            
                        
                    if event.key == K_x:
                        menu_flag = False
                        return units
                                 
            self.engine.surface.blit(self.engine.party_bg,(0,0))
            self.engine.surface.blit(self.engine.menu_board,(0,490))
           
            
            # Draw the panels
            for row in xrange(0,4):
                for column in xrange(0,8):
                    self.engine.surface.blit(self.engine.unit_tile,(158+70*column,142+70*row))
                
            # Draw the units
            counter_x = 0
            counter_y = 0
            for unit in units:
            
                self.engine.surface.blit(unit.image,(158+70*counter_x,140+70*counter_y),(0,0,35,35))
                self.engine.surface.blit(names[counter_x+8*counter_y][0],(175+70*counter_x-names[counter_x+8*counter_y][1],180+70*counter_y))
                
                counter_x += 1
                # starts the next row if counter exceeds 8
                if counter_x > 7:
                    counter_x = 0
                    counter_y += 1
            
            # Draw the cursor
            pygame.draw.circle(self.engine.surface, (0,0,0), (440+100*menu_pos[0],528+15*menu_pos[1]),5)
            self.engine.surface.blit(text_header, (440,500))
            [self.engine.surface.blit(text_criteria,(450,520+15*index)) for index,text_criteria in enumerate(text_criteria_list[:5])]
            [self.engine.surface.blit(text_criteria,(550,520+15*index)) for index,text_criteria in enumerate(text_criteria_list[5:])]
            
            
                    
            pygame.display.flip()
            self.engine.clock.tick(60)
    
    
    def unit_options(self,selected):
        """
        # Name: Unit_options
        # Purpose: Allows selection of various managment options for an individual unit
        # Inputs: None
        """
    
        menu_pos = 0
        menu_flag = True
        
        text_spells = self.engine.sfont.render("Spell Actions",True, (0,0,0))
        text_traits = self.engine.sfont.render("Swap Traits",True, (0,0,0))
        text_learn_traits = self.engine.sfont.render("Learn New Traits",True, (0,0,0))
        text_stats = self.engine.sfont.render("Stats",True, (0,0,0))
        text_cancel = self.engine.sfont.render("Cancel",True, (0,0,0))
        
        units = self.engine.player_units
        selected_unit = self.engine.player_units[selected]
        
        
        names = []
        # Compiles a list of names
        for unit in units:
            text_name = self.engine.sfont.render(unit.name,True,(255,255,255))
            names.append([text_name,text_name.get_width()/2])
        
        while menu_flag:
            
             # looks for event type data to select interaction
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if ( event.key == K_UP or event.key == K_LEFT ):
                        if menu_pos > 0:
                            menu_pos -= 1
                        elif menu_pos == 0:
                            menu_pos = 4
                    if ( event.key == K_DOWN or event.key == K_RIGHT ):
                        if menu_pos < 4:
                            menu_pos += 1
                        elif menu_pos == 4:
                            menu_pos = 0
                    if event.key == K_z or event.key == K_RETURN:
                        # Spells
                        if menu_pos == 0:
                            selected_unit.spell_swap_menu()
                        
                        # Swap Traits
                        elif menu_pos == 1:
                            selected_unit.trait_swap_menu()
                        
                        # Learn New Traits
                        elif menu_pos == 2:
                            self.engine.trait_learning_system.trait_learning_menu(selected_unit)
                        
                        # Stats
                        elif menu_pos == 3:
                            selected_unit.stats_loop(wm_flag = True)
                        
                        # Cancel
                        elif menu_pos == 4:
                            menu_flag = False
                            
                    if event.key == K_x:
                        menu_flag = False
                    
                    
            if menu_flag:
            
                self.engine.surface.blit(self.engine.party_bg,(0,0))
                self.engine.surface.blit(self.engine.menu_board,(0,490))
               
                
                # Draw the panels
                for row in xrange(0,4):
                    for column in xrange(0,8):
                        self.engine.surface.blit(self.engine.unit_tile,(158+70*column,142+70*row))
                    
                # Draw the units
                counter_x = 0
                counter_y = 0
                for unit in units:

                    self.engine.surface.blit(unit.image,(158+70*counter_x,140+70*counter_y),(0,0,35,35))
                    self.engine.surface.blit(names[counter_x+8*counter_y][0],(175+70*counter_x-names[counter_x+8*counter_y][1],180+70*counter_y))

                    counter_x += 1
                    # starts the next row if counter exceeds 8
                    if counter_x > 7:
                        counter_x = 0
                        counter_y += 1
                    
                pygame.draw.circle(self.engine.surface, (0,0,0), (440,512+15*menu_pos),5)
                self.engine.surface.blit(text_spells, (450,505))
                self.engine.surface.blit(text_traits, (450,520))
                self.engine.surface.blit(text_learn_traits, (450,535))
                self.engine.surface.blit(text_stats, (450,550))
                self.engine.surface.blit(text_cancel, (450,565))
                
                # Draw the cursor
                # If the selected position is in the range of the length of the units, draw the 
                # stats of the unit
                selected_unit.plot_stats()
                        
                pygame.display.flip()
                self.engine.clock.tick(60)
                
        
class Region(object):
    
    def __init__(self,engine,name,wm_coords,portrait):
        """
        # Function Name: __init__
        # Purpose: Intializes a region
        # Inputs: 
        #         engine - main system engine
        #         name - Region Name
        #         wm_coords - World map coordinates
        #         portrait - portrait image ID
        """
    
        self.engine = engine
        self.name = name
        self.wm_coords = wm_coords
        self.pix_coords = 35*wm_coords
        self.portrait = portrait
        
        # World Map Display Name
        self.name_small = self.engine.sfont.render(self.name,True,(0,0,0))
        self.name_big = self.engine.bfont.render(self.name,True,(0,0,0))
        # Gets the size of the name to center it
        name_small_halfwidth = self.name_small.get_width()/2
        self.name_coords = 35*(wm_coords+Vector2(0,1))+Vector2(17-name_small_halfwidth,0)

        # Subregions and locations
        self.subregions = []
        self.all_locations = {}
        self.all_locations_by_name = {}
       
        # Connection dot points 
        self.dot_positions = set()
        self.paths = {}
        # self.connection_paths = paths
        # self.vector_paths = {'up':Vector2(0,0),'down':Vector2(0,0),'left':Vector2(0,0),'right':Vector2(0,0)} 

        default_position = Vector2(0,0)
        self.player_pos = default_position
        self.player_sprite_group = pygame.sprite.RenderUpdates()
        self.player_sprite = WorldmapSprite( 35*self.player_pos, self.engine.player_img)
        self.player_sprite_group.add(self.player_sprite)
        self.background_surface = pygame.surface.Surface((840,630))


    def add_locations(self,locations):
        """
        # Function Name: add_locations
        # Purpose: Adds a set of locations
        # Inputs: locations - a list of location objects to be added to the 
        """
    
        for location in locations:
            # Adds the location to the dictionary
            self.all_locations[tuple(location.wm_coords)] = location
            self.all_locations_by_name[location.name] = location
            
            # Associates a location with the region
            location.region = self
            location.get_graphical()
            
            # if the location is flagged as an entrance location, set it to the 
            # # default entry way to the world map
            # if location.name == self.entrance:
                # location.entrance = True
                # self.entrance_coords = location.wm_coords
            if location.name in self.entrances.values():
                location.exit = True
            
            
    def initialize_paths(self,location_path_list):
    
        """
        # Function Name: initialize_paths
        # Purpose: Initializes the paths of each location
        # Input: Location path list - a list of the paths to be initialized
        """

        self.location_path_list = location_path_list
        
        # List of all dots to not include a range around the location
        remove_pts = []
        for location in self.all_locations.values():
            remove_pts.extend((tuple(location.wm_coords),tuple((location.wm_coords+Vector2(-1,1))),tuple((location.wm_coords+Vector2(0,1))),tuple((location.wm_coords+Vector2(0,1)))))

        for path in location_path_list:
        
            # Each path has two connecting points along with the direction key to press to get to each location
            direction_to_1 = path.locations.keys()[0]
            region_1 = path.locations[direction_to_1]
            direction_to_2 = path.locations.keys()[1]
            region_2 = path.locations[direction_to_2]

            # We assign the starting point's path directions to this path so it can be called when the character needs to move
            self.all_locations_by_name[region_2].paths[direction_to_1] = path
            self.all_locations_by_name[region_1].paths[direction_to_2] = path
            
            # Constructs destination vectors
            delta_pos = self.all_locations_by_name[region_1].wm_coords - self.all_locations_by_name[region_2].wm_coords
            path.vectors[direction_to_1] = delta_pos
            path.vectors[direction_to_2] = -1*(delta_pos.copy())
            
            # # Decomposes delta vector into path dots for the map
            for i in xrange(min(1,int(delta_pos.y)),max(int(delta_pos.y),0)):
                dot_vector = self.all_locations_by_name[region_2].wm_coords + Vector2(0,i)
                if tuple(dot_vector) not in remove_pts:
                    path.dot_positions.append(tuple(35*dot_vector))
                
            for j in xrange(min(1,int(delta_pos.x)),max(int(delta_pos.x),0)):
                dot_vector = self.all_locations_by_name[region_2].wm_coords + Vector2(j,delta_pos.y)
                if tuple(dot_vector) not in remove_pts:
                    path.dot_positions.append(tuple(35*dot_vector))
        
    def update_region_player_data(self):
        """
        # Function Name: update wm player data
        # Purpose: Updates the location of the player
        """
    
        self.engine.player.wm_data['in_region'] = True
        self.engine.player.wm_data['region_coords'] = tuple(self.player_pos)
            
    ############################################
    # Graphical Methods
    ############################################
    
    def render_background(self):
    
        """
        # Function Name: Render background to the background surface
        # Purpose: Draws the background map and the locations
        """
    
        self.background_surface.blit(self.engine.region_bg,(0,0))
        self.background_surface.blit(self.engine.menu_board,(0,490))
        
        # Directions
        text_directions1 = self.engine.sfont.render("A - Menu",True,(0,0,0))
        text_directions2 = self.engine.sfont.render("Z - Select",True,(0,0,0))
        text_directions3 = self.engine.sfont.render("X - Return",True,(0,0,0))
        
        self.background_surface.blit(self.engine.tiny_text_board,(735,0))
        self.background_surface.blit(text_directions1,(840-(text_directions1.get_width()+20),15))
        self.background_surface.blit(text_directions2,(840-(text_directions2.get_width()+20),30))
        self.background_surface.blit(text_directions3,(840-(text_directions3.get_width()+20),45))


        # Location dots and path points
        for location in self.all_locations.values():
            self.background_surface.blit(self.engine.location_circle,location.pix_coords,(location.type*35,0,35,35))
            self.background_surface.blit(location.name_small,location.name_coords)
            if location.has_event():
                 self.background_surface.blit(self.engine.event_icon,(location.pix_coords+Vector2(10,-5)))
            
        [self.background_surface.blit(self.engine.path_dot,dot_pos)  for path in self.location_path_list if path.unlock for dot_pos in path.dot_positions]

        # Location Description
        self.background_surface.blit(self.all_locations[tuple(self.player_pos)].name_big,(140,505))
        self.background_surface.blit(self.name_small,(140,525))
        if self.all_locations[tuple(self.player_pos)].desc != None:
            desc_lines = split_lines(self.all_locations[tuple(self.player_pos)].desc,40)
            [self.background_surface.blit(self.engine.sfont.render(line,True,(0,0,0)),(140,555+index*20)) for index,line in enumerate(desc_lines)]

        self.background_surface.blit(self.engine.portrait_catalog[self.all_locations[tuple(self.player_pos)].portrait],(10,500))

    def render_updated_map(self):

        """
        Purpose: Updates the background surface and draws everything
        """
        self.render_background()
        self.engine.surface.blit(self.background_surface, (0,0))
        self.player_sprite_group.update(self.player_pos*35)
        self.player_sprite_group.clear( self.engine.surface, self.background_surface)
        self.player_sprite_group.draw(self.engine.surface)
        pygame.display.flip()
        self.engine.clock.tick(60)

    def move_to(self,direction_vector):
        """
        # Function Name: move_to
        # Purpose: Moves the player to a new direction
        # DIrection vector - Relative direction to move from curren tposition
        """

        path = []
        if abs(direction_vector.x) > 0:
            path.append(Vector2(direction_vector.x,0))
        if abs(direction_vector.y) > 0:
            path.append(Vector2(0,direction_vector.y))
        
        # Smoothstep function
        # See: http://sol.gfxile.net/interpolation/
        smoothstep = lambda v: (v*v*(3-2*v))
        
        
        start_pos = 35*self.player_pos
        current_pos = 35*self.player_pos
        
        for next_step in path:
        
            # Sets frame count proportional to the square root of the magnitude of the step vector
            frame_count = int(10*next_step.get_magnitude()**(1.0/2))           
            
            for t in xrange(0,int(frame_count)+1):
                v = float(t)/float(frame_count)
                v = smoothstep(v)
                scale_term = int(35*v)
                intermediate_step = scale_term*next_step
                
                current_pos = start_pos + intermediate_step

                self.player_sprite_group.update(current_pos)
                self.player_sprite_group.clear( self.engine.surface, self.background_surface)
                rects = self.player_sprite_group.draw( self.engine.surface)
                
                pygame.display.update(rects)
                self.engine.clock.tick(60)
                
            start_pos = current_pos.copy()
            
        self.player_pos += direction_vector
        self.render_updated_map()
    
       
    ############################################
    # Interactive methods
    ############################################
       
    def user_input(self):
        """
        # Function Name: user input
        # Purpose: top level interaction. Checks if there is any input by the user
        #   and allows the user to move around the cursor with the arrow keys and select a unit
        #   if there is one under the mouse
        """
    
        directions = {K_LEFT:'left',K_RIGHT:'right',K_UP:'up',K_DOWN:'down'}
    
        for event in pygame.event.get():
            if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                exit()
            if event.type == KEYDOWN:
                if event.key == K_LEFT or event.key == K_RIGHT or event.key == K_UP or event.key == K_DOWN:
                    # Checks if the player is allowed to move:
                    #   a. A path exists
                    #   b. The path is unlocked
                    if directions[event.key] in self.all_locations[tuple(self.player_pos)].paths.keys() and self.all_locations[tuple(self.player_pos)].paths[directions[event.key]].unlock:
                        self.move_to(self.all_locations[tuple(self.player_pos)].paths[directions[event.key]].vectors[directions[event.key]])
                        self.update_region_player_data()
                        return 
                        
                if event.key == K_z or event.key == K_RETURN:
                    exit_flag = self.all_locations[tuple(self.player_pos)].menu_loop()
                    return exit_flag
                   
               # World Map Menu
                if event.key == K_a:
                    self.engine.fade_to('black')
                    confirm_load = self.wm_parent.wm_menu()
                    if confirm_load:
                        return 'load_player'
                    
                if event.key == K_x:
                    return 'jump2wm'

        self.render_updated_map()
        return False
            
    def navigate_loop(self, start_coords=None, incoming_direction=None):
    
        """
        # Function Name: Navigate Loop
        # Purpose: Player Interaction with the world map
        # Inputs: start_coords - Player starting coordinates
        #         incoming direction - Player's incoming direction if a location has two+ gates
        """
    
        menu_flag = True
        
        self.framenum = 0
        if start_coords:
            self.player_pos = Vector2(start_coords)
        else:
            # If a location only has one entrance or player is not coming from another location since they just started the game, jump them directly to that default/only entrance
            if len(self.entrances.values()) == 1 or not incoming_direction:
                self.player_pos = Vector2(tuple(self.all_locations_by_name[self.entrances['default']].wm_coords))
            else:
                self.player_pos = Vector2(tuple(self.all_locations_by_name[self.entrances[incoming_direction]].wm_coords))
        self.update_region_player_data()

        self.render_updated_map()
        
        while menu_flag:
        
            if self.framenum == 9:
                self.framenum = 0
            self.framenum += 1
            
            input_flag = self.user_input()
                    

            # If the location flag received, return to the world map
            if input_flag == 'jump2wm':
                menu_flag = False
                self.wm_parent.player_pos = Vector2(tuple(self.wm_coords))
                self.wm_parent.update_wm_player_data()
            # If load player flag received, load new map
            elif input_flag == 'load_player':
                return True

            self.player_sprite_group.update(self.player_pos*35)
            self.player_sprite_group.clear( self.engine.surface, self.background_surface)
            rects = self.player_sprite_group.draw(self.engine.surface)

            pygame.display.update(rects)
            self.engine.clock.tick(60)


# Connecting path object
class Path(object):

    def __init__(self,locations,prereq):
    
        """
        # Function Name; __init__
        # Purpose: Initializes a path object
        # Inputs:
        #           locations - where does this path connect to
        #           prereqs - events that must be completed for this path to open
        """
    
        self.locations = locations
        if prereq:
            self.prereq = prereq
            self.unlocked = False
        else:
            self.prereq = None
            self.unlocked = True
        self.vectors = {}
        self.dot_positions = []
        
     
    def __str__(self):
        """
        # function Name: __str__
        # Purpose: String representation of a path
        # Output: returns the two connections of the path, the prerequisites, and whether it is locked or not
        """

        return str([self.locations,self.prereqs,self.unlocked])
     
     
class Location(object):
    
    def __init__(self,name,desc,wm_coords,type,portrait):
        """
        # Function Name: __init__
        # Purpose: Intializes a location
        # Inputs: 
        #         Region - Parent region
        #         Name - Location's name
        #         Wm_coords - World map coordinates
        #         type - An integer that corresponds to the following types of location
        #                   0 - Neutral
        #                   1 - Danger Zone (Likely location for monster attacks and missions)
        #                   2 - Safe Zone (Towns and villages)
        #                   3 - Region exit / entrance point
        #         portrait - portrait image ID 
        """
    
        self.name = name
        self.desc = desc
        self.wm_coords = wm_coords
        self.pix_coords = 35*wm_coords + Vector2(0,5)
        self.type = type
        self.portrait = portrait
        self.all_events = [None,None,None,None]
        self.entrance = False
        self.exit = False
        self.paths = {} 
    def add_event(self,event):
        """
        # Function Name: add_event
        # Purpose: Adds an event to the region's list
        """
        if None in self.all_events:
            self.all_events[self.all_events.index(None)] = event
            event.location = self
    
    def has_event(self):
        """
        # Function Name: has_event
        # Purpose: Checks if this location has an any events
        # Output: Return True if location has events
        """
        for event in self.all_events:
            if event:
                return True
        # no event, return False
        return False
        
    def get_graphical(self):

        """
        # Function Name: get_graphical
        # Purpose: Generates the graphical objects for a location (note: requires association with a region)
        """
    
        # World Map Display objects
        
        self.name_small = self.region.engine.sfont.render(self.name,True,(0,0,0))
        self.name_big = self.region.engine.bfont.render(self.name,True,(0,0,0))
    
        # Gets the size of the name to center it
        name_small_halfwidth = self.name_small.get_width()/2
        self.name_coords = 35*(self.wm_coords+Vector2(0,1))+Vector2(17-name_small_halfwidth,5)
        
    def get_text_all_events(self):
        """
        # Function Name: get_text_all_events
        # Purpose: Generates the text objects from a location's events
        """
    
        text_event = []
        for event in self.all_events:
            if not event:
                text_event.append(self.region.engine.sfont.render('None',True,(100,100,100)))
            else:
                text_event.append(self.region.engine.sfont.render("["+event.type+"] "+event.name,True,(0,0,0)))
        return text_event

    def render_menu_options(self):
        """
        Purpose: Draws the locations event options to background surface
        """

        # Generate text objects
        if self.exit == True:
            text_exit = self.region.engine.sfont.render("Leave Region",True,(0,0,0))
        else:
            text_exit = self.region.engine.sfont.render("Leave Region",True,(100,100,100))

        text_cancel = self.region.engine.sfont.render("Cancel",True,(0,0,0))

        text_event = self.get_text_all_events()
        for i in (0,1,2,3):
            self.region.background_surface.blit(text_event[i], (450,505+15*i))
        self.region.background_surface.blit(text_exit, (450,565))
        self.region.background_surface.blit(text_cancel,(450,580))


    def render_updated_map(self):
        """
        Renders a full frame of the map and options menu
        """

        self.region.render_background()
        self.render_menu_options()
        self.region.engine.surface.blit(self.region.background_surface, (0,0))
        self.region.player_sprite_group.draw(self.region.engine.surface)
        pygame.display.flip()

        
    def menu_loop(self):
        """
        # Function Name: menu_loop
        # Purpose: Location level menu loop that allows for the selection of an event in a location.
        # Inputs: None
        # Outputs: exit_flag
        """
    

            
        menu_flag = True
        menu_pos = 0

        self.render_updated_map()

        while menu_flag:

            # User Input 
            for event in pygame.event.get():
                if event.type == QUIT or (event.type == KEYDOWN and event.key == K_F4 and bool(event.mod & KMOD_ALT)):
                    exit()
                if event.type == KEYDOWN:
                    if event.key == K_LEFT or event.key == K_UP:
                        if menu_pos > 0:
                            menu_pos -= 1
                        else:
                            menu_pos = 5
                    elif event.key == K_RIGHT or event.key == K_DOWN:
                        if menu_pos < 5:
                            menu_pos += 1
                        else:
                            menu_pos = 0
                        
                    if event.key == K_z or event.key == K_RETURN:
                    
                        # Selects event if it is available
                        if menu_pos >= 0 and menu_pos<=3 and self.all_events[menu_pos]:
                            
                                                
                            # Auto-save after mission if enabled
                            if self.region.engine.options.auto_save:
                                self.region.engine.player.save(5)
                                
                            self.all_events[menu_pos].execute()
                            
                            # Auto-save after mission if enabled
                            if self.region.engine.options.auto_save:
                                self.region.engine.player.save(5)
                                
                            # Refreshes text_events
                            text_event = self.get_text_all_events()

                            self.render_updated_map()
                                    
                        # Checks exit 
                        elif menu_pos == 4 and self.exit == True:
                            return 'jump2wm'
                        # Cancel
                        elif menu_pos == 5:
                            menu_flag = False
                    if event.key == K_x:
                        menu_flag = False
            
            if menu_flag:
                # Location menu
                self.region.engine.surface.blit(self.region.background_surface, (435, 505), (435, 505, 10, 125 ))
                pygame.draw.circle(self.region.engine.surface, (0,0,0), (440,513+15*menu_pos),5)
                
                pygame.display.update(pygame.Rect(435, 505, 10, 125 ))
                self.region.engine.clock.tick(60)


class WorldmapSprite(pygame.sprite.Sprite):

    def __init__(self, start_pos, image):
        """
        # Function: __init__
        # Purpose: Creates a sprite for the worldmap
        # Inputs:   map - worldmap object to associate with
        #           image - image to use in the worldmap sprite
        #
        """
        pygame.sprite.Sprite.__init__(self)
        self.image = image.subsurface((0,0,35,35))
        self.rect = self.image.get_rect()
        self.rect.topleft = tuple(start_pos)

    def update(self, new_position):
        """
        # Function: update
        # Purpose: Updates sprite to current location of the unit (pixel)
        """
        self.rect.topleft = tuple(new_position)
